#ifndef _BOARD_EASY334_H
#define _BOARD_EASY334_H
/******************************************************************************

                               Copyright (c) 2008
                            Infineon Technologies AG
                     Am Campeon 1-12; 81726 Munich, Germany

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

  ****************************************************************************
   Module      : board_easy50712.h
   Description :
 ****************************************************************************
   \file board_easy50712.h

   \remarks
      
   \note Changes:
*******************************************************************************/

/* ============================= */
/* Includes                      */
/* ============================= */
#include "device_vinetic_vip.h"

/* ============================= */
/* Debug interface               */
/* ============================= */

/* ============================= */
/* Global Defines                */
/* ============================= */

/* ============================= */
/* Global Variables              */
/* ============================= */
typedef enum
{
   SYSTEM_AM_16MOT_VIP = EASY334_ACCESS_16BIT_MOTOROLA,
   SYSTEM_AM_16INTMUX_VIP = EASY334_ACCESS_16BIT_INTELMUX,
   SYSTEM_AM_16INTDEMUX_VIP = EASY334_ACCESS_16BIT_INTELDEMUX,
   SYSTEM_AM_8MOT_VIP = EASY334_ACCESS_8BIT_MOTOROLA,
   SYSTEM_AM_8INTMUX_VIP = EASY334_ACCESS_8BIT_INTELMUX,
   SYSTEM_AM_8INTDEMUX_VIP = EASY334_ACCESS_8BIT_INTELDEMUX,
   SYSTEM_AM_8SPI_VIP = EASY334_ACCESS_SPI
} SYS_ACCESS_MODE_VIP;
enum { SYSTEM_AM_DEFAULT_VIP = SYSTEM_AM_16MOT_VIP };
enum { SYSTEM_AM_UNSUPPORTED_VIP = SYSTEM_AM_8SPI_VIP + 1 };

/** offset of EASY334 register used to setup PCM, cpld cmd 3 register offset */
#define CMDREG1_OFF_VIP                    EASY334_CPLD_CMDREG3_OFF
#define CMDREG2_OFF_VIP                    EASY334_CPLD_CMDREG2_OFF
#define CMDREG3_OFF_VIP                    EASY334_CPLD_CMDREG3_OFF
#define CMDREG4_OFF_VIP                    EASY334_CPLD_CMDREG4_OFF
#define CMDREG5_OFF_VIP                    EASY334_CPLD_CMDREG5_OFF

/** define if master - 1 or slave - 0 */
enum { MASTER_VIP = 0x010 }; /* bit 4 of register CMDREG3 */
/** Vinetic reset not active */
enum { VRES_VIP = 0x01 };
/** Which highway is selected A or B */
enum
{
   HIGHWAY_A = 0x40,
   HIGHWAY_B = 0x0
};
/** Micro controller via CPLD or MPC ~ VINETIC */
enum
{
   MICRO_CONTROLER_CPLD = 0x0,
   MICRO_CONTROLER_MPC_VIN = 0x20
};
/** Access mode with PCM */
enum
{
   MOTOR_16_BIT_AND_PCM = 0x0C,
   MOTOR_8_BIT_AND_PCM = 0x04,
   INTEL_16_BIT_DEMUX_AND_PCM = 0x0B,
   INTEL_8_BIT_DEMUX_AND_PCM = 0x03,
   INTEL_16_BIT_MUX_AND_PCM = 0x0A,
   INTEL_8_BIT_MUX_AND_PCM = 0x02
};
/** Used to clear bits in register */
enum { ACCESS_MODE_MASK = 0x0F };
/** How MCLK is defined */
enum
{
   MCLK_IS_DCL = 0x0,
   MCLK_SEL_IN_CMDREG5 = 0x80
};
/** Is IOM-2 connector clock input or output */ 
enum
{
   IOM_2_CLK_IN = 0x0,
   IOM_2_CLK_OUT = 0x20
};
/** Frequencies for clock generator */
enum
{
   FREQ_4096_KHZ_8_KHZ = 0x0,
   FREQ_2048_KHZ_8_KHZ = 0x1,
   FREQ_1536_KHZ_8_KHZ = 0x2,
   FREQ_8192_KHZ_8_KHZ = 0x3,
   FREQ_3072_KHZ_8_KHZ = 0x4,
   FREQ_1024_KHZ_8_KHZ = 0x5,
   FREQ_512_KHZ_8_KHZ = 0x6
};
/** Used to clear bits in register */
enum { FREQ_MASK = 0x0F };

/*
Master:
CMDREG1:                  - bit 7..5: Control bits for analog multiplexer
                                      (Line test support)
                                     000 - Voltage on RING channel B is
                                           measured
                                     001 - VBATH is measured
                                     ... - Just measuring
                          - bit 4: always 0
                          - bit 3: always 0
                          - bit 0: 1 - VRES, Vinetic Reset not active
                                   0 - VRES, Vinetic Reset active
CMDREG2: 0x2c (0010 1100) - bit 6: 0 - highway B conn to MPC, A select for
                                       measurments
                                   1 - highway A conn to MPC, B select for
                                       measurments
                            bit 5: 0 - micro controller interface conn via
                                       CPLD
                                   1 -  -||-                      conn 
                                       directly between MPC and VINETIC 
                            bit 3..0: 1100 - 16 bit Motor. mode + PCM
                                      0100 - 8 bit Motor. mode + PCM
                                      1011 - 16 bit Intel Demux. mode + PCM
                                      0011 - 8 bit Intel Demux. mode + PCM
                                      1010 - 16 bit Intel Multip. mode + PCM
                                      0010 - 8 bit Intel Multip. mode + PCM
                                      0001 - SCI + PCM
                                      1111 - IOM-2

CMDREG3: 0x31 (0011 0001) - bit 7: 0 - MCLK = DCL
                                   1 - MCLK selected in CMDREG5
                            bit 5: 0 - IOM-2 Connector is clock input
                                   1 -   -||-                   output
                            bit 4: 0 - Board in slave mode
                                   1 - Board in master mode
                            bit 3..0: Clock generator
                                              PCLK = DCL = PDC, PFS = FSC
                                      0000 -  4.096 MHz, 8kHz
                                      0001 -  2.048 MHz, 8kHz
                                      0010 -  1.536, 8
                                      0011 -  8.192, 8
                                      0100 -  3.072, 8
                                      0101 -  1.024, 8
                                      0110 -  512, 8
                                      ostalo reserved 
CMDREG4: 0x0 - Controls LEDs 0 off, 1 on :-)
CMDREG5: 0x0 - bit 7..5 - Master Clock Selection (CMDREG3, bit 7 must be 1)
                          000 - MCLK = 4.096 MHz
                          ...
               Other bits are used for DEBUG, HDLC, JTAG interfaces.

The configuration I am using for clock master mode:
CMDREG1: 0x01
CMDREG2: 0x2C
CMDREG3: 0x31
CMDREG4: 0x00
CMDREG5: 0x00
Slave:
CMDREG1: 0x01
CMDREG2: 0x2C
CMDREG3: 0x21
CMDREG4: 0x0
CMDREG5: 0x0
*/

/* ============================= */
/* Global Structures             */
/* ============================= */

/* ============================= */
/* Global function declaration   */
/* ============================= */

IFX_void_t BOARD_Easy334_RemoveBoard(BOARD_t* pBoard);
IFX_return_t BOARD_Easy334_Init(BOARD_t* pBoard, PROGRAM_ARG_t* pProgramArg);

#endif /* _BOARD_EASY334_H */
