/******************************************************************************

                               Copyright (c) 2008
                            Infineon Technologies AG
                     Am Campeon 1-12; 81726 Munich, Germany

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

  ****************************************************************************
   Module      : cid.c
   Date        : 2005-11-29
   Description : This file contains the implementation of the functions for
                 the tapi demo working with Caller ID.
   \file

   \remarks

   \note Changes:
*******************************************************************************/

/* ============================= */
/* Includes                      */
/* ============================= */

#include "tapidemo.h"
#include "cid.h"
#ifdef EASY336
#include "common.h"
#endif /* EASY336 */

/* ============================= */
/* Local structures              */
/* ============================= */

/** Structure holding CID for one user (phone number, date, user name) */
typedef struct _CID_USER_t
{
   IFX_uint32_t nMonth;
   IFX_uint32_t nDay;
   IFX_uint32_t nHour;
   IFX_uint32_t nMn;
   IFX_char_t sName[IFX_TAPI_CID_MSG_LEN_MAX];
   IFX_char_t sNumber[IFX_TAPI_CID_MSG_LEN_MAX];
} CID_USER_t;

/* ============================= */
/* Global Structures             */
/* ============================= */


/* ============================= */
/* Global function declaration   */
/* ============================= */


/* ============================= */
/* Local variable definition     */
/* ============================= */

/** Current CID standard to use */
static IFX_TAPI_CID_STD_t eCurrCID_Std = IFX_TAPI_CID_STD_ETSI_FSK;

/** Predefined CID data for 4 users */
const static CID_USER_t CID_USERS[CID_MAX_USERS] =
{
   {11, 3, 10, 44, "Lisa\0", "12345\0"},
   {11, 3, 10, 44, "Homer\0", "67890\0"},
   {11, 3, 10, 44, "March\0", "33333\0"},
   {11, 3, 10, 44, "Bart\0", "44444\0"}
};

/** CID standards with description */
const static CID_STANDARD_t CID_STANDARDS[CID_MAX_STANDARDS] =
{
   {IFX_TAPI_CID_STD_TELCORDIA, "TELCORDIA, Belcore, USA\0"},
   {IFX_TAPI_CID_STD_ETSI_FSK, "ETSI FSK, Europe\0"},
   {IFX_TAPI_CID_STD_ETSI_DTMF, "ETSI DTMF, Europe\0"},
   {IFX_TAPI_CID_STD_SIN, "SIN BT, Great Britain\0"},
   {IFX_TAPI_CID_STD_NTT, "NTT, Japan\0"},
};

/** CID standards with short description */
const CID_STANDARD_t CID_STANDARDS_SHORT[CID_MAX_STANDARDS] =
{
   {IFX_TAPI_CID_STD_TELCORDIA, "TELCORDIA\0"},
   {IFX_TAPI_CID_STD_ETSI_FSK, "ETSI FSK\0"},
   {IFX_TAPI_CID_STD_ETSI_DTMF, "ETSI DTMF\0"},
   {IFX_TAPI_CID_STD_SIN, "SIN BT\0"},
   {IFX_TAPI_CID_STD_NTT, "NTT\0"},
};

/* ============================= */
/* Local function definition     */
/* ============================= */


/* ============================= */
/* Global function definition    */
/* ============================= */


/**
   Prepare CID message.

   \param  nIdx - cid user index, by default its the same as channel number
   \param  pPhone - pointer to PHONE

   \return status (IFX_SUCCESS or IFX_ERROR)

   \remark
      This test assumes straights connections between phone and data channel
*/
IFX_return_t CID_SetupData(IFX_uint32_t nIdx,
                           PHONE_t* pPhone)
{
   IFX_uint8_t element = 0;
   struct tm* date_time = IFX_NULL;
   time_t time_in_sec;
   char temp_string[IFX_TAPI_CID_MSG_LEN_MAX];

   /** \todo CID_SetupData() according to CID types we have different
             structure of CID messages */
   /* check input parameters */
   TD_PTR_CHECK(pPhone, IFX_ERROR);
   TD_PARAMETER_CHECK((0 > nIdx), nIdx, IFX_ERROR);
   TD_PARAMETER_CHECK((CID_MAX_USERS < nIdx), nIdx, IFX_ERROR);

   memset(&pPhone->pCID_Msg->message[0], 0, sizeof(pPhone->pCID_Msg->message));

   /* current calendar time in seconds */
   time_in_sec = time(IFX_NULL);

/*   TAPIDEMO_PRINTF(("%s%ju secs since the Epoch %d\n",
           asctime(localtime(&time_in_sec)),
                   (uintmax_t)time_in_sec));*/

   /* converts calendar time into broke-down time */
   date_time = localtime(&time_in_sec);

/*   TAPIDEMO_PRINTF(("day %d, month %d, year %d, hour %d, min %d\n",
           date_time->tm_mday, date_time->tm_mon, date_time->tm_year,
           date_time->tm_hour, date_time->tm_min);*/


   /*if ((0 != CID_USERS[nIdx].nMonth) && (0 != CID_USERS[nIdx].nDay))*/
   if (IFX_NULL != date_time)
   {
      pPhone->pCID_Msg->message[element].date.elementType =
         IFX_TAPI_CID_ST_DATE;

      /* Only 1 - 12 allowed */
      pPhone->pCID_Msg->message[element].date.month =
         ((date_time->tm_mon + 1) % 13);

      /* Only 1 - 31 allowed */
      pPhone->pCID_Msg->message[element].date.day = (date_time->tm_mday % 32);
      /* Only 0 - 23 allowed */
      pPhone->pCID_Msg->message[element].date.hour = (date_time->tm_hour % 24);
      /* Only 0 - 59 allowed */
      pPhone->pCID_Msg->message[element].date.mn = (date_time->tm_min % 60);
      element++;
   }
   else
   {
      TRACE (TAPIDEMO, DBG_LEVEL_HIGH,
             ("Error, Setup datetime element in CID.\n"));
   }

   /* TRACE(TAPIDEMO, DBG_LEVEL_LOW, ("Setup number CLI element in CID.\n")); */

   /*if (IFX_NULL != CID_USERS[nIdx].sNumber)*/
   {
      pPhone->pCID_Msg->message[element].string.elementType =
         IFX_TAPI_CID_ST_CLI;

      /*snprintf(temp_string, IFX_TAPI_CID_MSG_LEN_MAX,
               "%d", pPhone->nPhoneNumber);*/

      sprintf(temp_string, "%d", (int) pPhone->nPhoneNumber);

      pPhone->pCID_Msg->message[element].string.len =
         strlen(temp_string);

      strncpy((char*)pPhone->pCID_Msg->message[element].string.element,
              temp_string,
              sizeof(pPhone->pCID_Msg->message[element].string.element));
      element++;
   }

   /*if (IFX_NULL != CID_USERS[nIdx].sName)*/
   {
      pPhone->pCID_Msg->message[element].string.elementType =
         IFX_TAPI_CID_ST_NAME;

      /*snprintf(temp_string, IFX_TAPI_CID_MSG_LEN_MAX,
               "%s#%d", pPhone->pBoard->pszBoardName, pPhone->nPhoneCh);*/
#ifndef TAPI_VERSION4
      sprintf(temp_string, "%s#%d",
              pPhone->pBoard->pszBoardName, (int) pPhone->nPhoneCh);
#else /* TAPI_VERSION4 */
      sprintf(temp_string, "%s#%d,%d",
                    pPhone->pBoard->pszBoardName,
                    pPhone->nDev,
                    pPhone->nPhoneCh);
#endif /* TAPI_VERSION4 */

      pPhone->pCID_Msg->message[element].string.len =
         strlen(temp_string);

      strncpy((char*)pPhone->pCID_Msg->message[element].string.element,
              temp_string,
              sizeof(pPhone->pCID_Msg->message[element].string.element));
      element++;
   }

   /* Put these elements in the cid message */
   if (0 < element)
   {
      pPhone->pCID_Msg->nMsgElements = element;
      return IFX_SUCCESS;
   }

   TRACE (TAPIDEMO, DBG_LEVEL_HIGH,
          ("Error, Phone No %d: Setup CID - failed.\n(File: %s, line: %d)\n",
           pPhone->nPhoneNumber, __FILE__, __LINE__));

   return IFX_ERROR;
} /* CidSetupData() */


/**
   If nMode == CID_DISPLAY_ALL:
   Displays cid structure, which has 3 elements:
   1.) date,
   2.) CLI number
   3.) name
   If nMode == CID_DISPLAY_SEND:
   Displays:
     -sending phone number,
     -cid standard name,
     -CLI number and name

   \param pPhone - pointer to PHONE

   \return output on screen
*/
IFX_void_t CID_Display(PHONE_t* pPhone, CID_DISPLAY_MODE_t nMode)
{
   /* check input arguments */
   if (IFX_NULL == pPhone)
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return;
   }
   switch (nMode)
   {
      case CID_DISPLAY_ALL:
         SEPARATE;
         /*TAPIDEMO_PRINTF(("There are %d usefull CIDs\n", MAX_SYS_LINE_CH));*/
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("Element count %d\n", pPhone->pCID_Msg->nMsgElements));
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("Day: %d, Month: %d, Hour: %d, Minutes: %d\n",
                pPhone->pCID_Msg->message[CID_IDX_DATE].date.day,
                pPhone->pCID_Msg->message[CID_IDX_DATE].date.month,
                pPhone->pCID_Msg->message[CID_IDX_DATE].date.hour,
                pPhone->pCID_Msg->message[CID_IDX_DATE].date.mn));

         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("CLI number is %s\n",
                pPhone->pCID_Msg->message[CID_IDX_CLI_NUM].string.element));

         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("Name is %s\n",
                pPhone->pCID_Msg->message[CID_IDX_NAME].string.element));
        SEPARATE;
         break;
      case CID_DISPALY_SEND:
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("Phone No %d: Sending CID (%s) data [%s, %s]\n",
                pPhone->nPhoneNumber,
                CID_STANDARDS_SHORT[eCurrCID_Std].sDescription,
                pPhone->pCID_Msg->message[CID_IDX_NAME].string.element,
                pPhone->pCID_Msg->message[CID_IDX_CLI_NUM].string.element));
         break;
      case CID_DISPLAY_SIMPLE:
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("Phone No %d: Set CID (%s) data [%s, %s]\n",
                pPhone->nPhoneNumber,
                CID_STANDARDS_SHORT[eCurrCID_Std].sDescription,
                pPhone->pCID_Msg->message[CID_IDX_NAME].string.element,
                pPhone->pCID_Msg->message[CID_IDX_CLI_NUM].string.element));
         break;
      default:
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Wrong CID_Display mode %d\n",
                nMode));
         break;
   }

} /* Show_CID_Number() */


/**
   Set CID standard to use

   \param nCID_Standard - CID standard number

   \return IFX_SUCCESS if OK, IFX_ERROR if errors
*/
IFX_return_t CID_SetStandard(IFX_int32_t nCID_Standard)
{
   /* set default value */
   IFX_TAPI_CID_STD_t cid_standard = IFX_TAPI_CID_STD_ETSI_FSK;

   if ((0 > nCID_Standard) || (CID_MAX_STANDARDS <= nCID_Standard))
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s), will use default %s CID standard. "
            "(File: %s, line: %d)\n",
            CID_STANDARDS[IFX_TAPI_CID_STD_ETSI_FSK].sDescription,
            __FILE__, __LINE__));
      /* Don't send error, but use default value, ETSI-FSK (Europe). */
   }
   else
   {
      cid_standard = (IFX_TAPI_CID_STD_t) (nCID_Standard);
   }

   eCurrCID_Std = cid_standard;

   TRACE(TAPIDEMO, DBG_LEVEL_LOW, ("Used CID standard %d => %s\n",
         (int) CID_STANDARDS[cid_standard].nNumber,
         CID_STANDARDS[cid_standard].sDescription));

   return IFX_SUCCESS;
} /* CID_GetStandard() */


/**
   Return CID standard to use

   \return CID standard
*/
IFX_TAPI_CID_STD_t CID_GetStandard(IFX_void_t)
{
   /* set standard */
   return eCurrCID_Std;
}


/**
   Configures driver for CID services

   \param pPhone - pointer to phone

   \return IFX_SUCCESS if OK, otherwise IFX_ERROR
*/
IFX_return_t CID_ConfDriver(PHONE_t* pPhone)
{
   IFX_TAPI_CID_CFG_t cid_std;
   IFX_TAPI_CID_STD_TYPE_t oFSK_cfg;
   IFX_TAPI_CID_FSK_CFG_t  oFSKConf;
   IFX_return_t ret;
   IFX_uint32_t nDataCh_FD;

   /* check input parameters */
   TD_PTR_CHECK(pPhone, IFX_ERROR);

   nDataCh_FD = pPhone->nPhoneCh_FD;
   if (0 > nDataCh_FD)
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }

   memset(&cid_std, 0, sizeof(cid_std));
   cid_std.nStandard = eCurrCID_Std;

   if (IFX_TAPI_CID_STD_ETSI_FSK == cid_std.nStandard)
   {
      memset(&oFSK_cfg, 0, sizeof(IFX_TAPI_CID_STD_TYPE_t));
      memset(&oFSKConf, 0, sizeof(IFX_TAPI_CID_FSK_CFG_t));
      /* add CID configuration */
      cid_std.cfg = &oFSK_cfg;
      /* set parameters */
      oFSK_cfg.etsiFSK.ackTone = 0;
      oFSK_cfg.etsiFSK.nAlertToneOffhook = 0;
      oFSK_cfg.etsiFSK.nAlertToneOnhook = 0;
      oFSK_cfg.etsiFSK.nETSIAlertNoRing = IFX_TAPI_CID_ALERT_ETSI_RP;
      oFSK_cfg.etsiFSK.nETSIAlertRing = IFX_TAPI_CID_ALERT_ETSI_RP;
      oFSK_cfg.etsiFSK.ringPulseTime = 500;
      /* add CID configuration */
      oFSK_cfg.etsiFSK.pFSKConf = &oFSKConf;
      /* set gains */
      oFSKConf.levelRX = -320;
      oFSKConf.levelTX = -100;
      oFSKConf.markRXOffhook = 55;
      oFSKConf.markRXOnhook = 150;
      oFSKConf.markTXOffhook = 80;
      oFSKConf.markTXOnhook = 180;
      oFSKConf.seizureRX = 200;
      oFSKConf.seizureTX = 300;
   }
#ifdef TAPI_VERSION4
   if (pPhone->pBoard->fSingleFD)
   {
      cid_std.dev = pPhone->nDev;
      cid_std.ch = pPhone->nPhoneCh;

      /* Selection of CID standard to be used */
      TD_ASSERT((ret = ioctl(nDataCh_FD, IFX_TAPI_CID_CFG_SET,
         (IFX_int32_t) &cid_std)) == IFX_SUCCESS, nDataCh_FD, pPhone->pBoard,
         cid_std.dev, cid_std.ch);
   } /* if (pPhone->pBoard->fSingleFD) */
   else
#endif /* TAPI_VERSION4 */
   {
      ret= ioctl(nDataCh_FD, IFX_TAPI_CID_CFG_SET,
                             (IFX_int32_t) &cid_std);
   } /* if (pPhone->pBoard->fSingleFD) */
   if (ret == IFX_ERROR)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("CID configuration set failed for this channel. \
            (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }

   return IFX_SUCCESS;
} /* CID_ConfDriver() */


/**
   Release CID.

   \param pCID_Msg - CID message to be cleaned up

   \return IFX_SUCCESS if OK, otherwise IFX_ERROR
*/
IFX_return_t CID_Release(IFX_TAPI_CID_MSG_t* pCID_Msg)
{
   TD_PTR_CHECK(pCID_Msg, IFX_ERROR);

   if (pCID_Msg->message->transparent.data != IFX_NULL)
   {
      free(pCID_Msg->message->transparent.data);
      pCID_Msg->message->transparent.data = IFX_NULL;
   }
   return IFX_SUCCESS;
} /* CID_Release() */


/**
   Sends a caller id sequence according to hook state, standard, ...

   \param nDataCh_FD - file descriptor of data channel, for Duslic-xT this is
                       phone channel file descriptor
   \param nHookState - state of hook ONHOOK or OFFHOOK
   \param pCID_Msg   - pointer to CID message
   \param pPhone     - pointer to phone structure

   \return IFX_SUCCESS if OK, otherwise IFX_ERROR
*/
IFX_return_t CID_Send(IFX_uint32_t nDataCh_FD, PHONE_t* pPhone,
                      IFX_TAPI_CID_HOOK_MODE_t nHookState,
                      IFX_TAPI_CID_MSG_t *pCID_Msg)
{
   IFX_int32_t i;
#ifdef TAPI_VERSION4
   IFX_TAPI_RING_t ring;
#endif
#if (defined(EASY336) || defined(XT16))
   IFX_boolean_t bCID_AS;
#endif
   IFX_return_t ret;
   struct tm* date_time = IFX_NULL;
   time_t time_in_sec;

   /* check input parameters */
   TD_PTR_CHECK(pPhone, IFX_ERROR);
   TD_PTR_CHECK(pCID_Msg, IFX_ERROR);

   /* Stop ringing before any cid transmission. */
#ifdef TAPI_VERSION4
   if (pPhone->pBoard->fSingleFD)
   {
      ring.ch = pPhone->nPhoneCh;
      ring.dev = pPhone->nDev;
      TD_ASSERT ((ret = ioctl(nDataCh_FD,
         IFX_TAPI_RING_STOP, (IFX_int32_t) &ring)) == IFX_SUCCESS,
         nDataCh_FD, pPhone->pBoard, ring.dev, ring.ch);
   } /* if (pPhone->pBoard->fSingleFD) */
   else
#endif
   {
      ret = ioctl(nDataCh_FD, IFX_TAPI_RING_STOP, 0);
   }/* if (pPhone->pBoard->fSingleFD) */
   if (IFX_ERROR == ret)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH, ("CID: ring stop error!\n"));
      return IFX_ERROR;
   }
   /* for all message elements */
   for (i=0; i< pPhone->pCID_Msg->nMsgElements; i++)
   {
      /* check if sending time */
      if (IFX_TAPI_CID_ST_DATE ==
          pCID_Msg->message[i].date.elementType)
      {
         /* current calendar time in seconds */
         time_in_sec = time(IFX_NULL);
         /* converts calendar time into broke-down time */
         date_time = localtime(&time_in_sec);
         /* set current time */
         if (IFX_NULL != date_time)
         {
            /* Only 1 - 12 allowed */
            pCID_Msg->message[i].date.month = ((date_time->tm_mon + 1) % 13);
            /* Only 1 - 31 allowed */
            pCID_Msg->message[i].date.day = (date_time->tm_mday % 32);
            /* Only 0 - 23 allowed */
            pCID_Msg->message[i].date.hour = (date_time->tm_hour % 24);
            /* Only 0 - 59 allowed */
            pCID_Msg->message[i].date.mn = (date_time->tm_min % 60);
         } /* if (IFX_NULL != date_time) */
      } /* check if sending time */
   } /* for all message elements */
   /* Message type is call setup. */
   pCID_Msg->messageType = IFX_TAPI_CID_MT_CSUP;
   if (IFX_TAPI_CID_HM_OFFHOOK == nHookState)
   {
      /* Offhook cid */
      TRACE(TAPIDEMO, DBG_LEVEL_LOW, ("CID: offhook tx\n"));
      pCID_Msg->txMode = IFX_TAPI_CID_HM_OFFHOOK;
   }
   else if (IFX_TAPI_CID_HM_ONHOOK == nHookState)
   {
      /* Onhook cid */
      TRACE(TAPIDEMO, DBG_LEVEL_LOW, ("CID: onhook tx\n"));
      pCID_Msg->txMode = IFX_TAPI_CID_HM_ONHOOK;
   }
   else
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("CID: Using wrong nHookState parameter!(File: %s, line: %d)\n",
             __FILE__, __LINE__));
      return IFX_ERROR;
   }

#if (defined(EASY336) || defined(XT16))
   if ((pCID_Msg->txMode == IFX_TAPI_CID_HM_ONHOOK &&
      CID_GetStandard() == IFX_TAPI_CID_STD_NTT) ||
      (pCID_Msg->txMode == IFX_TAPI_CID_HM_ONHOOK &&
      CID_GetStandard() != IFX_TAPI_CID_STD_NTT))
      bCID_AS = IFX_TRUE;
   else
      bCID_AS = IFX_FALSE;

   ret = Common_CID_Enable(pPhone, bCID_AS);
   if (ret == IFX_ERROR)
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
        ("Common_CID_Enable failed. (File: %s, line: %d)\n",
         __FILE__, __LINE__));
      return IFX_ERROR;
   }
#endif /* EASY336 || XT16 */

#if 1
   /* Transmit the caller id */
   /* This interfaces transmits CID message with handling of the complete
      sequence (alert, ack and ringing). */
#ifdef TAPI_VERSION4
   if (pPhone->pBoard->fSingleFD)
   {
      pCID_Msg->dev = pPhone->nDev;
      pCID_Msg->ch = pPhone->nPhoneCh;
      TD_ASSERT ((ret = ioctl(nDataCh_FD,
         IFX_TAPI_CID_TX_SEQ_START, (IFX_int32_t) pCID_Msg)) ==
         IFX_SUCCESS, nDataCh_FD, pPhone->pBoard, pCID_Msg->dev,
         pCID_Msg->ch);
   } /* if (pPhone->pBoard->fSingleFD) */
   else
#endif /* TAPI_VERSION4 */
   {
      ret= ioctl(nDataCh_FD, IFX_TAPI_CID_TX_SEQ_START, pCID_Msg);
   }/* if (pPhone->pBoard->fSingleFD) */
   if (IFX_ERROR == ret)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH, ("CID: seq tx error!\n"));
      return IFX_ERROR;
   }
#else /* #if 1 */
   /* At the moment only SEQ START is used. */

   /* Otherwise, if only the FSK transmission is required */
   /* This interfaces transmits CID message. The message data and type
      information are given by IFX_TAPI_CID_MSG_t. Only the pure sending
      of FSK or DTMF data is done, no alert or ringing. */
   if (IFX_ERROR == ioctl(nDataCh_FD, IFX_TAPI_CID_TX_INFO_START,
        (IFX_int32_t) pCID_Msg))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH, ("CID: seq tx error!\n"));
      return IFX_ERROR;
   }
#endif /* #if 1 */

   return IFX_SUCCESS;
} /* CID_Send() */


/**
   Start CID RX FSK receiver.

   \param nFD - data channel file descriptor
   \param eHookMode - hook mode

   \return IFX_SUCCESS if OK, otherwise IFX_ERROR
*/
IFX_return_t CID_StartFSK(IFX_int32_t nFD, IFX_TAPI_CID_HOOK_MODE_t eHookMode)
{
   IFX_TAPI_CID_CFG_t cid_std;
   IFX_return_t ret = IFX_SUCCESS;

   /* check input parameters */
   TD_PARAMETER_CHECK((nFD <= 0), nFD, IFX_ERROR);

   memset(&cid_std, 0, sizeof(cid_std));
   cid_std.nStandard = eCurrCID_Std;

#ifdef TAPI_VERSION4
   ret = ioctl(nFD, IFX_TAPI_CID_CFG_SET, (IFX_int32_t) &cid_std);
#endif /* TAPI_VERSION4 */

   if (IFX_SUCCESS != ret)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Could not configure CID for data channel %d."
            " (File: %s, line: %d)\n",
            nFD, __FILE__, __LINE__));
      return IFX_ERROR;
   }

   /* Start CID RX Engine */
   return ioctl(nFD, IFX_TAPI_CID_RX_START, (IFX_int32_t) eHookMode);
} /* CID_StartFSK() */


/**
   Stop CID RX FSK receiver.

   \param nFD - data channel file descriptor

   \return IFX_SUCCESS if OK, otherwise IFX_ERROR
*/
IFX_return_t CID_StopFSK(IFX_int32_t nFD)
{
   /* check input parameters */
   TD_PARAMETER_CHECK((nFD <= 0), nFD, IFX_ERROR);

   /* Stop CID RX Engine */
   return ioctl(nFD, IFX_TAPI_CID_RX_STOP, (IFX_int32_t) 0);
} /* CID_StopFSK() */


/**
   Prepare Transparent CID message

   \param pCID_Msg - CID message to be prepared
   \param pData - received data which was readed with FSK
   \param nLen - length of received data in bytes
   \param eHookMode - hook mode

   \return IFX_SUCCESS Transparent CID message created.
*/
IFX_return_t CID_PrepareTransparentMsg(IFX_TAPI_CID_MSG_t* pCID_Msg,
                                       IFX_char_t* pData,
                                       IFX_int32_t nLen,
                                       IFX_TAPI_CID_HOOK_MODE_t eHookMode)
{
   /* check input parameters */
   TD_PTR_CHECK(pCID_Msg, IFX_ERROR);
   TD_PTR_CHECK(pData, IFX_ERROR);
   TD_PARAMETER_CHECK((nLen <= 0), nLen, IFX_ERROR);
   TD_PARAMETER_CHECK((nLen > IFX_TAPI_CID_TX_SIZE_MAX), nLen, IFX_ERROR);

   TRACE(TAPIDEMO, DBG_LEVEL_LOW, ("Prepare transparent CID msg\n"));

   pCID_Msg->message->transparent.elementType = IFX_TAPI_CID_ST_TRANSPARENT;
   pCID_Msg->message->transparent.len = nLen;
   if (pCID_Msg->message->transparent.data != IFX_NULL)
   {
      free(pCID_Msg->message->transparent.data);
   }

   pCID_Msg->message->transparent.data = calloc(nLen, sizeof(IFX_char_t));

   if(IFX_NULL == pCID_Msg->message->transparent.data)
   {
       /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Memory not allocated for CID message element. "
            "(File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }

   memcpy(pCID_Msg->message->transparent.data, pData, nLen);

   pCID_Msg->txMode = eHookMode;
   pCID_Msg->messageType = IFX_TAPI_CID_MT_CSUP;
   pCID_Msg->nMsgElements = 1;


   return IFX_SUCCESS;
}


/**
   Read received CID data.

   \param nFD - data channel file descriptor
   \param pCID_Msg - CID message
   \param nLength - length of retrieved data which will be read

   \return IFX_SUCCESS, data was read with CID RX FSK and transparent CID
           message was prepared, otherwise IFX_ERROR
*/
IFX_return_t CID_ReadData(IFX_int32_t nFD, IFX_TAPI_CID_MSG_t* pCID_Msg,
                          IFX_int32_t nLength)
{
   IFX_return_t ret = IFX_SUCCESS;
   IFX_TAPI_CID_RX_DATA_t oCID_RX_Data;
   IFX_int32_t i = 0;
   IFX_char_t* received_data = IFX_NULL;
   IFX_int32_t rcv_data_size = 0;
   IFX_TAPI_CID_RX_STATUS_t oCID_Status = {0};

   /* check input parameters */
   TD_PTR_CHECK(pCID_Msg, IFX_ERROR);
   TD_PARAMETER_CHECK((nFD <= 0), nFD, IFX_ERROR);

   memset (&oCID_Status, 0, sizeof (IFX_TAPI_CID_RX_STATUS_t));
   memset (&oCID_RX_Data, 0, sizeof (IFX_TAPI_CID_RX_DATA_t));

   ret = ioctl(nFD, IFX_TAPI_CID_RX_STATUS_GET, (IFX_int32_t) &oCID_Status);

   if ((ret != IFX_SUCCESS)
       || (oCID_Status.nError != IFX_TAPI_CID_RX_ERROR_NONE)/*
       || (oCID_Status.nStatus != IFX_TAPI_CID_RX_STATE_DATA_READY)*/)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Error on IFX_TAPI_CID_RX_STATUS_GET. (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }

   received_data = calloc(nLength, 1);
   rcv_data_size = 0;

   do
   {
      ret = ioctl(nFD, IFX_TAPI_CID_RX_DATA_GET, (IFX_int32_t) &oCID_RX_Data);

      if (ret != IFX_SUCCESS)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Error on IFX_TAPI_CID_RX_DATA_GET. (File: %s, line: %d)\n",
               __FILE__, __LINE__));
         break;
      }

      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("Have read length %d CID data:\n", oCID_RX_Data.nSize));

      if (oCID_RX_Data.nSize > 0 &&
          oCID_RX_Data.nSize < IFX_TAPI_CID_RX_SIZE_MAX)
      {
         memcpy(received_data + rcv_data_size, oCID_RX_Data.data,
                oCID_RX_Data.nSize);
         rcv_data_size += oCID_RX_Data.nSize;
         nLength -= oCID_RX_Data.nSize;
      }
      else
      {
          TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                ("Error on IFX_TAPI_CID_RX_DATA_GET. (File: %s, line: %d)\n",
                __FILE__, __LINE__));

         break;
      }
   }
   while ((oCID_RX_Data.nSize > 0) && (nLength > 0));


   if (ret != IFX_ERROR)
   {
      ret = CID_PrepareTransparentMsg(pCID_Msg, received_data,
                                      rcv_data_size, IFX_TAPI_CID_HM_ONHOOK);
   }

   if (ret != IFX_ERROR)
   {
      for (i = 0; i < rcv_data_size; i++)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("0x%02X ", (IFX_uint8_t) received_data[i]));

         if ((i > 0) && ((i + 1) % 15 == 0))
         {
            TRACE(TAPIDEMO, DBG_LEVEL_LOW, ("\n"));
         }
      }
      TRACE(TAPIDEMO, DBG_LEVEL_LOW, ("\n"));
   }

   /* Free allocated memory */
   free(received_data);

   return ret;
}

