#ifndef _CID_H
#define _CID_H
/******************************************************************************

                               Copyright (c) 2008
                            Infineon Technologies AG
                     Am Campeon 1-12; 81726 Munich, Germany

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

  ****************************************************************************
   Module      : cid.h
   Date        : 2005-11-30
   Description : This file enchance tapi demo with CID.

   \file cid.h

   \remarks

   \note Changes:
*******************************************************************************/

/* ============================= */
/* Includes                      */
/* ============================= */

/* ============================= */
/* Global Defines                */
/* ============================= */

/* ============================= */
/* Local Macros & Definitions    */
/* ============================= */

/* ============================= */
/* Global Structures and enums   */
/* ============================= */

enum
{
   /** Number of CID elements, we use CLI number, name, date */
   CID_ELEM_COUNT  = 3,
   /** Index of element location */
   CID_IDX_DATE    = 0,
   CID_IDX_CLI_NUM = 1,
   CID_IDX_NAME    = 2
};

enum 
{
   /** Maximum number of users */
#if (!defined(EASY336) && !defined(XT16))
   CID_MAX_USERS = 4,
#else /* (!defined(EASY336) && !defined(XT16)) */
   CID_MAX_USERS = 1000,
#endif /* (!defined(EASY336) && !defined(XT16)) */
   /** Maximum number of standards */
   CID_MAX_STANDARDS = 5
};

typedef enum _CID_DISPLAY_MODE_t
{
   /** Display all cid data (date, CLI number, name). */
   CID_DISPLAY_ALL     = 0,
   /** Display sending cid data (CLI number, name). */
   CID_DISPALY_SEND    = 1,
   /** Display received cid data (CLI number, name). */
   CID_DISPALY_RECEIVE = 2,
   /** Print only CLI number and name */
   CID_DISPLAY_SIMPLE  = 3
} CID_DISPLAY_MODE_t;

/** Structure holding standard number and standard description */
typedef struct _CID_STANDARD_t
{
   IFX_int32_t  nNumber;
   IFX_char_t  *sDescription;
} CID_STANDARD_t;

/* ============================= */
/* Global Variable declaration   */
/* ============================= */

/** CID standards with short description */
extern const CID_STANDARD_t CID_STANDARDS_SHORT[CID_MAX_STANDARDS];

/* ============================= */
/* Global Variable definition    */
/* ============================= */

/* ============================= */
/* Global function declaration   */
/* ============================= */

IFX_return_t CID_SetupData(IFX_uint32_t nIdx, PHONE_t* pPhone);
IFX_void_t CID_Display(PHONE_t* pPhone, CID_DISPLAY_MODE_t nMode);
IFX_return_t CID_SetStandard(IFX_int32_t nCID_Standard);
IFX_TAPI_CID_STD_t CID_GetStandard(IFX_void_t);
IFX_return_t CID_ConfDriver(PHONE_t* pPhone);
IFX_return_t CID_Send(IFX_uint32_t nDataCh_FD, PHONE_t* pPhone,
                      IFX_TAPI_CID_HOOK_MODE_t nHookState,
                      IFX_TAPI_CID_MSG_t *pCID_Msg);
IFX_return_t CID_Release(IFX_TAPI_CID_MSG_t *pCID_Msg);
IFX_return_t CID_StartFSK(IFX_int32_t nFD, IFX_TAPI_CID_HOOK_MODE_t eHookMode);
IFX_return_t CID_StopFSK(IFX_int32_t nFD);
IFX_return_t CID_ReadData(IFX_int32_t nFD, IFX_TAPI_CID_MSG_t* pCID_Msg,
                          IFX_int32_t nLength);


#endif /* _CID_H */

