

#ifndef _COMMON_H
#define _COMMON_H
/******************************************************************************

                               Copyright (c) 2008
                            Infineon Technologies AG
                     Am Campeon 1-12; 81726 Munich, Germany

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

  ****************************************************************************
   Module      : common.h
   Description :
*******************************************************************************/
/**
   \file common.c
   \date 2005-10-01
   \brief Common for tapidemo enums, structure, macros and defines.
*/
/* ============================= */
/* Includes                      */
/* ============================= */

/** system includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef WIN32
#include <sys/ioctl.h>
#endif
#ifdef LINUX
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
/**
  * Added network support, TCP/IP, UDP, timers, sockets, ...
  */
/* #include <linux/errno.h> */
#include <errno.h>
#include <netdb.h>              /* h_errno */
#include <sys/socket.h>         /* socket, gethostbyaddr */
/* linux system includes */
#include <sys/ioctl.h>
#include <sys/types.h>
#include <getopt.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <time.h>
#include <math.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/poll.h>
/* tolower(), ... functions */
#include <ctype.h>
#include <net/if.h> /* Low level access to linux network devices */

#ifdef EASY336
#include <linux/ifx_svip_nat_io.h>
#endif

#include <syslog.h>

#endif /* LINUX */

#ifdef VXWORKS
   #include <vxWorks.h>
   #include <sockLib.h>
   #include <inetLib.h>
   #include <ioLib.h>
   #include <sys/ioctl.h>
   #include <sysLib.h>
   #include <tickLib.h>
   #include <taskLib.h>

   #include <hostLib.h>

   #include <time.h>
   #include <taskHookLib.h>

   #include <sigLib.h>
   #include <inetLib.h>
#ifdef EASY336
   #include <natProtocol.h>
#endif

#endif /* VXWORKS */


/** Drivers interfaces needed headers */
#include "ifx_types.h"
#ifdef LINUX
#include "tapidemo_config.h"
#endif

#ifdef HAVE_DRV_TAPI_HEADERS
   #include "drv_tapi_io.h"
#else
    #error "drv_tapi is missing"
#endif

#ifdef TAPI_VERSION4
#include "sys_debug.h"
#endif

#ifdef QOS_SUPPORT
/* QOS_INIT_SESSION defined in drv_qos.h */
    #ifdef HAVE_DRV_TAPI_HEADERS
       #include "drv_tapi_qos_io.h"
    #else
        #error "drv_tapi is missing"
    #endif
#endif /* QOS_SUPPORT */

#ifdef HAVE_DRV_TAPI_HEADERS
   #include "drv_tapi_event_io.h"
#else
    #error "drv_tapi is missing"
#endif

#if (defined(EASY336) || defined(XT16))
#include "lib_svip_rm.h"
#endif

extern size_t strnlen (__const char *__string, size_t __maxlen);

/* ============================= */
/* Debug interface               */
/* ============================= */

/** Trace redirection setting. */
typedef enum _TRACE_REDIRECTION_t
{
   /** No redirection */
   TRACE_REDIRECTION_NONE = 0,
   /** Redirection to file */
   TRACE_REDIRECTION_FILE = 1,
   /** Redirection to local syslog */
   TRACE_REDIRECTION_SYSLOG_LOCAL = 2,
   /** Redirection to remote syslog */
   TRACE_REDIRECTION_SYSLOG_REMOTE = 3
} TRACE_REDIRECTION_t;

#ifdef ABACUS_TEST
#define TAPIDEMO_MOD_FXO(message) \
        do \
        { \
           if ((pCtrlStatus->pProgramArg->nTestType == COM_MODULAR_TEST) && \
               (pCtrlStatus->pProgramArg->nSubTestType & COM_SUBTEST_FXO)) \
           { \
              TAPIDEMO_PRINTF(message); \
           } \
        } while(0);
#endif /* ABACUS_TEST */

#ifdef ABACUS_TEST
#define TAPIDEMO_PRINTF(message) \
      do \
      { \
          printf message; \
          if(COM_MODULAR_TEST == *nTestType) \
          { \
             if(NO_SOCKET != nComOutSocket)\
             { \
                  IFX_char_t tmp[512];\
                  COM_GENERATE_STRING message;\
                  COM_SendTraceMessage(DBG_LEVEL_OFF, tmp);\
             } \
          } \
      } while(0)
#else
#define TAPIDEMO_PRINTF(message) \
      do \
      { \
         printf message; \
      } while(0)
#endif /* ABACUS_TEST */

/* Tapi version 4 takes this enum from sys_debug.h */
#ifndef TAPI_VERSION4
/** Trace levels */
typedef enum _DEBUG_LEVEL_t
{
   /** None reporting */
   DBG_LEVEL_OFF = 4,
   /** Error reporting */
   DBG_LEVEL_HIGH = 3,
   /** Status reporting */
   DBG_LEVEL_NORMAL = 2,
   /** More info reporting */
   DBG_LEVEL_LOW = 1
} DEBUG_LEVEL_t;
#endif /* TAPI_VERSION4 */

/**
   Stores the trace string in temporal variable.

   \param ...     - variable number of arguments

   \return

   \remark
      Makro is used to remove parenthesis from TRACE macro's message argument
      Used only for sending Trace messages through communication socket.
*/
#ifdef VXWORKS
#define COM_GENERATE_STRING(ARGS...) sprintf(tmp, ## ARGS)
#else /* VXWORKS */
#define COM_GENERATE_STRING(...) sprintf(tmp,__VA_ARGS__)
#endif /* VXWORKS */

#ifdef ABACUS_TEST
#ifdef TAPI_VERSION4
   /* when making abacus tests we are using our own TRACE */
   #undef TRACE
#endif /* TAPI_VERSION4 */
/**
   Prints a trace message and sends it through communication socket.

   \param name     - Name of the trace group
   \param level    - level of this message
   \param message  - a printf compatible formated string + opt. arguments

   \return

   \remark
      The string will be redirected via printf if level is higher or equal to
      the actual level for this trace group ( please see SetTraceLevel ).
*/
#define TRACE(name,level,message) \
      do \
      { \
         if(level >= G_nTraceGroup##name) \
         { \
            printf message; \
         } \
         if((level >= *nComDbgLevel) \
            && (COM_MODULAR_TEST == *nTestType)) \
         { \
            if(NO_SOCKET != nComOutSocket)\
            { \
               IFX_char_t tmp[512];\
               COM_GENERATE_STRING message;\
               COM_SendTraceMessage(level, tmp);\
            } \
         } \
      } while(0)

/**
   Prints a trace message without sending it through communication socket.

   \param name     - Name of the trace group
   \param level    - level of this message
   \param message  - a printf compatible formated string + opt. arguments

   \return

   \remark
      The string will be redirected via printf if level is higher or equal to
      the actual level for this trace group ( please see SetTraceLevel ).
*/
#define TRACE_LOCAL(name,level,message) \
      do \
      { \
          if(level >= G_nTraceGroup##name) \
            { \
               printf message; \
            } \
      } while(0)
#else /* ABACUS_TEST */
#ifndef TAPI_VERSION4
/**
   Prints a trace message.

   \param name     - Name of the trace group
   \param level    - level of this message
   \param message  - a printf compatible formated string + opt. arguments

   \return

   \remark
      The string will be redirected via printf if level is higher or equal to
      the actual level for this trace group ( please see SetTraceLevel ).
*/
#define TRACE(name,level,message) \
      do \
      { \
         if(level >= G_nTraceGroup##name) \
         { \
            printf message; \
         } \
      } while(0)
#endif /* TAPI_VERSION4 */
/**
   Set the actual level of a trace group.

   \param name - Name of the trace group
   \param new_level - new trace level
*/
#endif /* ABACUS_TEST */
#ifndef TAPI_VERSION4
#define SetTraceLevel(name, new_level) {G_nTraceGroup##name = new_level;}

/**
   Create a log group.

   \param name - name of the log group

   \return

   \remark
      This has to be done once in the project. We do it in the file
      tapidemo.c. The default level of this log group is DBG_LEVEL_HIGH.
*/
#define CREATE_TRACE_GROUP(name) \
        IFX_uint32_t G_nTraceGroup##name = DBG_LEVEL_HIGH

/**
   Make a log group extern so other modules, files can use it also.

   \param name - name of the log group

   \return

   \remark
      This has to be done once in the project. We do it in the file
      tapidemo.c. The default level of this log group is DBG_LEVEL_HIGH.
*/
#define DECLARE_TRACE_GROUP(name) \
        extern IFX_uint32_t G_nTraceGroup##name

#endif /* TAPI_VERSION4 */

DECLARE_TRACE_GROUP(TAPIDEMO);


/* ============================= */
/* Debug interface               */
/* ============================= */

/* macros for asserts */
#ifdef LINUX
#define TD_ASSERT(expr, fd, pBoard, dev, ch) if(!(expr)) {                    \
      printf(#expr" failed in file \"%s\" on line %d, %s dev %d, "            \
      "ch %d\n",                                                              \
      __FILE__, __LINE__, pBoard->pszBoardName, dev, ch);                     \
      Common_ErrorHandle (fd, pBoard, dev);                                   \
   }
#if 0
   else                                                                       \
   {                                                                          \
     TRACE (TAPIDEMO, DBG_LEVEL_LOW, (#expr", %s dev %d, ch %d\n",            \
     pBoard->pszBoardName, dev, ch));                                         \
   }
#endif
#endif
#ifdef VXWORKS
#define TD_ASSERT(expr, fd, pBoard, dev, ch) if(!(expr)) {                    \
      IFX_char_t* sFileName = __FILE__;                                       \
      Common_StripPath(&sFileName);                                           \
      printf(#expr" failed in file \"%s\" on line %d, %s dev %d, "            \
      "ch %d\n",                                                              \
      sFileName, __LINE__, pBoard->pszBoardName, dev, ch);                    \
      Common_ErrorHandle (fd, pBoard, dev);                                   \
   }
#if 0
   else                                                                       \
   {                                                                          \
     TRACE (TAPIDEMO, DBG_LEVEL_LOW, (#expr", %s dev %d, ch %d\n",            \
     pBoard->pszBoardName, dev, ch));                                         \
   }
#endif
#endif

/** Check if pointer is NULL.
       expr  - condition to check
       param - parameter to print in case of expr being false
       ret   - value to retunr in case of error. */
#define TD_PTR_CHECK(ptr, ret) if(ptr == IFX_NULL) {                           \
        TRACE(TAPIDEMO, DBG_LEVEL_HIGH,                                        \
              ("Error, Invalid pointer ("#ptr"). (File: %s, line: %d)\n",      \
               __FILE__, __LINE__));                                           \
        return ret;                                                            \
        }

/** Check condition and print value of parameters that will help in debug.
       expr  - condition to check, if true then return ret
       param - parameter to print in case of expr being false
       ret   - value to retunr in case of error. */
#define TD_PARAMETER_CHECK(expr, param, ret) if(expr) {                        \
        TRACE(TAPIDEMO, DBG_LEVEL_HIGH,                                        \
              ("Error, Invalid parameter ("#param"=%d). "                      \
              "(File: %s, line: %d)\n",                                        \
               param, __FILE__, __LINE__));                                    \
        return ret;                                                            \
        }

/* Check if driver's function returned error. In case of error, last error 
   from drivers is read. 
   fct - driver's function 
   ret - could be TD_RETURN or TD_NO_RETURN. If it is set to TD_RETURN 
         then return is called in case of error */
#define TD_DRV_FUNC_ASSERT(expr, pPhone, fct, dev, ch, ret)                    \
        if (expr != IFX_SUCCESS)                                               \
        {                                                                      \
           TRACE(TAPIDEMO, DBG_LEVEL_HIGH,                                     \
                ("Error, Phone No %d: "#fct" failed. "                         \
                 "(%s, dev %d, ch %d) "                                        \
                 "(File: %s, line: %d)\n",                                     \
                 pPhone->nPhoneNumber, pPhone->pBoard->pszBoardName,           \
                 dev, ch, __FILE__, __LINE__));                                \
           Common_ErrorHandle (pPhone->pBoard->nDevCtrl_FD, pPhone->pBoard,    \
              dev);                                                            \
           if (ret)                                                            \
           {                                                                   \
              return IFX_ERROR;                                                \
           }                                                                   \
        }

/**
   Call ioctl with error handling

   \param nRet - ioctl return value
   \param nFd  - file descriptor on which ioctl will be used on
   \param nFct - ioctl number
   \param pArg - pointer to structure (ioctl parameter)
   \param nPhoneNo - phone number (NO_PHONE_NUMBER if not applicable)
   \param pBoard   - pointer to board
   \param ch   - channel number
   \param dev  - device number (only tapi v4, in v3 use 0)
   \param bReturn  - if set macro should end function
 
   */
#define TD_IOCTL(nRet, nFd, nFct, pArg, nPhoneNo, pBoard, ch, dev, bReturn)  \
        do \
        { \
           nRet = ioctl(nFd, nFct, pArg); \
           if (IFX_SUCCESS > nRet) \
           { \
              TRACE(TAPIDEMO, DBG_LEVEL_HIGH, \
                    ("Error, ioctl(%d, "#nFct", %d) failed, ch %d, dev %d\n", \
                     nFd, (IFX_int32_t) pArg, ch, dev)); \
              if (NO_PHONE_NUMBER != nPhoneNo && IFX_NULL != pBoard) \
              { \
                 TRACE(TAPIDEMO, DBG_LEVEL_HIGH, \
                       ("Error, Phone No %d, %s,\n", \
                        nPhoneNo, pBoard->pszBoardName)); \
              } \
              TRACE(TAPIDEMO, DBG_LEVEL_HIGH, \
                    ("Error, File: %s, line: %d.\n", \
                     __FILE__, __LINE__)); \
              if (IFX_NULL != pBoard) \
              { \
                 Common_ErrorHandle(pBoard->nDevCtrl_FD, pBoard, dev); \
              } \
              if (TD_RETURN == bReturn) \
              { \
                 return IFX_ERROR; \
              } \
           } \
        } while (0);

/* Check if function returned error. 
   fct - function 
   ret - could be TD_RETURN or TD_NO_RETURN. If it is set to TD_RETURN 
         then return is called in case of error */
#define TD_FUNC_ASSERT(expr, pPhone, fct, dev, ch, ret)                        \
        if (expr != IFX_SUCCESS)                                               \
        {                                                                      \
           TRACE(TAPIDEMO, DBG_LEVEL_HIGH,                                     \
                ("Error, Phone No %d: "#fct" failed. "                         \
                 "(%s, dev %d, ch %d) "                                        \
                 "(File: %s, line: %d)\n",                                     \
                 pPhone->nPhoneNumber, pPhone->pBoard->pszBoardName,           \
                 dev, ch, __FILE__, __LINE__));                                \
           if (ret)                                                            \
           {                                                                   \
              return IFX_ERROR;                                                \
           }                                                                   \
        }

/* Play call progres tone, check if successfull and print apropriate traces */
#define TD_CPTG_PLAY_CHECK(ret, pCtrl, pPhone, nTone, pszToneName) \
   ret = Common_TonePlay(pCtrl, pPhone, nTone); \
   if (IFX_SUCCESS == ret) \
   { \
      TRACE(TAPIDEMO, DBG_LEVEL_NORMAL, \
            ("Phone No %d: Play %s.\n", pPhone->nPhoneNumber, pszToneName)); \
      SUBTEST_CONFERENCE_BUSY_TRACE(nTone); \
   } \
   else \
   { \
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH, \
            ("Phone No %d: Failed to play %s. (File: %s, line: %d)\n", \
             pPhone->nPhoneNumber, pszToneName, __FILE__, __LINE__)); \
   }

/** Add fd and set new max fd if needed */
#define TD_FD_SET(nFd, pFds, nMaxFd) \
   FD_SET(nFd, pFds); \
   if (nMaxFd < nFd) \
   { \
      nMaxFd = nFd; \
   }

/** Reset timout flag */
#define TD_TIMEOUT_RESET(pCtrl)  { pCtrl->nUseTimeout = 0; }
/** Increase timout flag - another resource(phones or FXO) needes
    to use timeout */
#define TD_TIMEOUT_START(pCtrl)  { pCtrl->nUseTimeout++; }
/** Decrease timout flag - resource(phones or FXO) doesn't need timeout now */
#define TD_TIMEOUT_STOP(pCtrl)   if (pCtrl->nUseTimeout > 0)   \
                          {   pCtrl->nUseTimeout--;  }
/** Check if timeout should be used */
#define TD_TIMEOUT_CHECK(pCtrl)  (0 != pCtrl->nUseTimeout)

#ifdef VXWORKS
/** Get current time */
#define TD_TIMEOUT_GET_TIME(pTime) \
             clock_gettime(CLOCK_REALTIME, (struct timespec *) pTime)
#else /* VXWORKS */
/** Get current time */
#define TD_TIMEOUT_GET_TIME(pTime) \
             gettimeofday(pTime, IFX_NULL)
#endif

/* ============================= */
/* Global Defines                */
/* ============================= */

/* those two enums are different only by name
   so to simplyfy code define is used */
#ifdef TAPI_VERSION4
#define  TD_LINE_MODE_t    IFX_TAPI_LINE_MODE_t
#else /* TAPI_VERSION4 */
#define  TD_LINE_MODE_t    IFX_TAPI_LINE_FEED_t
#endif /* TAPI_VERSION4 */

typedef struct _PHONE_t PHONE_t;
typedef struct _FXO_t FXO_t;
typedef struct _BOARD_t BOARD_t;
typedef struct _CTRL_STATUS_t CTRL_STATUS_t;
typedef struct _PROGRAM_ARG_t PROGRAM_ARG_t;
typedef struct _TAPIDEMO_DEVICE_CPU_t TAPIDEMO_DEVICE_CPU_t;

/** UDP port for voice data (VOICE_UDP_PORT .. (VOICE_UDP_PORT + 
 * 2*MAX_SYS_CH_RES))
    According to the standard it is recomended to send RTP packets on even
    ports while odd ones are reserved for RTCP */
#ifdef EASY336
enum { VOICE_UDP_PORT = SVIP_UDP_FROM };
#else /* EASY336 */
enum { VOICE_UDP_PORT = 5002 };
#endif /* EASY336 */

/** UDP port for making/ending voip/pcm/... calls. */
enum { ADMIN_UDP_PORT = 5000 };

#ifdef USE_FILESYSTEM
enum { MAX_FULL_FILE_LEN = 256 };
#endif /* USE_FILESYSTEM */ 
/** Phone digit representation */
typedef enum
{
   DIGIT_ONE = 0x01,
   DIGIT_TWO = 0x02,
   DIGIT_THREE = 0x03,
   DIGIT_FOUR = 0x04,
   DIGIT_FIVE = 0x05,
   DIGIT_SIX = 0x06,
   DIGIT_SEVEN = 0x07,
   DIGIT_EIGHT = 0x08,
   /** Used to start PCM calls */
   DIGIT_NINE = 0x09,
   DIGIT_STAR = 0x0A,
   DIGIT_ZERO = 0x0B,
   /** Used to start conference, calling another party */
   DIGIT_HASH = 0x0C
} DIGIT_t;

/** Sign for no parameters for ioctl() call */
enum { NO_PARAM = 0 };

/** Flag for no device FD */
enum { NO_DEVICE_FD = -1 };

/** Flag for no socket */
enum { NO_SOCKET = -1 };

/** Flag for no file */
enum { NO_FILE = -1 };

/** Flag for no channel file descriptor */
enum { NO_CHANNEL_FD = -1 };

/** Flag for no channel */
enum { NO_CHANNEL = -1 };

/** Flag for no used channel */
enum { NO_USED_CHANNEL = -1 };

/** Flag for no board idx */
enum { NO_BOARD_IDX = -1 };

/** Flag for no phone number */
enum { NO_PHONE_NUMBER = -1 };

/** Flag for PCM call, will be send to external phone */
enum { CALL_FLAG_PCM = 0x80 };

/** Flag for VoIP call, will be send to external phone */
enum { CALL_FLAG_VOIP = 0x00 };

/** Flag for VoIP call, will be send to external phone */
enum { CALL_FLAG_UNKNOWN = 0xFF };

/** Flag indicating start of command message */
enum { COMM_MSG_START_FLAG = 0xFF };

/** Flag indicating end of command message */
enum { COMM_MSG_END_FLAG = 0x77 };

/** Max. phone number or ... than can be dialed */
enum { MAX_DIALED_NUM = 24 };

/** Phone number lenght for TAPI3, e.g. Danube board */
enum { TAPI3_PHONE_NUM_LEN = 1 };

 /** Phone number lenght for TAPI4, e.g. SVIP board */
enum { TAPI4_PHONE_NUM_LEN = 3 };

/** Min. value of phone number */
#ifdef TAPI_VERSION4
   enum { MIN_PHONE_NUM = 100 };
#else
   enum { MIN_PHONE_NUM = 1 };
#endif

/** Max. length of first 3 number of IP address for different network */
enum { DIFF_NET_IP_LEN = 12 };

/** First part of digit sequence used to define call type FXO {05<...>}. */
enum { PREFIX_FXO_CALL_DIGITS_CNT = 2 };

/** Use restor default option. */
enum { RESTORE_DEFAULT = -1 };

/** Length for 32 bit IP address in characters "xxx.xxx.xxx.xxx". */
enum { IP_ADDRESS_LEN = 16 };

/** Length for port number "nnnnn, 0..65535". */
enum { PORT_NUM_LEN = 5 };

/** call progress tones names */
typedef enum _CPT_NAME_t
{
   NO_TONE = -1,
   DIAL_TONE,
   SECOND_DIAL_TONE,
   BUSY_TONE,
   RINGBACK_TONE,
   CONGESTION_TONE,
   HOWLER_TONE,
   STOP_TONE,
   MAX_TONE
}CPT_NAME_t;

/** Call progress tones indexes for more check drv_tapi_tone.c
    or Predefined Tones Table in tapi documentation. */
typedef enum _TONE_IDX_t
{
   DIALTONE_IDX = 25,
   RINGBACK_TONE_IDX = 26,
   BUSY_TONE_IDX = 27
} TONE_IDX_t;

#ifdef EASY336
/** resource type */
typedef enum _RESOURCE_TYPE_t
{
   RES_NONE = 0x00,
   RES_VOIP_ID = 0x01,
   RES_CONN_ID = 0x02,
   RES_DET = 0x04,
   RES_GEN = 0x08,
   RES_LEC = 0x10,
   RES_CODEC = 0x20,
   /** all resources voip_id, conn_id, sig_det, sig_gen, lec and codec */
   RES_ALL = 0xFF
} RESOURCE_TYPE_t;
#endif /* EASY336 */

/** number of state machines */
enum { MAX_STATE_MACHINE = 2 };

#ifndef ETH_ALEN
/** Octets in one ethernet address */
#define ETH_ALEN 6
#endif

/** mask to get state info part */
#define TD_STATE_MASK            0xFFFF0000
/** if this flag is set then during this state packets are transmited between
    phones - used by tapidemo packet handling functions */
#define TD_STATE_DATA_TRANSMIT   0x00010000
/** if this flag is set then this state is used for fax, modem or t.38 */
#define TD_STATE_FAX_MODEM       0x00020000
/** if this flag is set then this state is used for conferences */
#define TD_STATE_CONFERENCE      0x00040000
/** if this flag is set then during this state
    tone is played on phone channel */
#define TD_STATE_TONE_PLAY       0x00080000

#ifdef ABACUS_TEST
/** if this flag is set then this state is used for BACUS_TEST */
#define TD_STATE_ABACUS          0x01000000
#endif /* ABACUS_TEST */

/** returns non zero value if nInfo flag is set
      \param nState - phone state
      \param nInfo  - state flag to check
 */
#define TD_GET_STATE_INFO(nState, nInfo) (((nState) & TD_STATE_MASK) & (nInfo))

/** States of state machine. Enums represent names of states.
    TD_STATE_... flags are set to notify about state aditional features,
    e.g. S_OUT_DIALTONE = 0x0001 | TD_STATE_TONE_PLAY - during this state tone
    is played. This can help to turn off tone when state is changed e.g.
    TD_GET_STATE_INFO(nState, TD_STATE_TONE_PLAY) will return non zero value
    if tone is played during this state. */
typedef enum
{
   /** Phone is on-hook. */
   S_READY           = 0x0000,
   /** Phone is off-hook, and tone is played(dialtone),
       waiting for dialed digits. */
   S_OUT_DIALTONE    = 0x0001 | TD_STATE_TONE_PLAY,
   /** Phone is off-hook, one or more digits were dialed,
       waiting for more digits to complete number that can be called. */
   S_OUT_DIALING     = 0x0002,
   /** FXO call in progress. */
   S_OUT_FXO_CALL    = 0x0003,
   /** Phone number was dialed waiting for peer confirmation. */
   S_OUT_CALLING     = 0x0004,
   /** Waiting for peer to make hook-off, tone is played(ringback). */
   S_OUT_RINGBACK    = 0x0005 | TD_STATE_TONE_PLAY,
   /** Connection in progress, data should be transmitted. */
   S_ACTIVE          = 0x0006 | TD_STATE_DATA_TRANSMIT,
   /** Connection failed or peer disconected, tone is played(busy tone). */
   S_BUSYTONE        = 0x0007 | TD_STATE_TONE_PLAY,
   /** Phone is ringing, waiting for hook-off to establish connection. */
   S_IN_RINGING      = 0x0008,
   /** Phone is waiting for dialed digits to establish new connection,
       phone is in conference, tone is played(dialtone). */
   S_CF_DIALTONE     = 0x0009 | TD_STATE_CONFERENCE | TD_STATE_TONE_PLAY,
   /** One or more digits were dialed, waiting for more digits to complete
       number that can be called, phone is in conference. */
   S_CF_DIALING      = 0x000A | TD_STATE_CONFERENCE,
   /** Phone number was dialed waiting for peer confirmation, phone is in
       conference. */
   S_CF_CALLING      = 0x000B | TD_STATE_CONFERENCE,
   /** Waiting for peer to make hook-off, phone is in conference,
      tone is played(ringback). */
   S_CF_RINGBACK     = 0x000C | TD_STATE_CONFERENCE | TD_STATE_TONE_PLAY,
   /** Conference connections in progress, data should be transmitted,
       phone is in conference. */
   S_CF_ACTIVE       = 0x000D | TD_STATE_DATA_TRANSMIT | TD_STATE_CONFERENCE,
   /** Connection failed , phone is in conference, tone is played(busy tone). */
   S_CF_BUSYTONE     = 0x000E | TD_STATE_CONFERENCE | TD_STATE_TONE_PLAY,
#ifdef ABACUS_TEST
   /** Tone playing is tested, ABACUS Test specific state. */
   S_TEST_TONE       = 0x000F | TD_STATE_ABACUS,
#endif /* ABACUS_TEST */
   /** fax/modem transmission in progress, data should be transmitted,
       fax/modem specific state. */
   S_FAX_MODEM       = 0x0010 | TD_STATE_DATA_TRANSMIT | TD_STATE_FAX_MODEM,
   /** t.38 transmission in progress, data should be transmitted,
       fax/modem specific state. */
   S_FAXT38          = 0x0011 | TD_STATE_DATA_TRANSMIT | TD_STATE_FAX_MODEM,
   /** t.38 transmission request recived, fax/modem specific state. */
   S_FAXT38_REQ      = 0x0012 | TD_STATE_FAX_MODEM,
   /** t.38 transmission end recived, fax/modem specific state. */
   S_FAXT38_END      = 0x0013 | TD_STATE_FAX_MODEM,
   /** Last state - used to mark end of state tables */
   S_END             = 0xFFFF
} STATE_MACHINE_STATES_t;

/** Internal events */
typedef enum
{
   IE_NONE = 0,
   IE_READY,
   IE_HOOKOFF,
   IE_HOOKON,
   IE_FLASH_HOOK,
   IE_DIALING,
   IE_RINGING,
   IE_RINGBACK,
   IE_ACTIVE,
   IE_CONFERENCE,
   IE_END_CONNECTION,
   IE_BUSY,
   IE_CNG_FAX,
   IE_CNG_MOD,
   IE_AM,
   IE_PR,
   IE_CED,
   IE_CED_END,
   IE_DIS,
   IE_V8BIS,
   IE_V21L,
   IE_V18A,
   IE_V27,
   IE_BELL,
   IE_V22,
   IE_V22ORBELL,
   IE_V32AC,
   IE_CAS_BELL,
   IE_V21H,
   IE_END_DATA_TRANSMISSION,
   IE_CONFIG,
   IE_CONFIG_DIALING,
   IE_CONFIG_HOOKON,
   IE_CONFIG_PEER,
   IE_EXT_CALL_SETUP,
   IE_EXT_CALL_PROCEEDING,
   IE_EXT_CALL_ESTABLISHED,
   IE_EXT_CALL_WRONG_NUM,
   IE_EXT_CALL_NO_ANSWER,
   IE_EXT_CALL_WORKAROUND_EASY508xx,
   IE_T38_ERROR,
   IE_T38_REQ,
   IE_T38_ACK,
   IE_T38_END,
#ifdef ABACUS_TEST
   IE_TEST_TONE,
#endif /* ABACUS_TEST */
   /** this event is never generated,
      it is only used to mark end of EVENT_ACTIONS_t table */
   IE_END
} INTERNAL_EVENTS_t;

/** structure with event and pointer to action that should be taken
    when this event coccures */
typedef struct _EVENT_ACTIONS_t
{
   INTERNAL_EVENTS_t nEvent;
   IFX_return_t (*pAction)();
} EVENT_ACTIONS_t;

/** structure with state and pointer to table of EVENT_ACTIONS_t
    with all handled events */
typedef struct _STATE_COMBINATION_t
{
   STATE_MACHINE_STATES_t nState;
   EVENT_ACTIONS_t* pEventActions;
} STATE_COMBINATION_t;

/** structure with current state
    and pointer to table of STATE_COMBINATION_t for all states */
typedef struct _STATE_MACHINES_t
{
   STATE_MACHINE_STATES_t nState;
   STATE_COMBINATION_t* pStateCombination;
} STATE_MACHINES_t;

/** types of state machines */
typedef enum _STATE_MACHINE_TYPE_t
{
   /** configuration state machine */
   CONFIG_SM = 0,
   /** FXS state machine */
   FXS_SM = 1
} STATE_MACHINE_TYPE_t;

/** call direction */
typedef enum _CALL_DIRECTION_TYPE_t
{
   /** no call direction - no call */
   CALL_DIRECTION_NONE = 0,
   /** transmit direction - set for caller phone */
   CALL_DIRECTION_TX,
   /** receive direction - set for called phone */
   CALL_DIRECTION_RX
} CALL_DIRECTION_TYPE_t;


/** max enum ID */
#define TD_MAX_ENUM_ID  0xffffffff
/** max length of name array */
#define TD_MAX_NAME     64

/** Max path length to download files. */
enum { MAX_PATH_LEN = 256 };

/** structure to assigne enum name to its value (id) */
typedef struct _TD_ENUM_2_NAME_t
{
   IFX_int32_t nID;
   IFX_char_t* rgName;
} TD_ENUM_2_NAME_t;

/** error code for Common_Enum2Name() function. This enums are used as 
    index for TD_rgEnum2NameErrorCode[] table. */
typedef enum _TD_ENUM_2_NAME_ERR_CODE_t 
{
   /** enum value not found in the table */
   ENUM_VALUE_NOT_FOUND = 0,
   /** NULL pointer */
   ENUM_NULL_POINTER
} TD_ENUM_2_NAME_ERR_CODE_t;

/** enum used by macros to decide if call return. */
typedef enum _TD_RETURN_CODE_t
{
   TD_NO_RETURN = 0,
   TD_RETURN = 1
} TD_RETURN_CODE_t;

/** number of star digits that must be used to dial phone number
    with full IP address */
#define NUM_OF_STARS_IN_IP 5

#if (!defined(EASY336) && !defined(XT16))
/** number of digits of phone number x */
#define TD_PHONE_NUM_LEN      1
#else /* (!defined(EASY336) && !defined(XT16)) */
/** number of digits of phone number xxx */
#define TD_PHONE_NUM_LEN      3
#endif /* (!defined(EASY336) && !defined(XT16)) */

#define TD_PHONE_STRING       "Phone No : " 

/** Default path to download files. */
#ifdef LINUX
#define DEF_DOWNLOAD_PATH        "/opt/ifx/downloads/"
#endif /* LINUX */
#ifdef VXWORKS
#define DEF_DOWNLOAD_PATH        "/romfs/"
#endif /* VXWORKS */

/** name of file with path to downloads (FW and BBD files) */
#define DOWNLOAD_FILE_NAME       ".tapidemorc"
/** name of variable with path to downloads in DOWNLOAD_FILE_NAME */
#define DOWNLOAD_VARIABLE       "DOWNLOADS"


/** max device name length with number e.g. /dev/vmmc10 */
#define MAX_DEV_NAME_LEN_WITH_NUMBER      50

/* --------------------------------------------------------- */
/*                    CONFERENCE START                       */
/* --------------------------------------------------------- */

/** representing local or remote call */
typedef enum _CALL_TYPE_t
{
   UNKNOWN_CALL_TYPE = 0,
   EXTERN_VOIP_CALL = 1,
   LOCAL_CALL = 2,
   PCM_CALL = 3,
   LOCAL_BOARD_CALL = 5,
   FXO_CALL = 6,
   LOCAL_PCM_CALL = 7,
   LOCAL_BOARD_PCM_CALL = 8,
   DTMF_COMMAND = 9
} CALL_TYPE_t;

/** Conference status */
/** Also used for flag if not in conference */
enum { NO_CONFERENCE = 0 };

#ifndef EASY336
/** Max peers in conference */
enum { MAX_PEERS_IN_CONF = 4 };
#else /* EASY336 */
/** Max peers in conference */
enum { MAX_PEERS_IN_CONF = 10 };
#endif /* EASY336 */

/** How many conferences can we have. Cannot be bigger than number of phones. */
#ifndef EASY336
enum { MAX_CONFERENCES = 2 };
#else /* EASY336 */
enum { MAX_CONFERENCES = 10 };
#endif /* EASY336 */

/** Flag for no free data channel */
enum { NO_FREE_DATA_CH = -1 };

/** Status no external peer in conference */
enum { NO_EXTERNAL_PEER = 0 };

/** Status no PCM peer in conference */
enum { NO_PCM_PEER = 0 };

/** Status no local peer in conference */
enum { NO_LOCAL_PEER = 0 };

/** No value is set */
enum { TD_NO_VALUE = -1 };

/** No device is specified */
enum { NO_DEV_SPECIFIED = 0 };

#ifdef TD_FAX_MODEM
/**
   Mode of FAX transmission.
 */
typedef enum _TD_FAX_MODE_t
{
   /** Only transparetn fax transmission can be used */
   TD_FAX_MODE_TRANSPARENT = 0,
   /** Call is started as T.38 transmission (Fax Data Pump) */
   TD_FAX_MODE_FDP,
   /** Fax transmission is configured according to detected signals */
   TD_FAX_MODE_DEAFAULT
} TD_FAX_MODE_t;
#endif /* TD_FAX_MODEM */

#if (defined(TD_FAX_MODEM) && defined(TD_FAX_T38))
/**
   FAX T.38 configuration data.
 */
 typedef struct _TD_T38_CONFIG_t
 {
#ifdef TD_T38_FAX_TEST
    /** fax stack session configuration */
    IFX_TAPI_T38_FAX_CFG_t oT38FaxCfg;
    /** fax data pump(fdp) configuration */
    IFX_TAPI_T38_FDP_CFG_t oT38FdpCfg;
#endif /* TD_T38_FAX_TEST */
    /** session configuration */
    IFX_TAPI_T38_SESS_CFG_t oT38SessCfg;
 } TD_T38_CONFIG_t;

#ifdef TD_T38_FAX_TEST
 /**
   Structure with FAX T.38 TEST data
 */
 typedef struct _TD_T38_FAX_TEST_CTRL_t
 {
    /** socket fd */
    IFX_int32_t nComSocket;
    /** flag to check if fax test should be use,
        set to false if fax test initialization fails */
    IFX_boolean_t bUseFaxTest;
 } TD_T38_FAX_TEST_CTRL_t;
#endif /* TD_T38_FAX_TEST */
#endif /* (defined(TD_FAX_MODEM) && defined(TD_FAX_T38)) */

/** Conference support */
/* no conferencing support in static RM */
typedef struct _CONFERENCE_t
{
   /** Conference status, IFX_TRUE this conference is active, IFX_FALSE
       is free for use. */
   IFX_boolean_t fActive;

   /** Number of external peers in conference */
   IFX_int32_t nExternalPeersCnt;

   /** Number of PCM peers (channels) in conference */
   IFX_int32_t nPCM_PeersCnt;

   /** Number of local peers in conference */
   IFX_int32_t nLocalPeersCnt;
} CONFERENCE_t;


/* --------------------------------------------------------- */
/*                    CONFERENCE END                         */
/* --------------------------------------------------------- */


/** Represents called peer */
typedef struct _PEER_t
{
   /** Called phone number */
   IFX_int32_t nPhoneNum;

   /** Structure for external phone */
   struct {
      /** Data channel number of this peer. */
      IFX_int32_t nCh;
      /** Addr we send data to */
      struct sockaddr_in oToAddr;
      /** Device number of this peer. */
      IFX_uint16_t nDev;
      /** MAC address of this peer. */
      IFX_uint8_t MAC[ETH_ALEN];
      /* Keeps dialed number, except network part -
         used in external connection */
      IFX_char_t rgCollectNumber[4];
   } oRemote;
   /** Structure for PCM phone */
   struct {
      /** Addr we send data to */
      struct sockaddr_in oToAddr;
   } oPCM;
   /** Pointer to connection structure - other local phone */
   struct {
      /** Pointer to other phone */
      struct _PHONE_t* pPhone;
      /** Flag status used for retrieving right timeslot index pair.
          Called phone represents slave, caller phone represents master.
          IFX_TRUE if we are slave, otherwise IFX_FALSE for master */
      IFX_boolean_t fSlave;
   } oLocal;
   /** Pointer to connection structure - other local phone*/
#ifndef EASY336
   struct {
      /** Pointer to used FXO */
      struct _FXO_t* pFXO;
      /** FXO channel number of this peer. */
      IFX_int32_t nFxoCh;
   } oFXO;
#endif /* EASY336 */
} PEER_t;


/** Represents connection */
typedef struct _CONNECTION_t
{
   /** Connection status, IFX_TRUE - active, IFX_FALSE - not active
       Set only for caller when changing state from ringback to active
       and ringing to active. */
   IFX_boolean_t fActive;

   /** Connection status, IFX_TRUE - active, IFX_FALSE - not active */
   /** workaround for new SM use fActive flag */
   IFX_boolean_t fPCM_Active;

   /* ---------------------------------------------------------------
          Information of channels, sockets used to call the phone
    */

   /** 2 - local, 1 - remote, 3 - pcm, ... */
   CALL_TYPE_t nType;

   /** For external actions (IP + port) */
   struct sockaddr_in oUsedAddr;

   /** For external actions (MAC) */
   IFX_uint8_t oUsedMAC[ETH_ALEN];

   /** Socket on which call to peer was made (only VoIP or PCM connections). */
   IFX_int32_t nUsedSocket;

   /** Channel on which call to peer was made (can be data, pcm).
       It depends of the flag fType. */
   IFX_int32_t nUsedCh;

   /** Channel file descriptor */
   IFX_int32_t nUsedCh_FD;

   /** External dialing status, IFX_TRUE this is external dialing */
   IFX_boolean_t bExtDialingIn;

#if (defined(TD_FAX_MODEM) && defined(TD_FAX_T38))
   /** Informs if peer finished T.38 transmission. IFX_TRUE - peer finished 
       T.38 transmission. */
   IFX_boolean_t bT38_PeerEndTransm;
#endif /* (defined(TD_FAX_MODEM) && defined(TD_FAX_T38)) */

   /** Time when phone started waiting on second peer in external call */
   struct timeval timeStartExtCalling;

#ifdef EASY336
   /** pointer to used voipID */
   IFX_int32_t voipID;
#endif /* EASY336 */

   /** Feature number to use/configure */
   IFX_int32_t nFeatureID;

   /** PCM call specific data */
   struct {
      /** Timeslot RX index */
      IFX_int32_t nTimeslot_RX;
      /** Timeslot TX index */
      IFX_int32_t nTimeslot_TX;
   } oPCM;


   /* -------------------------------------------
               Information of called phone
    */
   /** Called peer */
   PEER_t oConnPeer;

#ifdef QOS_SUPPORT
    struct {
        /** QOS_SUPPORT: 1 - session is started, 0 - session stopped */
        IFX_int32_t nInSession;
        /** Session pair */
        QOS_INIT_SESSION oPair;
    } oSession;
#endif

} CONNECTION_t;

/** Connection structure, represents abstraction object. Basically represents
    phone with phone channel and device on which phone events are recorded. */
struct _PHONE_t
{
#if (defined EASY336) || (defined XT16)
   /** connID */
   IFX_int32_t connID;
   /** lineID */
   IFX_int32_t lineID;
   /** array of voipIDs */
   IFX_int32_t *pVoipIDs;
   /** number of voipIDs in the array */
   IFX_int32_t nVoipIDs;
#endif /* (defined EASY336) || (defined XT16) */
   /** Device number */
   IFX_uint16_t nDev;
   /** Phone index */
   IFX_int32_t        nIdx;
   /** Phone number */
   IFX_int32_t        nPhoneNumber;
   /** Phone channel number */
   IFX_int32_t        nPhoneCh;
   /** Data channel number, phone channel is mapped to this data channel at
       start. Used for DTMF and other stuff. */
   IFX_int32_t        nDataCh;
   /** PCM channel number */
   IFX_int32_t        nPCM_Ch;
   /** Socket used for voice, data streaming */
   IFX_int32_t        nSocket;
   /** Device file descriptor this structure is connected to (events coming
       from phone channel. Basically first device is connected to first
       structure, second device to second structure, etc. */
   IFX_int32_t        nPhoneCh_FD;
   /** Device file descriptor this structure is connected to (events coming
       from data channel. */
   IFX_int32_t        nDataCh_FD;

   /** CID: description and number of source phone */
   IFX_TAPI_CID_MSG_t* pCID_Msg;

   /** Count of dialed numbers. */
   IFX_int32_t        nDialNrCnt;
   /** Array of dialed numbers. */
   IFX_char_t         pDialedNum[MAX_DIALED_NUM];

   /** Array of connections with other phones */
   CONNECTION_t rgoConn[MAX_PEERS_IN_CONF];

   /** Count of connections. This number - 1 represents number of
       active connections. Index to last one in array represents empty
       connection. */
   IFX_int32_t nConnCnt;

/* STATIC RM doesn't support conferencing */
   /** CONFERENCE, conference index which it belongs if greater than zero,
      otherwise does not belong to conference.
      When using to get conference structure
      use the nConfIdx value decrased by 1 */
   IFX_uint32_t nConfIdx;

   /** CONFERENCE, this phone started conference:
       IFX_FALSE - no,
       IFX_TRUE - yes */
   IFX_boolean_t nConfStarter;

   /** We called local peer?
       IFX_FALSE - no,
       IFX_TRUE - yes */
   IFX_boolean_t fLocalPeerCalled;

   /** We called external peer?
       IFX_FALSE - no,
       IFX_TRUE - yes */
   IFX_boolean_t fExtPeerCalled;

   /** We called PCM peer?
       IFX_FALSE - no,
       IFX_TRUE - yes */
   IFX_boolean_t fPCM_PeerCalled;

   /** If local call then this phone channel send us some action. Used when 
    * removing local phone from conference. */
   CONNECTION_t* pConnFromLocal;

   /* To which board belongs this phone */
   BOARD_t* pBoard;

   /** FXO, this phone started FXO call:
       IFX_FALSE - no,
       IFX_TRUE - yes */
   IFX_boolean_t fFXO_Call;

   /** To which FXO belongs this phone */
   FXO_t* pFXO;

   /** True if we are in connection and pressed digit 'STAR' to make some
       adjustments. Otherwise false. */
   IFX_boolean_t fConfTAPI_InConn;

   /** Holding true if phone is in wideband, otherwise false. This flag
       is also used when we end phone connection or exit from tapidemo to
       set back the default values (NarrowBand). */
   IFX_boolean_t fWideBand_Cfg;

   /** Holding true if phone is in wideband for PCM calls, otherwise false.
       This flag is also used when we end phone connection to set back the
       default values (NarrowBand). */
   IFX_boolean_t fWideBand_PCM_Cfg;

   /** Holding true if phone AGC is configurd, otherwise false. This flag
       is also used when we end phone connection or exit from tapidemo to
       set back the default values (no AGC). */
   IFX_boolean_t fAGC_Cfg;

#ifdef EASY336
   /** true if CID detection is enabled on phone */
   IFX_boolean_t fCID_Enabled;
#endif /* EASY336 */

   /** Current state of phone. */
   STATE_MACHINES_t rgStateMachine[MAX_STATE_MACHINE];

   /** New internal FXS event to handle */
   INTERNAL_EVENTS_t nIntEvent_FXS;

   /** New internal config event to handle */
   INTERNAL_EVENTS_t nIntEvent_Config;

   /** Caller or Called - transmit and receive direction */
   CALL_DIRECTION_TYPE_t nCallDirection;

   /** Signals enabled for phone */
   IFX_TAPI_SIG_DETECTION_t stSignal;

#ifdef EVENT_COUNTER
   /** Event counter */
   IFX_uint32_t nEventCounter;
#endif   
};


/** Connection structure, represents abstraction object. Basically represents
    FXO with device fd. */
struct _FXO_t
{
   /** FXO number (internal phone number or timeslot index) which follows
       the last phone number */
   IFX_int32_t        nFXO_Number;

   /** FXO channel */
   IFX_int32_t        nFxoCh;

   /** PCM channel number which is used for FXO connection. */
   IFX_int32_t        nPCM_Ch;

   /** Data channel used for CID RX between first and second ring.
       Afterthat this channel is released. */
   IFX_int32_t        nDataCh;

   /** Device file descriptor this structure is connected to (events coming
       from data channel. */
   IFX_int32_t        nFxoCh_FD;

   /** Current state of FXO. */
   STATE_MACHINE_STATES_t nState;

   /** CID message  */
   IFX_TAPI_CID_MSG_t oCID_Msg;

   /** Only one phone can be connected to this FXO. */
   CONNECTION_t oConn;

   /** Which phone is using this FXO line. */
   PHONE_t* pPhone;

   /** To which board belongs this FXO */
   BOARD_t* pBoard;

   /** Time when ringing event has occured. */
   struct timeval oRingEventTimeout;

   /** Indication that first ring stop event has occured. Its use for
       CID RX FSK support. */
   IFX_boolean_t fFirstStopRingEventOccured;

   /** Indication that CID FSK is started */
   IFX_boolean_t fCID_FSK_Started;

   /* FXO direction marker IFX_TRUE when incoming call, when not IFX_FALSE
      used to distinguish with connection structure use */
   IFX_boolean_t fIncomingFXO;
};


/** represents startup mapping table */
/* mapping is done by RM */
typedef struct _STARTUP_MAP_TABLE_t
{
   /** Phone channel, index */
   IFX_int32_t nPhoneCh;

   /** Connected data channel, just index, if -1 no connection to data
       channel */
   IFX_int32_t nDataCh;

   /** Connected PCM channel, just index, if -1 no connection to PCM
       channel.
       NOTICE: PCM channel is not mapped at start, but when PCM call
       is started.  */
   IFX_int32_t nPCM_Ch;

} STARTUP_MAP_TABLE_t;


/** Structure holding status for data channel */
typedef struct _VOIP_DATA_CH_t
{
   /* mapping managed by RM */
   /** IFX_FALSE if its free, not mapped. Otherwise IFX_TRUE if mapped. */
   IFX_boolean_t fMapped;

   /** Status of encoder/decoder. IFX_TRUE - started, IFX_FALSE - stopped */
   IFX_boolean_t fCodecStarted;

   /** IFX_FALSE - don't start codec, IFX_TRUE - start codec */
   IFX_boolean_t fStartCodec;

   /** IFX_TRUE - default VAD configuration,
      IFX_FALSE - VAD configuration changed */
   IFX_boolean_t fDefaultVad;

   /** Which type of coder to use. */
   IFX_TAPI_ENC_TYPE_t nCodecType;

   /** Length of frame in miliseconds OR packetisation time. */
   IFX_int32_t nFrameLen;

   /** ITU-T I366.2 Bit Alignment for G.726 codecs (IETF RFC3550),
       Used only by G.726. */
   IFX_TAPI_COD_AAL2_BITPACK_t nBitPack;

   /** Which type of coder was used before fax/modem connection*/
   IFX_TAPI_ENC_TYPE_t nPrevCodecType;

   /** What length of frame was used before fax/modem connection*/
   IFX_int32_t nPrevFrameLen;

   /** Settings of LEC before Fax/modem connection */
   IFX_TAPI_WLEC_CFG_t oPrevLECcfg;

   /** IFX_FALSE if no need to restore channel after fax/modem transmission */
   IFX_boolean_t fRestoreChannel;

   /**Setting of Jitter buffor before Fax/modem connection */
   IFX_TAPI_JB_CFG_t tapi_jb_PrevConf;

#ifdef TAPI_VERSION4
   /** Device number */
   IFX_uint16_t nDev;

   /** Channel number */
   IFX_uint16_t nCh;

   /** Port number */
   IFX_uint16_t nUDP_Port;
#endif /* TAPI_VERSION4 */
} VOIP_DATA_CH_t;


/** Structure holding status for socket */
typedef struct _SOCKET_STAT_t
{
   /** IFX_FALSE if its free, not used. Otherwise IFX_TRUE if used. */
   IFX_boolean_t fUsed;

   /** Handle to socket */
   IFX_int32_t nSocket;

   /** Port used */
   IFX_int32_t nPort;
} SOCKET_STAT_t;


/* --------------------------------------------------------- */
/*                 MULTIPLE BOARDS START                     */
/* --------------------------------------------------------- */

/** Max supporting boards */
enum { MAX_BOARDS = 4 };

/** Max supporting phones on all boards together */
#ifndef EASY336
enum { MAX_ALL_PHONES = 16 };
#else /* EASY336 */
enum { MAX_ALL_PHONES = 96 };
#endif /* EASY336 */

/** Types of board */
typedef enum
{
   TYPE_VINETIC = 0,
   TYPE_DANUBE = 1,
   TYPE_DUSLIC_XT = 2,
   TYPE_DUSLIC = 3,
   TYPE_VINAX = 4,
   TYPE_EASY508XX = 5,
   TYPE_SVIP = 6,
   TYPE_XT16 = 7,
   TYPE_TERIDIAN = 8,
   TYPE_VR9 = 9,
   TYPE_NONE = 10
} BOARD_TYPE_t;

/** Types of FXO devices */
typedef enum
{
   FXO_TYPE_DUSLIC = 1,
   FXO_TYPE_TERIDIAN = 2,
   FXO_TYPE_NONE = 3
} FXO_TYPE_t;

/** devices to which access is done by GPIO */
typedef enum
{
   /** Vinetic on the EASY50510 */
   GPIO_VINETIC_EASY50510 = 1,
   /** DuSLIC on the EASY50712 (Danube) */
   GPIO_DUSLIC_EASY50712 = 2,
   /** Teridian on the EASY50712 (Danube) */
   GPIO_TERIDIAN_EASY50712 = 3,
   /** DuSLIC on the EASY50510 */
   GPIO_DUSLIC_EASY50510 = 4,
   /** Teridian on the EASY508XX (AR9) board */
   GPIO_TERIDIAN_EASY508XX = 5,
   /** Teridian on the EASY80800 (VINAX) board */
   GPIO_TERIDIAN_EASY80800 = 6,
   /** Teridian on the extension board */
   GPIO_TERIDIAN_EXT = 7,
   /** Teridian on the EASY80800 (VINAX) board */
   GPIO_TERIDIAN_EASY80910 = 8,
   /** Unknown device and board type */
   GPIO_NONE = 9
} GPIO_ACCESS_DEV_t;

/** Types of devices */
typedef enum
{
   TAPIDEMO_DEV_CPU = 1,
   TAPIDEMO_DEV_FXO = 2,
   TAPIDEMO_DEV_NONE = 3
} TAPIDEMO_DEVICE_TYPE_t;


/** Max. len of device path (/dev/vin...) */
enum { DEV_NAME_LEN = 32 };

/** Structure representing FXO device. */
typedef struct _TAPIDEMO_DEVICE_FXO_t
{
   /** FXO device. Can be duslic or teridian */
   FXO_TYPE_t nFxoType;

   /** Country selection. */
   IFX_uint8_t nCountry;

   /** device irq number */
   IFX_int32_t nIrqNum;

   /** Pointer to function which sets direction of GPIO to out which is used
       for set/clear reset */
   IFX_return_t  (*SetPin_Direction_onGPIO)    (GPIO_ACCESS_DEV_t eType);

   /** Pointer to function which set/clear reset of device */
   IFX_return_t  (*Reset)                      (BOARD_t* pBoard,
                                               IFX_int32_t nArg,
                                               IFX_boolean_t hSetReset);
   /** Pointer to function which read verion */
   IFX_return_t (*Version)    (BOARD_t* pBoard);

   /** Pointer to function which returns last error */
   IFX_return_t   (*GetLastErr) (BOARD_t* pBoard);

} TAPIDEMO_DEVICE_FXO_t;

/** Structure representing CPU device. */
struct _TAPIDEMO_DEVICE_CPU_t
{
   /** FW files data */
   IFX_char_t* pszDRAM_FW_File;

   /** FW files code */
   IFX_char_t* pszPRAM_FW_File[2];

   /** Coefficient file (BBD, CRAM) */
   IFX_char_t* pszBBD_CRAM_File;

   /** Pointer to device dependent structure */
   IFX_void_t* pDevDependStuff;

   /** Access mode for device. */
   IFX_int32_t AccessMode;

   /** Device physical base address. */
   unsigned long nBaseAddress;

   /** Device irq number. */
   IFX_int32_t nIrqNum;

   /** Pointer to function which sets direction of GPIO to out which is used
       for set/clear reset */
   IFX_return_t  (*SetPin_Direction_onGPIO)    (GPIO_ACCESS_DEV_t eType);

   /** Pointer to function which set/clear reset of device */
   IFX_return_t  (*Reset)     (BOARD_t* pBoard, IFX_int32_t nArg,
                                               IFX_boolean_t hSetReset);

   /** Pointer to function which initializes device */
   IFX_return_t (*Init)       (BOARD_t* pBoard, IFX_char_t* pPath);

   /** Pointer to function which read verion */
   IFX_return_t (*Version)    (BOARD_t* pBoard);

   /** Pointer to function which Handles the DRIVER error codes and stack */
   IFX_void_t   (*DecodeErrno)(IFX_int32_t nCode);

   /** Pointer to function which returns number of last driver error */
   IFX_return_t   (*GetErrno) (IFX_int32_t nFd);

#ifdef ABACUS_TEST
   /** Pointer to function which fills in firmware pointers */
   IFX_return_t (*FillinFwPtr) (IFX_void_t* pDrvInit,
                                TAPIDEMO_DEVICE_CPU_t* pCpuDevice,
                                IFX_char_t* pPath);

   /** Pointer to function which releases firmware pointers */
   IFX_void_t  (*ReleaseFwPtr) (IFX_void_t* pDrvInit);
#endif /* ABACUS_TEST */
};

/** Structure representing device. */
typedef struct _TAPIDEMO_DEVICE_t
{
   /** device type (CPU or FXO) */
   TAPIDEMO_DEVICE_TYPE_t nType;

   /** Pointer to next device structure. */
   struct _TAPIDEMO_DEVICE_t *pNext;

   union
   {
      TAPIDEMO_DEVICE_CPU_t   stCpu;
      TAPIDEMO_DEVICE_FXO_t   stFxo;
   } uDevice;
} TAPIDEMO_DEVICE_t;

/** Structure representing board and info about it. */
struct _BOARD_t
{
   /** Board idx */
   IFX_int32_t nBoard_IDX;

   /** Board id, used to open right device */
   IFX_int32_t nID;

   /** Board type */
   BOARD_TYPE_t nType;

   /** Board name */
   const char* pszBoardName;

   /** Prefix used to adjust indention in traces */
   IFX_char_t* pIndentBoard;

   /** Prefix used to adjust indention in traces */
   IFX_char_t* pIndentPhone;

   /** Number of analog ch on board. */
   IFX_int32_t nMaxAnalogCh;

   /** Number of coder, signalling (data) ch on board. */
   IFX_int32_t nMaxCoderCh;

   /** Number of used channels for open/close, events, streaming, ...
       Used, because some systems have only analog channels, coder channels are
       not present. In SVIP this field is preserved, but not used so often
       as before, nMaxAnalogCh or nMaxCoderCh is used instead */
   IFX_int32_t nUsedChannels;

   /** Number of pcm ch on board */
   IFX_int32_t nMaxPCM_Ch;

   /** System file descriptor */
   IFX_int32_t nSystem_FD;

   /** Usage of Sys FD, 0 - not used */
   IFX_boolean_t nUseSys_FD;

   /** Device control file descriptor */
   IFX_int32_t nDevCtrl_FD;

   /** Channel file descriptors */
   IFX_int32_t* nCh_FD;

   /** Number of FXO (duslic or teridian) ch on board */
   IFX_int32_t nMaxFxoCh;

   /** FXO device control file descriptor */
   IFX_int32_t nFxoDevCtrl_FD;

   /** FXO channel file descriptors */
   IFX_int32_t* nFxoCh_FD;

   /** if IFX_TRUE, T.38 is supported */
   IFX_boolean_t nT38_Support;

#if (defined(TD_FAX_MODEM) && defined(TD_FAX_T38))
   /** Capabilities of the T.38 stack implementation. */
   IFX_TAPI_T38_CAP_t stT38Cap;

   /** T.38 configuration */
   TD_T38_CONFIG_t oT38_Config;
#endif /* (defined(TD_FAX_MODEM) && defined(TD_FAX_T38) ) */

   /** FXO device name without numbers, like /dev/vmmc, /dev/vin, ... */
   IFX_char_t pszFxoChDevName[DEV_NAME_LEN];

   /** FXO device control name numbers, like /dev/vmmc[board id]0, ... */
   IFX_char_t pszFxoCtrlDevName[DEV_NAME_LEN];

   /** used to store daa event data */
   IFX_TAPI_EVENT_t daaEvent;

   /** Device name without numbers, like /dev/vmmc, /dev/vin, ... */
   IFX_char_t pszChDevName[DEV_NAME_LEN];

   /** Device control name numbers, like /dev/vmmc[board id]0, ... */
   IFX_char_t pszCtrlDevName[DEV_NAME_LEN];

   /** System device name, like /dev/easy3332/[board id], ... */
   IFX_char_t pszSystemDevName[DEV_NAME_LEN];

   /** Number of chips on board */
   IFX_int32_t nChipCnt;

   /** Startup mapping table */
   const STARTUP_MAP_TABLE_t* pChStartupMapping;

#if defined(ABACUS_TEST) || defined(TD_T38_FAX_TEST)
   /** Custom mapping table */
   STARTUP_MAP_TABLE_t* pChCustomMapping;
#endif /* defined(ABACUS_TEST) || defined(TD_T38_FAX_TEST) */

#ifdef ABACUS_TEST
   /** Idication that Coefficient file (BBD, CRAM) was changed */
   IFX_boolean_t fBBD_Changed;
#endif /* ABACUS_TEST */

   /** Status array of data channels */
   VOIP_DATA_CH_t* pDataChStat;

   /** Default RTP and RTCP settings */
   IFX_TAPI_PKT_RTP_CFG_t rtpConf;

   /** Status of used pcm channels, IFX_FALSE - used, IFX_TRUE - not used */
   IFX_boolean_t* fPCM_ChFree;

   /** Socket on which we expect incoming data (events and voice)
       Also used to send data to external peer. There is linear
       transformation  between sockets and data channel. Socket
       with index 1 belongs to data channel 1, socket with index
       2 to data channel 2, etc. */
   IFX_int32_t* rgnSockets;

   /** Hold array of phones */
   PHONE_t* rgoPhones;

   /** Hold array of FXO-es */
   FXO_t* rgoFXOs;

   /** Pointer to control structure */
   CTRL_STATUS_t *pCtrl;

   /** Pointer to device info structure */
   TAPIDEMO_DEVICE_t *pDevice;

   /** used to store board event data */
   IFX_TAPI_EVENT_t tapiEvent;

#ifdef TAPI_VERSION4
   /** Holding true if there is a single file descriptor */
   IFX_boolean_t fSingleFD;

   /** Holding true if sockets are used for voice streaming through codecs */
   IFX_boolean_t fUseSockets;
#endif /* TAPI_VERSION4 */

   /** Pointer to function which initializes board */
   IFX_return_t (*Init)       (BOARD_t* pBoard, PROGRAM_ARG_t* pProgramArg);

   /** Pointer to function which release board structure */
   IFX_void_t   (*RemoveBoard) (BOARD_t* pBoard);

#ifdef EVENT_COUNTER
   /** Array of event counters. One counter per chip */
   IFX_uint32_t *pEventCounters;
#endif
};


/** Max combinations */
enum { MAX_COMBINATIONS = 15 };

/** Max boards in one combination */
enum { MAX_BOARD_IN_COMBINATION = MAX_BOARDS };

/** structure for different board combinations */
typedef struct _BOARD_COMBINATION_ELEM_t
{
   /** board ID */
   IFX_int32_t nBoard_ID;
   /** board type */
   BOARD_TYPE_t nBoardType;
} BOARD_COMBINATION_ELEM_t;

/** Structure holding combination */
typedef struct _BOARD_COMBINATION_t
{
   /** identifier */
   IFX_int32_t nIdx;
   /** number of boards */
   IFX_int32_t nBoardCnt;
   /** array with boards parameters */
   BOARD_COMBINATION_ELEM_t nBoards[MAX_BOARD_IN_COMBINATION];
} BOARD_COMBINATION_t;

/* --------------------------------------------------------- */
/*                 MULTIPLE BOARDS END                       */
/* --------------------------------------------------------- */


/** Flags for command line arguments */
typedef struct _ARG_FLAGS_t
{
   /** If 1 then fax channel enabled instaed of codecs */
   IFX_int32_t nFaxChannel : 1;
   /** If 1 then CID is send when calling other phone, but only local */
   IFX_int32_t nCID : 1;
   /** If 1 then conference is used */
   IFX_int32_t nConference : 1;
   /** If 1 then help is displayed */
   IFX_int32_t nHelp : 1;
   /** If 1 then wait after each state machine step */
   IFX_int32_t nWait : 1;
   /** If 1 then download CRAM file */
   /*IFX_int32_t nCRAM : 1;*/
   /** 1 if quality of service is used, otherwise 0 */
   IFX_int32_t nQos : 1;
   /** 1 if PCM is used and we are master, otherwise 0 */
   IFX_int32_t nPCM_Master : 1;
   /** 1 if PCM is used and we are slave, otherwise 0  */
   IFX_int32_t nPCM_Slave : 1;
   /** 1 for new encoder type, otherwise 0  */
   IFX_int32_t nEncTypeDef : 1;
   /** 1 for new packetisation time, otherwise 0  */
   IFX_int32_t nFrameLen : 1;
   /** 1 for calls to different network, otherwise 0  */
   IFX_int32_t nDiffNet : 1;
   /** 1 if FXO is used, otherwise 0 */
   IFX_int32_t nFXO : 1;
   /** 1 if local UDP will be tested, otherwise 0 - external UDP is used */
   IFX_int32_t nQOS_Local : 1;
   /** 1 if local loop PCM will be used, otherwise 0 */
   IFX_int32_t nPCM_Loop : 1;
   /** 1 to use coders for local connection, 0 to use phone mapping */
   IFX_int32_t nUseCodersForLocal : 1;
   /** 1 to use PCM wideband, 0 to use PCM narrowband */
   IFX_int32_t nUsePCM_WB : 1;
   /** 1 to run Tapidemo as daemon */
   IFX_int32_t nDaemon : 1;
   /** path to download has been changed */
   IFX_int32_t nUseCustomDownloadsPath : 1;
   /** file .tapidemorc was checked for existence */
   IFX_int32_t nTapidemorcFileChecked : 1;
   /** do not use LEC */
   IFX_int32_t nNoLEC : 1;
} ARG_FLAGS_t;


/** type of phone port */
typedef enum _TAPIDEMO_PORT_TYPE_t
{
   /** FXO port */
   PORT_FXO,
   /** FXS port */
   PORT_FXS
} TAPIDEMO_PORT_TYPE_t;

/** port data structure */
typedef struct _TAPIDEMO_PORT_DATA_t
{
   /** port type(fxo/fxs) */
   TAPIDEMO_PORT_TYPE_t nType;
   /** port data */
   union
   {
      /** FXS data */
      PHONE_t* pPhopne;
      /** FXO data */
      FXO_t*   pFXO;
   } uPort;
} TAPIDEMO_PORT_DATA_t;

#ifdef ABACUS_TEST
/** Test types performed by test script on Control PC*/
typedef enum _COM_TEST_TYPE_t
{
   COM_NO_TEST = -1,
   COM_MODULAR_TEST = 0,
   COM_BULK_CALL_TEST = 1 /* default */
} COM_TEST_TYPE_t;

/** Modular subtest types performed by test script on Control PC*/
typedef enum _COM_SUB_TEST_TYPE_t
{
   /** default */
   COM_SUBTEST_NONE              = 0x0000,
   COM_SUBTEST_HOOKSTATE         = 0x0001,
   COM_SUBTEST_HOOKFLASH         = 0x0002,
   COM_SUBTEST_PULSEDIGIT        = 0x0004,
   COM_SUBTEST_DTMFDIGIT         = 0x0008,
   COM_SUBTEST_CONFERENCE        = 0x0010,
   COM_SUBTEST_FXO               = 0x0020,
   COM_SUBTEST_FAX               = 0x0040,
   COM_SUBTEST_MODEM             = 0x0080,
   COM_SUBTEST_FAX_T_38          = 0x0100,
   COM_SUBTEST_SYSINIT           = 0x0200
} COM_SUB_TEST_TYPE_t;

/** send message to abacus that busy tone is played during conference test */
#define SUBTEST_CONFERENCE_BUSY_TRACE(nTone) \
   /* If it's a conference test... */ \
   if (BUSY_TONE_IDX == nTone) \
   { \
      if ((COM_MODULAR_TEST == pCtrl->pProgramArg->nTestType ) && \
          (pCtrl->pProgramArg->nSubTestType & COM_SUBTEST_CONFERENCE)) \
      { \
         TAPIDEMO_PRINTF(("CONFERENCE:BUSYTONE::%d", \
                          (IFX_int32_t) pPhone->nPhoneCh)); \
      } \
   }
#else /* ABACUS_TEST */
/** defined to get rid of compiler warnings */
#define SUBTEST_CONFERENCE_BUSY_TRACE(nTone) \
   if (0 > nTone) \
   { \
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH, ("Invalid Tone number\n")); \
   }
#endif /* ABACUS_TEST */


/** Demo structure holding command line arguments as flags */
struct _PROGRAM_ARG_t
{
   /* #warning karol -workaround for teridian extension board */
   IFX_boolean_t hUseDuslicFxo;
   /* end workaround */

    /** Set parameters */
    ARG_FLAGS_t oArgFlags;
    /** Contains IP address for connection in network format */
    struct sockaddr_in oMy_IP_Addr;
    /** Contains first three numbers of IP address when calling different
        network. Syntax: xxx.xxx.xxx{.yyy - added by us} */
    struct in_addr oDiffNetIP;
    /** Path to FW, BBD filenames to download */
    IFX_char_t sPathToDwnldFiles[MAX_PATH_LEN];
    /** Board combination */
    IFX_int32_t nBoardComb;
    /** Debug, trace level */
    DEBUG_LEVEL_t nDbgLevel;
#ifdef ABACUS_TEST
    /** Remote (send to control PC) debug, trace level */
    DEBUG_LEVEL_t nComDbgLevel;
    /** Test type beeing performed (bulk call or modular)*/
    COM_TEST_TYPE_t nTestType;
    /** Modular sub test beeing performed */
    IFX_uint32_t nSubTestType;
#endif
    /** CID standard */
    IFX_int32_t nCIDStandard;
    /** Encoder type */
    IFX_int32_t nEnCoderType;
    /** Packetisation time */
    IFX_int32_t nPacketisationTime;
    /** Voice Activity Detection (VAD) settings */
    IFX_TAPI_ENC_VAD_t nVadCfg;
    /** Trace redirection */
    TRACE_REDIRECTION_t nTraceRedirection;
};

/** Control structure for all channels, holding status of connections and
    also program arguments */
struct _CTRL_STATUS_t
{
   /* #warning karol -workaround for teridian extension board */
   IFX_boolean_t hUseDuslicFxo;
   /* end workaround */

   /** Holds array of boards */
   BOARD_t rgoBoards[MAX_BOARDS];
   /** Holds array of conferences */
   /* conferencing not supported by static RM */
   CONFERENCE_t rgoConferences[MAX_CONFERENCES];
   /** Conference index, if 0 no conference yet */
   IFX_uint32_t nConfCnt;
   /** Program arguments */
   PROGRAM_ARG_t* pProgramArg;
   /** Holds control device file descriptor, like for /dev/vin10, but
       highest file descriptor, used for FD_ISSET */
   IFX_int32_t rwd;
   /** Contains IP address for connection but as number */
   IFX_uint32_t nMy_IP_Addr;
   /** Contains our ip address */
   struct sockaddr_in oTapidemo_IP_Addr;
   /** Contains our MAC address */
   IFX_uint8_t oTapidemo_MAC_Addr[ETH_ALEN];
#if (defined EASY336 || defined XT16)
   /** Contains veth0 ip address */
   struct sockaddr_in oVeth0_IP_Addr;
   /** Contains veth0 MAC address */
   IFX_uint8_t oVeth0_MAC_Addr[ETH_ALEN];
#endif /* (defined EASY336 || defined XT16) */
   /** Board combination index */
   IFX_int32_t nBoardCombination;
   /** Boards in combination */
   IFX_int32_t nBoardCnt;
   /** Sum of all data channels on all boards. */
   IFX_int32_t nSumCoderCh;
   /** Sum of all analog channels (phones) on all boards. */
   IFX_int32_t nSumAnalogCh;
   /** Sum of all pcm channels on all boards. */
   IFX_int32_t nSumPCM_Ch;

   /** Socket on which we expect incoming data (events and voice)
       Also used to send data to external peer. */
   SOCKET_STAT_t* rgnSockets;

   /** Socket used for administration of calls (make/end/...) */
   IFX_int32_t nAdminSocket;
#ifdef ABACUS_TEST
   /** Socket used for communication with control PC */
   IFX_int32_t nComInSocket;
   /** Did the SUT send its ip address to control PC*/
   IFX_boolean_t fComIdentSent;
   /** Expected command Ids from Control PC */
   IFX_int32_t nExpectedID;
   /** Command Ids received from Control PC */
   IFX_int32_t nReceivedID;
   /** Used in verify system initialization */
   IFX_char_t aParametrs[20];
   /** Change address IP */
   IFX_boolean_t bChangeIPAddress;
   /** if IFX_TRUE use custom CPTs */
   IFX_boolean_t nUseCustomCpt;
   /** array with tapidemo version */
   IFX_char_t rgoTapidemoVersion[TD_MAX_NAME];
   /** array with ABACUS TEST version */
   IFX_char_t rgoAbacusTestVersion[TD_MAX_NAME];
   /** array with firmware version */
   IFX_char_t rgoFW_Version[TD_MAX_NAME];
#endif
#if (defined(TD_T38_FAX_TEST) && defined(TD_FAX_T38) && defined(AR9))
   /** fax test control structure */
   TD_T38_FAX_TEST_CTRL_t oFaxTestCtrl;
#endif /* (defined(TD_T38_FAX_TEST) && defined(TD_FAX_T38) && defined(AR9)) */
#ifdef TD_FAX_MODEM
   /* used in tapidemo fax transmission mode */
   TD_FAX_MODE_t nFaxMode;
#endif /* TD_FAX_MODEM */
   /* socket used for syslog */
   IFX_int32_t nSyslogSocket;
   /** socket used for syslog */
   IFX_int32_t nSyslogChildPid;
   /** set if internal event generated from local phone or fxo,
       this flag is set together with connection pointer on destination phone
       e.g. p_dst_phone->pConnFromLocal = p_conn */
   IFX_boolean_t bInternalEventGenerated;
   /** flag with number of resources(phones and FXO channels)
       that need timeout check */
   IFX_int32_t nUseTimeout;
};


/** Message used to communicate between boards */
typedef struct _COMM_MSG_t
{
   /** Mark start of message - must match, otherwise wrong message */
   IFX_uint8_t nMarkStart;

   /** Phone number of message sender */
   IFX_uint32_t nSenderPhoneNum;

   /** Phone number of message receiver */
   IFX_uint32_t nReceiverPhoneNum;

   /** Port number of message sender for VOICE streaming */
   IFX_int32_t nSenderPort;

   /** message sender MAC */
   IFX_uint8_t MAC[8];

   /** Timeslot RX index, used only in PCM connection */
   IFX_int32_t nTimeslot_RX;
   /** Timeslot TX index, used only in PCM connection */
   IFX_int32_t nTimeslot_TX;

   /** message sender board idx */
   IFX_int32_t nBoard_IDX;

   /** Action (OFF_HOOK, ...) */
   IFX_uint8_t nAction;

   /** 0x80 - PCM or 0x00 - VoIP call */
   IFX_uint8_t fPCM;

   /** Configure phone with some features (number defines which one)
       also nAction must be set to IE_CONFIG_PEER */
   IFX_int32_t nFeatureID;

   /** Mark end of message - must match, otherwise wrong message */
   IFX_uint8_t nMarkEnd;
} COMM_MSG_t;

#if defined(ABACUS_TEST) || defined(TD_T38_FAX_TEST)
/** Type of mapping */
typedef enum _MAPPING_TYPE_t { NO_MAPPING = -1 } MAPPING_TYPE_t;
#endif /* defined(ABACUS_TEST) || defined(TD_T38_FAX_TEST) */

/** Structure holding status for trace redirection */
typedef struct _TRACE_REDIRECTION_CTRL_t
{
   /** Contains IP address for connection to syslog */
   struct sockaddr_in oSyslog_IP_Addr;
   /** Path to log file*/
   IFX_char_t sPathToLogFiles[MAX_PATH_LEN];
   /** Socket used for communication with remote syslog */
   IFX_int32_t nSyslogSocket;
   /* Holding file descriptor for log file */
   IFX_int32_t nFileLogFD;
   /* Start redirecting log */
   IFX_boolean_t fStartRedirectLogs;
} TRACE_REDIRECTION_CTRL_t;

#define SEPARATE \
TAPIDEMO_PRINTF(("############################################\n"));

/* ============================= */
/* Global Variables              */
/* ============================= */

/** Frequencies strings */
extern const char * TAPIDEMO_PCM_FREQ_STR[];

/* ============================= */
/* Global Structures             */
/* ============================= */

 #ifdef EASY336
/** Structure for requesting allocation/deallocation of resources */
typedef struct _RESOURCES_t
{
   /* resources to get/release */
   IFX_int32_t nType;
   /* connID number to use */
   IFX_int32_t nConnID_Nr;
   /* voipID number to release */
   IFX_int32_t nVoipID_Nr;
} RESOURCES_t;
#endif

/* ============================= */
/* Global function declaration   */
/* ============================= */

/* fd operations */
extern IFX_return_t Common_Open(IFX_char_t* pszName);
extern IFX_return_t Common_Set_FDs(BOARD_t* pBoard);
extern IFX_void_t Common_Close_FDs(BOARD_t* pBoard);
extern IFX_int32_t Common_GetDeviceOfCh(BOARD_t* pBoard, IFX_int32_t nChNum);

/* FXO fd operations */
extern IFX_int32_t Common_GetFxoDevCtrlFd(BOARD_t* pBoard);
extern IFX_int32_t Common_GetFxoDevOfCh(BOARD_t* pBoard, IFX_int32_t nChNum);

extern IFX_return_t Common_Set_Fxo_FDs(BOARD_t* pBoard);
extern IFX_void_t Common_Close_Fxo_FDs(BOARD_t* pBoard);

/* board initialization */
extern IFX_return_t Common_RegisterBoard(BOARD_t* pBoard);
extern IFX_return_t Common_GetVersions(BOARD_t* pBoard);

extern IFX_return_t Common_GetCapabilities(BOARD_t* pBoard,
                                           IFX_int32_t *nPhoneChCnt,
                                           IFX_int32_t *nDataChCnt,
                                           IFX_int32_t *nPCM_ChCnt);
extern IFX_char_t* Common_GetTraceIndention(IFX_int32_t nSize);
#ifdef USE_FILESYSTEM
IFX_return_t Common_ReadBinFile(IFX_char_t* pszFullFilename,
                                IFX_uint8_t** pBuf,
                                IFX_int32_t* pSize);
IFX_return_t Common_CheckFile(IFX_char_t* pszFullFilename);
IFX_return_t Common_CreateFullFilename(IFX_char_t* pszFilename,
                                     IFX_char_t* pszPath,
                                     IFX_char_t* pszFullFilename);
IFX_boolean_t Common_FindPRAM_FW(TAPIDEMO_DEVICE_CPU_t* pCpuDevice, 
                                   IFX_char_t* pszPath, 
                                   IFX_char_t* pszFullFilename);
#ifndef VXWORKS
extern IFX_return_t Common_CheckDownloadPath(IFX_char_t* sPathToDwnldFiles,
                                             BOARD_t* pBoard,
                                             TAPIDEMO_DEVICE_CPU_t* pCpuDevice);
#endif /* VXWORKS */
#endif  /* USE_FILESYSTEM */

extern IFX_return_t Common_ConfigureTAPI(CTRL_STATUS_t* pCtrl,
                                         BOARD_t* pBoard,
                                         PHONE_t* pPhone,
                                         IFX_int32_t nAction,
                                         IFX_int32_t nDataCh_FD,
                                         IFX_boolean_t fConfigureAll,
                                         IFX_boolean_t fConfigureConn);

extern IFX_return_t Common_FillAddrByNumber(CTRL_STATUS_t* pCtrl,
                                            PHONE_t* pPhone,
                                            CONNECTION_t* pConn,
                                            IFX_uint32_t nIP_Addr,
                                            IFX_boolean_t fIsPCM_Num);
extern IFX_int32_t Common_DigitsToNum(IFX_char_t* prgnDialNum,
                                      IFX_int32_t nFirstDigit,
                                      IFX_int32_t nDialNrCnt);

extern CALL_TYPE_t Common_CheckDialedNumForFull_IP(IFX_char_t* prgnDialNum,
                                                   IFX_int32_t nDialNrCnt,
                                                   IFX_uint32_t* nIpAddr);
extern CALL_TYPE_t Common_CheckDialedNum(PHONE_t* pPhone,
                                         IFX_int32_t* nDialedNum,
                                         IFX_uint32_t* nIpAddr,
                                         CTRL_STATUS_t* pCtrl);

extern IFX_void_t Common_PrintCallProgress(STATE_MACHINE_STATES_t current_state,
                                           INTERNAL_EVENTS_t new_event,
                                           CALL_TYPE_t nOldCallType,
                                           PHONE_t* pPhone, CONNECTION_t* pConn,
                                           IFX_int32_t nUseCodersForLocal);
extern IFX_void_t Common_PrintEncoderType(CTRL_STATUS_t* pCtrl, PHONE_t* pPhone,
                                          CONNECTION_t* pConn);

extern IFX_return_t Common_JBStatistic(IFX_int32_t nDataCh,
                                       IFX_int32_t nDev,
                                       IFX_boolean_t fReset,
                                       IFX_boolean_t fShow,
                                       BOARD_t* pBoard);

extern IFX_return_t Common_RTCPStatistic(IFX_int32_t fd,
                                         IFX_boolean_t fReset,
                                         IFX_boolean_t fShow);

extern IFX_void_t Common_ErrorHandle(IFX_int32_t nFD, BOARD_t* pBoard, 
   IFX_uint16_t nDev);
#ifdef TAPI_VERSION4
#if (defined(EASY336) || defined(XT16))
extern IFX_return_t Common_PCM_ConfigSet(BOARD_t* pBoard,
                                         RM_SVIP_PCM_CFG_t* pcm_cfg);
#endif

#ifdef EASY336
extern IFX_return_t Common_ResourcesGet(PHONE_t* pPhone,
                                        const RESOURCES_t* pResources);
extern IFX_return_t Common_ResourcesRelease(PHONE_t* pPhone,
                                            const RESOURCES_t* pResources);
extern VOIP_DATA_CH_t* Common_GetCodec(CTRL_STATUS_t* pCtrl,
                                       IFX_int32_t voipID);
#endif
extern IFX_return_t Common_DTMF_Enable(CTRL_STATUS_t* pCtrl, PHONE_t* pPhone);
extern IFX_return_t Common_DTMF_Disable(CTRL_STATUS_t* pCtrl, PHONE_t* pPhone);
#ifdef EASY336
extern IFX_int32_t Common_ConnID_PhonesGet(CTRL_STATUS_t* pCtrl,
                                           IFX_int32_t connID,
                                           PHONE_t** pPhones[]);
#endif /* EASY336 */

extern IFX_return_t Common_PulseEnable(CTRL_STATUS_t* pCtrl, PHONE_t* pPhone);
extern IFX_return_t Common_PulseDisable(CTRL_STATUS_t* pCtrl, PHONE_t* pPhone);

extern IFX_return_t Common_CID_Enable(PHONE_t* pPhone, IFX_boolean_t bCID_AS);
extern IFX_return_t Common_CID_Disable(PHONE_t* pPhone);
extern IFX_return_t Common_SetChipCount(BOARD_t* pBoard);
#ifdef LINUX
extern IFX_return_t Common_WriteByte(IFX_uint32_t nAddr, IFX_uint8_t nValue);
#endif
#else /* TAPI_VERSION4 */
extern IFX_return_t Common_DTMF_SIG_Enable(CTRL_STATUS_t* pCtrl, 
   PHONE_t* pPhone);
extern IFX_return_t Common_DTMF_SIG_Disable(CTRL_STATUS_t* pCtrl, 
   PHONE_t* pPhone);
#endif /* TAPI_VERSION4 */
extern IFX_return_t Common_LineFeedSet(PHONE_t* pPhone, TD_LINE_MODE_t mode);
extern IFX_return_t Common_TonePlay(CTRL_STATUS_t* pCtrl, PHONE_t* pPhone,
                                    IFX_uint32_t nTone);

extern IFX_return_t Common_ToneStop(CTRL_STATUS_t* pCtrl, PHONE_t* pPhone);
extern IFX_return_t Common_LEC_Enable(CTRL_STATUS_t* pCtrl, PHONE_t* pPhone);
extern IFX_return_t Common_LEC_Disable(CTRL_STATUS_t* pCtrl, PHONE_t* pPhone);
extern IFX_return_t Common_LEC_On_PCM_Ch(IFX_int32_t nPCM_Ch, 
   IFX_enDis_t nEnable, BOARD_t* pBoard);
extern IFX_return_t Common_SetVad(PHONE_t* pPhone, BOARD_t* pBoard,
                                   IFX_int32_t nVadCfg);
extern IFX_void_t Common_StripPath (IFX_char_t** sPath);

extern TAPIDEMO_DEVICE_t * Common_GetDevice_CPU(TAPIDEMO_DEVICE_t* pDevice);
extern TAPIDEMO_DEVICE_t * Common_GetDevice_FXO(TAPIDEMO_DEVICE_t* pDevice,
                                                 FXO_TYPE_t nFxoType);
extern IFX_return_t Common_AddDevice(BOARD_t* pBoard,
                                      TAPIDEMO_DEVICE_t* pDevice);
extern IFX_return_t Common_RemoveDeviceList(BOARD_t* pBoard);
extern IFX_char_t * Common_Enum2Name(IFX_int32_t nEnum, 
                                    TD_ENUM_2_NAME_t *pEnumName);
extern IFX_int32_t Common_Name2Enum(IFX_char_t* pString,
                                    TD_ENUM_2_NAME_t *pEnumName,
                                    IFX_boolean_t bPrintError);
IFX_void_t Common_CheckTraceLen(IFX_char_t* pBuffer);
#endif /* _COMMON_H */

