


/******************************************************************************

                               Copyright (c) 2008
                            Infineon Technologies AG
                     Am Campeon 1-12; 81726 Munich, Germany

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

  ****************************************************************************
   Module      : lib_tapi_signal.c
   Desription  :
*******************************************************************************/
/**
   \file lib_tapi_signal.c
   \date
   \brief Enabling/disable signal detections, functions to prepare
          phone for fax/modem transmission.
*/

/* ============================= */
/* Includes                      */
/* ============================= */

#ifdef VXWORKS
   #include "vxWorks.h"
   #include <vxWorks.h>
   #include <stdio.h>
   #include <stdlib.h>
   #include <ioLib.h>
   #include <sys/ioctl.h>
   #include <logLib.h>
   #include <string.h>
#endif /* VXWORKS */

#include "ifx_types.h"
#ifdef LINUX
#include "tapidemo_config.h"
#endif

#include "lib_tapi_signal.h"
#include "tapidemo.h"
#include "state_trans.h"
#include "voip.h"
#include "abstract.h"

/* ============================= */
/* Defines                       */
/* ============================= */

/* ============================= */
/* Local variable definition     */
/* ============================= */

#ifdef TD_FAX_MODEM
/** Array used to assign names to the signal values reported by the TAPI */
static TD_ENUM_2_NAME_t TD_rgSig[] =
{
   {IFX_TAPI_SIG_NONE, "IFX_TAPI_SIG_NONE"},
   {IFX_TAPI_SIG_DISRX, "IFX_TAPI_SIG_DISRX"},
   {IFX_TAPI_SIG_DISTX, "IFX_TAPI_SIG_DISTX"},
   {IFX_TAPI_SIG_DIS, "IFX_TAPI_SIG_DIS"},
   {IFX_TAPI_SIG_CEDRX, "IFX_TAPI_SIG_CEDRX"},
   {IFX_TAPI_SIG_CEDTX, "IFX_TAPI_SIG_CEDTX"},
   {IFX_TAPI_SIG_CED, "IFX_TAPI_SIG_CED"},
   {IFX_TAPI_SIG_CNGFAXRX, "IFX_TAPI_SIG_CNGFAXRX"},
   {IFX_TAPI_SIG_CNGFAXTX, "IFX_TAPI_SIG_CNGFAXTX"},
   {IFX_TAPI_SIG_CNGFAX, "IFX_TAPI_SIG_CNGFAX"},
   {IFX_TAPI_SIG_CNGMODRX, "IFX_TAPI_SIG_CNGMODRX"},
   {IFX_TAPI_SIG_CNGMODTX, "IFX_TAPI_SIG_CNGMODTX"},
   {IFX_TAPI_SIG_CNGMOD, "IFX_TAPI_SIG_CNGMOD"},
   {IFX_TAPI_SIG_PHASEREVRX, "IFX_TAPI_SIG_PHASEREVRX"},
   {IFX_TAPI_SIG_PHASEREVTX, "IFX_TAPI_SIG_PHASEREVTX"},
   {IFX_TAPI_SIG_PHASEREV, "IFX_TAPI_SIG_PHASEREV"},
   {IFX_TAPI_SIG_AMRX, "IFX_TAPI_SIG_AMRX"},
   {IFX_TAPI_SIG_AMTX, "IFX_TAPI_SIG_AMTX"},
   {IFX_TAPI_SIG_AM, "IFX_TAPI_SIG_AM"},
   {IFX_TAPI_SIG_TONEHOLDING_ENDRX, "IFX_TAPI_SIG_TONEHOLDING_ENDRX"},
   {IFX_TAPI_SIG_TONEHOLDING_ENDTX, "IFX_TAPI_SIG_TONEHOLDING_ENDTX"},
   {IFX_TAPI_SIG_TONEHOLDING_END, "IFX_TAPI_SIG_TONEHOLDING_END"},
   {IFX_TAPI_SIG_CEDENDRX, "IFX_TAPI_SIG_CEDENDRX"},
   {IFX_TAPI_SIG_CEDENDTX, "IFX_TAPI_SIG_CEDENDTX"},
   {IFX_TAPI_SIG_CEDEND, "IFX_TAPI_SIG_CEDEND"},
   {IFX_TAPI_SIG_CPTD, "IFX_TAPI_SIG_CPTD"},
   {IFX_TAPI_SIG_V8BISRX, "IFX_TAPI_SIG_V8BISRX"},
   {IFX_TAPI_SIG_V8BISTX, "IFX_TAPI_SIG_V8BISTX"},
   {IFX_TAPI_SIG_CIDENDTX, "IFX_TAPI_SIG_CIDENDTX"},
   {IFX_TAPI_SIG_DTMFTX, "IFX_TAPI_SIG_DTMFTX"},
   {IFX_TAPI_SIG_DTMFRX, "IFX_TAPI_SIG_DTMFRX"},
   {TD_MAX_ENUM_ID, "IFX_TAPI_SIG_t"}
};

static TD_ENUM_2_NAME_t TD_rgSigExt[] =
{
   {IFX_TAPI_SIG_EXT_NONE, "IFX_TAPI_SIG_EXT_NONE"},
   {IFX_TAPI_SIG_EXT_V21LRX, "IFX_TAPI_SIG_EXT_V21LRX"},
   {IFX_TAPI_SIG_EXT_V21LTX, "IFX_TAPI_SIG_EXT_V21LTX"},
   {IFX_TAPI_SIG_EXT_V21L, "IFX_TAPI_SIG_EXT_V21L"},
   {IFX_TAPI_SIG_EXT_V18ARX, "IFX_TAPI_SIG_EXT_V18ARX"},
   {IFX_TAPI_SIG_EXT_V18ATX, "IFX_TAPI_SIG_EXT_V18ATX"},
   {IFX_TAPI_SIG_EXT_V18A, "IFX_TAPI_SIG_EXT_V18A"},
   {IFX_TAPI_SIG_EXT_V27RX, "IFX_TAPI_SIG_EXT_V27RX"},
   {IFX_TAPI_SIG_EXT_V27TX, "IFX_TAPI_SIG_EXT_V27TX"},
   {IFX_TAPI_SIG_EXT_V27, "IFX_TAPI_SIG_EXT_V27"},
   {IFX_TAPI_SIG_EXT_BELLRX, "IFX_TAPI_SIG_EXT_BELLRX"},
   {IFX_TAPI_SIG_EXT_BELLTX, "IFX_TAPI_SIG_EXT_BELLTX"},
   {IFX_TAPI_SIG_EXT_BELL, "IFX_TAPI_SIG_EXT_BELL"},
   {IFX_TAPI_SIG_EXT_V22RX, "IFX_TAPI_SIG_EXT_V22RX"},
   {IFX_TAPI_SIG_EXT_V22TX, "IFX_TAPI_SIG_EXT_V22TX"},
   {IFX_TAPI_SIG_EXT_V22, "IFX_TAPI_SIG_EXT_V22"},
   {IFX_TAPI_SIG_EXT_V22ORBELLRX, "IFX_TAPI_SIG_EXT_V22ORBELLRX"},
   {IFX_TAPI_SIG_EXT_V22ORBELLTX, "IFX_TAPI_SIG_EXT_V22ORBELLTX"},
   {IFX_TAPI_SIG_EXT_V22ORBELL, "IFX_TAPI_SIG_EXT_V22ORBELL"},
   {IFX_TAPI_SIG_EXT_V32ACRX, "IFX_TAPI_SIG_EXT_V32ACRX"},
   {IFX_TAPI_SIG_EXT_V32ACTX, "IFX_TAPI_SIG_EXT_V32ACTX"},
   {IFX_TAPI_SIG_EXT_V32AC, "IFX_TAPI_SIG_EXT_V32AC"},
   {IFX_TAPI_SIG_EXT_CASBELLRX, "IFX_TAPI_SIG_EXT_CASBELLRX"},
   {IFX_TAPI_SIG_EXT_CASBELLTX, "IFX_TAPI_SIG_EXT_CASBELLTX"},
   {IFX_TAPI_SIG_EXT_CASBELL, "IFX_TAPI_SIG_EXT_CASBELL"},
   {IFX_TAPI_SIG_EXT_V21HRX, "IFX_TAPI_SIG_EXT_V21HRX"},
   {IFX_TAPI_SIG_EXT_V21HTX, "IFX_TAPI_SIG_EXT_V21HTX"},
   {IFX_TAPI_SIG_EXT_V21H, "IFX_TAPI_SIG_EXT_V21H"},
   {IFX_TAPI_SIG_EXT_VMD, "IFX_TAPI_SIG_EXT_VMD"},
   {TD_MAX_ENUM_ID, "IFX_TAPI_SIG_EXT_t"}
};

IFX_return_t SIGNAL_PrintNames(IFX_TAPI_SIG_DETECTION_t *pSignal, 
                               IFX_char_t* pTraceIndention);
#endif /* defined TD_FAX_MODEM */

/* ============================= */
/* Local function definition     */
/* ============================= */

/**
   Function to enable the detection of a signal and report via TAPI exception.

   \param   fd      -  file descriptor to tapi channel
   \param   pSignal  -  pointer to signal to enable
   \param   pPhone   - pointer to phone

   \return  status (IFX_SUCCESS or IFX_ERROR)

   \remark
*/
IFX_return_t SIGNAL_EnableDetection(IFX_int32_t fd, 
                                    IFX_TAPI_SIG_DETECTION_t *pSignal,
                                    PHONE_t* pPhone)
{
   IFX_return_t ret = IFX_SUCCESS;
#ifdef TD_FAX_MODEM
   IFX_int32_t nDataCh = -1;

   TD_PTR_CHECK(pPhone, IFX_ERROR);
   TD_PTR_CHECK(pSignal, IFX_ERROR);
   TD_PARAMETER_CHECK((fd <= 0), fd, IFX_ERROR);

   TRACE(TAPIDEMO, DBG_LEVEL_LOW,
         ("Phone No %d: Enable signal detectors:\n", pPhone->nPhoneNumber));

   SIGNAL_PrintNames(pSignal, pPhone->pBoard->pIndentPhone);
#ifndef TAPI_VERSION4
   /* get channel number */
   nDataCh = ABSTRACT_GetDataChannelOfFD(fd, pPhone->pBoard);
#endif /* TAPI_VERSION4 */
   TD_PARAMETER_CHECK(((nDataCh < 0) || (nDataCh >= pPhone->pBoard->nMaxCoderCh)),
                       nDataCh, IFX_ERROR);

   /* check if there is any signal for enabling. */
   if ((pSignal->sig != 0) || (pSignal->sig_ext != 0))
   {
      ret = ioctl(fd, IFX_TAPI_SIG_DETECT_ENABLE, (IFX_int32_t) pSignal);
      TD_DRV_FUNC_ASSERT(ret, pPhone, IFX_TAPI_SIG_DETECT_ENABLE,
                         0, nDataCh, TD_RETURN);
   }
#endif /* TD_FAX_MODEM */
   return ret;
} /* SIGNAL_EnableDetection() */


/**
   Function to disable the detection of a signal

   \param   fd       -  file descriptor to tapi channel
   \param   pSignal  -  pointer to signal to disable
   \param   pPhone   - pointer to phone

   \return  status (IFX_SUCCESS or IFX_ERROR)

   \remark
*/
IFX_return_t SIGNAL_DisableDetection(IFX_int32_t fd, 
                                     IFX_TAPI_SIG_DETECTION_t *pSignal,
                                     PHONE_t* pPhone)
{
   IFX_return_t ret = IFX_SUCCESS;
#ifdef TD_FAX_MODEM
   IFX_int32_t nDataCh = -1;

   TD_PTR_CHECK(pPhone, IFX_ERROR);
   TD_PTR_CHECK(pSignal, IFX_ERROR);
   TD_PARAMETER_CHECK((fd <= 0), fd, IFX_ERROR);

   TRACE(TAPIDEMO, DBG_LEVEL_LOW,
         ("Phone No %d: Disable signal detectors:\n", pPhone->nPhoneNumber));
   SIGNAL_PrintNames(pSignal, pPhone->pBoard->pIndentPhone);

#ifndef TAPI_VERSION4
   /* get channel number */
   nDataCh = ABSTRACT_GetDataChannelOfFD(fd, pPhone->pBoard);
#endif /* TAPI_VERSION4 */
   TD_PARAMETER_CHECK(((nDataCh < 0) || 
      (nDataCh >= pPhone->pBoard->nMaxCoderCh)), nDataCh, IFX_ERROR);

   /* check if there is any signal for disabling. */
   if ((pSignal->sig != 0) || (pSignal->sig_ext != 0))
   {
      ret = ioctl(fd, IFX_TAPI_SIG_DETECT_DISABLE, (IFX_int32_t) pSignal);
      TD_DRV_FUNC_ASSERT(ret, pPhone, IFX_TAPI_SIG_DETECT_DISABLE,
                         0, nDataCh, TD_RETURN);
   }
#endif /* TD_FAX_MODEM */
   return ret;
} /* SIGNAL_DisableDetection() */


/**
   Function to switch a transmission channel to a uncompressed codec.

   \param   pPhone   -  pointer to PHONE_t
   \param   pConn    - pointer to phone connection

   \return  status (IFX_SUCCESS or IFX_ERROR)

   \remark
   This function switches a transmission channel to G711 coded, sets
   a fixed jitter buffer, disables silence suppresion and comfort noise
   generation. Further event transmission and auto suppresion are turned off.
*/
IFX_return_t SIGNAL_ClearChannel(PHONE_t *pPhone, CONNECTION_t* pConn)
{
   IFX_return_t ret = IFX_SUCCESS;
#ifdef TD_FAX_MODEM
   IFX_TAPI_ENC_CFG_t cod_type_parm = {0};
   IFX_TAPI_JB_CFG_t tapi_jb_conf;
   BOARD_t *pBoard = IFX_NULL;

   TD_PTR_CHECK(pPhone, IFX_ERROR);
   pBoard = pPhone->pBoard;
   /* check if coder channels are avaible and structure allocated */
   if (0 >= pBoard->nMaxCoderCh || IFX_NULL == pBoard->pDataChStat)
   {
      return ret;
   }

   TRACE(TAPIDEMO, DBG_LEVEL_LOW, ("Phone No %d: SIGNAL_ClearChannel()\n", 
                                   pPhone->nPhoneNumber));

   if (pBoard->pDataChStat[pConn->nUsedCh].fCodecStarted == IFX_TRUE)
   {
      /* switch to G.711 A-Law VBD*/
      cod_type_parm.nEncType = IFX_TAPI_ENC_TYPE_ALAW_VBD;
      cod_type_parm.nFrameLen = IFX_TAPI_COD_LENGTH_10;

      pBoard->pDataChStat[pConn->nUsedCh].nCodecType = cod_type_parm.nEncType;
      pBoard->pDataChStat[pConn->nUsedCh].nFrameLen = cod_type_parm.nFrameLen;
      pBoard->pDataChStat[pConn->nUsedCh].fStartCodec = IFX_TRUE;

      /* stop currently started code */
      if (VOIP_StopCodec(pConn->nUsedCh, pConn, pPhone->pBoard) == IFX_ERROR)
         return IFX_ERROR;
      /* start codec with new configuration */
      if (VOIP_StartCodec(pConn->nUsedCh, pConn, pPhone->pBoard) == IFX_ERROR)
         return IFX_ERROR;

      /* set fixed jitter buffer */
      memset(&tapi_jb_conf, 0x00, sizeof(IFX_TAPI_JB_CFG_t));

      tapi_jb_conf.nJbType      =  IFX_TAPI_JB_TYPE_FIXED;
      tapi_jb_conf.nPckAdpt     =  IFX_TAPI_JB_PKT_ADAPT_DATA;
      tapi_jb_conf.nScaling     =  0x16;   /* Scaling factor */
      tapi_jb_conf.nInitialSize =  0x0190; /* Inital JB size 50 ms */
      tapi_jb_conf.nMinSize     =  0x00A0; /* Min.   JB size 20 ms */
      tapi_jb_conf.nMaxSize     =  0x0320; /* Max.   JB size 100 ms */

      ret = ioctl(pPhone->nDataCh_FD, IFX_TAPI_JB_CFG_SET, 
                  (IFX_int32_t) &tapi_jb_conf);
      TD_DRV_FUNC_ASSERT(ret, pPhone, IFX_TAPI_JB_CFG_SET, 0, pPhone->nDataCh, 
                         TD_RETURN);

      /* Disable silence suppression and confort noise generation */
      ret = ioctl(pConn->nUsedCh_FD, IFX_TAPI_ENC_VAD_CFG_SET, 
                  IFX_TAPI_ENC_VAD_NOVAD);
      TD_DRV_FUNC_ASSERT(ret, pPhone, IFX_TAPI_ENC_VAD_CFG_SET, 0, 
                         pConn->nUsedCh, TD_RETURN);
   }
   else
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Phone No %d: Codec not started. (File: %s, line: %d)\n",
             pPhone->nPhoneNumber, __FILE__, __LINE__));
   }
#ifdef ABACUS_TEST
   COM_AbacusTestTraces(pPhone->pBoard->pCtrl, pPhone, pConn,
                        ATTP_CLEAR_CHANNEL);
#endif /* ABACUS_TEST */
#endif /* TD_FAX_MODEM */
   return ret;
} /* SIGNAL_ClearChannel() */

/**
   Function to disable the non-linear-processor of the VINETIC

   \param   pPhone      -  pointer to PHONE

   \return  status (IFX_SUCCESS or IFX_ERROR)

   \remark
*/
IFX_return_t SIGNAL_DisableNLP(PHONE_t *pPhone)
{
   IFX_return_t ret = IFX_SUCCESS;
#ifdef TD_FAX_MODEM
   IFX_int32_t ch = -1;
   IFX_int32_t fd = -1;
   IFX_TAPI_WLEC_CFG_t tapi_lec_conf;
   BOARD_t *pBoard = IFX_NULL;

   TD_PTR_CHECK(pPhone, IFX_ERROR);

   pBoard = pPhone->pBoard;
   ch = pPhone->nPhoneCh;
   fd = pPhone->nPhoneCh_FD;
   TRACE(TAPIDEMO, DBG_LEVEL_LOW,
         ("Phone No %d: SIGNAL_DisableNLP()\n", pPhone->nPhoneNumber));

   memset(&tapi_lec_conf, 0, sizeof(IFX_TAPI_WLEC_CFG_t));

   /* get wlec configuration */
   ret = ioctl(pPhone->nPhoneCh_FD, IFX_TAPI_WLEC_PHONE_CFG_GET,
               (IFX_int32_t) &tapi_lec_conf);
   TD_DRV_FUNC_ASSERT(ret, pPhone, IFX_TAPI_WLEC_PHONE_CFG_GET,
                      0, pPhone->nPhoneCh, TD_RETURN);

   /* Switch the NLP off*/
   tapi_lec_conf.nType = IFX_TAPI_WLEC_TYPE_NFE;
   tapi_lec_conf.bNlp = IFX_TAPI_WLEC_NLP_OFF;
   ret = ioctl(pPhone->nPhoneCh_FD, IFX_TAPI_WLEC_PHONE_CFG_SET,
               (IFX_int32_t) &tapi_lec_conf);
   TD_DRV_FUNC_ASSERT(ret, pPhone, IFX_TAPI_WLEC_PHONE_CFG_SET,
                      0, pPhone->nPhoneCh, TD_RETURN);

#ifdef ABACUS_TEST
   COM_AbacusTestTraces(pPhone->pBoard->pCtrl, pPhone, &pPhone->rgoConn[0],
                        ATTP_DISABLE_NLP);
#endif /* ABACUS_TEST */
#endif /* TD_FAX_MODEM */
   return ret;
} /* SIGNAL_DisableNLP() */


/* ============================= */
/* Global function definition    */
/* ============================= */

#ifdef TD_FAX_MODEM
/**
   Helper function to display the name assigned to a signal value

   \param   pSignal  -  pointer to structure wth signals

   \return  status (IFX_SUCCESS or IFX_ERROR)
*/
IFX_return_t SIGNAL_PrintNames(IFX_TAPI_SIG_DETECTION_t *pSignal, 
                               IFX_char_t* pTraceIndention)
{
   IFX_TAPI_SIG_DETECTION_t stSignal={0};
   IFX_uint32_t nTempSignal=0;
   IFX_uint32_t nMask = 0;

   TD_PTR_CHECK(pSignal, IFX_ERROR)
   memcpy((IFX_char_t *)&stSignal, (IFX_char_t *)pSignal, 
           sizeof(IFX_TAPI_SIG_DETECTION_t));

   /* set first bit in signal mask */
   nMask = 1;
   while (stSignal.sig != 0)
   {
      /* check if nMask flag is set */
      nTempSignal = stSignal.sig & nMask;
      if (nTempSignal != 0)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("%s- %s\n", pTraceIndention, 
                Common_Enum2Name(nTempSignal, TD_rgSig)));
      }
      /* remove this signal flag */
      stSignal.sig &= ~nMask;
      /* set mask to next signal flag */
      nMask <<= 1;
   } /* while() */

   /* set first bit in signal mask */
   nMask = 1;
   while (stSignal.sig_ext != 0)
   {
      /* check if nMask flag is set */
      nTempSignal = stSignal.sig_ext & nMask;
      if (nTempSignal != 0)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("%s- %s\n", pTraceIndention,
                Common_Enum2Name(nTempSignal, TD_rgSigExt)));
      }
      /* remove this signal flag */
      stSignal.sig_ext &= ~nMask;
      /* set mask to next signal flag */
      nMask <<= 1;
   } /* while() */

   return IFX_SUCCESS;
} /* SIGNAL_PrintNames() */
#endif /* TD_FAX_MODEM */

/**
   Setup initial signal detection for fax/modem transmission.

   \param   fd       - file descriptor to tapi channel
   \param   pPhone   - pointer to phone

   \return  status (IFX_SUCCESS or IFX_ERROR)

   \remark
   This function is called when the calling subscriber moves to active state.
*/
IFX_return_t SIGNAL_Setup(IFX_int32_t fd, PHONE_t* pPhone)
{
   IFX_return_t ret = IFX_SUCCESS;
#ifdef TD_FAX_MODEM

   TD_PTR_CHECK(pPhone, IFX_ERROR);
   TD_PARAMETER_CHECK((fd <= 0), fd, IFX_ERROR);

   /* signals that detection should be enabled */
   pPhone->stSignal.sig = IFX_TAPI_SIG_CNGMOD | IFX_TAPI_SIG_CED |
                          IFX_TAPI_SIG_CEDEND | IFX_TAPI_SIG_AM |
                          IFX_TAPI_SIG_V8BISRX | IFX_TAPI_SIG_V8BISTX |
                          IFX_TAPI_SIG_DIS | IFX_TAPI_SIG_CNGFAX |
                          IFX_TAPI_SIG_PHASEREV;
   pPhone->stSignal.sig_ext = IFX_TAPI_SIG_EXT_V21L | IFX_TAPI_SIG_EXT_V18A |
      IFX_TAPI_SIG_EXT_V27 | IFX_TAPI_SIG_EXT_BELL |
      IFX_TAPI_SIG_EXT_V22 | IFX_TAPI_SIG_EXT_V22ORBELL |
      IFX_TAPI_SIG_EXT_V32AC | IFX_TAPI_SIG_EXT_CASBELL |
      IFX_TAPI_SIG_EXT_V21H;
   ret = SIGNAL_EnableDetection(fd, &pPhone->stSignal, pPhone);
#endif /* TD_FAX_MODEM */
   return ret;
} /* SIGNAL_Setup() */

/**
   reset signal handler
   This function is called e.g. on hook interrupts

   \param   fd  -  file descriptor to tapi data channel
   \param   pPhone   - pointer to phone

   \return  status (IFX_SUCCESS or IFX_ERROR)
*/
IFX_return_t SIGNAL_HandlerReset(IFX_int32_t fd, PHONE_t* pPhone)
{
   IFX_return_t ret= IFX_SUCCESS;
#ifdef TD_FAX_MODEM

   TD_PTR_CHECK(pPhone, IFX_ERROR);
   TD_PARAMETER_CHECK((fd <= 0), fd, IFX_ERROR);

   ret = SIGNAL_DisableDetection(fd, &pPhone->stSignal, pPhone);
   pPhone->stSignal.sig = 0;
   pPhone->stSignal.sig_ext = 0;
#endif /* TD_FAX_MODEM */
   return(ret);
} /* SIGNAL_HandlerReset() */

/**
   Function to restore transmission channel from a uncompressed codec.

   \param   pPhone      -  pointer to PHONE

   \return  status (IFX_SUCCESS or IFX_ERROR)

   \remark
   This function switches a transmission channel to codec used previously by it
*/
IFX_return_t SIGNAL_RestoreChannel(PHONE_t *pPhone, CONNECTION_t* pConn)
{
   IFX_return_t ret = IFX_SUCCESS;
#ifdef TD_FAX_MODEM
   IFX_TAPI_ENC_CFG_t cod_type_parm = {0};
   BOARD_t *pBoard = IFX_NULL;

   TD_PTR_CHECK(pPhone, IFX_ERROR);
   pBoard = pPhone->pBoard;
   /* check if coder channels are avaible and structure allocated */
   if (0 >= pBoard->nMaxCoderCh || IFX_NULL == pBoard->pDataChStat)
   {
      return ret;
   }

   /* Check if channel setting were changed. */
   if (IFX_TRUE != pBoard->pDataChStat[pConn->nUsedCh].fRestoreChannel)
   {
      return ret;
   }

   TRACE(TAPIDEMO, DBG_LEVEL_LOW, ("Phone No %d: Restore Channel\n", 
                                   pPhone->nPhoneNumber));

   /* Restore previous codec settings */
   pBoard->pDataChStat[pConn->nUsedCh].nCodecType = 
      pBoard->pDataChStat[pPhone->nDataCh].nPrevCodecType;
   pBoard->pDataChStat[pConn->nUsedCh].nFrameLen = 
      pBoard->pDataChStat[pPhone->nDataCh].nPrevFrameLen;
   pBoard->pDataChStat[pConn->nUsedCh].fStartCodec = IFX_TRUE;

   cod_type_parm.nEncType = pBoard->pDataChStat[pConn->nUsedCh].nPrevCodecType;
   cod_type_parm.nFrameLen = pBoard->pDataChStat[pConn->nUsedCh].nPrevFrameLen;
   /* Change codec to previous. */
   ret = ioctl(pConn->nUsedCh_FD, IFX_TAPI_ENC_CFG_SET,(int) &cod_type_parm);
   TD_DRV_FUNC_ASSERT(ret, pPhone, IFX_TAPI_ENC_CFG_SET,
                      0, pConn->nUsedCh, TD_RETURN);

   /* Restore previous settings for Jitter buffer */
    ret = ioctl(pConn->nUsedCh_FD, IFX_TAPI_JB_CFG_SET,
       (IFX_int32_t) &pBoard->pDataChStat[pConn->nUsedCh].tapi_jb_PrevConf);
    TD_DRV_FUNC_ASSERT(ret, pPhone, IFX_TAPI_JB_CFG_SET,
                      0, pConn->nUsedCh, TD_RETURN);

    /* Restore previous settings for LEC */
    ret = ioctl(pPhone->nPhoneCh_FD, IFX_TAPI_WLEC_PHONE_CFG_SET,
                (IFX_int32_t) &pBoard->pDataChStat[pConn->nUsedCh].oPrevLECcfg);
    TD_DRV_FUNC_ASSERT(ret, pPhone, IFX_TAPI_WLEC_PHONE_CFG_SET,
                      0, pConn->nUsedCh, TD_RETURN);
#ifdef ABACUS_TEST
    COM_AbacusTestTraces(pPhone->pBoard->pCtrl, pPhone, pConn,
                         ATTP_LEC_ENABLE);
#endif /* ABACUS_TEST */

   /* enable silence suppression and confort noise generation */
   ret = ioctl(pConn->nUsedCh_FD, IFX_TAPI_ENC_VAD_CFG_SET,
               pBoard->pCtrl->pProgramArg->nVadCfg);
   TD_DRV_FUNC_ASSERT(ret, pPhone, IFX_TAPI_ENC_VAD_CFG_SET,
                      0, pConn->nUsedCh, TD_RETURN);

#if 0
   /* Enable high-pass filter */
   ret = ioctl(pConn->nUsedCh_FD, IFX_TAPI_COD_DEC_HP_SET, IFX_TRUE);
   TD_DRV_FUNC_ASSERT(ret, pPhone, IFX_TAPI_COD_DEC_HP_SET,
                      0, pConn->nUsedCh, TD_RETURN);
#endif

   if (IFX_SUCCESS == ret)
   {
      pBoard->pDataChStat[pConn->nUsedCh].fRestoreChannel = IFX_FALSE;
#ifdef ABACUS_TEST
      COM_AbacusTestTraces(pPhone->pBoard->pCtrl, pPhone, pConn,
                           ATTP_RESTORE_CHANNEL);
#endif /* ABACUS_TEST */
   }
#endif /* TD_FAX_MODEM */
   return ret;
} /* SIGNAL_RestoreChannel() */

/**
   Store current settings like used coder, LEC configuration etc. It is used
   for fax/modem to restore configuration when fax/modem transmision is
   finished.

   \param  pPhone - pointer to PHONE
   \param  pConn  - pointer to phone connection

   \return status (IFX_SUCCESS or IFX_ERROR)
   \remark
*/
IFX_return_t SIGNAL_StoreSettings(PHONE_t *pPhone, CONNECTION_t* pConn)
{
   IFX_return_t ret = IFX_SUCCESS;
   BOARD_t *pBoard = IFX_NULL;
   IFX_int32_t ch = -1;

   if (IFX_NULL == pPhone)
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Invalid input argument(s). (pPhone=%p) "
             "(File: %s, line: %d)\n",
             pPhone, __FILE__, __LINE__));
      return IFX_ERROR;
   }
   /* get board */
   pBoard = pPhone->pBoard;
   /* get channel */
   ch = pPhone->nDataCh;
   /* check if coder channels are avaible and structure allocated */
   if ((0 >= pBoard->nMaxCoderCh) || (IFX_NULL == pBoard->pDataChStat))
   {
      return ret;
   }

   /* Save current configuration*/
   pBoard->pDataChStat[pConn->nUsedCh].fRestoreChannel = IFX_TRUE;

   /* Save codec and length */
   pBoard->pDataChStat[pConn->nUsedCh].nPrevCodecType =
       pBoard->pDataChStat[pConn->nUsedCh].nCodecType;
   pBoard->pDataChStat[pConn->nUsedCh].nPrevFrameLen =
       pBoard->pDataChStat[pConn->nUsedCh].nFrameLen;
   /* Save LEC settings of phone ch in data channel statistics*/
   ret = ioctl(pConn->nUsedCh_FD, IFX_TAPI_WLEC_PHONE_CFG_GET,
               &pBoard->pDataChStat[pConn->nUsedCh].oPrevLECcfg);
   if (ret == IFX_ERROR)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Phone No %d: Error, IFX_TAPI_WLEC_PHONE_CFG_SET failed.\n",
             pPhone->nPhoneNumber));
#ifdef TAPI_VERSION4
      Common_ErrorHandle(pBoard->nDevCtrl_FD, pBoard, 
         pBoard->pDataChStat[pConn->nUsedCh].nDev);
#else
      Common_ErrorHandle(pBoard->nDevCtrl_FD, pBoard, pPhone->nDev);
#endif
   }

   return ret;
} /* SIGNAL_StoreSettings() */

#if defined(TAPI_VERSION4) && defined(EASY336)
#if 0
IFX_return_t SIGNAL_GetCoderRu(RM_SVIP_RU_t* pCodRu, PHONE_t* pPhone)
{
   TD_PTR_CHECK(pPhone, IFX_ERROR);
   TD_PTR_CHECK(pCodRu, IFX_ERROR);
   TD_PARAMETER_CHECK((1 != pPhone->nVoipIDs), pPhone->nVoipIDs);

   if (IFX_ERROR == SVIP_RM_VoipIdRUCodGet(pPhone->pVoipIDs[pPhone->nVoipIDs-1],
                                            pCodRu))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Error, Phone No %d: SVIP_RM_VoipIdRUCodGet failed. "
             "(File: %s, line: %d)\n",
             pPhone->nPhoneNumber, __FILE__, __LINE__));
   }

}
#endif /* 0 */
#endif /* defined(TAPI_VERSION4) && defined(EASY336) */

