


/******************************************************************************

                               Copyright (c) 2008
                            Infineon Technologies AG
                     Am Campeon 1-12; 81726 Munich, Germany

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

  ****************************************************************************/

/**
   \file tapidemo.c
   \date 2003-03-24
   \brief Demonstrates a simple PBX with voice over IP on Linux or VxWorks OS.

   \note Changes:
      Date: 28.11.2005 it is working on new board with new drivers, kernel.
                       CID structure changed.
                       Removed EXCHFD_SUPPORT, its obsolete.
                       Extracted QOS, TAPI SIGNALLING into different files.
      Date: 15.12.2005 Added conference support.
      Date: 21.04.2006 Major changes, added PCM, Features support. Also added
                       Event handling (bitfields or messages).
      Date: 07.12.2006 Major changes, support for FXO, DANUBE with CPE
                       extension board, multiple boards, parsing argumens in
                       VxWorks.
      Date: 31.05.2007 Added support for board DuSLIC-XT.
*/


/* ============================= */
/* Includes                      */
/* ============================= */

#ifdef VXWORKS
#include <selectLib.h>
#include <inetLib.h>
#endif

#ifdef LINUX
#include "tapidemo_config.h"
#endif

#include "tapidemo_version.h"
#include "tapidemo.h"
#include "abstract.h"
#include "analog.h"
#include "cid.h"
#include "conference.h"
#include "common.h"
#include "event_handling.h"
#include "feature.h"
#include "parse_cmd_arg.h"
#include "pcm.h"
#include "qos.h"
#include "state_trans.h"
#include "lib_tapi_signal.h"
#include "voip.h"

#ifdef FXO
   #include "common_fxo.h"
#endif /* FXO */

#ifdef EASY336
#include "lib_svip.h"
#include "lib_svip_rm.h"
#include "board_easy336.h"
#ifdef WITH_VXT
#include "board_xt16.h"
#endif /* WITH_VXT */
#endif /* EASY336 */

#ifdef XT16
#include "lib_svip.h"
#include "board_easy336.h"
#include "board_xt16.h"
#endif

#if (defined(TD_T38_FAX_TEST) && defined(TD_FAX_T38) && defined(AR9))
#include "./t38_fax_test/td_t38_fax_test.h"
#endif /* (defined(TD_T38_FAX_TEST) && defined(TD_FAX_T38) && defined(AR9)) */

/* ============================= */
/* Debug interface               */
/* ============================= */

CREATE_TRACE_GROUP(TAPIDEMO);

/* ============================= */
/* Global Structures             */
/* ============================= */
#ifdef ABACUS_TEST
IFX_int32_t nVerifySystemInitializtion;
/** pointer to tapidemo control structure */
CTRL_STATUS_t* pCtrlStatus = IFX_NULL;
#endif /* ABACUS_TEST */

/* ============================= */
/* Global function declaration   */
/* ============================= */

/* ============================= */
/* Local variable definition     */
/* ============================= */

const IFX_char_t EOL = '\n';

/** Holding information about each connection */
static CTRL_STATUS_t oCtrlStatus;
/** Holding program arguments */
static PROGRAM_ARG_t oProgramArg;

/** List of combinations */
const BOARD_COMBINATION_t BOARD_COMBINATIONS[MAX_COMBINATIONS] =
{
   /* comb idx, board cnt, list of boards
                           {board id, type, sub type, init (normal/gpio) } */
   { ONE_CPE, 1, { {1, TYPE_VINETIC},
                   {0}, {0}, {0} } },
   { ONE_VIP, 1, { {1, TYPE_VINETIC},
                   {0}, {0}, {0} } },
   { ONE_VMMC, 1, { {1, TYPE_DANUBE},
                    {0}, {0}, {0} } },
   { ONE_VMMC_ONE_CPE, 2, { {1, TYPE_DANUBE},
                            {1, TYPE_VINETIC},
                            {0}, {0} } },
   { ONE_DXT, 1, { {1, TYPE_DUSLIC_XT},
                   {0}, {0}, {0} } },
   { TWO_DXTS, 2, { {1, TYPE_DUSLIC_XT},
                    {2, TYPE_DUSLIC_XT},
                    {0}, {0} } },
   { ONE_VMMC_VINAX, 1, { {1, TYPE_VINAX},
                    {0}, {0}, {0} } },
   { ONE_VMMC_EASY508XX, 1, { {1, TYPE_EASY508XX},
                    {0}, {0}, {0} } },
   { X_SVIP, 1, { {1, TYPE_SVIP},
                  {0}, {0}, {0} } },
   { X_SVIP_X_XT16, 2, { {1, TYPE_SVIP},
                         {1, TYPE_XT16},
                         {0}, {0} } },
   { X_XT16, 1, { {1, TYPE_XT16},
                    {0}, {0}, {0} } },
   { ONE_EASY50712_ONE_TERIDIAN73M1866, 2, { {1, TYPE_DANUBE},
                                           {1, TYPE_TERIDIAN},
                                           {0}, {0} } },
   { ONE_EASY508XX_ONE_TERIDIAN73M1866, 2, { {1, TYPE_EASY508XX},
                                           {1, TYPE_TERIDIAN},
                                           {0}, {0} } },
   { ONE_EASY80800_ONE_TERIDIAN73M1866, 2, { {1, TYPE_VINAX},
                                           {1, TYPE_TERIDIAN},
                                           {0}, {0} } },
   { ONE_EASY80910, 1, { {1, TYPE_VR9},
                       {0}, {0}, {0} } },
};

/** Contains board combination names. */
const IFX_char_t* BOARD_COMB_NAMES[MAX_COMBINATIONS] =
{
   "EASY3332",
   "EASY334",
   "EASY50712",
   "EASY50712 with EASY50510",
   "EASY3201",
   "EASY3201 with EASY3111",
   "EASY80800",
   "EASY508xx",
   "EASY336",
   "EASY336 and xT-16s",
   "ONE xT-16",
   "EASY50712 with TERIDIAN73M1866",
   "EASY508xx with TERIDIAN73M1866",
   "EASY80800 with TERIDIAN73M1866",
   "EASY80910"
};

#ifdef EASY336
#ifndef ABACUS_TEST
/** PCM configuration */
static RM_SVIP_PCM_CFG_t pcm_cfg;
#else /* ABACUS_TEST */
/** PCM configuration */
RM_SVIP_PCM_CFG_t pcm_cfg;
#endif /* ABACUS_TEST */
#endif

#ifdef LINUX
/** Max log's length */
enum { MAX_LOG_LEN = 2048 };

/** Max message's length for external syslog: 2048+5 
    Look into function TAPIDEMO_SendLogRemote() for more details */
enum { MAX_EXT_SYSLOG_MSG_LEN = 2053 };

#endif

/** Holding trace redirection settings */
TRACE_REDIRECTION_CTRL_t oTraceRedirection;

/** Keep descriptor of pipe, which will be used to communication between process */
IFX_int32_t nPipeFd[2];

/* ============================= */
/* Local function declaration    */
/* ============================= */
static IFX_return_t TAPIDEMO_Setup(BOARD_t* pBoard,
                                   CTRL_STATUS_t* pCtrl,
                                   IFX_int32_t* nPhoneNum,
                                   IFX_int32_t* pFXO_Num);
static IFX_return_t TAPIDEMO_StateMachine(CTRL_STATUS_t* pCtrl);
IFX_int32_t TAPIDEMO_InitAdminSocket(IFX_void_t);
IFX_return_t TAPIDEMO_HandleCallSeq(CTRL_STATUS_t* pCtrl);
#if defined(LINUX)
IFX_int32_t TAPIDEMO_InitSyslogSocket(IFX_void_t);
IFX_return_t TAPIDEMO_SendLogRemote(IFX_char_t* pBuffer, IFX_int32_t nSyslogSocket);
IFX_return_t TAPIDEMO_SendLog(IFX_char_t* pMsg);
IFX_return_t TAPIDEMO_PreapareLogRedirection(IFX_void_t);
IFX_void_t TAPIDEMO_RunLogRedirection(IFX_int32_t nPipeFd);
IFX_return_t TAPIDEMO_CreateDaemon(CTRL_STATUS_t* pCtrl);
#endif

#if (!defined LINUX) && (!defined TAPI_VERSION4)
extern gethostname();
extern inet_aton();
extern sendto();
extern recvfrom();
extern socket();
extern bind();
#endif /* LINUX */


/* ============================= */
/* Local function definition     */
/* ============================= */

/* SIGNAL HANDLER */
static IFX_return_t TAPIDEMO_RegisterSigHandler(IFX_void_t);
/*static IFX_void_t TAPIDEMO_CleanUp(IFX_int32_t nSignal);*/
static void TAPIDEMO_CleanUp(int nSignal);
#ifdef LINUX
static void TAPIDEMO_ChildCleanUp(int nSignal, siginfo_t* oSigInfo, void* oVoid);
#endif /* LINUX */

#ifdef LINUX

#elif VXWORKS

/** Holds application task ID. */
static IFX_int32_t nTapiDemo_TID = -1;

/**
   This function is called after task is deleted in clear up everything.

   \param pTcb - pointer to deleted task's WIND_TCB
*/
void deleteHook(WIND_TCB* pTcb)
{
   WIND_TCB* evlog_tcb = IFX_NULL;

   evlog_tcb = taskTcb(nTapiDemo_TID);
   if (evlog_tcb == pTcb)
   {
      /* Only call this when our task is killed */
      TRACE(TAPIDEMO, DBG_LEVEL_LOW, ("deleteHook()\n"));

      TAPIDEMO_CleanUp(0);

      if (taskDeleteHookDelete((FUNCPTR) deleteHook) != OK)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Err delete task handler when task is deleted. "
               "(File: %s, line: %d)\n",
               __FILE__, __LINE__));
      }
   }
}
#endif /* VXWORKS */


/**
   Register signal handler which is used when task is deleted from LINUX by
   'kill' or CTRL + C command and from VXWORKS by 'td' command. So task can
   cleanup everything.

   \return IFX_SUCCESS on success, otherwise IFX_ERROR
*/
static IFX_return_t TAPIDEMO_RegisterSigHandler(IFX_void_t)
{
   TRACE(TAPIDEMO, DBG_LEVEL_LOW, ("TAPIDEMO_RegisterSigHandler()\n"));
#ifdef LINUX
   /*sighandler_t sig_handler = TAPIDEMO_CleanUp;*/
   struct sigaction act;

   act.sa_handler = TAPIDEMO_CleanUp;
   act.sa_flags = SA_NODEFER; /* Do not prevent signal from being blocked. */
   if (sigaction(SIGABRT, &act, IFX_NULL) < 0) /* CTRL + C */
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Err adding signal handler. "
            "(File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }
   if (sigaction(SIGQUIT, &act, IFX_NULL) < 0)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Err adding signal handler. "
            "(File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }
   if (sigaction(SIGINT, &act, IFX_NULL) < 0)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Err adding signal handler. "
            "(File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }
   if (sigaction(SIGTERM, &act, IFX_NULL) < 0)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Err adding signal handler. "
            "(File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }
   act.sa_handler = IFX_NULL;
   act.sa_sigaction = TAPIDEMO_ChildCleanUp;
   act.sa_flags |= SA_SIGINFO;
   if (sigaction(SIGCHLD, &act, IFX_NULL) < 0)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Err adding signal handler. "
            "(File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }
#elif VXWORKS
   if (taskDeleteHookAdd((FUNCPTR) deleteHook) != OK)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Err adding delete task handler. "
            "(File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }
#endif /* VXWORKS */

   return IFX_SUCCESS;
}


/**
   Clean up memory after task ends, close all open connections, ...

   \param nSignal - in LINUX signal number in VXWORKS not used
*/
/*static IFX_void_t TAPIDEMO_CleanUp(IFX_int32_t nSignal)*/
static void TAPIDEMO_CleanUp(int nSignal)
{
   IFX_int32_t i = 0;
   IFX_int32_t ch = 0;
   PHONE_t* pPhone = IFX_NULL;
   BOARD_t* pBoard = IFX_NULL;
#ifdef EASY336
   SVIP_RM_Status_t ret;
#endif
#ifdef ABACUS_TEST
   TAPIDEMO_DEVICE_t *pDevice = IFX_NULL;
   TAPIDEMO_DEVICE_CPU_t *pCpuDevice = IFX_NULL;
#endif /* ABACUS_TEST */

#ifdef LINUX
   if ((nSignal == SIGINT) || (nSignal == SIGTERM) || (nSignal == SIGSTOP)
       || (nSignal == SIGABRT) ||(nSignal == SIGQUIT) ||(nSignal == SIGCHLD))
   {
#endif /* LINUX */
      TRACE(TAPIDEMO, DBG_LEVEL_NORMAL, ("TAPIDEMO_CleanUp()\n"));

#if (defined(TD_T38_FAX_TEST) && defined(TD_FAX_T38) && defined(AR9))
      TD_T38_FAX_TEST_DeInit(&oCtrlStatus);
#endif /* (defined(TD_T38_FAX_TEST) && defined(TD_FAX_T38) && defined(AR9)) */

      /* Cleanup stuff. */
      for (i = 0; i < oCtrlStatus.nBoardCnt; i++)
      {
         pBoard = &oCtrlStatus.rgoBoards[i];

         if (pBoard == IFX_NULL)
         {
            continue;
         }

         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("Clean all phones for this board\n"));

         for (ch = 0; ch < pBoard->nMaxAnalogCh; ch++)
         {
            pPhone = &pBoard->rgoPhones[ch];

            if (pPhone == IFX_NULL)
            {
               continue;
            }

            /* End with all connections, ringing, phone calls,
               cleanup mappings. Set phone to state S_READY.*/
            if ((&pPhone->rgoConn[0] != IFX_NULL) &&
                (pPhone->rgStateMachine[FXS_SM].nState != S_READY))
            {
               TRACE(TAPIDEMO, DBG_LEVEL_LOW,
                     ("Remove all connections for this phone\n"));

               pPhone->nIntEvent_FXS = IE_HOOKON;
               while (pPhone->nIntEvent_FXS != IE_NONE)
               {
                  ST_HandleState_FXS(&oCtrlStatus, pPhone,
                                     &pPhone->rgoConn[0]);
               }
            }

            /* Set line feed to disabled */
            Common_LineFeedSet(pPhone, IFX_TAPI_LINE_FEED_DISABLED);

            if ((pPhone->pCID_Msg != IFX_NULL) && (pPhone->pCID_Msg->message != IFX_NULL))
            {
               free(pPhone->pCID_Msg->message);
               pPhone->pCID_Msg->message = IFX_NULL;
            }

            if (pPhone->pCID_Msg != IFX_NULL)
            {
               TRACE(TAPIDEMO, DBG_LEVEL_LOW,
                     ("Release CID messages from phones.\n"));
               free(pPhone->pCID_Msg);
               pPhone->pCID_Msg = IFX_NULL;
            }
         } /* for (ch = 0; ... */

         if (oCtrlStatus.pProgramArg->oArgFlags.nFXO)
         {
            /* Channels are already open */
            for (ch = 0; ch < pBoard->nMaxFxoCh; ch++)
            {
               /* release memory */
               free(pBoard->rgoFXOs[ch].oCID_Msg.message);
            } /* for */
         }

         if (oCtrlStatus.pProgramArg->oArgFlags.nQos)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_LOW, ("QOS_CleanUp()\n"));
            /* QoS Support, cleanup for all boards. */
            QOS_CleanUpSession(&oCtrlStatus, oCtrlStatus.pProgramArg);
         }

         VOIP_Release(oCtrlStatus.pProgramArg, pBoard);

         /* \todo Only one should be set to master? */
         if ((oCtrlStatus.pProgramArg->oArgFlags.nPCM_Master)
             || (oCtrlStatus.pProgramArg->oArgFlags.nPCM_Slave))
         {
            PCM_Release(oCtrlStatus.pProgramArg, pBoard);
         }

#ifndef EASY336
         /* Unmapp all mapped channels */
         TRACE(TAPIDEMO, DBG_LEVEL_LOW, ("Unmap default mapped channels.\n"));
         ABSTRACT_UnmapDefaults(&oCtrlStatus, pBoard);
#endif

#ifdef EASY336
         ret = SVIP_RM_DeInit(&pcm_cfg);
         if (ret != SVIP_RM_Success)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                 ("Resource manager deinit failed with error code %d. "
                  "(File: %s, line: %d)\n",
                  ret, __FILE__, __LINE__));
         }
         free (pcm_cfg.pDevsSVIP);
         free (pcm_cfg.pDevsXT16);
#endif

         TRACE(TAPIDEMO, DBG_LEVEL_LOW, ("Close opened file descriptors.\n"));
         /* Close devices, sockets, ... for all boards. */
         Common_Close_FDs(pBoard);

         if (oCtrlStatus.pProgramArg->oArgFlags.nFXO)
         {
            Common_Close_Fxo_FDs(pBoard);
         }

#ifdef ABACUS_TEST
         if ((pDevice = Common_GetDevice_CPU(pBoard->pDevice)) == IFX_NULL)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                 ("Err, CPU device not found. (File: %s, line: %d)\n",
                  __FILE__, __LINE__));
         }
         else
         {
            pCpuDevice = &pDevice->uDevice.stCpu;

            /* when changing country settings memory is allocated
               for new BBD file name, it should be freed */
            if ((pCpuDevice->pszBBD_CRAM_File != IFX_NULL) &&
                (IFX_TRUE ==pBoard->fBBD_Changed))
            {
               free(pCpuDevice->pszBBD_CRAM_File);
            }
         }
#endif /* ABACUS_TEST */

         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("Remove board %s.\n", pBoard->pszBoardName));
         /* Release boards and free memory. */
         pBoard->RemoveBoard(pBoard);

         pBoard = IFX_NULL;
      } /* for (i = 0; ... */

      TRACE(TAPIDEMO, DBG_LEVEL_LOW, ("Close opened sockets.\n"));

      if (!oCtrlStatus.pProgramArg->oArgFlags.nQos)
      {
         if (oCtrlStatus.nSumCoderCh > 0)
         {
            VOIP_ReleaseSocket(&oCtrlStatus);
         }
      }

      if (oCtrlStatus.rgnSockets != IFX_NULL)
      {
         free(oCtrlStatus.rgnSockets);
         oCtrlStatus.rgnSockets = IFX_NULL;
      }

      /* Close socket used for call administration */
      close(oCtrlStatus.nAdminSocket);
      oCtrlStatus.nAdminSocket = NO_SOCKET;

#ifdef ABACUS_TEST
      TRACE(TAPIDEMO, DBG_LEVEL_LOW, ("Close communication sockets.\n"));
      COM_CleanUp(&oCtrlStatus);
#endif /* ABACUS_TEST */

#ifdef LINUX
      if (oCtrlStatus.pProgramArg->nTraceRedirection ==
          TRACE_REDIRECTION_SYSLOG_LOCAL)
      {
         closelog();
      }
      else if (oCtrlStatus.pProgramArg->nTraceRedirection ==
               TRACE_REDIRECTION_SYSLOG_REMOTE)
      {
         close(oTraceRedirection.nSyslogSocket);
         oTraceRedirection.nSyslogSocket = NO_SOCKET;
      }
      else if (oCtrlStatus.pProgramArg->nTraceRedirection ==
               TRACE_REDIRECTION_FILE)
      {
         close(oTraceRedirection.nFileLogFD);
         oTraceRedirection.nFileLogFD = NO_FILE;
      }

      /* End with task */
      exit(0);
   }
#endif /* LINUX */
}

#ifdef LINUX
/**
   Handle SIGCHLD signal, if trace redirection child process died,
   then end application.

   \param nSignal    - in LINUX signal number in VXWORKS not used
   \param pSigInfo   - structure holding signal information
   \param pVoid      - this field is not used
*/
static void TAPIDEMO_ChildCleanUp(int nSignal, siginfo_t* pSigInfo, void* pVoid)
{
   TD_PTR_CHECK(pSigInfo,);

   /* make sure that signal name is SIGCHLD */
   if (nSignal == SIGCHLD)
   {
      /* if using trace redirection then child pid is set */
      if (oCtrlStatus.nSyslogChildPid != 0)
      {
         /* check if trace redirection process */
         if (oCtrlStatus.nSyslogChildPid == pSigInfo->si_pid)
         {
            /* end tapidemo */
            TAPIDEMO_CleanUp(nSignal);
         } /* if (oCtrlStatus.nSyslogChildPid == pSigInfo->si_pid) */
      } /* if (oCtrlStatus.nSyslogChildPid != 0) */
   } /* if (nSignal == SIGCHLD) */

} /* TAPIDEMO_ChildCleanUp() */
#endif /* LINUX */


/**
   Turn on services specified by user.

   \param pCtrl - program control structure
*/
IFX_return_t TAPIDEMO_RunServices(CTRL_STATUS_t* pCtrl)
{
   PROGRAM_ARG_t* pProgramArg;

   TD_PTR_CHECK(pCtrl, IFX_ERROR);
   TD_PTR_CHECK(pCtrl->pProgramArg, IFX_ERROR);

   /* set pointer to program settings structure */
   pProgramArg = pCtrl->pProgramArg;

   SetTraceLevel(TAPIDEMO, pProgramArg->nDbgLevel);
#ifdef EASY336
   SVIP_RM_SetTraceLevel(pProgramArg->nDbgLevel);
#endif

   if (pProgramArg->oArgFlags.nQos)
   {
      QOS_TurnServiceOn(pProgramArg);
   }

   if ((pProgramArg->oArgFlags.nQos) &&
       (pProgramArg->oArgFlags.nQOS_Local))
   {
      /* With local UDP redirection conference should be disabled. */
      pProgramArg->oArgFlags.nConference = 0;
   }

   if (pProgramArg->oArgFlags.nCID)
   {
      CID_SetStandard(pProgramArg->nCIDStandard);
   }

   if (pProgramArg->oArgFlags.nEncTypeDef)
   {
      VOIP_SetEncoderType(pProgramArg->nEnCoderType);

      /* User didn't set frame len, so use default one */
      if (!pProgramArg->oArgFlags.nFrameLen)
      {
         VOIP_SetDefaultFrameLen();
      }
   }

   if (pProgramArg->oArgFlags.nFrameLen)
   {
      VOIP_SetFrameLen(pProgramArg->nPacketisationTime);
   }
#ifdef FXO
   if (pProgramArg->oArgFlags.nFXO)
   {
      if (pProgramArg->oArgFlags.nPCM_Slave)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("The PCM i/f can not be configured as a SLAVE when FXO is used.\n "
             "PCM i/f will be set as MASTER\n"));
         pProgramArg->oArgFlags.nPCM_Slave = 0;
      }
      /* General FXO intialization if needed */
      pProgramArg->oArgFlags.nPCM_Master = 1;
   }
#endif /* FXO */
   return IFX_SUCCESS;
}


/**
   Displays calling sequence.

   \param pCtrl - handle to control status structure

   \return always IFX_SUCCESS
*/
static IFX_int32_t TAPIDEMO_Call_Info(CTRL_STATUS_t* pCtrl)
{
   IFX_int32_t i = 0;
   IFX_int32_t j = 0;
   BOARD_t* pBoard = IFX_NULL;
   IFX_char_t* sTableSeparate = IFX_NULL;
#ifdef FXO
   FXO_t* pFXO;
   IFX_int32_t nFXO_Number = 0;
#endif /* FXO */

#ifndef TAPI_VERSION4
   sTableSeparate = "------------------------------------------------";
#else /* TAPI_VERSION4 */
   sTableSeparate = "----------------------------------------------------------";
#endif /*  TAPI_VERSION4 */

#ifndef TAPI_VERSION4
   TAPIDEMO_PRINTF(("Dialing sequences for different call types:\n"));
   TAPIDEMO_PRINTF((" a) Local connection: <c>\n"));
   TAPIDEMO_PRINTF((" b) Extern connection: <0iiic>\n"));
   TAPIDEMO_PRINTF((" c) PCM connection: <09iiic>\n"));
   TAPIDEMO_PRINTF((" d) FXO connection: <05d>\n"));
   TAPIDEMO_PRINTF((" e) FEATURE selection first digit must be 'STAR (*),\n"));
   TAPIDEMO_PRINTF(("    afterthat feature number (4 digits).\n"));
   TAPIDEMO_PRINTF((" f) CONFERENCING first digit must be 'HASH (#)',\n"));
   TAPIDEMO_PRINTF(("    afterthat normal dialing sequence,\n"));
   TAPIDEMO_PRINTF((" g) Extern connection with full ip:\n"
                    "    <0*xxx*xxx*xxx*xxx*dcc> - each part separated with STAR (*),\n"));
   TAPIDEMO_PRINTF(("\n"));
   TAPIDEMO_PRINTF(("Legend:\n"));
   TAPIDEMO_PRINTF(("<c> - channel number started from 1 (1 means channel 0).\n"));
   TAPIDEMO_PRINTF(("<iii> - lowest part of external board IP address (a.b.c.<iii>)\n"));
   TAPIDEMO_PRINTF(("<d> - FXO channel, started from 1 (1 means channel 0).\n"));
   TAPIDEMO_PRINTF(("<xxx*xxx*xxx*xxx> - external board IP address "
                    "(xxx.xxx.xxx.xxx)\n"));
#else
   TAPIDEMO_PRINTF((" a) Local connection: <dcc>\n"));
   TAPIDEMO_PRINTF((" b) Extern connection: <0iiidcc>\n"));
   TAPIDEMO_PRINTF((" c) PCM connection: <09iiidcc>\n"));
   TAPIDEMO_PRINTF((" d) FXO connection: <05Ddcc>\n"));
   TAPIDEMO_PRINTF((" e) FEATURE selection first digit must be 'STAR (*),\n"));
   TAPIDEMO_PRINTF(("    afterthat feature number (4 digits).\n"));
   TAPIDEMO_PRINTF((" f) CONFERENCING first digit must be 'HASH (#)',\n"));
   TAPIDEMO_PRINTF(("    afterthat normal dialing sequence,\n"));
   TAPIDEMO_PRINTF((" g) Extern connection with full ip:\n"
                    "    <0*xxx*xxx*xxx*xxx*dcc> - each part separated with STAR (*).\n"));
   TAPIDEMO_PRINTF(("\n"));
   TAPIDEMO_PRINTF(("Legend:\n"));
   TAPIDEMO_PRINTF(("<d> - device number started from 1 (1 means device 0).\n"));
   TAPIDEMO_PRINTF(("<cc> - channel number started from 1\n"
                    "       (01 means channel 0, 02 means channel 1, etc.).\n"));
   TAPIDEMO_PRINTF(("<iii> - lowest part of external board IP address (a.b.c.<iii>)\n"));
   TAPIDEMO_PRINTF(("<D> - FXO channel, only one digit at the moment\n"));
   TAPIDEMO_PRINTF(("<xxx*xxx*xxx*xxx> - external board IP address "
                    "(xxx.xxx.xxx.xxx)\n"));
#endif /* TAPI_VERSION4 */
   TAPIDEMO_PRINTF(("\n"));

   TAPIDEMO_PRINTF(("\n"));
   TAPIDEMO_PRINTF(("FXS channels calling table:\n"));
   TAPIDEMO_PRINTF((" %s\n", sTableSeparate));
#ifndef TAPI_VERSION4
   TAPIDEMO_PRINTF(("| Dial |          Board          | Phone channel |\n"));
#else
   TAPIDEMO_PRINTF(("| Dial |          Board          | Device | Phone channel  |\n"));
#endif
   TAPIDEMO_PRINTF(("|%s|\n", sTableSeparate));
   for (j = 0; j < pCtrl->nBoardCnt; j++)
   {
      pBoard = &pCtrl->rgoBoards[j];
      for (i = 0; i < pBoard->nMaxAnalogCh; i++)
      {
#ifndef TAPI_VERSION4
         TAPIDEMO_PRINTF(("|  %3d |     %15s     |      %3d      |\n",
                          (int) pBoard->rgoPhones[i].nPhoneNumber,
                          pBoard->pszBoardName, (int) i));
#else
         TAPIDEMO_PRINTF(("|  %03d |     %15s     |   %2d   |      %3d       |\n",
                          pBoard->rgoPhones[i].nPhoneNumber,
                          pBoard->pszBoardName,
                          pBoard->rgoPhones[i].nDev,
                          pBoard->rgoPhones[i].nPhoneCh));
#endif
      }
   }
   TAPIDEMO_PRINTF((" %s\n", sTableSeparate));
#ifndef TAPI_VERSION4
#ifdef FXO
   /* if using FXO */
   if (pCtrl->pProgramArg->oArgFlags.nFXO)
   {
      /* for all boards */
      for (j = 0; j < pCtrl->nBoardCnt; j++)
      {
         pBoard = &pCtrl->rgoBoards[j];
         /* check if fxo structure allocated */
         if (IFX_NULL != pBoard->rgoFXOs)
         {
            for (i = 0; i < pBoard->nMaxFxoCh; i++)
            {
               nFXO_Number += pBoard->nMaxFxoCh;
            }
         }
      }
      if (0 == nFXO_Number)
      {
         TAPIDEMO_PRINTF(("No FXO channels.\n"));
      } /* if (0 == nFXO_Number) */
      else
      {
         TAPIDEMO_PRINTF(("\n"));
         TAPIDEMO_PRINTF(
            ("FXO channels calling table:\n"));
         TAPIDEMO_PRINTF(
            (" %s\n", sTableSeparate));
         TAPIDEMO_PRINTF(
            ("| Dial |          Board          |     FXO ch    |\n"));
         TAPIDEMO_PRINTF(("|%s|\n", sTableSeparate));
         /* for all boards */
         for (j = 0; j < pCtrl->nBoardCnt; j++)
         {
            pBoard = &pCtrl->rgoBoards[j];
            /* check if fxo structure allocated */
            if (IFX_NULL != pBoard->rgoFXOs)
            {
               for (i = 0; i < pBoard->nMaxFxoCh; i++)
               {
                  pFXO = &pBoard->rgoFXOs[i];
                  TAPIDEMO_PRINTF(
                     ("|  05%d |     %15s     |      %3d      |\n",
                      pFXO->nFXO_Number, pBoard->pszBoardName, pFXO->nFxoCh));
               }
            }
         } /* for all boards */
         TAPIDEMO_PRINTF((" %s\n", sTableSeparate));
         TAPIDEMO_PRINTF(("\n"));
      }
   } /* if (pCtrl->pProgramArg->oArgFlags.nFXO) */
#endif /* FXO */
#endif /* TAPI_VERSION4 */

   return IFX_SUCCESS;
} /* TAPIDEMO_Call_Info() */

#ifdef STREAM_1_1
/**
   Clear phone buffer containing DTMF, pulse digits.

   \param nDataCh_FD  - file descriptor for data channel
*/
static IFX_void_t TAPIDEMO_ClearDigitBuff(IFX_int32_t nDataCh_FD)
{
   IFX_int32_t ready = 0;
   IFX_int32_t digit = 0;


   /* Check for DTMF digit */
   /* Check the DTMF dialing status, but on data device not control device */
   do
   {
      ready = 0;
      ioctl(nDataCh_FD, IFX_TAPI_TONE_DTMF_READY_GET, (IFX_int32_t) &ready);
      if (1 == ready)
      {
         /* Digit arrived or was in buffer */
         ioctl(nDataCh_FD, IFX_TAPI_TONE_DTMF_GET, (IFX_int32_t) &digit);
      }
      else
      {
         /* No digit in the buffer, buffer is cleared */
         break;
      }
   }
   while (1 == ready);

   /* Check for Pulse digit */
   /* Check the Pulse dialing status, but on data device not control device */
   do
   {
      ready = 0;
      ioctl(nDataCh_FD, IFX_TAPI_PULSE_READY, (IFX_int32_t) &ready);
      if (1 == ready)
      {
         /* Digit arrived or was in buffer */
         ioctl(nDataCh_FD, IFX_TAPI_PULSE_GET, (IFX_int32_t) &digit);
      }
      else
      {
         /* No digit in the buffer, buffer is cleared */
         break;
      }
   }
   while (1 == ready);

   TRACE(TAPIDEMO, DBG_LEVEL_LOW,
        ("Buffer with DTMF, pulse digits cleared, flushed\n"));

} /* TAPIDEMO_ClearDigitBuff() */
#endif /* STREAM_1_1 */


/**
   Initialize main structures used for testing

   \param pBoard      - board handle
   \param pCtrl       - pointer to status control structure
   \param nPhoneNum   - number of phone
   \param pFXO_Num    - number of FXO

   \return IFX_SUCCESS if everything OK, otherwise IFX_ERROR
*/
static IFX_return_t TAPIDEMO_Setup(BOARD_t* pBoard,
                                   CTRL_STATUS_t* pCtrl,
                                   IFX_int32_t* nPhoneNum,
                                   IFX_int32_t* pFXO_Num)
{
   IFX_int32_t ch = 0;
#ifndef TAPI_VERSION4
   IFX_TAPI_SIG_DETECTION_t dtmfDetection;
#endif

   if ((IFX_NULL == pCtrl) || (IFX_NULL == pCtrl->pProgramArg) ||
       (IFX_NULL == pBoard))
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }

   /* Add default ip address and demo arguments */
   pCtrl->oTapidemo_IP_Addr = pCtrl->pProgramArg->oMy_IP_Addr;
   pCtrl->nMy_IP_Addr =
      ntohl((int)pCtrl->pProgramArg->oMy_IP_Addr.sin_addr.s_addr) & 0xffffff00;

   if (pBoard->nMaxCoderCh > 0)
   {
      if (IFX_SUCCESS != VOIP_Init(pCtrl->pProgramArg, pBoard))
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH, ("VOIP_Init failed\n"));
         return IFX_ERROR;
      }
   }

   /* PCM is global and initialized in main() function at the moment */

#ifdef STREAM_1_1
   ABSTRACT_UnmapDefaults(pCtrl, pBoard);
#endif /* STREAM_1_1 */

   /* Do startup mapping of channels */
   ABSTRACT_DefaultMapping(pCtrl, pBoard);

   /* Channels are already open */
   for (ch = 0; ch < pBoard->nMaxAnalogCh; ch++)
   {
      /* Prepare structure */
      pBoard->rgoPhones[ch].nIdx = ch;
      /* reset event fields */
      pBoard->rgoPhones[ch].nIntEvent_FXS = IE_NONE;
      pBoard->rgoPhones[ch].nIntEvent_Config = IE_NONE;
#ifndef TAPI_VERSION4
      pBoard->rgoPhones[ch].nPhoneNumber = *nPhoneNum;
      *nPhoneNum += 1;
#else
      pBoard->rgoPhones[ch].nPhoneNumber = 
         (pBoard->rgoPhones[ch].nDev + 1) * 100 + 
         pBoard->rgoPhones[ch].nPhoneCh + 1 + *nPhoneNum;
#endif

      /* Now setup FDs and sockets, because startup mapping is done. */
      pBoard->rgoPhones[ch].nPhoneCh_FD =
         Common_GetDeviceOfCh(pBoard, pBoard->rgoPhones[ch].nPhoneCh);

      if (pBoard->nMaxCoderCh > 0)
      {
         pBoard->rgoPhones[ch].nDataCh_FD = 
            Common_GetDeviceOfCh(pBoard, pBoard->rgoPhones[ch].nDataCh);

#ifdef TAPI_VERSION4
         if (pBoard->fUseSockets)
#endif
         {
            /* Set sockets */
            pBoard->rgoPhones[ch].nSocket = 
               VOIP_GetSocket(pBoard->rgoPhones[ch].nDataCh, pBoard);
         }
      }

      /* Add also board to phone */
      pBoard->rgoPhones[ch].pBoard = pBoard;

      pBoard->rgoPhones[ch].pCID_Msg = calloc(1, sizeof(IFX_TAPI_CID_MSG_t));

      if (IFX_NULL != pBoard->rgoPhones[ch].pCID_Msg)
      {
         pBoard->rgoPhones[ch].pCID_Msg->message = calloc(CID_ELEM_COUNT,
                                           sizeof(IFX_TAPI_CID_MSG_ELEMENT_t));
      }
      else
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Error on calloc for phone->pCID_Msg"
                "File: %s, line: %d)\n",
                 __FILE__, __LINE__));
      }

      if (oCtrlStatus.pProgramArg->oArgFlags.nCID)
      {
#ifndef TAPI_VERSION4
         CID_SetupData(ch, &pBoard->rgoPhones[ch]);
#else
         CID_SetupData(pBoard->rgoPhones[ch].nPhoneCh, &pBoard->rgoPhones[ch]);
#endif

         /* Shows CID data */
         CID_Display(&pBoard->rgoPhones[ch], CID_DISPLAY_SIMPLE);

         /* Configure CID for this channel */
         if (IFX_SUCCESS != CID_ConfDriver(&pBoard->rgoPhones[ch]))
         {
#ifdef TAPI_VERSION4
            if (pBoard->rgoPhones[ch].pBoard->fSingleFD)
            {
               TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                     ("CID initialization failed for device %d, channel %d. "
                     "File: %s, line: %d)\n",
                      pBoard->rgoPhones[ch].nDev, 
                      pBoard->rgoPhones[ch].nPhoneCh, __FILE__, __LINE__));
            }
            else
#endif
            {
               TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("CID initialization failed for channel %d. "
                  "File: %s, line: %d)\n",
                  (int)ch, __FILE__, __LINE__));
            }
            Common_Close_FDs(pBoard);
            return IFX_ERROR;
         }
      }
      /* set initial states of state machines */
      pBoard->rgoPhones[ch].rgStateMachine[FXS_SM].nState = S_READY;
      pBoard->rgoPhones[ch].rgStateMachine[FXS_SM].pStateCombination = ST_rgFXSStates;
      pBoard->rgoPhones[ch].nCallDirection = CALL_DIRECTION_NONE;
      pBoard->rgoPhones[ch].rgStateMachine[CONFIG_SM].nState = S_READY;
      pBoard->rgoPhones[ch].rgStateMachine[CONFIG_SM].pStateCombination = ST_rgConfigStates;

   } /* for */

#ifdef TAPI_VERSION4
   if (pBoard->nMaxAnalogCh != 0 && &pBoard->rgoPhones[pBoard->nMaxAnalogCh-1] 
      != IFX_NULL)
      *nPhoneNum = (pBoard->rgoPhones[pBoard->nMaxAnalogCh-1].nDev + 1) * 100;
#endif

   if (pCtrl->pProgramArg->oArgFlags.nFXO)
   {
      /* Channels are already open */
      for (ch = 0; ch < pBoard->nMaxFxoCh; ch++)
      {
         /* Prepare FXO_t structure */
         pBoard->rgoFXOs[ch].nFxoCh = ch;
         pBoard->rgoFXOs[ch].nState = S_READY;
         pBoard->rgoFXOs[ch].fFirstStopRingEventOccured = IFX_FALSE;
         pBoard->rgoFXOs[ch].fCID_FSK_Started = IFX_FALSE;
         pBoard->rgoFXOs[ch].nFXO_Number = *pFXO_Num;
         /* increase FXO number */
         *pFXO_Num++;
         pBoard->rgoFXOs[ch].oCID_Msg.message = calloc(1, sizeof(IFX_TAPI_CID_MSG_ELEMENT_t));

         /* Now setup FDs and sockets, because startup mapping is done. */
         pBoard->rgoFXOs[ch].nFxoCh_FD =
            Common_GetFxoDevOfCh(pBoard, pBoard->rgoFXOs[ch].nFxoCh);
         /* Add also board to fxo */
         pBoard->rgoFXOs[ch].pBoard = pBoard;
      } /* for */
   }

#ifdef EASY336
   if (Common_PCM_ConfigSet(pBoard, &pcm_cfg) != IFX_SUCCESS)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("PCM configuration set failed "
            "(File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }
#endif

#ifndef TAPI_VERSION4
   memset(&dtmfDetection, 0, sizeof (IFX_TAPI_SIG_DETECTION_t));
   /* Stop detection of DTMF tones from local interface */
   dtmfDetection.sig = IFX_TAPI_SIG_DTMFTX;
   if (0 < pBoard->nMaxCoderCh)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("%s: Disable DTMF detection on all data channels \n",
             pBoard->pszBoardName));
   }
   for (ch = 0; ch < pBoard->nMaxCoderCh; ch++)
   {
      if (IFX_SUCCESS != ioctl(pBoard->nCh_FD[ch], IFX_TAPI_SIG_DETECT_DISABLE,
                               (IFX_int32_t) &dtmfDetection))
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: IFX_TAPI_SIG_DETECT_DISABLE failed (File: %s, line: %d)\n",
            __FILE__, __LINE__));
         return IFX_ERROR;
      }
   }
#endif /* TAPI_VERSION4 */
   return IFX_SUCCESS;
} /* TAPIDEMO_Setup() */

/**
   Set file descriptors set for select().

   \param pCtrl   - pointer to control structure
   \param pRdFds  - pointer to file descriptors set

   \return IFX_SUCCESS if everything OK, otherwise IFX_ERROR
*/
static IFX_return_t TAPIDEMO_SetFdSet(CTRL_STATUS_t* pCtrl, fd_set* pRdFds)
{
   IFX_int32_t i, j;
   IFX_int32_t fd_ch, fd_dev_ctrl;
   BOARD_t* pBoard;
   IFX_int32_t nSocket;

   /* check input parameters */
   TD_PTR_CHECK(pCtrl, IFX_ERROR);
   TD_PTR_CHECK(pRdFds, IFX_ERROR);

   /* reset structure */
   FD_ZERO(pRdFds);

   /* for all boards */
   for (j = 0; j < pCtrl->nBoardCnt; j++)
   {
      /* get board */
      pBoard = &pCtrl->rgoBoards[j];

      /* Add all board file descriptor */
      fd_dev_ctrl = pBoard->nDevCtrl_FD;
      if (0 < fd_dev_ctrl)
      {
         TD_FD_SET(fd_dev_ctrl, pRdFds, pCtrl->rwd);
      }
      else
      {
         TAPIDEMO_DEVICE_t  *pDevice = IFX_NULL;
         /* some boards does not have control device file descriptor
           (e.g. board_teridian73M1866). */
         if ((pDevice = Common_GetDevice_CPU(pBoard->pDevice)) != IFX_NULL)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("Error, Control device not open, %s. (File: %s, line: %d)\n",
                   pBoard->pszBoardName, __FILE__, __LINE__));
            return IFX_ERROR;
         }
      }

      /* set fds responsible for communication/calls that use data channels
         e.g. voip connections */
#ifdef TAPI_VERSION4
      if (!pBoard->fSingleFD)
#endif
      {
         /* Set up file descriptor table without sockets */
         for (i = 0; i < pBoard->nMaxCoderCh; i++)
         {
            fd_ch = Common_GetDeviceOfCh(pBoard, i);
            if (fd_ch > 0)
            {
#ifdef STREAM_1_1
               /* Clear DTMF buffer for each phone channel */
               TAPIDEMO_ClearDigitBuff(fd_ch);
#endif /* STREAM_1_1 */
               /* Put file descriptors into the read set */
               TD_FD_SET(fd_ch, pRdFds, pCtrl->rwd);
#ifdef TAPI_VERSION4
               if (pBoard->fUseSockets)
#else /* TAPI_VERSION4 */
               if (!pCtrl->pProgramArg->oArgFlags.nQos)
#endif /* TAPI_VERSION4 */
               {
                  /* Set possible sockets */
                  nSocket = VOIP_GetSocket(i, pBoard);
                  if (NO_SOCKET != nSocket)
                  {
                     TD_FD_SET(nSocket, pRdFds, pCtrl->rwd);
                  }
               }
            } /* if (fd_ch > 0) */
            else
            {
               TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                     ("Error, File descriptor for channel %d not open. "
                      "(File: %s, line: %d)\n",
                      (int) i, __FILE__, __LINE__));
               return IFX_ERROR;
            }
         } /* for - go through all coder channels */
      } /* if (!pBoard->fSingleFD */

      /* add FXO device control fd */
      if (oCtrlStatus.pProgramArg->oArgFlags.nFXO)
      {
         if (0 < pBoard->nMaxFxoCh)
         {
            fd_dev_ctrl = Common_GetFxoDevCtrlFd(pBoard);
            /* Add config fd to fd set */
            if (0 < fd_dev_ctrl)
            {
               TD_FD_SET(fd_dev_ctrl, pRdFds, pCtrl->rwd);
            }
         }
      }
   } /* for - go through all boards */

   /* Also add administration socket for handling calls */
   if (pCtrl->nAdminSocket == NO_SOCKET)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Error, Couldn't get administration socket. "
             "(File: %s, line: %d)\n",
             __FILE__, __LINE__));
      /* Without this socket we cannot make calls. */
      return IFX_ERROR;
   }
   TD_FD_SET(pCtrl->nAdminSocket, pRdFds, pCtrl->rwd);

#ifdef ABACUS_TEST
   /* Add communication socket for handling PC's requests */
   if(NO_SOCKET != pCtrl->nComInSocket)
   {
      TD_FD_SET(pCtrl->nComInSocket, pRdFds, pCtrl->rwd);
   }
   pCtrl->bChangeIPAddress = IFX_FALSE;
#endif /* ABACUS_TEST */

#if (defined(TD_T38_FAX_TEST) && defined(TD_FAX_T38) && defined(AR9))
   /* Add communication socket for handling PC's requests */
   if(NO_SOCKET != pCtrl->oFaxTestCtrl.nComSocket)
   {
      /* check if using T.38 fax test, if not then leave function */
      TD_T38_FAX_TEST_CHECK_IF_WORKING(&pCtrl->oFaxTestCtrl);
      /* add socket */
      TD_FD_SET(pCtrl->oFaxTestCtrl.nComSocket, pRdFds, pCtrl->rwd);
   }
#endif /* (defined(TD_T38_FAX_TEST) && defined(TD_FAX_T38) && defined(AR9)) */

   return IFX_SUCCESS;
}

/**
   Implements a simple state machine for PBX

   \param  pCtrl - pointer to status control structure

   \return status (IFX_SUCCESS or IFX_ERROR)

   \remark
      Actions are done from other sources, for example other channels
      wanting to call one channel. This can also be from a remote side via
      sockets.
      Hook status, digit status, line status is only checked on
      analog lines, that mean phone channel.
*/
static IFX_return_t TAPIDEMO_StateMachine(CTRL_STATUS_t* pCtrl)
{
   IFX_return_t ret = IFX_SUCCESS;
#ifdef FXO
   IFX_int32_t nFxoCh = 0;
#endif /* FXO */
   PHONE_t* pPhone = IFX_NULL;
   FXO_t* pFXO = IFX_NULL;
   CONNECTION_t* pConn = IFX_NULL;

   IFX_boolean_t bIsFxo = IFX_FALSE;

   BOARD_t* pBoard = IFX_NULL;
#ifdef FXO
   BOARD_t* pRingBoard = IFX_NULL;
#endif /* FXO */
   IFX_int32_t nBoard, nPhone, nConn;
#if (!defined EASY336 && !defined TAPI_VERSION4)
   IFX_int32_t nDataCh;
   IFX_int32_t nSocket = NO_SOCKET;
   IFX_int32_t nDataChFd;
#endif /* (!defined EASY336 && !defined TAPI_VERSION4) */
#if !defined(STREAM_1_1) && !defined(VXWORKS) && defined(FXO)
   IFX_int32_t nFxoDevCtrl_FD;
#endif /* !defined(STREAM_1_1) && !defined(VXWORKS) && defined(FXO) */
   fd_set rfds, trfds;
   /* Wait forever by default */
   struct timeval* pTimeOut = IFX_NULL;
   struct timeval curr_time = {0};
   struct timeval sel_timeout;
   IFX_int32_t time_diff = 0;

   /* check input parameters */
   TD_PTR_CHECK(pCtrl, IFX_ERROR);

   /*****************************
      PREPARE STRUCTURES 
    *****************************/
   memset(&sel_timeout, 0, sizeof(sel_timeout));
   pCtrl->bInternalEventGenerated = IFX_FALSE;
   /* do not use timeout untill it is needed */
   TD_TIMEOUT_RESET(pCtrl);

   /********************************************************
         sets the set of file descriptors for select
    ********************************************************/
   if (IFX_ERROR == TAPIDEMO_SetFdSet(pCtrl, &rfds))
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Error, Failed to set fd_set for select. (File: %s, line: %d)\n",
             __FILE__, __LINE__));
      return IFX_ERROR;
   }
   /********************************************************
         wait for events and handle them
    ********************************************************/
   while (1)
   {
      /* check if timeout chould be used */
      if (TD_TIMEOUT_CHECK(pCtrl))
      {
         /* time out used for external calls dialing and FXO calls */
         pTimeOut = &sel_timeout;
         /* check if timeout time elapsed */
         if (0 == sel_timeout.tv_sec && 0 == sel_timeout.tv_usec)
         {
            /* set time out (modified in select for LINUX) used for FXO calls
               and external call dialing (for dialing timeout). */
            sel_timeout.tv_sec   = SELECT_TIMEOUT_SEC;
            sel_timeout.tv_usec  = SELECT_TIMEOUT_USEC;
         }
      }
      else
      {
         /* select should wait without timeout */
         pTimeOut = IFX_NULL;
      }
#ifdef ABACUS_TEST
      /* if request to change IP address via TCL scripts*/
      if (IFX_TRUE == pCtrl->bChangeIPAddress)
      {
          if (IFX_ERROR == COM_ResetCommunicationSocket(pCtrl, &rfds))
          {
               return IFX_ERROR;
          }
      } /* if (IFX_TRUE == pCtrl->bChangeIPAddress) */
#endif /* ABACUS_TEST */

      /* Update the local file descriptor by the copy in the task parameter */
      memcpy((void *) &trfds, (void*) &rfds, sizeof(fd_set));
      /* wait for event or timeout */
      ret = select(pCtrl->rwd + 1, &trfds, IFX_NULL, IFX_NULL, pTimeOut);
      /* check if select waked up by event */
      if (0 < ret)
      {
/* start         HANDLING events from administration sockets            start */
         /* Check socket for handling calls between boards */
         if (FD_ISSET(pCtrl->nAdminSocket, &trfds))
         {
            /* Got something on admin socket
               (actions sent by TAPIDEMO_SetAction()) */
            TRACE(TAPIDEMO, DBG_LEVEL_LOW, ("New action on ADMIN socket\n"));
            TAPIDEMO_HandleCallSeq(pCtrl);
            /* check if internal event generated */
            if (IFX_TRUE == pCtrl->bInternalEventGenerated)
            {
               EVENT_InternalEventHandle(pCtrl);
            }
         }
/*   end         HANDLING events from administration socket             end   */
#ifdef ABACUS_TEST
/* start     HANDLING events from ABACUS TESTs communication socket     start */
         /* ABACUS TESTs message handling e.g. choosing test */
         if(NO_SOCKET != pCtrl->nComInSocket)
         {
            /* Check if there is a command pending on communication socket */
            if (FD_ISSET(pCtrl->nComInSocket, &trfds))
            {
               /* Receive command */
               TRACE(TAPIDEMO, DBG_LEVEL_LOW,
                     ("Command on abacus test communication socket.\n"));
               COM_ReceiveCommand(pCtrl);
            }
         } /* if(NO_SOCKET != pCtrl->nComInSocket) */
/*   end     HANDLING events from ABACUS TESTs communication socket     end   */
#endif /* ABACUS_TEST */

#if (defined(TD_T38_FAX_TEST) && defined(TD_FAX_T38) && defined(AR9))
         /* Add communication socket for handling PC's requests */
         if(NO_SOCKET != pCtrl->oFaxTestCtrl.nComSocket)
         {
            /* Check if there is a command pending on communication socket */
            if (FD_ISSET(pCtrl->oFaxTestCtrl.nComSocket, &trfds))
            {
               /* Receive command */
               /*TRACE(TAPIDEMO, DBG_LEVEL_LOW,
                     ("Command on fax test communication socket.\n"));*/
               TD_T38_FAX_TEST_HandleSocketData(pCtrl);
            }
         }
#endif /* (defined(TD_T38_FAX_TEST) && defined(TD_FAX_T38) && defined(AR9)) */
         /* for all boards */
         for(nBoard = 0; nBoard < pCtrl->nBoardCnt; nBoard++)
         {
            /* get board */
            pBoard = &pCtrl->rgoBoards[nBoard];
/* start             Handling tapi events on FXS                        start */
            /* some boards have only FXO channels
               and don't use pBoard->nDevCtrl_FD only nFxoDevCtrl_FD,
               if pBoard->nDevCtrl_FD not set skip this part without error */
            if (0 < pBoard->nDevCtrl_FD) 
            {
               /* check device file descriptor */
               if (FD_ISSET(pBoard->nDevCtrl_FD, &trfds))
               {
                  /* get event */
                  ret = EVENT_Check(pBoard->nDevCtrl_FD, pBoard, CHECK_EVENTS_ON_FXS);
                  if (IFX_ERROR == ret)
                  {
                     /* no event for this device */
                     TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                           ("Error, Board %s failed to get event. "
                            "(File: %s, line: %d)\n",
                            pBoard->pszBoardName, __FILE__, __LINE__));
                     continue;
                  }
                  /* Get event owner. Checking if event is dedicated for phone
                     or for fxo. There are some situations that data channel
                     is used to detect CPT for fxo. In such situation event
                     is dedicated for fxo */
                  if (IFX_ERROR == EVENT_GetOwnerOfEvent(pBoard, CHECK_EVENTS_ON_FXS,
                                      &pPhone, &pConn, &bIsFxo, &pFXO))
                  {
                     /* failed to get owner or event doesn't need handling */
                     /*TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                           ("Error, Failed to get event owner."
                            " (File: %s, line: %d)\n",
                            __FILE__, __LINE__));*/
                     break;
                  }
                  /* event for phone channel */
                  if (IFX_FALSE == bIsFxo)
                  {
                     /* check if pointers are set */
                     if ((IFX_NULL == pPhone) || (IFX_NULL == pConn))
                     {
                        TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                              ("Error, Invalid data: phone(%d) and connection(%d)."
                               " (File: %s, line: %d)\n", (IFX_int32_t)pPhone,
                               (IFX_int32_t)pConn, __FILE__, __LINE__));
                        break;
                     }
                     /* handle detected event */
                     EVENT_Handle(pPhone, pCtrl, pBoard);
                     /* if internal event generated then handle it
                        (both config and FXS state machine ) */
                     while (IE_NONE != pPhone->nIntEvent_Config)
                     {
                        ST_HandleState_Config(pCtrl, pPhone, pConn);
                     }
                     while (IE_NONE != pPhone->nIntEvent_FXS)
                     {
                        ST_HandleState_FXS(pCtrl, pPhone, pConn);
                     }
                  }
                  /* event for fxo channel */
                  else if (IFX_TRUE == bIsFxo)
                  {
                     if (IFX_NULL == pFXO)
                     {
                        TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                              ("Error, Failed to get fxo."
                               " (File: %s, line: %d)\n", __FILE__, __LINE__));
                        break;
                     }
                     /* handle event on data channel connected to FXO */
                     EVENT_Handle_DataCH(pFXO->nDataCh, IFX_NULL, pBoard, pFXO);
                  }
                  else
                  {
                     TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                           ("Error, Unpredicted behaviour of EVENT_GetOwnerOfEvent()."
                            " (File: %s, line: %d)\n", __FILE__, __LINE__));
                  }
                  /* check if internal event generated */
                  if (IFX_TRUE == pCtrl->bInternalEventGenerated)
                  {
                     EVENT_InternalEventHandle(pCtrl);
                  }
               } /* fd is set */
            } /* if pBoard->nDevCtrl_FD is used for this board */
/*   end             Handling tapi events on FXS                        end   */
/* start             Handling tapi events on FXO                        start */
#if !defined(STREAM_1_1) && !defined(VXWORKS) && defined(FXO)
            /* FXO events handling - check if FXO enabled */
            if (oCtrlStatus.pProgramArg->oArgFlags.nFXO)
            {
               /* get fxo device fd */
               nFxoDevCtrl_FD = Common_GetFxoDevCtrlFd(pBoard);
               /* not all boards have FXO device,
                  but no error should be printed because of that */
               if (0 < nFxoDevCtrl_FD)
               {
                  /* check if fd is set */
                  if (FD_ISSET(nFxoDevCtrl_FD, &trfds))
                  {
                     /* get FXO event */
                     ret = EVENT_Check(nFxoDevCtrl_FD, pBoard, CHECK_EVENTS_ON_FXO);
                     if (IFX_ERROR == ret)
                     {
                        /* no event for this device */
                        TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                              ("Error, Board %s failed to get event. "
                               "(File: %s, line: %d)\n",
                               pBoard->pszBoardName, __FILE__, __LINE__));
                        continue;
                     }
                     /* get owner of event, here can be only fxo */
                     if (IFX_ERROR == EVENT_GetOwnerOfEvent(pBoard, CHECK_EVENTS_ON_FXO,
                                         &pPhone, &pConn, &bIsFxo, &pFXO))
                     {
                        /* failed to get fxo or event doesn't need handling */
                        /*TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                              ("Error, Failed to get event owner."
                               " (File: %s, line: %d)\n",
                               __FILE__, __LINE__));*/
                        break;
                     }
                     if (IFX_TRUE != bIsFxo)
                     {
                        TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                              ("Error, Not a FXO event. (File: %s, line: %d)\n",
                               __FILE__, __LINE__));
                        break;
                     }
                     if (IFX_NULL == pFXO)
                     {
                        TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                              ("Error, Failed to Get FXO. (File: %s, line: %d)\n",
                               __FILE__, __LINE__));
                        break;
                     }
                     /* handle FXO event */
                     if (IFX_SUCCESS == EVENT_FXO_Handle(pFXO, pCtrl, pBoard))
                     {
                        /* EVENT_FXO_Handle() can generate internal events on
                           phones, check if such events are generated
                           and handle them */
                        if (IFX_TRUE == pCtrl->bInternalEventGenerated)
                        {
                           EVENT_InternalEventHandle(pCtrl);
                        }
                     }
                  } /* nFxoDevCtrl_FD is set */
               } /* if (0 < nFxoDevCtrl_FD) */
            } /* if (oCtrlStatus.pProgramArg->oArgFlags.nFXO) */
#endif /* !defined(STREAM_1_1) && !defined(VXWORKS) && defined(FXO) */
         }/* for all boards */
/*   end             Handling tapi events on FXO                        end   */
/* start      HANDLING voip packets on data channels and sockets        start */
/* packet handling must after event handling otherwise there are errors
   from drivers eg. INFO: Buffer pool growth limit reached */
         /* for all boards */
         for(nBoard = 0; nBoard < pCtrl->nBoardCnt; nBoard++)
         {
            /* get board */
            pBoard = &pCtrl->rgoBoards[nBoard];
#ifdef TAPI_VERSION4
            /* check if single fd is used */
            if (!pBoard->fSingleFD)
#endif
            {
#if (!defined EASY336 && !defined TAPI_VERSION4)
               /* for all data channels */
               for (nDataCh = 0; nDataCh < pBoard->nMaxCoderCh; nDataCh++)
               {
                  /* HANDLING PACKETS FROM DATA CHANNELS */
                  /* get data channel fd to check for voice packets */
                  nDataChFd = VOIP_GetFD_OfCh(nDataCh, pBoard);
                  /* check if valid data chennel fd was returned */
                  if (NO_DEVICE_FD == nDataChFd)
                  {
                     TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                           ("Error, Board %s Failed to get fd for "
                            "data channel %d. (File: %s, line: %d)\n",
                            pBoard->pszBoardName, nDataCh,
                            __FILE__, __LINE__));
                     break;
                  }
                  /* check if packets can be read from data channel  */
                  if (FD_ISSET(nDataChFd, &trfds))
                  {
                     /* get connection that uses this data channel */
                     if (IFX_SUCCESS == ABSTRACT_GetConnOfDataCh(&pPhone, &pConn,
                                                                 pBoard, nDataCh))
                     {
                        /* We received data from phone */
                        VOIP_HandleData(pCtrl, pPhone, nDataChFd, pConn,
                                        QOS_HandleService(pCtrl, pConn));
                     }
                     else
                     {
                      /*  TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                              ("Error, Unable to handle packets on Data channel"
                               "%d - discard data. (File: %s, line: %d)\n",
                               nDataCh, __FILE__, __LINE__)); */
                        /* because connection is not stopped at the same time on
                           both phones, packet from ended connection can 
                           be still received and must be discarded */
                        VOIP_DiscardData(nDataChFd);
                     }
                  } /* check if data channel is readable*/
                  /* HANDLING PACKETS FROM SOCKETS */
                  /* check if QoS is used */
                  if (!pCtrl->pProgramArg->oArgFlags.nQos)
                  {
                     /* get socket */
                     nSocket = VOIP_GetSocket(nDataCh, pBoard);
                     /* check if valid data chennel fd was returned */
                     if (NO_SOCKET == nSocket)
                     {
                        TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                              ("Error, Board %s Failed to get socket for "
                               "data channel %d. (File: %s, line: %d)\n",
                               pBoard->pszBoardName, nDataCh,
                               __FILE__, __LINE__));
                        break;
                     }
                     /* check if packets can be read from socket */
                     if (FD_ISSET(nSocket, &trfds))
                     {
                        pPhone = ABSTRACT_GetPHONE_OfSocket(nSocket, &pConn, pBoard);
                        if (IFX_NULL == pPhone || IFX_NULL == pConn)
                        {
                         /*  TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                                 ("Error, %s not found for socket %d. "
                                  "(File: %s, line: %d)\n",
                                  IFX_NULL == pPhone ? "Phone" : "Connection",
                                  nSocket, __FILE__, __LINE__));*/
                           /* because connection is not stopped at the same time
                              on both phones, packet from ended connection can 
                              be still received and must be discarded */
                           VOIP_DiscardData(nSocket);
                        }
                        else
                        {
                           /* We received data over socket, external phone */
                           VOIP_HandleSocketData(pPhone, pConn,
                                                 QOS_HandleService(pCtrl, pConn));
                        }
                     } /* check if socket is readable*/
                  } /* check if QoS is used */
               } /* for all data channels */
#endif /* (!defined EASY336 && !defined TAPI_VERSION4) */
            } /* if (!pBoard->fSingleFD) */
/*   end      HANDLING voip packets on data channels and sockets        end   */
         }/* for all boards */
      } /* if (0 < ret)*/
      /* check if select waked up by timeot */
      else if (0 == ret)
      {
/* start                        GET TIME                                start */
         /* Signal timeouts are handled via tick counter */
         if (TD_TIMEOUT_GET_TIME(&curr_time) != 0)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("Error gettimeofday(). (File: %s, line: %d)\n",
                   __FILE__, __LINE__));
            continue;
         }
/*   end                        GET TIME                                end   */
         /* for all boards */
         for(nBoard = 0; nBoard < pCtrl->nBoardCnt; nBoard++)
         {
            /* get board */
            pBoard = &pCtrl->rgoBoards[nBoard];
/* start            check timeout for external call dialing             start */
            /* for all phones on board */
            for (nPhone=0; nPhone < pBoard->nMaxAnalogCh; nPhone++)
            {
               pPhone = &pBoard->rgoPhones[nPhone];
               /* for all connections */
               for (nConn=0; nConn < pPhone->nConnCnt; nConn++)
               {
                  pConn = &pPhone->rgoConn[nConn];
                  /* check connection (must be external call in dialing) */
                  if ((pConn->nType == EXTERN_VOIP_CALL) &&
                      (pConn->bExtDialingIn == IFX_TRUE) &&
                      ((pConn->timeStartExtCalling.tv_sec != 0) ||
                       (pConn->timeStartExtCalling.tv_usec != 0)))
                  {
                     /* check time difference */
                     time_diff = curr_time.tv_sec - pConn->timeStartExtCalling.tv_sec;
                     if (time_diff > TIMEOUT_FOR_EXTERNAL_CALL)
                     {
                        TRACE(TAPIDEMO, DBG_LEVEL_LOW,
                              ("Phone No %d: Timeout diff %d sec for external connection\n"
                               "%selapsed, start to play busy tone.\n",
                               pPhone->nPhoneNumber, (int) time_diff,
                               pPhone->pBoard->pIndentPhone));
                        /* check phone state */
                        if ((pPhone->rgStateMachine[FXS_SM].nState == S_OUT_DIALING) ||
                            (pPhone->rgStateMachine[FXS_SM].nState == S_CF_DIALING))
                        {
                           pPhone->nIntEvent_FXS = IE_EXT_CALL_NO_ANSWER;
                           ST_HandleState_FXS(pCtrl, pPhone, pConn);
                        } /* if */
                        else
                        {
                           TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                                 ("Error, Phone No %d: Detected incorrect event for state: %s\n",
                                  (IFX_int32_t) pPhone->nPhoneNumber,
                                  Common_Enum2Name(pPhone->rgStateMachine[FXS_SM].nState,
                                                   TD_rgStateName)));
                        } /* check phone state */
                     } /* check time difference */
                  } /* check connection - if (pConn->... == ...) */
               }/* for all connections */
            }/* for all phones on board */
/*   end            check timeout for external call dialing             end   */
#ifdef FXO
/* start                check timeout for FXO ringing                   start */
            /* check if FXO is enabled */
            if (oCtrlStatus.pProgramArg->oArgFlags.nFXO)
            {
               /* for all fxo channels */
               for (nFxoCh = 0; nFxoCh < pBoard->nMaxFxoCh; nFxoCh++)
               {
                  /* get fxo of channel */
                  pFXO = ABSTRACT_GetFXO(nFxoCh, pBoard);
                  if (pFXO == IFX_NULL)
                  {
                     TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                           ("FXO not found for fxo channel %d. "
                            "(File: %s, line: %d)\n",
                            (int) nFxoCh, __FILE__, __LINE__));
                     break;
                  }
                  /* if ringing on FXO channel */
                  if (S_IN_RINGING == pFXO->nState)
                  {
                     /* Check if timeout for ringing event - it means
                        caller phone went ONHOOK and ringing must be stopped on
                        all the called phones. */
                     time_diff = curr_time.tv_sec - pFXO->oRingEventTimeout.tv_sec;
                     /* check timeout */
                     if (time_diff > TIMEOUT_FOR_ONHOOK)
                     {
                        /* if timeout occured then FXO call has ended */
                        TRACE(TAPIDEMO, DBG_LEVEL_LOW,
                              ("Timeout diff %d sec for ringing elapsed\n"
                               "on fxo ch %d, stop it on all called phones.\n",
                               (int) time_diff, (int) nFxoCh));
                        /* set board that should be ringing */
                        pRingBoard = pBoard;
                        /* check if phones are present on this board */
                        if ((pBoard->nMaxAnalogCh == 0) && (pBoard->nBoard_IDX > 0))
                        {
                           /* This is the extension board without phone channels.
                              Stop ringing on the evaluation board. */
                           pRingBoard = &pBoard->pCtrl->rgoBoards[0];
                        }
                        IFX_boolean_t fFxoEnd = IFX_FALSE;
                        /* for all phones on ring board */
                        for (nPhone=0; nPhone < pRingBoard->nMaxAnalogCh; nPhone++)
                        {
                           /* get phone */
                           pPhone = &pRingBoard->rgoPhones[nPhone];
                           /* check if phone ringing caused by fxo call */
                           if ((pPhone->rgoConn[0].nType == FXO_CALL) &&
                               (pPhone->rgStateMachine[FXS_SM].nState == S_IN_RINGING))
                           {
                              fFxoEnd = IFX_TRUE;
                              /* fxo call ended - stop ringing */
                              pPhone->nIntEvent_FXS = IE_READY;
                              ST_HandleState_FXS(pCtrl, pPhone, &pPhone->rgoConn[0]);
                           } /* if ringing uring FXO call*/
                        } /*for all phones on ring board */
                        /* check if any phone was ringing */
                        if (IFX_TRUE != fFxoEnd)
                        {
                           PHONE_t oDummyPhone;
                           FXO_EndConnection(pCtrl->pProgramArg, pFXO,
                                             &oDummyPhone, oDummyPhone.rgoConn,
                                             pBoard);
                        }
                     } /* if (time_diff > TIMEOUT_FOR_ONHOOK) */
                  } /* if (S_IN_RINGING == pFXO->fRingStopOccured) */
               } /* for all FXOs on board */
            } /* if (oCtrlStatus.pProgramArg->oArgFlags.nFXO) */
/*   end                check timeout for FXO ringing                   end   */
#endif /* FXO */
         } /* for all boards */
      } /* ret of select() */
      /* if ret < 0 select returned error */
      else if (0 > ret)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Error in select(). (File: %s, line: %d)\n",
              __FILE__, __LINE__));
      }
   } /* while */

   return IFX_SUCCESS;
} /* TAPIDEMO_StateMachine() */


/**
   Clear phone after connection.

   \param  pCtrl  - pointer to status control structure
   \param  pPhone - pointer to PHONE
   \param  pConn  - pointer to phone connection
   \param  nInternalEvent - new event information
   \param  nDataCh_FD - data channel file descriptor
   \param  pBoard - board handle

   \return Phone is cleared.
*/
IFX_void_t TAPIDEMO_ClearPhone(CTRL_STATUS_t* pCtrl, PHONE_t* pPhone,
                               CONNECTION_t* pConn,
                               INTERNAL_EVENTS_t nInternalEvent,
                               BOARD_t* pBoard)
{
   IFX_int32_t fd_data_ch = -1;

   if ((IFX_NULL == pCtrl) || (IFX_NULL == pPhone) || (IFX_NULL == pBoard))
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return;
   }

   TRACE(TAPIDEMO, DBG_LEVEL_LOW,
         ("Phone No %d: TAPIDEMO_ClearPhone.\n",
          pPhone->nPhoneNumber));

   /* set connection pointer for FXO_CALL*/
   if ((pPhone->fFXO_Call == IFX_TRUE) ||
       /* if pointer noot set and connection in progress */
       ((pPhone->nConnCnt == 1) && (pConn == IFX_NULL)))
   {
      pConn = &pPhone->rgoConn[0];
   }
   /************************
    TURN OFF SPECIAL FEATURES
    ************************/
   /* VAD works only for Data channels */
   if (0 < pBoard->nMaxCoderCh)
   {
      Common_SetVad(pPhone, pBoard, RESTORE_DEFAULT);
   }
   /* if wideband was set then change back narrowband
      if AGC was enabled then disable AGC
      AGC must be disabled before disabling encoder */
   if ((pBoard != IFX_NULL) && (pConn != IFX_NULL))
   {
      /* Set back to default features */
      if ((pPhone->fWideBand_Cfg) || (pPhone->fAGC_Cfg))
      {
         /* Set line type, coders, ... to some default values */
         if (pBoard->nMaxCoderCh > 0)
         {
            fd_data_ch = VOIP_GetFD_OfCh(pConn->nUsedCh, pBoard);
         }
         else if (pBoard->nMaxAnalogCh > 0)
         {
            fd_data_ch = Common_GetDeviceOfCh(pBoard, pConn->nUsedCh);
         }
         else
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("Board %s without coders and analog channels. "
                   "(File: %s, line: %d)\n",
                   pBoard->pszBoardName, __FILE__, __LINE__));
         }
         /* change back to NarrowBand if WideBand was enabled */
         if (pPhone->fWideBand_Cfg)
         {
            Common_ConfigureTAPI(pCtrl, pBoard, pPhone, 2000,
                                 fd_data_ch, IFX_TRUE, IFX_FALSE);
            pPhone->fWideBand_Cfg = IFX_FALSE;
         }
         /* disable AGC if enabled */
         if (pPhone->fAGC_Cfg)
         {
            Common_ConfigureTAPI(pCtrl, pBoard, pPhone, 2010,
                                 fd_data_ch, IFX_TRUE, IFX_FALSE);
            pPhone->fAGC_Cfg = IFX_FALSE;
         }
      } /* if ((pPhone->fWideBand_Cfg) || (pPhone->fAGC_Cfg)) */

   } /* if ((pBoard != IFX_NULL) && (pConn != IFX_NULL)) */

   /************************
      END CONNECTION
    ************************/
   if (IFX_NULL != pConn)
   {
      if (UNKNOWN_CALL_TYPE == pConn->nType)
      {
         pConn = &pPhone->rgoConn[0];
      }
/* CONFERENCE CONFERENCE CONFERENCE CONFERENCE CONFERENCE CONFERENCE */
      /* Conference clean up*/
      if (pPhone->nConfStarter)
      {
         if (IFX_ERROR == CONFERENCE_End(pCtrl, pPhone,
                             &pCtrl->rgoConferences[pPhone->nConfIdx - 1]))
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("Failed to end conference properly. Could cause "
                   "mapping related problems. (File: %s, line: %d)\n",
                   __FILE__, __LINE__));
         }
      }
#if !defined(STREAM_1_1) && !defined(VXWORKS) && defined(FXO)
/* FXO CALL FXO CALL FXO CALL FXO CALL FXO CALL FXO CALL FXO CALL FXO CALL */
      else if (FXO_CALL == pConn->nType)
      {
         if (oCtrlStatus.pProgramArg->oArgFlags.nFXO)
         {
            if (pPhone->fFXO_Call)
            {
               /* stop FXO call */
               if (pPhone->pFXO == IFX_NULL)
               {
                  /* should not happend*/
                  TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                        ("Error, PHONE_t:pFXO is NULL. (File: %s, line: %d)\n",
                         __FILE__, __LINE__));
               }
               else if ((pPhone->pFXO->nState == S_ACTIVE) ||
                        (pPhone->pFXO->nState == S_IN_RINGING))
               {
                  FXO_EndConnection(pCtrl->pProgramArg, pPhone->pFXO,
                                    pPhone, pConn, pBoard);
               }
               else
               {
                  TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                        ("Wrong FXO state. (File: %s, line: %d)\n",
                         __FILE__, __LINE__));
               }
            }
            else
            {
               TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                     ("pPhone->fFXO_Call is not set. (File: %s, line: %d)\n",
                      __FILE__, __LINE__));
            }
         }
         else
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("FXO calls are not enbled. (File: %s, line: %d)\n",
                   __FILE__, __LINE__));
         }
      }
#endif /* !defined(STREAM_1_1) && !defined(VXWORKS) && defined(FXO) */
/* PCM CALLS  PCM CALLS PCM CALLS PCM CALLS PCM CALLS PCM CALLS PCM CALLS */
#ifndef EASY336
      else if (PCM_CALL == pConn->nType ||
               LOCAL_PCM_CALL == pConn->nType ||
               LOCAL_BOARD_PCM_CALL == pConn->nType)
      {
         TAPIDEMO_PORT_DATA_t oPortData;
         oPortData.nType = PORT_FXS;
         oPortData.uPort.pPhopne = pPhone;
         if (PCM_EndConnection(&oPortData, pCtrl->pProgramArg, pBoard, pConn))
         {
#if (!defined(DXT) && !defined(XT16))
            /* Unmap phone and pcm channels */
            PCM_MapToPhone(pConn->nUsedCh,
                           pPhone->nPhoneCh,
                           IFX_FALSE, pBoard);
#endif /* !DXT && !XT16 */
         }
      }
#endif /* EASY336 */
      else
      {
/* CALL THAT USES CODECS CALL THAT USES CODECS CALL THAT USES CODECS */
         /* Stop coder if voip call or local call that uses codec*/
         if ((EXTERN_VOIP_CALL == pConn->nType) ||
             (LOCAL_BOARD_CALL == pConn->nType) ||
             ((LOCAL_CALL == pConn->nType) &&
              (pCtrl->pProgramArg->oArgFlags.nUseCodersForLocal)))
         {
            if (IFX_TRUE == pConn->bExtDialingIn)
            {
               /* Clean external connection */
               pConn->bExtDialingIn = IFX_FALSE;
               TD_TIMEOUT_STOP(pCtrl);
            }
            /* codec is started when connection is active */
            if (IFX_TRUE == pConn->fActive)
            {
#ifndef TAPI_VERSION4
               VOIP_StopCodec(pConn->nUsedCh, pConn, pBoard);
#else
#ifdef EASY336
               TRACE(TAPIDEMO, DBG_LEVEL_LOW,
                     ("pPhone->nVoipIDs: %d\n", pPhone->nVoipIDs));
               /* if conference then codec is stopped in CONFERENCE_End() */
               if ((0 < pPhone->nVoipIDs) &&
                   (IFX_FALSE == pPhone->nConfStarter))
               {
                  if (!pCtrl->pProgramArg->oArgFlags.nFaxChannel)
                  {
                     if (pBoard->fSingleFD)
                     {
                        /* check if vad settings were changed */
                        Common_SetVad(pPhone, pBoard, RESTORE_DEFAULT);
                     }
                     VOIP_StopCodec(pConn->nUsedCh, pConn,
                        ABSTRACT_GetBoard(pCtrl, IFX_TAPI_DEV_TYPE_VIN_SUPERVIP));
                  }
                  else
                  {
                     VOIP_FaxStop(pCtrl, pConn,
                        ABSTRACT_GetBoard(pCtrl, IFX_TAPI_DEV_TYPE_VIN_SUPERVIP));
                  }
               } /* if (0 < pPhone->nVoipIDs) */
#endif /* EASY336 */
#endif /* TAPI_VERSION4 */
#ifdef QOS_SUPPORT
               if (pCtrl->pProgramArg->oArgFlags.nQos)
               {
                  /* QoS Support */
                  QOS_StopSession(pConn, pPhone);
               }
#endif /* QOS_SUPPORT */
/* TAPI SIGNAL SUPPORT TAPI SIGNAL SUPPORT TAPI SIGNAL SUPPORT */
#ifdef TAPI_VERSION4
               if (!pPhone->pBoard->fSingleFD)
#endif
               {
                  /* fax modem only avaible for voip calls */
                  if (EXTERN_VOIP_CALL == pConn->nType)
                  {
                     /* reset signal handlers */
                     SIGNAL_HandlerReset(pPhone->nDataCh_FD, pPhone);
                  }
               }
            } /* if (IFX_TRUE == pConn->fActive) */

         } /* connection that uses codecs */
#ifndef EASY336
/* LOCAL CALL LOCAL CALL LOCAL CALL LOCAL CALL LOCAL CALL LOCAL CALL */
         else if (LOCAL_CALL == pConn->nType)
         {
            /* Unmap only on caller side, because its sufficient to map it also
               on caller side. */
            if (pPhone->fLocalPeerCalled)
            {
               /* Unmap phone channels */
               ALM_MapToPhone(pPhone->nPhoneCh,
                              pConn->oConnPeer.oLocal.pPhone->nPhoneCh,
                              IFX_FALSE, pPhone->pBoard);
            }
         }  /* if ((LOCAL_CALL != pConn->nType) */
#endif
      }
#ifdef EASY336
      pConn->voipID = SVIP_RM_UNUSED;
#endif /* EASY336 */
      /* connection is no longer active */
      pConn->fActive = IFX_FALSE;
   } /* if */

   /************************
      RESET PHONE STRUCTURE
    ************************/
   if (nInternalEvent != IE_BUSY)
   {
      pPhone->nCallDirection = CALL_DIRECTION_NONE;
   }
   /* Here we should also clear connections. */
   pPhone->nDialNrCnt = 0;
   pPhone->fExtPeerCalled = IFX_FALSE;
   pPhone->fPCM_PeerCalled = IFX_FALSE;
   pPhone->fLocalPeerCalled = IFX_FALSE;
   pPhone->nConnCnt = 0;
   pPhone->fConfTAPI_InConn = IFX_FALSE;
   pPhone->fFXO_Call = IFX_FALSE;

   /************************
   TURN OFF UNNEEDED RESOURCES
    ************************/
   /* phone goes on hook - turn off unneeded resources and change line feed */
   if (nInternalEvent == IE_READY)
   {
#ifdef TAPI_VERSION4
      if (!pBoard->fSingleFD)
#endif /* TAPI_VERSION4 */
      {
#ifndef TAPI_VERSION4
         /* disable the DTMF detection */
         Common_DTMF_SIG_Disable(pCtrl, pPhone);
         /* Disable the LEC*/
         Common_LEC_Disable(pCtrl, pPhone);
#endif /* TAPI_VERSION4 */
      }
      /* Set line in standby */
      if (IFX_ERROR == Common_LineFeedSet(pPhone, IFX_TAPI_LINE_FEED_STANDBY))
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Err, line feed set failed for ch %d, %s."
                "(File: %s, line: %d)\n",
                pPhone->nPhoneCh, pBoard->pszBoardName, __FILE__, __LINE__));
      } /* if (IFX_ERROR == Common_LineFeedSet()) */
   } /* if (nInternalEvent == IE_READY) */

} /* TAPIDEMO_ClearPhone() */

/**
   SetAction    : set action of the phone we are talking to.

   \param pCtrl   - pointer to status control structure
   \param pPhone  - pointer to PHONE
   \param pConn   - pointer to phone connection
   \param nAction - which action to send to other phone

   \return       IFX_SUCCESS - no error, otherwise IFX_ERROR
*/
IFX_return_t TAPIDEMO_SetAction(CTRL_STATUS_t* pCtrl,
                                PHONE_t* pPhone,
                                CONNECTION_t* pConn,
                                IFX_int32_t nAction)
{
   PHONE_t* p_dst_phone = IFX_NULL;
   IFX_int32_t ret = IFX_SUCCESS;
   struct sockaddr_in to_addr;
   COMM_MSG_t msg = {0};
   CONNECTION_t* p_conn = IFX_NULL;
   IFX_char_t* psActionName = IFX_NULL;

   /* check input parameter */
   TD_PTR_CHECK(pCtrl, IFX_ERROR);
   TD_PTR_CHECK(pPhone, IFX_ERROR);
   TD_PTR_CHECK(pConn, IFX_ERROR);

   /* set pointer string with action name */
   psActionName = Common_Enum2Name(nAction, TD_rgIE_EventsName);
   /* check if name was set, if no name is set,
      then invalid value of TD_rgIE_EventsName is set */
   TD_PTR_CHECK(psActionName, IFX_ERROR);

   /* prepare message end and start */
   msg.nMarkStart = COMM_MSG_START_FLAG;
   msg.nMarkEnd = COMM_MSG_END_FLAG;

   /* For this two actions no phone and connection are selected,
      pPhone and pConn are phony only for message send use. */
   if ((nAction == IE_EXT_CALL_PROCEEDING) ||
       (nAction == IE_EXT_CALL_WRONG_NUM))
   {
      /* Set action to external phone, that tried to call phone on this board,
         tell him that phone number is incomplete or wrong */
      /* set call type */
      if (EXTERN_VOIP_CALL == pConn->nType)
      {
         msg.fPCM = CALL_FLAG_VOIP;
      }
      else if (PCM_CALL == pConn->nType)
      {
         msg.fPCM = CALL_FLAG_PCM;
      }
      else
      {
         msg.fPCM = CALL_FLAG_UNKNOWN;
      }
      msg.nAction = nAction;
      msg.nFeatureID = pConn->nFeatureID;

      /* called phone info */
      msg.nBoard_IDX = NO_BOARD_IDX;
      msg.nSenderPhoneNum = pPhone->nPhoneNumber;
      /* no real connection set port number to 0*/
      msg.nSenderPort = 0;
      /* copy board MAC address */
      memcpy(msg.MAC, pConn->oUsedMAC, ETH_ALEN);

      /* caller phone info */
      msg.nReceiverPhoneNum = pConn->oConnPeer.nPhoneNum;
      /* set peer address */
      memset(&to_addr, 0, sizeof(struct sockaddr_in));
      to_addr.sin_addr = pConn->oConnPeer.oRemote.oToAddr.sin_addr;
      to_addr.sin_family = pConn->oConnPeer.oRemote.oToAddr.sin_family;
      to_addr.sin_port = htons(ADMIN_UDP_PORT);

      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("Sending MSG using socket %d to %s:%d:\n"
             " - call type %s\n"
             " - sender MAC %02X:%02X:%02X:%02X:%02X:%02X\n"
             " - sender phone number %d\n"
             " - receiver phone number %d\n"
             " - new internal event %s\n"
             " - feature id %d\n",
             (int) pCtrl->nAdminSocket,
             inet_ntoa(to_addr.sin_addr), ntohs(to_addr.sin_port),
             (CALL_FLAG_PCM == msg.fPCM)? "PCM" :
             (CALL_FLAG_VOIP == msg.fPCM) ? "EXTERN" : "UNKNOWN",
             msg.MAC[0],
             msg.MAC[1],
             msg.MAC[2],
             msg.MAC[3],
             msg.MAC[4],
             msg.MAC[5],
             msg.nSenderPhoneNum, msg.nReceiverPhoneNum,
             psActionName, msg.nFeatureID));

      ret = sendto(pCtrl->nAdminSocket, (IFX_char_t *) &msg, sizeof(msg), 0,
                   (struct sockaddr *) &to_addr,
                    sizeof(to_addr));
      if (sizeof(msg) != ret)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Error sending data (command %d byte(s)), %s. "
               "(File: %s, line: %d)\n",
               sizeof(msg), strerror(errno), __FILE__, __LINE__));
         return IFX_ERROR;
      }
   }
   else if (nAction == IE_EXT_CALL_WORKAROUND_EASY508xx)
   {
      /* It is workaround, which avoid problem with first external/PCM connection 
      to EASY508xx. 

      We need to send any packet to external board ESY508xx before starting 
      to establish external connection with it. If we don't do it, first packet 
      during first external connection will be not detected. 
      In this point, we neither know with what type of board we make connection 
      nor if it is first connection, so we have to send this packet before every 
      exernal connection. This packet will be just ignore by TAPIDEMO 
      on secont board  */

      memset(&to_addr, 0, sizeof(struct sockaddr_in));

      /* set call type */
      if (EXTERN_VOIP_CALL == pConn->nType)
      {
         msg.fPCM = CALL_FLAG_VOIP;
         to_addr.sin_addr = pConn->oConnPeer.oRemote.oToAddr.sin_addr;
         to_addr.sin_family = pConn->oConnPeer.oRemote.oToAddr.sin_family;
         to_addr.sin_port = htons(ADMIN_UDP_PORT);
      }
      else if (PCM_CALL == pConn->nType)
      {
         msg.fPCM = CALL_FLAG_PCM;
         to_addr.sin_addr = pConn->oConnPeer.oPCM.oToAddr.sin_addr;
         to_addr.sin_family = pConn->oConnPeer.oPCM.oToAddr.sin_family;
         to_addr.sin_port = htons(ADMIN_UDP_PORT);
      }
      else
      {
         msg.fPCM = CALL_FLAG_UNKNOWN;
      }

      msg.nAction = nAction;
      msg.nFeatureID = 0;

      /* called phone info */
      msg.nBoard_IDX = NO_BOARD_IDX;
      msg.nSenderPhoneNum = 0;
      /* no real connection set port number to 0*/
      msg.nSenderPort = 0;
      /* copy board MAC address */
      memcpy(msg.MAC, pConn->oUsedMAC, ETH_ALEN);

      /* caller phone info */
      msg.nReceiverPhoneNum = 0;

      ret = sendto(pCtrl->nAdminSocket, (IFX_char_t *) &msg, sizeof(msg), 0,
                   (struct sockaddr *) &to_addr,
                    sizeof(to_addr));
      if (sizeof(msg) != ret)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Error sending data (command %d byte(s)), %s. "
               "(File: %s, line: %d)\n",
               sizeof(msg), strerror(errno), __FILE__, __LINE__));
         return IFX_ERROR;
      }
   } /* else if (nAction == IE_EXT_CALL_WORKAROUND_EASY508xx) */
   else
   {
#ifndef TAPI_VERSION4
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("Phone No %d: SetAction: ConnType %s, Action %s\n"
             "%s(PCM Ch %d, DataCh %d, PhoneCh %d)\n",
             (int) pPhone->nPhoneNumber, 
             Common_Enum2Name(pConn->nType, TD_rgCallTypeName), psActionName, 
             pPhone->pBoard->pIndentPhone,
             (int) pPhone->nPCM_Ch, (int) pPhone->nDataCh,(int) pPhone->nPhoneCh));
#else /* TAPI_VERSION4 */
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("Phone No %d: %s device %d, PhoneCh %d,\n"
             "%sSetAction: ConnType %s, Action %s\n",
             (int) pPhone->nPhoneNumber,
             pPhone->pBoard->pszBoardName, pPhone->nDev, pPhone->nPhoneCh, 
             pPhone->pBoard->pIndentPhone, 
             Common_Enum2Name(pConn->nType, TD_rgCallTypeName), psActionName));
#endif /* TAPI_VERSION4 */

      if ((LOCAL_CALL == pConn->nType) || (LOCAL_BOARD_CALL == pConn->nType)||
          (LOCAL_PCM_CALL == pConn->nType) ||
          (LOCAL_BOARD_PCM_CALL == pConn->nType))
      {
         /* Set action to local phone, for example tell him that
            we are calling */
         p_dst_phone = pConn->oConnPeer.oLocal.pPhone;

         if (IFX_NULL == p_dst_phone)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("Pointer to destination phone is NULL, target phone is missing!"
                   " (File: %s, line: %d)\n",
                   __FILE__, __LINE__));
            return IFX_ERROR;
         }

#ifndef TAPI_VERSION4
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("SetAction on destination PhoneNo %d: DataCh %d, PhoneCh %d\n",
                (IFX_int32_t)p_dst_phone->nPhoneNumber,
                (int) p_dst_phone->nDataCh, (int) p_dst_phone->nPhoneCh));
#else /* TAPI_VERSION4 */
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("SetAction on destination PhoneNo %d: %s device %d, ch %d\n",
                (IFX_int32_t)p_dst_phone->nPhoneNumber,
                p_dst_phone->pBoard->pszBoardName,
                p_dst_phone->nDev, p_dst_phone->nPhoneCh));
#endif /* TAPI_VERSION4 */

         /* More phones can call us */
         /* Status from where we got command. */
         /* Called phone has always only one connection with caller phone */
         p_dst_phone->nIntEvent_FXS = nAction;
         /* set internal event flag */
         pCtrl->bInternalEventGenerated = IFX_TRUE;
         p_conn = ABSTRACT_GetConnOfLocal(p_dst_phone, pPhone, pConn->nType);
         if (IFX_NULL == p_conn)
         {
            /* Could not get phone connection */
            return IFX_ERROR;
         }
         /* internal event is generated on phone */
         p_dst_phone->pConnFromLocal = p_conn;

         /* set connection data only on first messsage */
         if (IE_RINGING == nAction)
         {
            p_conn->nType = pConn->nType;
            p_conn->oConnPeer.nPhoneNum = pPhone->nPhoneNumber;
            p_conn->oConnPeer.oLocal.pPhone = pPhone;
            p_conn->nFeatureID = pConn->nFeatureID;
      
            if ((LOCAL_PCM_CALL == pConn->nType) ||
                (LOCAL_BOARD_PCM_CALL == pConn->nType))
            {
               /* Caller phone is sending action */
               p_conn->oConnPeer.oLocal.fSlave = IFX_TRUE;
               /* cross timeslots numbers */
               p_conn->oPCM.nTimeslot_RX = pConn->oPCM.nTimeslot_TX;
               p_conn->oPCM.nTimeslot_TX = pConn->oPCM.nTimeslot_RX;
      
               p_conn->nUsedCh = p_dst_phone->nPCM_Ch;
               /* phone fd and pcm fd are the same */
               p_conn->nUsedCh_FD = p_dst_phone->nPhoneCh_FD;
            } /* if ((LOCAL_PCM_CALL == pConn->nType) */
            else if (LOCAL_CALL == pConn->nType)
            {
               p_conn->nUsedCh = p_dst_phone->nPhoneCh;
               p_conn->nUsedCh_FD = p_dst_phone->nPhoneCh_FD;
            }
            else if (LOCAL_BOARD_CALL == pConn->nType)
            {
               p_conn->nUsedCh = p_dst_phone->nDataCh;
               p_conn->nUsedCh_FD = p_dst_phone->nDataCh_FD;
            } /* if ((LOCAL_PCM_CALL == pConn->nType) */
         } /* if (IE_RINGING == nAction) */
      } /* if ((LOCAL_CALL == pConn->nType) ... */
      else if (EXTERN_VOIP_CALL == pConn->nType)
      {
         /* Set action to external phone, for example tell him that
            we are calling */
         msg.fPCM = CALL_FLAG_VOIP;
         msg.nAction = nAction;
         msg.nFeatureID = pConn->nFeatureID;

         /* message sender info */
         msg.nBoard_IDX = pPhone->pBoard->nBoard_IDX;
         msg.nSenderPhoneNum = pPhone->nPhoneNumber;
         msg.nSenderPort = ntohs(pConn->oUsedAddr.sin_port);
         memcpy(msg.MAC, pConn->oUsedMAC, ETH_ALEN);

         /* Receiver phone number */
         msg.nReceiverPhoneNum = pConn->oConnPeer.nPhoneNum;
         /* set receiver address */
         memset(&to_addr, 0, sizeof(struct sockaddr_in));
         to_addr.sin_addr = pConn->oConnPeer.oRemote.oToAddr.sin_addr;
         to_addr.sin_family = pConn->oConnPeer.oRemote.oToAddr.sin_family;
         to_addr.sin_port = htons(ADMIN_UDP_PORT);

         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("%s: Sending MSG using socket %d to %s:%d:\n"
                "%s  - call type %s\n"
                "%s  - sender UDP port %d\n"
                "%s  - sender MAC %02X:%02X:%02X:%02X:%02X:%02X\n"
                "%s  - sender phone number %d\n"
                "%s  - receiver phone number %d\n"
                "%s  - new internal event %s\n"
                "%s  - board idx %d\n"
                "%s  - feature id %d\n",
                pPhone->pCID_Msg->message[CID_IDX_NAME].string.element,
                (int) pCtrl->nAdminSocket,
                inet_ntoa(to_addr.sin_addr), ntohs(to_addr.sin_port),
                pPhone->pBoard->pIndentBoard,(CALL_FLAG_PCM == msg.fPCM)? "PCM" :
                (CALL_FLAG_VOIP == msg.fPCM) ? "EXTERN" : "UNKNOWN",
                pPhone->pBoard->pIndentBoard, msg.nSenderPort,
                pPhone->pBoard->pIndentBoard, msg.MAC[0],
                msg.MAC[1],
                msg.MAC[2],
                msg.MAC[3],
                msg.MAC[4],
                msg.MAC[5],
                pPhone->pBoard->pIndentBoard, msg.nSenderPhoneNum, 
                pPhone->pBoard->pIndentBoard, msg.nReceiverPhoneNum,
                pPhone->pBoard->pIndentBoard, psActionName,
                pPhone->pBoard->pIndentBoard, (int) msg.nBoard_IDX,
                pPhone->pBoard->pIndentBoard, msg.nFeatureID));

#ifdef ABACUS_TEST
         if ((pCtrl->pProgramArg->nTestType == COM_MODULAR_TEST) &&
             (pCtrl->pProgramArg->nSubTestType == COM_SUBTEST_NONE))
         {
            if (IE_ACTIVE == msg.nAction)
            {
               IFX_char_t buf[100];
               sprintf(buf,
                       "UDP_PORT_2_PHONE_MAPPING:"
                       "{CHANNEL %d PHONE %d UDP_PORT %d}%s",
#ifdef TAPI_VERSION4
                       pPhone->nPhoneNumber % 100 - 1,
#else /* TAPI_VERSION4 */
                       (IFX_int32_t) pConn->nUsedCh,
#endif /* TAPI_VERSION4 */
                       (IFX_int32_t) msg.nSenderPhoneNum,
                       (IFX_int32_t) msg.nSenderPort,
                       COM_MESSAGE_ENDING_SEQUENCE);
               COM_SendMessage(nComOutSocket, buf);
            }
         }
#endif /* ABACUS_TEST */

#if (defined(TD_T38_FAX_TEST) && defined(TD_FAX_T38) && defined(AR9))
         TD_T38_FAX_TEST_SET_ACTION(pCtrl, msg);
#endif /* (defined(TD_T38_FAX_TEST) && defined(TD_FAX_T38) && defined(AR9)) */

         ret = sendto(pCtrl->nAdminSocket, (IFX_char_t *) &msg, sizeof(msg), 0,
                      (struct sockaddr *) &to_addr, sizeof(to_addr));
         if (sizeof(msg) != ret)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("Error sending data (command %d byte(s)), %s. "
                   "(File: %s, line: %d)\n",
                   sizeof(msg), strerror(errno), __FILE__, __LINE__));
            return IFX_ERROR;
         }
      } /* else if (EXTERN_VOIP_CALL == pConn->nType) */
      else if (PCM_CALL == pConn->nType)
      {
         msg.nAction = nAction;
         /* send board MAC address */
         memcpy(msg.MAC, pCtrl->oTapidemo_MAC_Addr, ETH_ALEN);
         /** \todo: check if remote side handles port in host format */
         msg.nSenderPort = ntohs(pConn->oUsedAddr.sin_port);
         msg.fPCM = CALL_FLAG_PCM;
         msg.nBoard_IDX = pPhone->pBoard->nBoard_IDX;
         /* Get timeslot RX index */
         msg.nTimeslot_RX = pConn->oPCM.nTimeslot_RX;
         /* Get timeslot TX index */
         msg.nTimeslot_TX = pConn->oPCM.nTimeslot_TX;

         msg.nFeatureID = pConn->nFeatureID;
         msg.nReceiverPhoneNum = pConn->oConnPeer.nPhoneNum;
         msg.nSenderPhoneNum = pPhone->nPhoneNumber;

         /* Set action to PCM phone, for example tell him that we are calling */
         memset(&to_addr, 0, sizeof(struct sockaddr_in));
         to_addr.sin_addr = pConn->oConnPeer.oPCM.oToAddr.sin_addr;
         to_addr.sin_family = pConn->oConnPeer.oPCM.oToAddr.sin_family;
         to_addr.sin_port = htons(ADMIN_UDP_PORT);

#ifndef TAPI_VERSION4
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("%s: Sending MSG using PCM ch %d, socket %d to %s:%d:\n"
                "%s  - call type %s\n"
                "%s  - sender phone number %d\n"
                "%s  - receiver phone number %d\n"
                "%s  - new internal event %s\n"
                "%s  - board idx %d\n"
                "%s  - timeslot RX %d\n"
                "%s  - timeslot TX %d\n",
                pPhone->pCID_Msg->message[CID_IDX_NAME].string.element,
                (int) pConn->nUsedCh, (int) pCtrl->nAdminSocket,
                inet_ntoa(to_addr.sin_addr), ntohs(to_addr.sin_port),
                pPhone->pBoard->pIndentBoard,
                (CALL_FLAG_PCM == msg.fPCM)? "PCM" :
                (CALL_FLAG_VOIP == msg.fPCM) ? "EXTERN" : "UNKNOWN",
                pPhone->pBoard->pIndentBoard, msg.nSenderPhoneNum, 
                pPhone->pBoard->pIndentBoard, msg.nReceiverPhoneNum,
                pPhone->pBoard->pIndentBoard, psActionName,
                pPhone->pBoard->pIndentBoard,(int) msg.nBoard_IDX,
                pPhone->pBoard->pIndentBoard, msg.nTimeslot_RX, 
                pPhone->pBoard->pIndentBoard, msg.nTimeslot_TX));
#endif /* TAPI_VERSION4 */

         ret = sendto(pCtrl->nAdminSocket, (IFX_char_t *) &msg, sizeof(msg), 0,
                      (struct sockaddr *) &to_addr, sizeof(to_addr));

         if (sizeof(msg) != ret)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("Error sending data (command %d byte), %s. "
                   "(File: %s, line: %d)\n",
                   sizeof(msg), strerror(errno), __FILE__, __LINE__));
            return IFX_ERROR;
         }
      } /* else if (PCM_CALL == pConn->nType) */
      else if (pConn->nType == FXO_CALL)
      {
         /* Nothing at the moment */
      } /* else if (pConn->nType == FXO_CALL) */
      else
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Wrong call type, %d. (File: %s, line: %d)\n",
                (int) pConn->nType, __FILE__, __LINE__));
      }
   } /* if(nAction == IE_EXT_CALL_PROCEEDING||nAction == IE_EXT_CALL_WRONG_NUM)*/

   return IFX_SUCCESS;
} /* TAPIDEMO_SetAction() */

/**
   Initialize administration socket which will be used to manage phone
   connections.

   \return socket number or NO_SOCKET if error.
*/
IFX_int32_t TAPIDEMO_InitAdminSocket(IFX_void_t)
{
   IFX_int32_t s;
   IFX_char_t ip_tos = 0x68;
   struct sockaddr_in my_addr;

   s = socket(PF_INET, SOCK_DGRAM, 0);
   if (NO_SOCKET == s)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Can't create admin socket, %s. (File: %s, line: %d)\n",
             strerror(errno), __FILE__, __LINE__));
      return NO_SOCKET;
   }

   /* Socket Option Level IP, IP TOS field shall be set */
   setsockopt (s, IPPROTO_IP, IP_TOS, &ip_tos, sizeof(ip_tos));

   bzero((IFX_char_t*) &my_addr, (IFX_int32_t) sizeof(my_addr));
   my_addr.sin_family = AF_INET;
   my_addr.sin_port = htons(ADMIN_UDP_PORT);

   if (oCtrlStatus.pProgramArg->oMy_IP_Addr.sin_addr.s_addr)
   {
      my_addr.sin_addr.s_addr =
         oCtrlStatus.pProgramArg->oMy_IP_Addr.sin_addr.s_addr;
   }
   else
   {
      my_addr.sin_addr.s_addr = htonl(INADDR_ANY);
   }

   if (-1 == (bind(s, (struct sockaddr *) &my_addr, sizeof(struct sockaddr))))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Can't bind admin socket to port, %s. (File: %s, line: %d)\n",
            strerror(errno), __FILE__, __LINE__));
      close(s);
      return NO_SOCKET;
   }

#ifndef VXWORKS
   /* make the socket non blocking */
   fcntl(s, F_SETFL, O_NONBLOCK);
#endif /* VXWORKS */

   TRACE(TAPIDEMO, DBG_LEVEL_LOW, ("ADMIN socket is %d\n", (int) s));

   return s;
} /* TAPIDEMO_InitAdminSocket() */

/**
   Handle data from ADMIN socket.

   \param pCtrl  - pointer to status control structure

   \return IFX_SUCCESS message received and handled otherwise IFX_ERROR
*/
IFX_return_t TAPIDEMO_HandleCallSeq(CTRL_STATUS_t* pCtrl)
{
   static IFX_char_t buf[MAX_PACKET_SIZE];
   IFX_int32_t size = 0;
   IFX_int32_t ret = 0;
   IFX_int32_t event = 0;
   IFX_uint8_t from_MAC[ETH_ALEN];
   IFX_int32_t from_port = 0;
   IFX_int32_t board_idx = 0;
   COMM_MSG_t msg = {0};
   struct sockaddr_in addr_from;
   PHONE_t* phone = IFX_NULL;
   CONNECTION_t* conn = IFX_NULL;
   IFX_boolean_t bFreeConn = IFX_FALSE;
   IFX_int32_t feature_id = 0;
   /* message receiver and sender phone numbers */
   IFX_int32_t nReciverPhoneNum;
   IFX_int32_t nSenderPhoneNum;
   IFX_char_t* psEventName = IFX_NULL;
#ifdef EASY336
   SVIP_libStatus_t SVIP_libRet;
#endif

   /* check input arguments */
   TD_PTR_CHECK(pCtrl, IFX_ERROR);

   size = sizeof(addr_from);

   /* Get data from socket */
   ret = recvfrom(pCtrl->nAdminSocket, buf, sizeof(buf), 0,
                  /* Got data from address:port */
                  (struct sockaddr *) &addr_from,
                  /* Must be like that, otherwise you get compiler
                     warnings */
                  (int *) (IFX_int32_t) &size);

   if (0 > ret)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("No data read from ADMIN socket, %s. (File: %s, line: %d)\n",
             strerror(errno), __FILE__, __LINE__));
      return IFX_ERROR;
   }
   else if (0 == ret)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_LOW, ("No data read from ADMIN socket.\n"));
      return IFX_SUCCESS;
   }
   else if (sizeof(msg) != ret)
   {
      /* Could happen if wrong data was sent on this port possible voice packet
         or different versions of tapidemo with different message length. */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH, ("Got wrong data on ADMIN socket.\n"));
      return IFX_ERROR;
   }
#ifdef ABACUS_TEST
   /* change peer address */
   if (IFX_ERROR == COM_AbacusSetGetPort(&addr_from, ABACUS_TEST_IN_PROGRESS))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Error in Abacus Test function. (File: %s, line: %d)\n",
             __FILE__, __LINE__));
      return IFX_ERROR;
   }
#endif

   /* When we reach this point, we received our message, hopefully :-) */
   memcpy(&msg, &buf[0], sizeof(msg));

   if ((COMM_MSG_START_FLAG != msg.nMarkStart) ||
       (COMM_MSG_END_FLAG != msg.nMarkEnd))
   {
      /* If we came here then wrong message was received */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Received wrong msg from ADMIN socket %d. (File: %s, line: %d)\n",
             (int) pCtrl->nAdminSocket, __FILE__, __LINE__));
      return IFX_ERROR;
   }

   /* event to send to phone */
   event = msg.nAction;

   /* set pointer string with action name */
   psEventName = Common_Enum2Name(event, TD_rgIE_EventsName);

   /* check if name was set, if name was not set, then invalid value
      of TD_rgIE_EventsName is set */
   TD_PTR_CHECK(psEventName, IFX_ERROR);

   if(event  == IE_EXT_CALL_WORKAROUND_EASY508xx)
   {
      /* It is workaround, 
      which avoid problem with first external/PCM connection to EASY508xx.
      We just ignore this type of action */
      return IFX_SUCCESS;
   }

   /* get feature id */
   feature_id = msg.nFeatureID;

   /* Caller stuff */
   /* get board index (used to get board from pCtrl->rgoBoards[]) */
   board_idx = msg.nBoard_IDX;
   /* get message sending phone number */
   nSenderPhoneNum = msg.nSenderPhoneNum;
   memcpy(from_MAC, msg.MAC, ETH_ALEN);
   from_port = msg.nSenderPort;

   /* Called stuff */
   /* get message receiving phone number */
   nReciverPhoneNum = msg.nReceiverPhoneNum;

   /* Print message data */
   TRACE(TAPIDEMO, DBG_LEVEL_LOW,
         ("Received MSG from %s:%d with:\n"
          " - call type %s\n",
          inet_ntoa(addr_from.sin_addr), ntohs(addr_from.sin_port),
          (CALL_FLAG_PCM == msg.fPCM)? "PCM" : (CALL_FLAG_VOIP == msg.fPCM) ?
          "EXTERN" : "UNKNOWN"));
   /* messages currently are received only for PCM_CALL and EXTERN_VOIP_CALL,
      UDP port is only sent during EXTERN_VOIP_CALL, but if some error ocurres
      e.g. wrong call type, then print this data also. */
   if (CALL_FLAG_PCM != msg.fPCM)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            (" - sender UDP port %d\n", msg.nSenderPort));

   }
   TRACE(TAPIDEMO, DBG_LEVEL_LOW,
         (" - sender MAC %02X:%02X:%02X:%02X:%02X:%02X\n"
          " - message sender phone number %d\n"
          " - message receiver phone number %d\n"
          " - board idx %d\n"
          " - new internal event %s\n"
          " - feature id %d\n",
          msg.MAC[0],
          msg.MAC[1],
          msg.MAC[2],
          msg.MAC[3],
          msg.MAC[4],
          msg.MAC[5],
          msg.nSenderPhoneNum, msg.nReceiverPhoneNum,
          msg.nBoard_IDX,
          psEventName,
          msg.nFeatureID));
   /* messages currently are received only for PCM_CALL and EXTERN_VOIP_CALL,
      timeslots are only sent during PCM_CALL, but if some error ocurres
      e.g. wrong call type, then print this data also. */
   if (CALL_FLAG_VOIP != msg.fPCM)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            (" - timeslot RX %d\n"
             " - timeslot TX %d\n",
             msg.nTimeslot_RX, msg.nTimeslot_TX));
   }

   /* Check if phone number is avaible */
   phone = ABSTRACT_GetPHONE_OfCalledNum(pCtrl, msg.nReceiverPhoneNum);

   /* caller is dialing number but, didn't finish dialing full number */
   if (IFX_NULL == phone)
   {
      /* If receive call setup messsage, send answer */
      if (IE_EXT_CALL_SETUP == event)
      {
         /* phoney phone and connection to allow usage of TAPIDEMO_SetAction() */
         CONNECTION_t oConnTmp;
         PHONE_t oPhoneTmp;
         /* Received called number is incomplete or incorrect */
         /* Allocate temporary connection structure and send msg to caller */
         bFreeConn = IFX_TRUE;
         conn = &oConnTmp;
         memset(conn, 1, sizeof(CONNECTION_t));

         /* Allocate temporary phone structure and send msg to caller */
         phone = &oPhoneTmp;
         memset(phone, 1, sizeof(PHONE_t));

         conn->nType = EXTERN_VOIP_CALL;
         memcpy(conn->oConnPeer.oRemote.MAC, from_MAC, ETH_ALEN);
         conn->oConnPeer.oRemote.oToAddr = addr_from;
         conn->oConnPeer.nPhoneNum = msg.nSenderPhoneNum;

#ifndef EASY336
         memcpy(conn->oUsedMAC, pCtrl->oTapidemo_MAC_Addr, ETH_ALEN);
#else
         SVIP_libRet = SVIP_libMAC_AddressForVoIPGet(addr_from.sin_addr.s_addr, 
            conn->oUsedMAC);
         if (SVIP_libRet != SVIP_libStatusOk)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Err: SVIP_libMAC_AddressForVoIPGet returned error %d "
                "(File: %s, line: %d)\n",
                SVIP_libRet, __FILE__, __LINE__));
            return IFX_ERROR;
         }
#endif
         conn->oUsedAddr.sin_addr = pCtrl->oTapidemo_IP_Addr.sin_addr;

         /* We don't know phone number, so we don't know below values */
         conn->nUsedSocket = NO_SOCKET;
         conn->nUsedCh_FD = NO_CHANNEL_FD;
         conn->nUsedCh = NO_USED_CHANNEL;
         conn->oUsedAddr.sin_port = ADMIN_UDP_PORT;

         /* We need to send correct called phone number, it's needed in
         function ABSTRACT_GetConnOfPhone for second phone */
         phone->nPhoneNumber = nReciverPhoneNum;

         /* It is workaround, 
         which avoid problem with first external/PCM connection to EASY508xx */
         if (IFX_ERROR == TAPIDEMO_SetAction(pCtrl, phone, conn, 
                                             IE_EXT_CALL_WORKAROUND_EASY508xx))
         {
            return IFX_ERROR;
         }

         if (MIN_PHONE_NUM <= msg.nReceiverPhoneNum)
         {
            /* Received phone number is not correct, if another digit will be dialed, then dialed phone number ,
            inform about it phone on second board */
            if (IFX_ERROR == TAPIDEMO_SetAction(pCtrl, phone, conn, 
                                                IE_EXT_CALL_WRONG_NUM))
            {
               return IFX_ERROR;
            }
         }
         else
         {
            /* Received phone number is not completed,
            so inform about it second board */
            if (IFX_ERROR == TAPIDEMO_SetAction(pCtrl, phone, conn, 
                                                IE_EXT_CALL_PROCEEDING))
            {
               return IFX_ERROR;
            }
         } /* if (pCtrl->nSumAnalogCh < msg.nCalledPhoneNum) */
         return IFX_SUCCESS;
      }
      else
      {
         /* For rest cases print error */
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Error get phone. (File: %s, line: %d)\n",
                __FILE__, __LINE__));
         return IFX_ERROR;
      } /* if (IE_EXT_CALL_SETUP == event) */
   } /* if (IFX_NULL == phone) */
   else
   {
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("%s: Received event %s on ADMIN socket\n",
             phone->pCID_Msg->message[CID_IDX_NAME].string.element,
             psEventName));

      if (CALL_FLAG_PCM == msg.fPCM)
      {
         /** \todo: this function should be reviewed */
         conn = ABSTRACT_GetConnOfPhone(phone, PCM_CALL,
                                        nReciverPhoneNum, nSenderPhoneNum,
                                        addr_from, &bFreeConn);
         if (conn == IFX_NULL)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("Error get connection. (File: %s, line: %d)\n",
                   __FILE__, __LINE__));
            return IFX_ERROR;
         } /* if (conn == IFX_NULL) */

         if (bFreeConn)
         {
            conn->nType = PCM_CALL;
            conn->oConnPeer.oPCM.oToAddr = addr_from;
            conn->oConnPeer.oPCM.oToAddr.sin_port = from_port;
            conn->oConnPeer.nPhoneNum = msg.nSenderPhoneNum;

            conn->oUsedAddr.sin_addr = pCtrl->oTapidemo_IP_Addr.sin_addr;
            conn->nUsedSocket = phone->nSocket;
            conn->nUsedCh_FD = PCM_GetFD_OfCh(phone->nPCM_Ch, phone->pBoard);
            conn->nUsedCh = phone->nPCM_Ch;
            conn->nFeatureID = feature_id;
         } /* if (free_conn) */

         /* Set new event for phone */
         phone->nIntEvent_FXS = event;

         if (!pCtrl->pProgramArg->oArgFlags.nPCM_Master)
         {
            if (IE_RINGING == event || IE_RINGBACK == event)
            {
               /* Set PCM timeslots */
               conn->oPCM.nTimeslot_RX = msg.nTimeslot_TX;
               conn->oPCM.nTimeslot_TX = msg.nTimeslot_RX;
            }
         }

#ifndef TAPI_VERSION4
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("%s: PCM connection on channel %d:\n"
                "%s  - used socket: %d\n" 
                "%s  - timeslot RX %d TX %d\n"
                "%s  - to: %s:%d\n", 
                phone->pCID_Msg->message[CID_IDX_NAME].string.element,
                (int) conn->nUsedCh,
                phone->pBoard->pIndentBoard,
                (int) conn->nUsedSocket,
                phone->pBoard->pIndentBoard,
                conn->oPCM.nTimeslot_RX, conn->oPCM.nTimeslot_TX,
                phone->pBoard->pIndentBoard,
                inet_ntoa(conn->oUsedAddr.sin_addr),
                ntohs(conn->oUsedAddr.sin_port)));
#endif /* TAPI_VERSION4 */
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("%s  - from %s:%d\n",
                phone->pBoard->pIndentBoard,
                inet_ntoa(conn->oConnPeer.oPCM.oToAddr.sin_addr),
                ntohs(conn->oConnPeer.oPCM.oToAddr.sin_port)));
      } /* if (CALL_FLAG_PCM == msg.fPCM) */
      else if (CALL_FLAG_VOIP == msg.fPCM)
      {
         /* note: this function should be reviewed */
         conn = ABSTRACT_GetConnOfPhone(phone, EXTERN_VOIP_CALL,
                                        nReciverPhoneNum, nSenderPhoneNum,
                                        addr_from, &bFreeConn);
         if (IFX_NULL == conn)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("Error get connection. (File: %s, line: %d)\n",
                   __FILE__, __LINE__));
            return IFX_ERROR;
         } /* if (IFX_NULL == conn) */

         if (bFreeConn)
         {
            conn->nType = EXTERN_VOIP_CALL;

            memcpy(conn->oConnPeer.oRemote.MAC, from_MAC, ETH_ALEN);
            conn->oConnPeer.oRemote.oToAddr = addr_from;
            conn->oConnPeer.nPhoneNum = msg.nSenderPhoneNum;

#ifndef EASY336
            memcpy(conn->oUsedMAC, pCtrl->oTapidemo_MAC_Addr, ETH_ALEN);
#else
            SVIP_libRet = SVIP_libMAC_AddressForVoIPGet(
               addr_from.sin_addr.s_addr, conn->oUsedMAC);
            if (SVIP_libRet != SVIP_libStatusOk)
            {
               TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("Err: SVIP_libMAC_AddressForVoIPGet returned error %d "
                   "(File: %s, line: %d)\n",
                   SVIP_libRet, __FILE__, __LINE__));
               return IFX_ERROR;
            }
#endif
            conn->oUsedAddr.sin_addr = pCtrl->oTapidemo_IP_Addr.sin_addr;
            conn->nUsedSocket = phone->nSocket;
#ifdef TAPI_VERSION4
            if (!phone->pBoard->fSingleFD)
#endif
            {
               conn->nUsedCh_FD = VOIP_GetFD_OfCh(phone->nDataCh,
                                                  phone->pBoard);
            }
            conn->nUsedCh = phone->nDataCh;

#ifdef TAPI_VERSION4
            if (phone->pBoard->fUseSockets)
#endif /* TAPI_VERSION4 */
            {
               if (!pCtrl->pProgramArg->oArgFlags.nQos)
               {
                  conn->oUsedAddr.sin_port =
                     VOIP_GetSocketPort(conn->nUsedSocket, pCtrl);
               }
            }
         } /* if (free_conn) */

#if (defined(TD_T38_FAX_TEST) && defined(TD_FAX_T38) && defined(AR9))
         if (IE_RINGBACK == event || IE_RINGING == event)
         {
            TD_T38_FAX_TEST_HANDLE_CALL_SEQUENCE(pCtrl, msg, phone);
         }
#endif /* (defined(TD_T38_FAX_TEST) && defined(TD_FAX_T38) && defined(AR9)) */
         if (IE_EXT_CALL_SETUP == event)
         {
            /* It is workaround, 
            which avoid problem with first external connection to EASY508xx */
            if (IFX_ERROR == TAPIDEMO_SetAction(pCtrl, phone, conn, 
                                                IE_EXT_CALL_WORKAROUND_EASY508xx))
            {
               return IFX_ERROR;
            }
            /* Received phone number is correct,
               so inform about it second phone */
            if (IFX_ERROR == TAPIDEMO_SetAction(pCtrl, phone, conn, 
                                                IE_EXT_CALL_ESTABLISHED))
            {
               return IFX_ERROR;
            }

            return IFX_SUCCESS;
         } /* if (IE_EXT_CALL_SETUP == event) */
         /* save port number, caller sends it with IE_RINGING,
            called phone sends it with IE_ACTIVE action */
         else if (IE_ACTIVE == event || IE_RINGING == event)
         {
            conn->oConnPeer.oRemote.oToAddr.sin_port = htons(from_port);
         }

         memcpy(conn->oConnPeer.oRemote.MAC, from_MAC, ETH_ALEN);
         conn->nFeatureID = feature_id;

         /* Set new event for phone */
         phone->nIntEvent_FXS = event;

         if ((phone->rgStateMachine[CONFIG_SM].nState != S_READY) ||
             (event == IE_CONFIG_PEER))
         {
            phone->nIntEvent_Config = event;
         }

         if (IE_ACTIVE == event)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_LOW,
                  ("%s: VOIP connection on channel %d:\n",
                   phone->pCID_Msg->message[CID_IDX_NAME].string.element,
                   phone->nPhoneCh));
#ifndef TAPI_VERSION4
            TRACE(TAPIDEMO, DBG_LEVEL_LOW,
                  ("%s - used socket: %d\n",
                   phone->pBoard->pIndentBoard, 
                   conn->nUsedSocket));
#else /* TAPI_VERSION4 */
            TRACE(TAPIDEMO, DBG_LEVEL_LOW,
                  ("%s - device: %d\n",
                   phone->pBoard->pIndentBoard, 
                   phone->nDev));
#endif /* TAPI_VERSION4 */
            TRACE(TAPIDEMO, DBG_LEVEL_LOW,
                  ("%s - to: %s:%d\n",
                   phone->pBoard->pIndentBoard, 
                   inet_ntoa(conn->oUsedAddr.sin_addr),
                   ntohs(conn->oUsedAddr.sin_port)));
            TRACE(TAPIDEMO, DBG_LEVEL_LOW,
                  ("%s - from: %s:%d\n",
                   phone->pBoard->pIndentBoard, 
                   inet_ntoa(conn->oConnPeer.oRemote.oToAddr.sin_addr),
                   ntohs(conn->oConnPeer.oRemote.oToAddr.sin_port)));
         }
      } /* if (CALL_FLAG_PCM == msg.fPCM) */
      else
      {
         /* If we came here then wrong message was received */
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Received Message with UNKNOWN call flag. "
                "(File: %s, line: %d)\n", __FILE__, __LINE__));
      }

      /* Action from other phone received, handle it. */
      while ( phone->nIntEvent_Config != IE_NONE )
      {
         ST_HandleState_Config(pCtrl, phone, conn);
      }

      while ( phone->nIntEvent_FXS!= IE_NONE)
      {
         ST_HandleState_FXS(pCtrl, phone, conn);
      }

      return IFX_SUCCESS;
   } /* if (IFX_NULL == phone) */
   
} /* ADMIN_HandleCallSeq() */

/** ---------------------------------------------------------------------- */


#ifdef LINUX
/**
   Initialize socket for syslog.

   \return socket number or NO_SOCKET if error.
*/
IFX_int32_t TAPIDEMO_InitSyslogSocket(IFX_void_t)
{

   IFX_int32_t s;

   struct sockaddr_in my_addr;
   bzero((IFX_char_t*) &my_addr, (IFX_int32_t) sizeof(my_addr));

   my_addr.sin_addr.s_addr = INADDR_ANY;
   my_addr.sin_port = 0;
   my_addr.sin_family = AF_INET;

   s = socket(AF_INET, SOCK_DGRAM, 0);
   if (NO_SOCKET == s)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("InitSyslogSocket: Can't create UDP socket, %s. "
             "(File: %s, line: %d)\n",
             strerror(errno), __FILE__, __LINE__));
      return NO_SOCKET;
   }

   if (-1 == (bind(s, (struct sockaddr *) &my_addr, sizeof(struct sockaddr))))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("InitSyslogSocket: Can't bind to port, %s. (File: %s, line: %d)\n",
             strerror(errno), __FILE__, __LINE__));
      close(s);
      return NO_SOCKET;
   }

   return s;
}

/**
   Send data to remote syslog

   \param pBuffer - pointer to message
   \param nSyslogSocket - socket used for communication with remote Syslog

   \return  IFX_SUCCESS if everything OK, otherwise IFX_ERROR
*/
IFX_return_t TAPIDEMO_SendLogRemote(IFX_char_t* pBuffer,
                                    IFX_int32_t nSyslogSocket)
{
   IFX_int32_t ret = 0;
   IFX_int32_t nLength;

   /* It is used to build external syslog message */
   IFX_char_t pExtSyslogMsg[MAX_EXT_SYSLOG_MSG_LEN] = {0};

   struct sockaddr_in server_addr;

   TD_PTR_CHECK(pBuffer, IFX_ERROR);

   bzero((IFX_char_t*) &server_addr, (IFX_int32_t) sizeof(server_addr));

   server_addr.sin_addr.s_addr =
      oTraceRedirection.oSyslog_IP_Addr.sin_addr.s_addr;
   server_addr.sin_port = oTraceRedirection.oSyslog_IP_Addr.sin_port;
   server_addr.sin_family = AF_INET;

   /* Correct syslog message has The Priority part (PRI) at the beginning.*/
   /* The PRI part MUST have three, four, or five characters and will be */
   /* bound with angle brackets as the first and last characters. */
   /* The code set used in this part MUST be seven-bit ASCII in an eight-bit field. */
   /* The number contained within these angle brackets is known as */ 
   /* the Priority value and represents both the Facility and */
   /* Severity. The Priority value is calculated by first multiplying the Facility */
   /* number by 8 and then adding the numerical value of the Severity. */
   /* Source: RFC3164 */

   /* In TAPIDEMO, Facility is set on 1 (user-level messages) and */
   /* Severity is set on 7 (Debug: debug-level messages): 1*8+7 = 15 */
   /* We need buffer with lenght + PRI' length + '\n' */

   /* Add PRI part */
   strcpy(pExtSyslogMsg, "<15>");
   /* Add message part */
   strcat(pExtSyslogMsg, pBuffer);
   nLength = strlen(pExtSyslogMsg);

   ret = sendto(nSyslogSocket,  pExtSyslogMsg, nLength, 0,
                (struct sockaddr *) &server_addr, sizeof(server_addr));

   if (ret != nLength)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Error calling sendto(), %s, socket %d"
             " (File: %s, line: %d)\n",
             strerror(errno), nSyslogSocket,
             __FILE__, __LINE__));

      return IFX_ERROR;
   }

   return IFX_SUCCESS;
}

/**
   Split data on smaller packets and send them.
   Data are splitted according to end of line. Every line is separate packet.

   \param pBuffer - pointer to message

   \return  IFX_SUCCESS if everything OK, otherwise IFX_ERROR
*/
IFX_return_t TAPIDEMO_SplitAndSendLog(IFX_char_t* pBuffer)
{
   IFX_return_t ret;
   IFX_char_t* pPresentCh;
   IFX_char_t* pPreviusCh;

   TD_PTR_CHECK(pBuffer, IFX_ERROR);
   TD_PARAMETER_CHECK((MAX_LOG_LEN <= strlen(pBuffer)),
                      strlen(pBuffer), IFX_ERROR);

   /* It is used to keep split data */
   IFX_char_t pTmpBuffer[MAX_LOG_LEN] = {0};

   /* If there isn't char '\n' at the end of string, some data can be missed */
   pBuffer[strlen(pBuffer)] = EOL;

   /* Locate first occurrence of character '\n' */
   pPresentCh = strchr(pBuffer, EOL);
   pPreviusCh = pBuffer;
   
   while (pPresentCh != NULL)
   {
      if (pPresentCh-pPreviusCh+1 < MAX_LOG_LEN)
      {
         strncpy(pTmpBuffer, pPreviusCh, pPresentCh-pPreviusCh+1);

         if (oCtrlStatus.pProgramArg->nTraceRedirection ==
             TRACE_REDIRECTION_SYSLOG_LOCAL)
         {
            /* Redirect log to local syslog */
            syslog(LOG_DEBUG, pTmpBuffer);
         } /* TRACE_REDIRECTION_SYSLOG_LOCAL */
         else if (oCtrlStatus.pProgramArg->nTraceRedirection ==
                  TRACE_REDIRECTION_SYSLOG_REMOTE)
         {
            /* Redirect log to remote syslog */
            ret = TAPIDEMO_SendLogRemote(pTmpBuffer, 
                                         oTraceRedirection.nSyslogSocket);
            if (ret == IFX_ERROR)
            {
               TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                     ("Error sending log to remote syslog. "
                      "(File: %s, line: %d)\n",
                      __FILE__, __LINE__));
               return IFX_ERROR;
            }
         } /* TRACE_REDIRECTION_SYSLOG_REMOTE */
         else if (oCtrlStatus.pProgramArg->nTraceRedirection ==
                  TRACE_REDIRECTION_FILE)
         {
            ret = write(oTraceRedirection.nFileLogFD, pTmpBuffer, strlen(pTmpBuffer));
            if (0 > ret)
            {
               TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                     ("Error writing log to file, %s. (File: %s, line: %d)\n",
                      strerror(errno), __FILE__, __LINE__));
               return IFX_ERROR;
            }
         } /* TRACE_REDIRECTION_FILE */
         else
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("Unknow type of log redirection (File: %s, line: %d)\n",
                   __FILE__, __LINE__));
            return IFX_ERROR;
         }
         /* Clean tmp buffer */
         memset(pTmpBuffer, 0, MAX_LOG_LEN);
      }
      else
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Message too long. (File: %s, line: %d)\n",
                __FILE__, __LINE__));
      }

      pPreviusCh = pPresentCh+1;
      /* Locate next occurrence of character '\n' */
      pPresentCh = strchr(pPreviusCh, EOL);
   } /* while */

   return IFX_SUCCESS;
}
/**
   Prepare logs' redirection, e.g. set up syslog, open file.

   \return IFX_SUCCESS if everything OK, otherwise IFX_ERROR
*/
IFX_return_t TAPIDEMO_PreapareLogRedirection(IFX_void_t)
{
   if (oCtrlStatus.pProgramArg->nTraceRedirection ==
       TRACE_REDIRECTION_SYSLOG_LOCAL)
   {
      /* Start redirecting logs to local syslog */
      openlog("Tapidemo", LOG_PID, LOG_USER);

      oTraceRedirection.fStartRedirectLogs = IFX_TRUE;

      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("Start redirecting logs to local syslog\n"));
   }
   else if (oCtrlStatus.pProgramArg->nTraceRedirection ==
            TRACE_REDIRECTION_SYSLOG_REMOTE)
   {
      /* Start redirecting logs to remote syslog */
      IFX_int32_t s;
      s = TAPIDEMO_InitSyslogSocket();
      if (s == NO_SOCKET)
      {
         oTraceRedirection.fStartRedirectLogs = IFX_TRUE;

         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Can not redirect logs to remote syslog "
                "on address %s:%d\n",
                inet_ntoa(oTraceRedirection.oSyslog_IP_Addr.sin_addr),
                oTraceRedirection.oSyslog_IP_Addr.sin_port));
         return IFX_ERROR;
      }
      else
      {
         oTraceRedirection.fStartRedirectLogs = IFX_TRUE;

         oTraceRedirection.nSyslogSocket = s;
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("Start redirecting logs to remote syslog "
                "on address %s:%d\n",
                inet_ntoa(oTraceRedirection.oSyslog_IP_Addr.sin_addr),
                oTraceRedirection.oSyslog_IP_Addr.sin_port));
      }
   }
   else if (oCtrlStatus.pProgramArg->nTraceRedirection ==
            TRACE_REDIRECTION_FILE)
   {
      /* Start redirecting logs to file */
      IFX_char_t full_filename[256] = {0};
      strcpy(full_filename, oTraceRedirection.sPathToLogFiles);
      oTraceRedirection.nFileLogFD =
         open(full_filename, O_CREAT | O_WRONLY | O_APPEND, 0666);
      if (oTraceRedirection.nFileLogFD < 0)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Can not create / open log file, %s\n"
                "File: %s, line: %d)\n",
                 strerror(errno), __FILE__, __LINE__));
         return IFX_ERROR;
      }
      else
      {
         oTraceRedirection.fStartRedirectLogs = IFX_TRUE;

         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("Start redirecting logs to file %s ",
                full_filename));
      }
   }
   else
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Error on setting redirection. "
             "File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }
   return IFX_SUCCESS;
}

/**
   Run logs' redirection - read data from pipe and redirect them.
 
   \param nPipeFd - pipe file descriptor

*/
IFX_void_t TAPIDEMO_RunLogRedirection(IFX_int32_t nPipeFd)
{
   /* Read data from pipe and redirect them. */
   IFX_int32_t bytes_read;
   IFX_char_t buf[MAX_LOG_LEN] = {0};

   IFX_return_t ret = IFX_SUCCESS;
   fd_set fds;
   FD_ZERO(&fds);
   FD_SET(nPipeFd, &fds); 

   while(1)
   {
      ret = select(nPipeFd+1, &fds, IFX_NULL, IFX_NULL, IFX_NULL);
      if (ret > 0) 
      {
         if(FD_ISSET(nPipeFd, &fds))
         {
            /* Read data */
            bytes_read = read(nPipeFd, buf, MAX_LOG_LEN - 1);

            /* Check how many bytes were read */
            if (0 > bytes_read)
            {
               TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                     ("Message read error - end task. (File: %s, line: %d)\n",
                      __FILE__, __LINE__));
               break;
            }
            else if (0 == bytes_read)
            {
               TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                     ("No data was read from pipe - end task.\n"));
               break;
            }
            else if (MAX_LOG_LEN > bytes_read) 
            {
               /* Check msg length */
               if (MAX_LOG_LEN < strlen(buf))
               {
                  TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                        ("Message too long. (File: %s, line: %d)\n",
                         __FILE__, __LINE__));
               }
               else
               {
                  /* Split and send log to new destination */
                  TAPIDEMO_SplitAndSendLog(buf);
               }
            }
            else
            {
                  TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                        ("Received too long logs. (File: %s, line: %d)\n",
                         __FILE__, __LINE__));
            } /* if (0 > bytes_read) */

            /* Clean up buffer */
            memset(&buf, 0, MAX_LOG_LEN);
         } /* if(FD_ISSET(nPipeFd, &fds)) */
      }
      else if (0 == ret)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Timeout on select(). (File: %s, line: %d)\n",
              __FILE__, __LINE__));
         break;
      }
      else
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Error in select(). (File: %s, line: %d)\n",
              __FILE__, __LINE__));
         break;
      } /* if (ret > 0) */
   } /* while(1) */

   /* If read returns zero, it means pipe has been closed by second process */
   close(nPipeFd);

   /* End with task */
   exit(0);
}

/**
   Run Tapidemo as daemon

   \return IFX_SUCCESS if everything OK, otherwise IFX_ERROR
*/
IFX_return_t TAPIDEMO_CreateDaemon(CTRL_STATUS_t* pCtrl)
{
   /* Start creating daemon */
   pid_t pid1, pid2, sid;

   /* Prepare pipe, which will be used to communication between process */
   if (0 != pipe(nPipeFd))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Can not create pipe. Error on pipe %s\n"
             "File: %s, line: %d)\n",
             strerror(errno), __FILE__, __LINE__));
      return IFX_ERROR;
   }

   /* Create new process - TAPIDEMO */
   pid2  = fork();
   if (pid2 < 0)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Can not create daemon. Error on fork %s\n"
             "File: %s, line: %d)\n",
             strerror(errno), __FILE__, __LINE__));

      return IFX_ERROR;
   }
   else if (pid2 != 0)
   {
      /* Close parent process */
      close(nPipeFd[0]);
      close(nPipeFd[1]);
      exit(0);
   }

   /* Create new process responsible for logs' redirection */
   pid1  = fork();
   if (pid1 < 0)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Can not create daemon. Error on fork %s\n"
             "File: %s, line: %d)\n",
             strerror(errno), __FILE__, __LINE__));
      return IFX_ERROR;
   }
   else if (pid1 == 0)
   {
      /* Child process - just read data from pipe and redirect it */

      /* We only read from pipe, so we can close descriptor 
         responsible for writing */
      close(nPipeFd[1]);
      /* Waits in while(read()) for data on pipe, on data read error child
         process calls exit(0) */
      TAPIDEMO_RunLogRedirection(nPipeFd[0]);
   }
   /* save child pid */
   pCtrl->nSyslogChildPid = pid1;
   /* This proces doesn't need to read from pipe, so we can close descriptor 
      responsible for reading. */
   close(nPipeFd[0]);

   /* Redirect stdout to logger via pipe */
   if (-1 == dup2(nPipeFd[1],STDOUT_FILENO))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Can not redirect standard output. Error on dup2 %s\n"
             "File: %s, line: %d)\n",
             strerror(errno), __FILE__, __LINE__));

      return IFX_ERROR;
   }

   /* Redirect stderr to logger via pipe */
   if (-1 == dup2(nPipeFd[1],STDERR_FILENO))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Can not redirect standard error. Error on dup2 %s\n"
             "File: %s, line: %d)\n",
             strerror(errno), __FILE__, __LINE__));

      return IFX_ERROR;
   }

   /* Start new session */
   sid = setsid();
   umask(0);

   return IFX_SUCCESS;
}
#endif /* defined(LINUX) */

/* ============================= */
/* Global function definition    */
/* ============================= */

/**
   Get ip address of board and set it.

   \param pProgramArg - pointer to program arguments structure

   \return IFX_ERROR if an error, otherwise IFX_SUCCESS (no error occurred)
*/
IFX_int32_t TAPIDEMO_SetDefaultAddr(PROGRAM_ARG_t* pProgramArg)
{
#ifdef EASY336
   SVIP_libStatus_t SVIP_libRet;
   IFX_char_t pIP[INET_ADDR_LEN];
#else
   /* In man page of hostname is written, that host names are limited
      to 255 bytes. */
   IFX_char_t buffer[256];
#ifdef LINUX
   /** Value for local ip address */
   const IFX_char_t* LOCAL_IP_ADDRESS = "127.0.0.1";
   struct hostent* he = IFX_NULL;
   IFX_int32_t sockt_fd = 0;
   struct ifreq ifr;
   IFX_boolean_t bAskSocket = IFX_FALSE;
#else /* LINUX */
   IFX_int32_t addr = 0;
#endif /* LINUX */
#endif /* EASY336 */

   if (IFX_NULL == pProgramArg)
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Invalid input argument(s). (File: %s, line: %d)\n",
             __FILE__, __LINE__));
      return IFX_ERROR;
   }

   if (pProgramArg->oMy_IP_Addr.sin_addr.s_addr != 0)
   {
      return IFX_SUCCESS;
   }

#ifndef EASY336
   if (gethostname(buffer, sizeof(buffer)) != 0)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Err: gethostname(), %s. (File: %s, line: %d)\n",
             strerror(errno), __FILE__, __LINE__));
      return IFX_ERROR;
   }

   TRACE(TAPIDEMO, DBG_LEVEL_LOW, ("gethostname() returned %s.\n", buffer));

#ifdef LINUX
   he = gethostbyname(buffer);

   if (IFX_NULL == he)
   {
      bAskSocket = IFX_TRUE;
   }
   else
   {
      if (he->h_addr_list[0] != IFX_NULL)
      {
         pProgramArg->oMy_IP_Addr.sin_addr.s_addr =
            *(IFX_uint32_t *)he->h_addr_list[0];

         if (strcmp(inet_ntoa(pProgramArg->oMy_IP_Addr.sin_addr),
                    LOCAL_IP_ADDRESS) == 0)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
                  ("Local address (127.0.0.1) read, "
                   "will now try to check eth0 for right address.\n"
                   "On error local address will be used.\n"));
            bAskSocket = IFX_TRUE;
         }
      }
   }

   if (bAskSocket)
   {
      sockt_fd = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP);
      if (sockt_fd == -1)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Err: socket(), %s. (File: %s, line: %d)\n",
                strerror(errno), __FILE__, __LINE__));
         return IFX_ERROR;
      }

      memset(&ifr, 0, sizeof (struct ifreq));
      strcpy (ifr.ifr_name, "eth0");
      ifr.ifr_addr.sa_family = AF_INET;
      if (ioctl(sockt_fd, SIOCGIFADDR, &ifr) == -1)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Err with ioctl cmd SIOCGIFADDR %s. "
                "(File: %s, line: %d)\n",
                strerror(errno), __FILE__, __LINE__));
         close(sockt_fd);
         return IFX_ERROR;
      }

      pProgramArg->oMy_IP_Addr.sin_addr.s_addr =
         ((struct sockaddr_in *) &ifr.ifr_addr)->sin_addr.s_addr;

      close(sockt_fd);
   } /* if (bAskSocket) */

  TRACE(TAPIDEMO, DBG_LEVEL_LOW,
        ("Default IP Address for host %s is %s\n",
         buffer, inet_ntoa(pProgramArg->oMy_IP_Addr.sin_addr)));

#else /* LINUX */

   /* There is also possibility to use resolvGetHostByName(), but
      in file configAll.h support for library resolvLib,
      #define INCLUDE_DNS_RESOLVER must be added and new image
      prepared */
   addr = hostGetByName(buffer);

   pProgramArg->oMy_IP_Addr.sin_addr.s_addr = addr;

   TRACE(TAPIDEMO, DBG_LEVEL_LOW,
         ("Default IP Address for host %s is %s\n",
          buffer, inet_ntoa(pProgramArg->oMy_IP_Addr.sin_addr)));
#endif /* LINUX */

#else /* EASY336 */
   SVIP_libRet = SVIP_libGetAddress("eth0", pIP, IFX_NULL);
   if (SVIP_libRet != SVIP_libStatusOk)
      return IFX_ERROR;
#ifdef VXWORKS
   if (inet_aton(pIP, &pProgramArg->oMy_IP_Addr.sin_addr) == ERROR)
      return IFX_ERROR;
#endif
#ifdef LINUX
   if (inet_aton(pIP, &pProgramArg->oMy_IP_Addr.sin_addr) == 0)
      return IFX_ERROR;
#endif

   TRACE(TAPIDEMO, DBG_LEVEL_LOW,
         ("Default IP-Address is %s\n", pIP));
#endif /* EASY336 */

   return IFX_SUCCESS;
} /* TAPIDEMO_SetDefaultAddr() */

/**
   Get MAC address of board and set it.

   \return IFX_ERROR if an error, otherwise IFX_SUCCESS (no error occurred)
*/
static IFX_int32_t TAPIDEMO_SetMACAddr()
{
#ifndef EASY336
#ifdef LINUX
   IFX_int32_t sockt_fd = 0;
   struct ifreq ifr;
#endif /* LINUX */

#ifdef LINUX

   sockt_fd = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP);
   if (sockt_fd == -1)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Err: socket(), %s. (File: %s, line: %d)\n",
             strerror(errno), __FILE__, __LINE__));
      return IFX_ERROR;
   }

   memset(&ifr, 0, sizeof (struct ifreq));
   strcpy (ifr.ifr_name, "eth0");
   ifr.ifr_addr.sa_family = AF_INET;
   if (ioctl(sockt_fd, SIOCGIFHWADDR, &ifr) == -1)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Err with ioctl cmd SIOCGIFHWADDR %s. "
             "(File: %s, line: %d)\n",
             strerror(errno), __FILE__, __LINE__));
      close(sockt_fd);
      return IFX_ERROR;
   }

   memcpy(oCtrlStatus.oTapidemo_MAC_Addr, ifr.ifr_hwaddr.sa_data, ETH_ALEN);

   close(sockt_fd);

#endif /* LINUX */
#else
   SVIP_libStatus_t SVIP_libRet;

   SVIP_libRet = SVIP_libGetAddress("eth0", IFX_NULL,
      oCtrlStatus.oTapidemo_MAC_Addr);
   if (SVIP_libRet != SVIP_libStatusOk)
      return IFX_ERROR;

#endif /* EASY336 */

   return IFX_SUCCESS;
} /* TAPIDEMO_SetMACAddr() */

#if (defined EASY336 || defined XT16)
/**
   Get IP and MAC addresses of veth0 interface of board and set it.

   \return IFX_ERROR if an error, otherwise IFX_SUCCESS (no error occurred)
*/
static IFX_return_t TAPIDEMO_SetVeth0Addr()
{
#ifdef EASY336
   SVIP_libStatus_t SVIP_libRet;
   IFX_char_t pIP[INET_ADDR_LEN];

   SVIP_libRet = SVIP_libGetAddress("veth0", pIP,
      oCtrlStatus.oVeth0_MAC_Addr);
   if (SVIP_libRet != SVIP_libStatusOk)
      return IFX_ERROR;
#ifdef VXWORKS
   if (inet_aton(pIP, &oCtrlStatus.oVeth0_IP_Addr.sin_addr) == ERROR)
      return IFX_ERROR;
#endif
#ifdef LINUX
   if (inet_aton(pIP, &oCtrlStatus.oVeth0_IP_Addr.sin_addr) == 0)
      return IFX_ERROR;
#endif

   TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
         ("veth0 IP-Address: %s, MAC %02X:%02X:%02X:%02X:%02X:%02X\n", pIP,
          oCtrlStatus.oVeth0_MAC_Addr[0],
          oCtrlStatus.oVeth0_MAC_Addr[1],
          oCtrlStatus.oVeth0_MAC_Addr[2],
          oCtrlStatus.oVeth0_MAC_Addr[3],
          oCtrlStatus.oVeth0_MAC_Addr[4],
          oCtrlStatus.oVeth0_MAC_Addr[5]));
#endif /* EASY336 */

   return IFX_SUCCESS;
} /* TAPIDEMO_SetVeth0Addr() */
#endif

/**
   Check board combination.

   \return IFX_SUCCESS if successful, otherwise IFX_ERROR.

   \remarks
*/
IFX_return_t TAPIDEMO_CheckBoardCombination()
{
   IFX_boolean_t hCheckPcmIfMode = IFX_FALSE;

   switch (oCtrlStatus.nBoardCombination)
   {
      case ONE_CPE:
      {
#ifndef HAVE_DRV_EASY3332_HEADERS
         return IFX_ERROR;
#endif
         break;
      }

      case ONE_VIP:
      {
#ifndef HAVE_DRV_EASY3332_HEADERS
         return IFX_ERROR;
#endif
         break;
      }

      case ONE_VMMC:
      {
#if !defined(DANUBE) || !defined(HAVE_DRV_VMMC_HEADERS)
         return IFX_ERROR;
#endif
         break;
      }

      case ONE_VMMC_ONE_CPE:
      {
#if !defined(DANUBE) || !defined(HAVE_DRV_VINETIC_HEADERS) || !defined(HAVE_DRV_VMMC_HEADERS)
         return IFX_ERROR;
#endif
         hCheckPcmIfMode = IFX_TRUE;
         break;
      }

      case ONE_DXT:
      case TWO_DXTS:
      {
#ifndef DXT
         return IFX_ERROR;
#endif
         break;
      }

      case ONE_VMMC_VINAX:
      {
#if !defined(VINAX) || !defined(HAVE_DRV_VMMC_HEADERS)
         return IFX_ERROR;
#endif
         break;
      }

      case ONE_VMMC_EASY508XX:
      {
#if !defined(AR9) || !defined(HAVE_DRV_VMMC_HEADERS)
         return IFX_ERROR;
#endif
         break;
      }

      case X_SVIP:
      {
#ifndef EASY336
         return IFX_ERROR;
#endif
         break;
      }

      case X_SVIP_X_XT16:
      {
#if (!defined(EASY336) || !defined(WITH_VXT))
         return IFX_ERROR;
#endif
         break;
      }

      case X_XT16:
      {
#ifndef XT16
         return IFX_ERROR;
#endif
         break;
      }

      case ONE_EASY50712_ONE_TERIDIAN73M1866:
      {
#if !defined(DANUBE) || !defined(HAVE_DRV_VMMC_HEADERS) || !defined(HAVE_DRV_TERIDIAN_HEADERS)
         return IFX_ERROR;
#endif
         hCheckPcmIfMode = IFX_TRUE;
         break;
      }

      case ONE_EASY508XX_ONE_TERIDIAN73M1866:
      {
#if !defined(AR9) || !defined(HAVE_DRV_VMMC_HEADERS) || !defined(HAVE_DRV_TERIDIAN_HEADERS)
         return IFX_ERROR;
#endif
         hCheckPcmIfMode = IFX_TRUE;
         break;
      }

      case ONE_EASY80800_ONE_TERIDIAN73M1866:
      {
#if !defined(VINAX) || !defined(HAVE_DRV_VMMC_HEADERS) || !defined(HAVE_DRV_TERIDIAN_HEADERS)
         return IFX_ERROR;
#endif
         hCheckPcmIfMode = IFX_TRUE;
         break;
      }
   } /* switch */

   if (hCheckPcmIfMode)
   {
      if (oCtrlStatus.pProgramArg->oArgFlags.nPCM_Slave)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Main board - the PCM i/f can not be configured as a SLAVE when \n"
             "extension board is used.\n"
             "PCM i/f will be set as MASTER\n"));
         oCtrlStatus.pProgramArg->oArgFlags.nPCM_Slave = 0;
      }
      oCtrlStatus.pProgramArg->oArgFlags.nPCM_Master = 1;
   }
   return IFX_SUCCESS;
}
/**
   Set default program settings.

   \param  pCtrl   - pointer to status control structure
   \param pProgramArg - program arguments

   \return none
*/
IFX_void_t TAPIDEMO_SetDeafault(CTRL_STATUS_t* pCtrl,
                                PROGRAM_ARG_t* pProgramArg)
{
   /* Trace level is set to high by default */
   SetTraceLevel(TAPIDEMO, DBG_LEVEL_HIGH);
   pProgramArg->nDbgLevel = DBG_LEVEL_HIGH;
   /* Trace redirection is disabled by default */
   pProgramArg->nTraceRedirection = TRACE_REDIRECTION_NONE;
   oTraceRedirection.fStartRedirectLogs = IFX_FALSE;
   oTraceRedirection.nFileLogFD = NO_FILE;
   oTraceRedirection.nSyslogSocket = NO_SOCKET;
   /* set board type */
   pProgramArg->nBoardComb = UNSUPPORTED_COMB;
   /* Voice Activity Detection */
   pProgramArg->nVadCfg = DEFAULT_VAD_CFG;
   /* CID is turned on by default and set to ETSI FSK */
   pProgramArg->oArgFlags.nCID = 1;
   pProgramArg->nCIDStandard = IFX_TAPI_CID_STD_ETSI_FSK;
#ifdef ABACUS_TEST
   pProgramArg->nComDbgLevel = DBG_LEVEL_OFF;
   pProgramArg->nTestType = COM_NO_TEST;
   pProgramArg->nSubTestType = COM_SUBTEST_NONE;
   pCtrl->nComInSocket = NO_SOCKET;
   nComDbgLevel = &pProgramArg->nComDbgLevel;
   nVerifySystemInitializtion = 0;
   pCtrlStatus = pCtrl;
#endif /* ABACUS_TEST */
#ifdef TD_FAX_MODEM
   pCtrl->nFaxMode = TD_FAX_MODE_DEAFAULT;
#endif /* TD_FAX_MODEM */
   /* set default options */
#if (!defined XT16)
   /* Turn conferencing on */
   pProgramArg->oArgFlags.nConference = 1;
#else
   /* enable local PCM loop */
   pProgramArg->oArgFlags.nPCM_Loop = 1;
   pProgramArg->oArgFlags.nPCM_Master = 1;
#endif /* XT16 */
#ifdef EASY336
   /* enable PCM slave */
   pProgramArg->oArgFlags.nPCM_Slave = 1;
#endif

#ifdef DXT
   /* EASY3201 by default uses PCM local loop */
   pProgramArg->oArgFlags.nPCM_Master = 1;
   pProgramArg->oArgFlags.nPCM_Slave = 0;
   pProgramArg->oArgFlags.nPCM_Loop = 1;
#endif /* DXT */

   /* Default path to download files */
   strncpy(pProgramArg->sPathToDwnldFiles, DEF_DOWNLOAD_PATH, MAX_PATH_LEN - 1);

   /* karol -workaround for teridian extension board */
   pCtrl->hUseDuslicFxo = IFX_TRUE;
   pProgramArg->hUseDuslicFxo = IFX_TRUE;

   return;
}
/**
   Main function.

   \param argc - number of arguments
   \param argv - array of arguments

   \return 0 or err code
*/
#if defined(LINUX)

int main(int argc, char *argv [])
{

#else /* defined(LINUX) */

IFX_int32_t tapidemo(char* cmd_args)
{
   IFX_int32_t argc = 0;
   IFX_char_t** argv = IFX_NULL;

#endif /* defined(LINUX) */

   int ret = OK;
   IFX_int32_t i = 0/*, j = 0, k = 0*/;
   const BOARD_COMBINATION_ELEM_t* board_list = IFX_NULL;
   /*BOARD_t* pBoard = IFX_NULL;*/
   IFX_int32_t phone_num = 0, nFXO_Num = 0;
   IFX_boolean_t check_arg = IFX_FALSE;
   IFX_int32_t j = 0;
   BOARD_t* pBoard = IFX_NULL;
   IFX_int32_t tmp_coder_sum = 0;
   /** Board combination */
    IFX_int32_t nDefaultBoardComb = UNSUPPORTED_COMB;

   /* Reset parameters */
   memset((IFX_uint8_t *) &oProgramArg, 0, sizeof(PROGRAM_ARG_t));
   /* Reset control status */
   memset(&oCtrlStatus, 0, sizeof(CTRL_STATUS_t));
   /* Set all variables and all used structures to zero. */
   memset(&oTraceRedirection, 0, sizeof(TRACE_REDIRECTION_t));

   /* set pointer to program arguments structure */
   oCtrlStatus.pProgramArg = &oProgramArg;

#ifdef ABACUS_TEST
   /* set test type pointer */
   nTestType = &oCtrlStatus.pProgramArg->nTestType;
#endif /* ABACUS_TEST */

#if defined(VXWORKS)

   /* Set application priority level */
   taskPrioritySet((int) IFX_NULL, (int) 40);

   /* Get TID of application */
   nTapiDemo_TID = taskIdSelf();

   /* Read arguments and setup application accordingly */
   if ((cmd_args != IFX_NULL) && (strlen(cmd_args) > 0))
   {
      argc = GetArgCnt(cmd_args, GetCmdArgsArr());
      argv = GetCmdArgsArr();
      check_arg = IFX_TRUE;
   }

#elif LINUX

   if (argc > 1)
   {
      check_arg = IFX_TRUE;
   }

#endif /* VXWORKS */
   /* Set some default arguments */
   TAPIDEMO_SetDeafault(&oCtrlStatus, oCtrlStatus.pProgramArg);

   /* end workaround */
   if (check_arg)
   {
      if (ReadOptions(argc, argv, oCtrlStatus.pProgramArg) != IFX_SUCCESS)
      {
         /* Wrong arguments */
         TAPIDEMO_PRINTF((" Wrong command line argument(s).\n"));
         return -1;
      }
      /* karol -workaround for teridian extension board */
      oCtrlStatus.hUseDuslicFxo = oProgramArg.hUseDuslicFxo;
      /* end workaround */
   }

#ifdef LINUX
   if (oProgramArg.nTraceRedirection != TRACE_REDIRECTION_NONE)
   {
      /* Set up log redirection - e.g. sockets, syslog or open file */
      if (TAPIDEMO_PreapareLogRedirection() != IFX_SUCCESS)
      {
         /* Can't start redirecting logs */
         TAPIDEMO_PRINTF((" Can not start redirecting logs.\n"));
         return -1;
      }
   }

   if (oProgramArg.oArgFlags.nDaemon)
   {
      /* Create process responsible for log redirection and TAPIDEMO daemon */
      if (TAPIDEMO_CreateDaemon(&oCtrlStatus) != IFX_SUCCESS)
      {
         /* Can't create daamon */
         TAPIDEMO_PRINTF((" Can not run Tapidemo as a daemon.\n"));
         return -1;
      }
   }
#endif /* defined(LINUX) */

   /* Print version */
   TAPIDEMO_PRINTF(("TAPI Demo Version %s\n", TAPIDEMO_VERSION));
#ifdef ABACUS_TEST
   TAPIDEMO_PRINTF(("ABACUS test version %s\n", ABACUS_TEST_VERSION));
   /* copy version strings to control structure */
   strncpy(oCtrlStatus.rgoTapidemoVersion, TAPIDEMO_VERSION, TD_MAX_NAME);
   strncpy(oCtrlStatus.rgoAbacusTestVersion, ABACUS_TEST_VERSION, TD_MAX_NAME);
#endif /* ABACUS_TEST */

#if (defined(TD_T38_FAX_TEST) && defined(TD_FAX_T38) && defined(AR9))
   /* FAX_TEST version */
   TD_T38_FAX_TEST_PrintVersion();
#endif /* (defined(TD_T38_FAX_TEST) && defined(TD_FAX_T38) && defined(AR9)) */

   PrintUsedOptions(&oProgramArg);

   /* Set default IP-Address  */
   TAPIDEMO_SetDefaultAddr(oCtrlStatus.pProgramArg);

   /* Set combination according to master board */
#ifdef EASY3332
   /* set default board combination */
   nDefaultBoardComb = ONE_CPE;
#elif EASY334
   /* set default board combination */
   nDefaultBoardComb = ONE_VIP;
#elif VMMC
   /* set default board combination */
   nDefaultBoardComb = ONE_VMMC;
#elif DXT
   /* set default board combination */
   nDefaultBoardComb = ONE_DXT;
#elif VINAX
   /* set default board combination */
   nDefaultBoardComb = ONE_VMMC_VINAX;
#elif AR9
   /* set default board combination */
   nDefaultBoardComb = ONE_VMMC_EASY508XX;
#elif VR9
   /* set default board combination */
   nDefaultBoardComb = ONE_EASY80910;
#elif EASY336
#ifdef WITH_VXT
   /* set default board combination */
   nDefaultBoardComb = X_SVIP_X_XT16;
#else /* WITH_VXT */
   /* set default board combination */
   nDefaultBoardComb = X_SVIP;
#endif /* WITH_VXT */
#elif XT16
   /* set default board combination */
   nDefaultBoardComb = X_XT16;
#else /* XT16 */
   /* Unknown, use first one */
   nDefaultBoardComb = ONE_CPE;
#endif /* EASY3332 */

#ifndef EASY3332
   if (ONE_CPE == nDefaultBoardComb)
   {
      TAPIDEMO_PRINTF(("Compiled for unknown board combination, use [%s]",
                       BOARD_COMB_NAMES[nDefaultBoardComb]));
   }
   else
#endif
   {
      TAPIDEMO_PRINTF(("Compiled for %s",
                       BOARD_COMB_NAMES[nDefaultBoardComb]));
   }
#ifdef FXO
   /* compiled with FXO */
   TAPIDEMO_PRINTF((" with FXO"));
#endif
   /* end trace */
   TAPIDEMO_PRINTF((".\n"));

   /* set board combination if not set with program args */
   if (UNSUPPORTED_COMB == oCtrlStatus.pProgramArg->nBoardComb)
   {
      oCtrlStatus.pProgramArg->nBoardComb = nDefaultBoardComb;
   }

   /* print used board combination */
   TAPIDEMO_PRINTF(("Used board(s) %s", BOARD_COMB_NAMES[oCtrlStatus.pProgramArg->nBoardComb]));
#ifdef FXO
   /* compiled with FXO */
   TAPIDEMO_PRINTF((", FXO %s", oCtrlStatus.pProgramArg->oArgFlags.nFXO ? "on" : "off"));
#endif
   /* end trace */
   TAPIDEMO_PRINTF((".\n"));

#ifndef FXO
   if ( oCtrlStatus.pProgramArg->oArgFlags.nFXO )
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("FXO support is disabled. Please rebuild "
             "and make sure all related header files are available.\n"
             "File: %s, line: %d)\n",
             __FILE__, __LINE__));
   }
#endif

   if (oCtrlStatus.pProgramArg->oArgFlags.nHelp)
   {
      return ProgramHelp();
   }

   /* Register signal handler */
   ret = TAPIDEMO_RegisterSigHandler();

   if (ret != IFX_SUCCESS)
   {
      /* Could not register signal handler */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Could not register signal handler. (File: %s, line: %d)\n",
            __FILE__, __LINE__));

      return IFX_ERROR;
   }

   /* initialize arrary with pointers to codecs names */
   VOIP_SetCodecNames();

   /* Turn on services */
   TAPIDEMO_RunServices(&oCtrlStatus);

   /* Set MAC-Address  */
   TAPIDEMO_SetMACAddr();

   TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
         ("Using IP-Address: %s, MAC %02X:%02X:%02X:%02X:%02X:%02X\n",
          inet_ntoa(oProgramArg.oMy_IP_Addr.sin_addr),
          oCtrlStatus.oTapidemo_MAC_Addr[0],
          oCtrlStatus.oTapidemo_MAC_Addr[1],
          oCtrlStatus.oTapidemo_MAC_Addr[2],
          oCtrlStatus.oTapidemo_MAC_Addr[3],
          oCtrlStatus.oTapidemo_MAC_Addr[4],
          oCtrlStatus.oTapidemo_MAC_Addr[5]));

#if (defined EASY336 || defined XT16)
   /* get IP and MAC addresses of veth0 interface */
   TAPIDEMO_SetVeth0Addr();
#endif /* (defined EASY336 || defined XT16) */

   /* Set combination index */
   oCtrlStatus.nBoardCombination = oCtrlStatus.pProgramArg->nBoardComb;
   oCtrlStatus.nBoardCnt =
      BOARD_COMBINATIONS[oCtrlStatus.nBoardCombination].nBoardCnt;

   if (TAPIDEMO_CheckBoardCombination() == IFX_ERROR)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("This board combination isn't supported, please rebuild "
             "and make sure all related header files are available.\n"
             "File: %s, line: %d)\n",
              __FILE__, __LINE__));
      TAPIDEMO_CleanUp(SIGQUIT);
      exit(1);
   }

#ifdef ABACUS_TEST
   /* Create communication socket for receiving commands from control PC*/
   oCtrlStatus.nComInSocket =
      COM_InitCommunicationSocket(oProgramArg.oMy_IP_Addr.sin_addr.s_addr,
                                  COM_COMMAND_PORT, COM_DIRECTION_IN);
   /* Set SUT's identification flag to FALSE*/
   oCtrlStatus.fComIdentSent = IFX_FALSE;

#endif /* ABACUS_TEST */
#if (defined(TD_T38_FAX_TEST) && defined(TD_FAX_T38) && defined(AR9))
   /* Initialize FAX T.38 test */
   TD_T38_FAX_TEST_Init(&oCtrlStatus);
#endif /* (defined(TD_T38_FAX_TEST) && defined(TD_FAX_T38) && defined(AR9)) */

   /* If support for multiple boards then initialize all boards. */
   for (i = 0; i < oCtrlStatus.nBoardCnt; i++)
   {
      board_list =
          &BOARD_COMBINATIONS[oCtrlStatus.nBoardCombination].nBoards[i];

      /* Set board type */
      pBoard = &oCtrlStatus.rgoBoards[i];
      pBoard->nID = board_list->nBoard_ID;
      pBoard->nType = board_list->nBoardType;
      pBoard->nBoard_IDX = i;
      pBoard->pCtrl = &oCtrlStatus;

      if (Common_RegisterBoard(pBoard) == IFX_ERROR)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Err, Board registration failed (Idx %d). (File: %s, line: %d)\n",
               pBoard->nBoard_IDX, __FILE__, __LINE__));

         /* Cleanup previous stuff */
#ifdef VXWORKS
         abort();
#else /* VXWORKS */
         TAPIDEMO_CleanUp(SIGABRT);
#endif /* VXWORKS */
         return ERROR;
      }


      TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
            ("----- Start board %s initialization. -----\n",
             pBoard->pszBoardName));
      /* Init board */
      if (pBoard->Init(pBoard, &oProgramArg) == IFX_ERROR)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Err, Board Initialization failed (Idx %d). (File: %s, line: %d)\n",
               pBoard->nBoard_IDX, __FILE__, __LINE__));

         /* Cleanup previous stuff */
#ifdef VXWORKS
         abort();
#else /* VXWORKS */
         TAPIDEMO_CleanUp(SIGABRT);
#endif /* VXWORKS */
         return ERROR;
      }
      TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
            ("-----      Board %s initialized.     -----\n",
             pBoard->pszBoardName));

      /* Indention = length of board name + length of ": " */
      pBoard->pIndentBoard = 
         Common_GetTraceIndention(strlen(pBoard->pszBoardName) + 2);

     /* Indention = length of phone name + length of phone number ": " */
     pBoard->pIndentPhone = 
        Common_GetTraceIndention(strlen(TD_PHONE_STRING) + TD_PHONE_NUM_LEN);

      /* Save sum of all channels */
      oCtrlStatus.nSumCoderCh += pBoard->nMaxCoderCh;
      oCtrlStatus.nSumAnalogCh += pBoard->nMaxAnalogCh;
      oCtrlStatus.nSumPCM_Ch += pBoard->nMaxPCM_Ch;
   }

   /* Prepare socket for making/ending calls. */
   oCtrlStatus.nAdminSocket = TAPIDEMO_InitAdminSocket();

   if (oCtrlStatus.nSumCoderCh > 0)
   {
      /* Prepare UDP sockets for VoIP external calls */
      oCtrlStatus.rgnSockets = calloc(oCtrlStatus.nSumCoderCh,
                                      sizeof(SOCKET_STAT_t));
      if (oCtrlStatus.rgnSockets == IFX_NULL)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("ERR allocate memory for UDP sockets. (File: %s, line: %d)\n",
               __FILE__, __LINE__));
         /* Cleanup previous stuff */
#ifdef VXWORKS
         abort();
#else /* VXWORKS */
         TAPIDEMO_CleanUp(SIGABRT);
#endif /* VXWORKS */
      }

      if (!oCtrlStatus.pProgramArg->oArgFlags.nQos)
      {
         IFX_int32_t k = 0;

         tmp_coder_sum = 0;
         j = 0;
         pBoard = &oCtrlStatus.rgoBoards[j];
         tmp_coder_sum = pBoard->nMaxCoderCh;
         for (i = 0; i < oCtrlStatus.nSumCoderCh; i++, k++)
         {
            /* Also initialize sockets */
#ifdef TAPI_VERSION4
            if (!pBoard->fUseSockets)
            {
               oCtrlStatus.rgnSockets[i].nSocket = NO_SOCKET;
            }
            else /* if (!pBoard->fUseSockets) */
#endif /* TAPI_VERSION4 */
            {
               oCtrlStatus.rgnSockets[i].nSocket =
                  VOIP_InitUdpSocket(oCtrlStatus.pProgramArg, i,
                                     &oCtrlStatus.rgnSockets[i].nPort);
               if (NO_SOCKET == oCtrlStatus.rgnSockets[i].nSocket)
               {
                  /* Error initializing VoIP sockets */
                  /* Cleanup previous stuff */
   #ifdef VXWORKS
                  abort();
   #else /* VXWORKS */
                  TAPIDEMO_CleanUp(SIGABRT);
   #endif /* VXWORKS */
                  return IFX_ERROR;
               }

                /* Connect this sockets for each board, data ch of this board is
                  index in socket array. */
               if (tmp_coder_sum <= i)
               {
                  j++;
                  pBoard = &oCtrlStatus.rgoBoards[j];
                  tmp_coder_sum += pBoard->nMaxCoderCh;
                  k = 0;
               }
               pBoard->rgnSockets[k] = oCtrlStatus.rgnSockets[i].nSocket;
            } /* if (!pBoard->fUseSockets) */
         } /* for */
      } /* if (!oCtrlStatus.pProgramArg->oArgFlags.nQos) */
   } /* if (oCtrlStatus.nSumCoderCh > 0) */

   /* Holds called phone number, which will be increased in *_Setup() */
#ifndef TAPI_VERSION4
   phone_num = 1;
   nFXO_Num = 1;
#endif

   /* If support for multiple boards then initialize all boards. */
   for (i = 0; i < oCtrlStatus.nBoardCnt; i++)
   {
      /* Prepare objects, structure */
      if (IFX_ERROR == TAPIDEMO_Setup(&oCtrlStatus.rgoBoards[i],
                                      &oCtrlStatus, &phone_num, &nFXO_Num))
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("TAPIDEMO_Setup() failed. (File: %s, line: %d)\n",
               __FILE__, __LINE__));
         /* Cleanup previous stuff */
#ifdef VXWORKS
         abort();
#else /* VXWORKS */
         TAPIDEMO_CleanUp(SIGABRT);
#endif /* VXWORKS */
         return ERROR;
      }

      /* If support for multiple boards then initialize all boards. */
   } /* for all boards*/

#ifndef EASY336
   if (oCtrlStatus.nSumPCM_Ch > 0)
   {
      if (PCM_Init(&oProgramArg) != IFX_SUCCESS)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("PCM_Init() failed. (File: %s, line: %d)\n",
               __FILE__, __LINE__));

         /* Cleanup previous stuff */
#ifdef VXWORKS
         abort();
#else /* VXWORKS */
         TAPIDEMO_CleanUp(SIGABRT);
#endif /* VXWORKS */
         return IFX_ERROR;
      } /* if */
   } /* if (oCtrlStatus.nSumPCM_Ch > 0) */
#endif /* EASY336 */

#ifdef EASY336
   if (IFX_ERROR == BOARD_Easy336_Init_RM(&oCtrlStatus, &pcm_cfg))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("DEVICE_SVIP_Init_RM() failed. (File: %s, line: %d)\n",
             __FILE__, __LINE__));
      /* Cleanup previous stuff */
#ifdef VXWORKS
      abort();
#else /* VXWORKS */
      TAPIDEMO_CleanUp(SIGABRT);
#endif /* VXWORKS */
      return IFX_ERROR;
   }
#endif /* EASY336 */

   TAPIDEMO_Call_Info(&oCtrlStatus);
   /* Start actual program */
   TAPIDEMO_StateMachine(&oCtrlStatus);

   /* This part never executes. */

   return ret;
} /* main() */


