

/******************************************************************************

                               Copyright (c) 2008
                            Infineon Technologies AG
                     Am Campeon 1-12; 81726 Munich, Germany

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

  ****************************************************************************/

/**
   \file common.c
   \date 2005-10-01
   \brief Common stuff like read file, open/close FDs and wrapper to board
          interface.
*/

/* ============================= */
/* Includes                      */
/* ============================= */
#include "common.h"
#include "tapidemo.h"
#include "feature.h"
#include "voip.h"
#include "abstract.h"
#include "qos.h"
#include "state_trans.h"
#ifdef LINUX
#include <sys/mman.h>
#endif
#ifdef VXWORKS
#include <sys/stat.h>
#endif

/* Include board specific commands, structures */

#if EASY3332
#include "board_easy3332.h"
#endif /* EASY3332 */

#if EASY50510
#include "board_easy50510.h"
#endif /* EASY50510 */

#ifdef VMMC
#include "board_easy50712.h"
#endif /* VMMC */

#ifdef EASY336
#include "board_easy336.h"
#ifdef WITH_VXT
#include "board_xt16.h"
#endif /* WITH_VXT */
#endif /* EASY336 */

#ifdef XT16
#include "board_xt16.h"
#endif /* XT16 */

#ifdef VINAX
#include "board_easy80800.h"
#endif /* VINAX */

#ifdef AR9
#include "board_easy508xx.h"
#endif /* AR9 */

#ifdef VR9
#include "board_easy80910.h"
#endif /* AR9 */

#ifdef TERIDIAN_FXO
   #include "board_teridian73M1866.h"
#endif /* TERIDIAN_FXO */

#if (!defined(EASY336) || !defined(XT16))
/* Prototype to remove compiler warning. */
extern IFX_int32_t vmmc_reset_device(IFX_uint32_t fdCtrlDev);
#endif

#ifdef DXT
#include "board_easy3201.h"
#include "board_easy3111.h"
#endif /* DXT */

#ifdef FXO
#include "common_fxo.h"
#endif /* FXO */

#include "pcm.h"

/* ============================= */
/* Defines                       */
/* ============================= */

/** Max trace lenght in one line,
    This value is used to check if traces aren't too long. */
enum { MAX_TRACE_LENGTH  = 75 };

/** Max length of trace indentation. */
#define TD_MAX_INDENT      20

/* ============================= */
/* Local variable definition     */
/* ============================= */

IFX_char_t acTraceIndentTable[TD_MAX_INDENT];

#ifdef TAPI_VERSION4
static const IFX_char_t* LINE_MODES[] =
{
   "IFX_TAPI_LINE_FEED_ACTIVE",
   "IFX_TAPI_LINE_FEED_ACTIVE_REV",
   "IFX_TAPI_LINE_FEED_STANDBY",
   "IFX_TAPI_LINE_FEED_HIGH_IMPEDANCE",
   "IFX_TAPI_LINE_FEED_DISABLED",
   "IFX_TAPI_LINE_FEED_GROUND_START",
   "IFX_TAPI_LINE_FEED_NORMAL_AUTO",
   "IFX_TAPI_LINE_FEED_REVERSED_AUTO",
   "IFX_TAPI_LINE_FEED_NORMAL_LOW",
   "IFX_TAPI_LINE_FEED_REVERSED_LOW",
   "IFX_TAPI_LINE_FEED_RING_BURST",
   "IFX_TAPI_LINE_FEED_RING_PAUSE",
   "IFX_TAPI_LINE_FEED_METER",
   "IFX_TAPI_LINE_FEED_ACTIVE_LOW",
   "IFX_TAPI_LINE_FEED_ACTIVE_BOOSTED",
   "IFX_TAPI_LINE_FEED_ACT_TESTIN",
   "IFX_TAPI_LINE_FEED_DISABLED_RESISTIVE_SWITCH",
   "IFX_TAPI_LINE_FEED_PARKED_REVERSED",
   "IFX_TAPI_LINE_FEED_ACTIVE_RES_NORMAL",
   "IFX_TAPI_LINE_FEED_ACTIVE_RES_REVERSED",
   "IFX_TAPI_LINE_FEED_ACT_TEST",
   "IFX_TAPI_LINE_FEED_NLT"
};
#endif /* TAPI_VERSION4 */

/* ============================= */
/* Global variable definition    */
/* ============================= */
#ifdef EASY336
extern IFX_TAPI_ENC_TYPE_t eCurrEncType;
#endif /* EASY336 */

const char * TAPIDEMO_PCM_FREQ_STR[] =
{
   "512", /* IFX_TAPI_PCM_IF_DCLFREQ_512 */
   "1024", /* IFX_TAPI_PCM_IF_DCLFREQ_1024 */
   "1536", /* IFX_TAPI_PCM_IF_DCLFREQ_1536 */
   "2048", /* IFX_TAPI_PCM_IF_DCLFREQ_2048 */
   "4096", /* IFX_TAPI_PCM_IF_DCLFREQ_4096 */
   "8192", /* IFX_TAPI_PCM_IF_DCLFREQ_8192 */
   "16384" /* IFX_TAPI_PCM_IF_DCLFREQ_16384 */
};

/* table with error strings for Common_Enum2Name() function. */
IFX_char_t* TD_rgEnum2NameErrorCode[] = 
{
   "enum value not found",
   "error, NULL pointer"
};

/* ============================= */
/* Local function declaration    */
/* ============================= */

IFX_return_t Common_JBStatistic(IFX_int32_t nDataCh,
                                IFX_int32_t nDev,
                                IFX_boolean_t fReset,
                                IFX_boolean_t fShow,
                                BOARD_t* pBoard);
IFX_return_t Common_TonePlay(CTRL_STATUS_t* pCtrl, PHONE_t* pPhone,
                             IFX_uint32_t nTone);
IFX_return_t Common_ToneStop(CTRL_STATUS_t* pCtrl, PHONE_t* pPhone);

/* ============================= */
/* Local function definition     */
/* ============================= */

#ifdef USE_FILESYSTEM
/**
   Create full filename (path + file's name)
  
   \param \param pszFilename  - file's name
   \param pszPath - path to file
   \param pszFullFilename - path with filename
 
   \return  status (IFX_SUCCESS or IFX_ERROR)
*/
IFX_return_t Common_CreateFullFilename(IFX_char_t* pszFilename,
                                       IFX_char_t* pszPath, 
                                       IFX_char_t* pszFullFilename)
{

   TD_PTR_CHECK(pszFilename, IFX_ERROR);
   TD_PTR_CHECK(pszPath, IFX_ERROR);
   TD_PARAMETER_CHECK((0 == strlen(pszFilename)), strlen(pszFilename), IFX_ERROR);
   TD_PARAMETER_CHECK((0 == strlen(pszPath)), strlen(pszPath), IFX_ERROR);

   if ((strlen(pszPath) + strlen(pszFilename)) < MAX_PATH_LEN)
   {
      strcpy(pszFullFilename, pszPath);
      strcat(pszFullFilename, pszFilename);
   }
   else
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Path [%s] + Filename [%s] too long (max. %d). "
            "(File: %s, line: %d)\n",
            pszPath, pszFilename, MAX_PATH_LEN, __FILE__, __LINE__));
      return IFX_ERROR;
   }

   return IFX_SUCCESS;
} /* Common_CreateFullFilename */

/**
   Check if file / symbolic link exists

   \param pszFullFilename  - binary file name with path

   \return IFX_SUCCESS if everything OK, otherwise IFX_ERROR
*/
IFX_return_t Common_CheckFile(IFX_char_t* pszFullFilename)
{
   TD_PTR_CHECK(pszFullFilename, IFX_ERROR);

   struct stat st;
   IFX_int32_t ret;

   ret = stat(pszFullFilename, &st);
   if(0 != ret)
   {
      /* Path is incorrect */
      return IFX_ERROR;
   }

   if(S_ISREG(st.st_mode))
   {
      /* Path focuses on regular file */ 
      return IFX_SUCCESS;
   }
   else if(S_ISLNK(st.st_mode))
   {
      /* Path focuses on symbolic link */
      return IFX_SUCCESS;
   }
   else
   {
      return IFX_ERROR;
   }
   
   return IFX_SUCCESS;
} /* Common_CheckFile */

/**
   Find PRAM_FW
  
   \param pCpuDevice - pointer to CPU device
   \param pszPath - path to file
   \param pszFullFilename - path with filename
 
   \return  Return IFX_TRUE is PRAM_FW file is found or IFX_FALSE if not
*/

IFX_boolean_t Common_FindPRAM_FW(TAPIDEMO_DEVICE_CPU_t* pCpuDevice, 
                                 IFX_char_t* pszPath, 
                                 IFX_char_t* pszFullFilename)
{
   IFX_return_t ret;
   IFX_boolean_t bFileFound = IFX_FALSE;

   TD_PTR_CHECK(pCpuDevice, IFX_FALSE);
   TD_PTR_CHECK(pszPath, IFX_FALSE);
   TD_PARAMETER_CHECK((0 == strlen(pszPath)), strlen(pszPath), IFX_FALSE);

   /* Find PRAM_FW file with default new name */
   ret = Common_CreateFullFilename(pCpuDevice->pszPRAM_FW_File[0], pszPath, pszFullFilename);
   if(IFX_SUCCESS == ret)
   {
      ret = Common_CheckFile(pszFullFilename);
      if(IFX_SUCCESS == ret)
      {
         /* Find PRAM_FW file with default name */
         bFileFound = IFX_TRUE;
      }
   }

   if(IFX_TRUE != bFileFound)
   {
      /* Find PRAM_FW file with default old name */
      ret = Common_CreateFullFilename(pCpuDevice->pszPRAM_FW_File[1], pszPath, pszFullFilename);
      if(IFX_SUCCESS == ret)
      {
          ret = Common_CheckFile(pszFullFilename);
         if(IFX_SUCCESS == ret)
         {
            /* Find PRAM_FW file with alternative name */
            bFileFound = IFX_TRUE;
         }
      }
   } /* if(IFX_TRUE != bFileFound) */

   return bFileFound;
} /* Common_FindPRAM_FW */

/**
   parses file and fill in given buffer

   \param pszFilename  - binary file name
   \param pBuf - buffer in which the file should be written
   \param pSize - size of read buffer

   \return IFX_SUCCESS if successful, otherwise IFX_ERROR.
*/
IFX_return_t Common_ReadBinFile(IFX_char_t* full_filename,
                                IFX_uint8_t** pBuf,
                                IFX_int32_t* pSize)
{
   IFX_int32_t fd;
   IFX_return_t ret = IFX_SUCCESS;

   TD_PTR_CHECK(full_filename, IFX_ERROR);
   TD_PTR_CHECK(pBuf, IFX_ERROR);
   TD_PTR_CHECK(pSize, IFX_ERROR);
   TD_PARAMETER_CHECK((0 == strlen(full_filename)), strlen(full_filename), IFX_ERROR);

   *pSize = 0;

   fd = open(full_filename, O_RDONLY, 0x644);
   if (0 >= fd)
   {
      /* No file by this name ? */
      return IFX_ERROR;
   }

   /* Get file size */
   if (IFX_SUCCESS == ret)
   {
      *pSize = lseek(fd, 0, SEEK_END);
   }

   if ((IFX_ERROR != ret) && (0 >= *pSize))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Error get size of file %s, %s. (File: %s, line: %d)\n",
            full_filename, strerror(errno), __FILE__, __LINE__));
      ret = IFX_ERROR;
   }

   if (ret != IFX_ERROR)
   {
      /* Move pointer back to start of file, otherwise ZERO bytes will
         be read. */
      lseek(fd, 0, SEEK_SET);
   }

   /* Allocate buffer memory if ptr is nil */
   if ((IFX_SUCCESS == ret) && (IFX_NULL == *pBuf))
   {
      *pBuf = (IFX_uint8_t *) malloc(*pSize);
      if (IFX_NULL == *pBuf)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("No memory for file %s, %s. (File: %s, line: %d)\n",
               full_filename, strerror(errno), __FILE__, __LINE__));
         ret = IFX_ERROR;
      }
   }
   else
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Memmory not allocated, pointer isn't NULL for file %s."
             "(File: %s, line: %d)\n",
             full_filename, __FILE__, __LINE__));
      ret = IFX_ERROR;
   }

   /* Now read file into buffer */
   if ((IFX_SUCCESS == ret) && (read(fd, (char*)*pBuf, *pSize) < *pSize))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Error while reading file %s, %s. \
           (File: %s, line: %d)\n",
            full_filename, strerror(errno), __FILE__, __LINE__));
      ret = IFX_ERROR;

   }

   /* \todo could also free buffer here on error */
   if (ret != IFX_SUCCESS)
   {
      free(*pBuf);
      *pBuf = IFX_NULL;
   }

   if (fd >= 0)
   {
      close(fd);
   }

   return ret;
} /* ReadBinFile() */

#ifndef VXWORKS
/**
   Read value for specified string from file and verify data.

   \param pFieldName - name of field to search
   \param pFieldValue - value found after pFieldName in file
   \param fd         - file descriptor

   \return IFX_SUCCESS if everything OK, otherwise IFX_ERROR
   \remarks
*/
IFX_return_t Common_ParseRcFile(IFX_char_t* pFieldName, IFX_char_t* pFieldValue,
                              IFX_int32_t nFd)
{
   IFX_char_t pBuff[MAX_PATH_LEN], pTmpBuff[MAX_PATH_LEN];
   IFX_int32_t nReadCharNum, nFieldNameLen, nCharsInBuff = 0;
   IFX_int32_t i = 0, j = 0;
   IFX_boolean_t fEndOfFileReached = IFX_FALSE, fNewLine = IFX_TRUE;

   /* check input arguments */
   if ((IFX_NULL == pFieldName) || (IFX_NULL == pFieldValue) || (0 >= nFd))
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }

   /* get searched filed name length */
   nFieldNameLen = strlen(pFieldName);
   /* check field name length */
   if ((0 >= nFieldNameLen) || (MAX_PATH_LEN <= (nFieldNameLen + 2)))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Field name %s is invalid. (File: %s, line: %d)\n",
             pFieldName, __FILE__, __LINE__));
      return IFX_ERROR;
   }

   /* search for field name in read data */
   while (1)
   {
      /* check if end of file reached */
      if (IFX_FALSE == fEndOfFileReached)
      {
         /* read data from file
            in first run nCharsInBuff is equal 0 */
         nReadCharNum = read(nFd, &pBuff[nCharsInBuff - i],
                             (MAX_PATH_LEN - nCharsInBuff));
         /* check if read was successfull */
         if (0 > nReadCharNum)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("Failed to read from fd %d. (File: %s, line: %d)\n",
                   nFd, __FILE__, __LINE__));
            return IFX_ERROR;
         }
         /* check if end of file reached */
         else if (nReadCharNum < (MAX_PATH_LEN - nCharsInBuff))
         {
            /* end of file reached - do not read from file anymore */
            fEndOfFileReached = IFX_TRUE;
         }
         /* number of data in buffer had increased */
         nCharsInBuff += nReadCharNum;

      } /* if (IFX_FALSE == fEndOfFileReached) */

      /* check if number of chars in buffer is bigger than
         length of field name plus two chars: =" */
      if ((nFieldNameLen + 2) >= nCharsInBuff)
      {
         /* number of bytes is too small to hold pFieldName with it's value */
         return IFX_ERROR;
      }
      /* field name can be only found at new line start
         check if buffer is starting with searched string followed by: =" */
      if ((IFX_TRUE == fNewLine) &&
          (0 == strncmp(pFieldName, pBuff, nFieldNameLen)) &&
          ('=' == pBuff[nFieldNameLen]) &&
          ('"' == pBuff[nFieldNameLen + 1]))
      {
         /* copy data until '"' or until end of buffer */
         for (i=(nFieldNameLen + 2); i<nCharsInBuff;i++)
         {
            /* check if invalid char */
            if (('\n' == pBuff[i]) ||
                ('\0' == pBuff[i]))
            {
               TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                     ("Ivalid char (%d) found. (File: %s, line: %d)\n",
                      (IFX_int32_t) pBuff[i], __FILE__, __LINE__));
               return IFX_ERROR;
            }
            /* copied data ends before '"' */
            else if ('"' != pBuff[i])
            {
               /* copy this char */
               pFieldValue[j] = pBuff[i];
               j++;
            }
            else
            {
               /* reached end of data, set string end */
               pFieldValue[j] = '\0';
               /* check number of data copied to pFieldValue */
               if (j > 0)
               {
                  /* value for pFieldName was found in file */
                  return IFX_SUCCESS;
               }
               else
               {
                  /* failed to retrive value for pFieldName from file */
                  return IFX_ERROR;
               } /* if (j > 0) */
            } /* if ('"' != pBuff[i]) */
         } /* for (i=(nFieldNameLen + 2), j=0; i<nCharsInBuff;i++, j++) */
         /* failed to retrive value for pFieldName from file
            value for pFiledName can be too long - '"' not found */
         return IFX_ERROR;
      } /* check if buffer is starting with searched string followed by: =" */
      else
      {
         /* buffer is not set to new line begining */
         fNewLine = IFX_FALSE;
         /* for all chars in buffer */
         for (i=0; i<nCharsInBuff; i++)
         {
            /* search for end of line */
            if ('\n' == pBuff[i])
            {
               /* end of line found */
               fNewLine = IFX_TRUE;
               break;
            }
         } /* for (i=0; i<nCharsInBuff; i++) */

         /* copy unchecked data to temporary buffer */
         memcpy(pTmpBuff, &pBuff[i+1], (nCharsInBuff - i));
         /* copy unchecked data to begining of buffer */
         memcpy(pBuff, pTmpBuff, (nCharsInBuff - i));
         /* number of read from file chars has chenged */
         nCharsInBuff -= i;

      }/* check if buffer is starting with searched string followed by: =" */
   } /* while (1) */

   return IFX_ERROR;
}

/**
   Check if path specified in file is valid and change path to downloads.

   \param sPathToDwnldFiles   - pointer to path string
   \param pBoard              - board handle
   \param pCpuDevice          - pointer to device

   \return IFX_SUCCESS if everything OK, otherwise IFX_ERROR
   \remarks
*/
IFX_return_t Common_CheckDownloadPath(IFX_char_t* sPathToDwnldFiles,
                                      BOARD_t* pBoard,
                                      TAPIDEMO_DEVICE_CPU_t* pCpuDevice)
{
   IFX_int32_t fd;
   IFX_return_t ret = IFX_ERROR;
   IFX_return_t ret2 = IFX_ERROR;
   IFX_char_t psPath[MAX_PATH_LEN], psFile[MAX_PATH_LEN];
   IFX_char_t* pUserName;
    /* Formated user name */
   IFX_char_t psUserName[MAX_USERNAME_LEN];
   IFX_int32_t nUserNameLength;
   IFX_char_t* pHomeDir = "home";
   IFX_char_t* pBackslash = "/";
   /* True if firmware file is found, else FALSE */
   IFX_boolean_t bPRAM_Found = IFX_FALSE;
   /* True if all required BIN files are found, else FALSE */
   IFX_boolean_t bAllBinFound = IFX_TRUE;
   /* Use to check if new download path is directory */
   struct stat st;

   TD_PTR_CHECK(sPathToDwnldFiles, IFX_ERROR);
   TD_PTR_CHECK(pBoard, IFX_ERROR);

   /* check if path to download directory was changed */
   if ((pBoard->pCtrl->pProgramArg->oArgFlags.nUseCustomDownloadsPath) ||
       (pBoard->pCtrl->pProgramArg->oArgFlags.nTapidemorcFileChecked))
   {
      /* path to downloads already changed or .tapidemorc file isn't valid */
      return IFX_SUCCESS;
   }
   /* from now on treat .tapidemorc as checked or invalid */
   pBoard->pCtrl->pProgramArg->oArgFlags.nTapidemorcFileChecked = IFX_TRUE;

   /* get user name */
   pUserName = getenv("USER");
   /* check return value */
   if (IFX_NULL == pUserName)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Failed to get user name. (File: %s, line: %d)\n",
             __FILE__, __LINE__));
      return IFX_ERROR;
   }
   /* to print user name
   TRACE(TAPIDEMO, DBG_LEVEL_LOW, ("User name: %s\n", pUserName));*/
   /* get length of user name */
   nUserNameLength = strlen(pUserName);
   /* check if user name length is valid */
   if ((0 >= nUserNameLength) || (MAX_USERNAME_LEN <= nUserNameLength))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Invalid user name. (File: %s, line: %d)\n",
             __FILE__, __LINE__));
      return IFX_ERROR;
   }

   /* String must be formated */
   snprintf(psUserName, MAX_USERNAME_LEN, "%s", pUserName); 

   /* check if root - root home directory is "/root/" */
   if(0 == strcmp("root", psUserName))
   {
      /* check file name with path length */
      if (MAX_PATH_LEN <=
          (strlen(psUserName) + strlen(DOWNLOAD_FILE_NAME) + 
           2*strlen(pBackslash)))
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("File name with path \"/%s/%s\" is too long.\n"
                "(File: %s, line: %d)\n",
                psUserName, DOWNLOAD_FILE_NAME, __FILE__, __LINE__));
         return IFX_ERROR;
      }
      else
      {
         snprintf(psFile, MAX_PATH_LEN, "%s%s%s%s", 
                  pBackslash, psUserName, 
                  pBackslash, DOWNLOAD_FILE_NAME);
      }
   }
   /* if not root - user home directory is "/home/user_name/ */
   else
   {
      /* check file name with path length */
      if (MAX_PATH_LEN <=
          (strlen(psUserName) + strlen(DOWNLOAD_FILE_NAME) + 
           strlen(pHomeDir) + 3*strlen(pBackslash)))
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("File name with path \"/home/%s/%s\" is too long.\n"
                "(File: %s, line: %d)\n",
                psUserName, DOWNLOAD_FILE_NAME, __FILE__, __LINE__));
         return IFX_ERROR;
      }
      else
      {
         /* get file name with path */
         snprintf(psFile, MAX_PATH_LEN, "%s%s%s%s%s%s", 
                  pBackslash, pHomeDir, 
                  pBackslash, psUserName, 
                  pBackslash, DOWNLOAD_FILE_NAME);
      }
   } /* check if root user */
   /* open file with path to download files */

   if(strnlen(psFile, MAX_PATH_LEN) == MAX_PATH_LEN)
   {
      TRACE (TAPIDEMO, DBG_LEVEL_HIGH,
             ("Path name is incorrect. It is either too long or "
              "it is non null terminated string\n(File: %s, line: %d)\n",
              __FILE__, __LINE__));

      return IFX_ERROR;
   }

   fd = open(psFile, O_RDONLY);
   if (0 > fd)
   {
      /* failed to open file */
      TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
            ("Optional file %s is not found.\n",
             psFile));
      return IFX_ERROR;
   }
   else
   {
      TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
            ("Optional file %s is found.\n", psFile));
   }

   /* read DOWNLOAD_VARIABLE value from file */
   ret = Common_ParseRcFile(DOWNLOAD_VARIABLE, psPath, fd);
   /* file descriptor is no longer needed */
   close(fd);
   /* check if read from file was successfull */
   if (IFX_SUCCESS != ret)
   {
      /* failed to read from file */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Failed to read \"%s\" from file %s - fd %d.\n"
             "(File: %s, line: %d)\n",
             DOWNLOAD_VARIABLE, DOWNLOAD_FILE_NAME, fd, __FILE__, __LINE__));
      return IFX_ERROR;
   }

   /* Check if psPath is directory */
   ret = stat(psPath, &st);
   if((-1 == ret) || !S_ISDIR(st.st_mode))
   {
       TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("The path \"%s\" read from file %s is not a directory.\n",
             psPath, DOWNLOAD_FILE_NAME));
       return IFX_ERROR;
   }

   /* Check if all required BIN files can be found in new path */
   ret = Common_CreateFullFilename(pCpuDevice->pszBBD_CRAM_File, psPath, psFile);
   if (ret == IFX_SUCCESS) 
   {
      ret2 = Common_CheckFile(psFile);
   }
   if (IFX_ERROR == ret || IFX_ERROR == ret2) 
   {
      bAllBinFound = IFX_FALSE;
   }

#if EASY3332
   ret = Common_CreateFullFilename(pCpuDevice->pszDRAM_FW_File, psPath, psFile);
   if (ret== IFX_SUCCESS) 
   {
      ret2 = Common_CheckFile(psFile);
   }
   if (IFX_ERROR == ret || IFX_ERROR == ret2) 
   {
      bAllBinFound = IFX_FALSE;
   }
#endif

   bPRAM_Found = Common_FindPRAM_FW(pCpuDevice, psPath, psFile);
   if(IFX_FALSE == bPRAM_Found)
   {
      bAllBinFound = IFX_FALSE;
   }

   if(IFX_TRUE == bAllBinFound)
   {
      /* All required BIN files are found. Set new path to download files. */
      strcpy(sPathToDwnldFiles, psPath);
      pBoard->pCtrl->pProgramArg->oArgFlags.nUseCustomDownloadsPath = IFX_TRUE;

      TRACE(TAPIDEMO, DBG_LEVEL_NORMAL, 
            ("Checking download path %s from file %s...ok.\n", 
             psPath, DOWNLOAD_FILE_NAME)); 
   }
   else
   {
      pBoard->pCtrl->pProgramArg->oArgFlags.nUseCustomDownloadsPath = IFX_TRUE;

      TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
            ("Checking download path %s from file %s...fail.\n",
             psPath, DOWNLOAD_FILE_NAME));
      return IFX_ERROR;
   }

   TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
         ("Using new download path: %s.\n", 
          sPathToDwnldFiles)); 


   return IFX_SUCCESS;
}
#endif /* VXWORKS */
#endif /* USE_FILESYSTEM */

/**
   Opens specified device.

   \param pszName - pointer to device name

   \return device file descriptor or -1 if fails to open.

   \remark
*/
IFX_return_t Common_Open(IFX_char_t* pszName)
{
#ifdef VXWORKS
   return open(pszName, O_RDWR, 0x644);
#else /* VXWORKS */
   return open(pszName, O_RDWR);
#endif /* VXWORKS */
} /* Common_Open() */

#ifdef EASY336
/**
   Adds a voipID to the array.

   \param pPhone - pointer to phone
   \param nVoipID - voipID to add

   \return IFX_SUCCESS if successful, otherwise IFX_ERROR.

   \remark
      Must be done at start.
*/
static IFX_return_t Common_VoipID_Add(PHONE_t* pPhone, IFX_int32_t nVoipID)
{
   IFX_int32_t *pVoipIDs;

   if (pPhone == IFX_NULL)
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
           __FILE__, __LINE__));
      return IFX_ERROR;
   }
   pVoipIDs = realloc (pPhone->pVoipIDs, ++pPhone->nVoipIDs *
      sizeof(IFX_int32_t));
   if (pVoipIDs == IFX_NULL)
   {
      return IFX_ERROR;
   }
   pPhone->pVoipIDs = pVoipIDs;
   pPhone->pVoipIDs[pPhone->nVoipIDs - 1] = nVoipID;

   return IFX_SUCCESS;
}

/**
   Removes a voipID from the array.

   \param pPhone - pointer to phone
   \param nVoipID - voipID to remove

   \return IFX_SUCCESS if successful, otherwise IFX_ERROR.

   \remark
      Must be done at start.
*/
static IFX_return_t Common_VoipID_Remove(PHONE_t* pPhone, IFX_int32_t nVoipID)
{
   IFX_int32_t i;
   IFX_boolean_t bFound = IFX_FALSE;

   if (pPhone == IFX_NULL)
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
           __FILE__, __LINE__));
      return IFX_ERROR;
   }

   for (i = 0; i < pPhone->nVoipIDs; i++)
   {
      if (pPhone->pVoipIDs[i] == nVoipID)
      {
         pPhone->pVoipIDs[i] = pPhone->pVoipIDs[pPhone->nVoipIDs - 1];
         bFound = IFX_TRUE;
         break;
      }
   }

   if (bFound)
   {
      pPhone->pVoipIDs = realloc (pPhone->pVoipIDs, --pPhone->nVoipIDs *
         sizeof(IFX_int32_t));
      return IFX_SUCCESS;
   }

      return IFX_ERROR;
}
#endif /* EASY336 */

/* ============================= */
/* Global function definition    */
/* ============================= */

/**
   Sets file descriptors.

   \param pBoard - pointer to board

   \return IFX_SUCCESS if successful, otherwise IFX_ERROR.

   \remarks
      On error, CommonClose_FDs() must be called to close all open fd-s.
      On SVIP called twice. In first call it initializes device
      FD to be able to do capabilities listing. During this pass
      pBoard->nUsedChannels == 0 and channel FDs aren't initialized. In
      second call channel FDs get initialized.
*/
IFX_return_t Common_Set_FDs(BOARD_t* pBoard)
{
   IFX_uint8_t i = 0, opened = 0;
   IFX_char_t sdev_name[MAX_DEV_NAME_LEN_WITH_NUMBER];

   if (pBoard == IFX_NULL)
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
           __FILE__, __LINE__));
      return IFX_ERROR;
   }

   /*if ((TYPE_DANUBE != pBoard->nType) || (TYPE_VINAX != pBoard->nType))*/
   if (pBoard->nUseSys_FD)
   {
      /* for VMMC fdSys is not necessary. */
      if (0 >= pBoard->nSystem_FD)
      {
         /* Get system fd */
         pBoard->nSystem_FD = Common_Open(pBoard->pszSystemDevName);

         if (0 >= pBoard->nSystem_FD)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                 ("Invalid system file descriptor (%s). (File: %s, line: %d)\n",
                  pBoard->pszSystemDevName, __FILE__, __LINE__));
            return IFX_ERROR;
         }
         opened++;
      }
      else
      {
         opened++;
      }
   }
   else
   {
      /* As if we open it. */
      opened++;
   }

   if (0 >= pBoard->nDevCtrl_FD)
   {
      /* get vinetic control fd */
      pBoard->nDevCtrl_FD = Common_Open(pBoard->pszCtrlDevName);

      if (0 >= pBoard->nDevCtrl_FD)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Invalid device control file descriptor (%s).\n"
               "(File: %s, line: %d)\n",
               pBoard->pszCtrlDevName, __FILE__, __LINE__));
         return IFX_ERROR;
      }
      opened++;
   }
   else
   {
      opened++;
   }

#ifdef TAPI_VERSION4
   if (!pBoard->fSingleFD)
#endif /* TAPI_VERSION4 */
   {
      for (i = 0; i < pBoard->nUsedChannels; i++)
      {
         if (0 >= pBoard->nCh_FD[i])
         {
            /* len of pBoard->pszChDevName string + 2 * 10(max len of int) */
            if (MAX_DEV_NAME_LEN_WITH_NUMBER > strlen(pBoard->pszChDevName) + 2)
            {
               sprintf(sdev_name, "%s%d%d", (char *) pBoard->pszChDevName,
                                         (int) pBoard->nID,
                                         (int) (i + 1));
            }
            else
            {
               TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                    ("Buffer overflow, array index of 'sdev_name' "
                     "may be outside the bounds. (File: %s, line: %d)\n",
                     __FILE__, __LINE__));
            }

            pBoard->nCh_FD[i] = open(sdev_name, O_RDWR, 0x644);

            if (0 >= pBoard->nCh_FD[i])
            {
               TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                    ("Invalid channel file descriptor (ch%d, %s)."
                     "(File: %s, line: %d)\n",
                     i, sdev_name, __FILE__, __LINE__));
               return IFX_ERROR;
            }
            opened++;
         }
      }

      /* Check opened fds */
      if (opened != (pBoard->nUsedChannels + 2))
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("One or more fds actually in use ... breaking."
               "(File: %s, line: %d)\n",
               __FILE__, __LINE__));
         return IFX_ERROR;
      }
   } /* if (!pBoard->fSingleFD) */
   /* len of pBoard->pszChDevName string + 10(max len of int) + 1('x' char) */
   if(MAX_DEV_NAME_LEN_WITH_NUMBER > (strlen(pBoard->pszChDevName) + 10 + 1))
   {
      sprintf(sdev_name, "%s%dx", (char *) pBoard->pszChDevName,
              (IFX_int32_t) pBoard->nID);
   }
   else
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Buffer overflow, array index of 'sdev_name' "
            "may be outside the bounds. (File: %s, line: %d)\n",
            __FILE__, __LINE__));
   }

   TRACE(TAPIDEMO, DBG_LEVEL_LOW,
         ("Device (%s) and channel (%s) fds opened.\n",
          pBoard->pszCtrlDevName, sdev_name));

   return IFX_SUCCESS;
} /* CommonSet_FDs() */


/**
   Close file descriptors.

   \param pBoard - pointer to board

   \return none
*/
IFX_void_t Common_Close_FDs(BOARD_t* pBoard)
{
   IFX_int32_t i = 0;


   if (pBoard == IFX_NULL)
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
           __FILE__, __LINE__));
      return;
   }

   /*if ((TYPE_DANUBE != pBoard->nType) || (TYPE_VINAX != pBoard->nType))*/
   if (pBoard->nUseSys_FD)
   {
      /* for VMMC fdSys is not necessary. */
      if (0 < pBoard->nSystem_FD)
      {
         close(pBoard->nSystem_FD);
         pBoard->nSystem_FD = -1;
      }
   }

   if (0 < pBoard->nDevCtrl_FD)
   {
      close(pBoard->nDevCtrl_FD);
      pBoard->nDevCtrl_FD = -1;
   }

#ifdef TAPI_VERSION4
   if (!pBoard->fSingleFD)
#endif /* TAPI_VERSION4 */
   {
      for (i = 0; i < pBoard->nUsedChannels; i++)
      {
         if (pBoard->nCh_FD[i] > 0)
         {
            close(pBoard->nCh_FD[i]);
            pBoard->nCh_FD[i] = -1;
         }
      } /* for */
   } /* if (!pBoard->fSingleFD) */

   TRACE(TAPIDEMO, DBG_LEVEL_NORMAL, ("Device and channel fds closed.\n"));
} /* CommonClose_FDs() */


/**
   Sets FXO (duslic or teridian) file descriptors.

   \param pBoard - pointer to board

   \return IFX_SUCCESS if successful, otherwise IFX_ERROR.

   \remarks
      On error, CommonClose_FDs() must be called to close all open fd-s.
*/
IFX_return_t Common_Set_Fxo_FDs(BOARD_t* pBoard)
{
   IFX_uint8_t i = 0, opened = 0;
   IFX_char_t sdev_name[MAX_DEV_NAME_LEN_WITH_NUMBER];

   if (pBoard == IFX_NULL)
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
           __FILE__, __LINE__));
      return IFX_ERROR;
   }

   if (0 >= pBoard->nFxoDevCtrl_FD)
   {
      pBoard->nFxoDevCtrl_FD = Common_Open(pBoard->pszFxoCtrlDevName);
      if (0 >= pBoard->nFxoDevCtrl_FD)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Invalid FXO device control file descriptor (%s)."
               "(File: %s, line: %d)\n",
               pBoard->pszFxoCtrlDevName, __FILE__, __LINE__));
         return IFX_ERROR;
      }
      opened++;
   }

#ifdef TAPI_VERSION4
   if (!pBoard->fSingleFD)
#endif
   {
      for (i = 0; i < pBoard->nMaxFxoCh; i++)
      {
         if (0 >= pBoard->nFxoCh_FD[i])
         {
            /* len of pBoard->pszChDevName string + 2 * 10(max len of int) */
            if (MAX_DEV_NAME_LEN_WITH_NUMBER > 
            strlen(pBoard->pszFxoChDevName) + 2 * 10)
            {
               sprintf(sdev_name, "%s%d%d", (char *) pBoard->pszFxoChDevName,
                                         (int) pBoard->nID,
                                         (int) (i + 1));
            }
            else
            {
               TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                    ("Buffer overflow, array index of 'sdev_name' "
                     "may be outside the bounds. (File: %s, line: %d)\n",
                     __FILE__, __LINE__));
            }

            pBoard->nFxoCh_FD[i] = open(sdev_name, O_RDWR, 0x644);

            if (0 >= pBoard->nFxoCh_FD[i])
            {
               TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                    ("Invalid FXO channel file descriptor (ch%d, %s)."
                     "(File: %s, line: %d)\n",
                     i, sdev_name, __FILE__, __LINE__));
               return IFX_ERROR;
            }
            /* to print device name and fd number
            TRACE(TAPIDEMO, DBG_LEVEL_LOW,
                 ("%s, %s -> fd %d.\n",
                  pBoard->pszBoardName, sdev_name, pBoard->nFxoCh_FD[i]));*/
            opened++;
         }
      }

      /* Check opened fds */
      if (opened != (pBoard->nMaxFxoCh + 1))
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("One or more FXO fds actually in use ... breaking."
               "(File: %s, line: %d)\n",
               __FILE__, __LINE__));
         return IFX_ERROR;
      }
   } /* if (!pBoard->fSingleFD) */
   /* len of pBoard->pszChDevName string + 10(max len of int) + 1('x' char) */
   if(MAX_DEV_NAME_LEN_WITH_NUMBER > (strlen(pBoard->pszFxoChDevName) + 10 + 1))
   {
      sprintf(sdev_name, "%s%dx", (char *) pBoard->pszFxoChDevName,
           (int) pBoard->nID);
   }
   else
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Buffer overflow, array index of 'sdev_name' "
            "may be outside the bounds. (File: %s, line: %d)\n",
            __FILE__, __LINE__));
   }

   TRACE(TAPIDEMO, DBG_LEVEL_LOW,
         ("FXO device (%s) and channel (%s) fds opened.\n",
          pBoard->pszFxoCtrlDevName, sdev_name));

   return IFX_SUCCESS;
} /* Common_Set_Fxo_FDs() */

/**
   Close FXO (duslic or teridian) file descriptors.

   \param pBoard - pointer to board

   \return none
*/
IFX_void_t Common_Close_Fxo_FDs(BOARD_t* pBoard)
{
   IFX_int32_t i = 0;


   if (pBoard == IFX_NULL)
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
           __FILE__, __LINE__));
      return;
   }

   if (0 < pBoard->nFxoDevCtrl_FD)
   {
      close(pBoard->nFxoDevCtrl_FD);
      pBoard->nFxoDevCtrl_FD = -1;
   }

#ifdef TAPI_VERSION4
   if (!pBoard->fSingleFD)
#endif
   {
      for (i = 0; i < pBoard->nMaxFxoCh; i++)
      {
         if (pBoard->nFxoCh_FD[i] > 0)
         {
            close(pBoard->nFxoCh_FD[i]);
            pBoard->nFxoCh_FD[i] = -1;
         }
      } /* for */
   } /* if (!pBoard->fSingleFD) */

   TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
         ("FXO device and channel fds closed.\n"));
} /* Common_Close_Fxo_FDs() */

/**
   Return file descriptor for specific channel (it makes no difference if
                                                phone or data)

   \param pBoard - pointer to board
   \param nChNum - channel number (starting from 0)

   \return -1 if err or device not open, otherwise file descriptor
*/
IFX_int32_t Common_GetDeviceOfCh(BOARD_t* pBoard, IFX_int32_t nChNum)
{
   if ((pBoard == IFX_NULL) || (0 > nChNum) || (pBoard->nUsedChannels < nChNum))
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
           __FILE__, __LINE__));
      return IFX_ERROR;
   }
#ifdef TAPI_VERSION4
   if (pBoard->fSingleFD)
      return pBoard->nDevCtrl_FD;
   else
#endif
      return pBoard->nCh_FD[nChNum];

} /* Common_GetDeviceOfCh() */


/**
   Return file descriptor for FXO device control fd.

   \param pBoard - pointer to board

   \return -1 if err or device not open, otherwise file descriptor.
*/
IFX_int32_t Common_GetFxoDevCtrlFd(BOARD_t* pBoard)
{
   if (pBoard == IFX_NULL)
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
           __FILE__, __LINE__));
      return IFX_ERROR;
   }

   return pBoard->nFxoDevCtrl_FD;
} /* Common_GetFxoDevCtrlFd() */

/**
   Return file descriptor for FXO specific channel

   \param pBoard - pointer to board
   \param nChNum - channel number (starting from 0)

   \return -1 if err or device not open, otherwise file descriptor
*/
IFX_int32_t Common_GetFxoDevOfCh(BOARD_t* pBoard, IFX_int32_t nChNum)
{
   if ((pBoard == IFX_NULL) || (0 > nChNum) || (pBoard->nMaxFxoCh < nChNum))
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
           __FILE__, __LINE__));
      return IFX_ERROR;
   }

   return pBoard->nFxoCh_FD[nChNum];
} /* Common_GetFxoDevOfCh() */

/**
   Read version after succesfull initialization and display it.

   \param pBoard - pointer to board

   \return IFX_SUCCESS if successful, otherwise IFX_ERROR.

   \remark
      This must be done only after a successfull basic device initialization
*/
IFX_return_t Common_GetVersions(BOARD_t* pBoard)
{
   IFX_return_t ret = IFX_SUCCESS;

   return ret;
} /* Common_GetVersions() */

/**
   Board registration.

   \param pBoard - pointer to board

   \return IFX_SUCCESS if successful, otherwise IFX_ERROR.

   \remark
*/
IFX_return_t Common_RegisterBoard(BOARD_t* pBoard)
{
   IFX_return_t ret = IFX_SUCCESS;

   if (pBoard == IFX_NULL)
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
           __FILE__, __LINE__));
      return IFX_ERROR;
   }

   switch (pBoard->nType)
   {
#if EASY3332
      case TYPE_VINETIC:
         ret = BOARD_Easy3332_Register(pBoard);
      break;
#endif /* EASY3332  */
#if EASY50510
      case TYPE_VINETIC:
         ret = BOARD_Easy50510_Register(pBoard);
      break;
#endif /* EASY50510  */
#if VMMC
      case TYPE_DANUBE:
         ret = BOARD_Easy50712_Register(pBoard);
      break;
#endif /* VMMC */
#if DXT
      case TYPE_DUSLIC_XT:
         if (0 == (pBoard->nID - 1))
         {
            ret = BOARD_Easy3201_Register(pBoard);
         }
         else
         {
            ret = BOARD_Easy3111_Register(pBoard);
         }
      break;
#endif /* DXT */
#if VINAX
      case TYPE_VINAX:
         ret = BOARD_Easy80800_Register(pBoard);
      break;
#endif /* VINAX */
#if AR9
      case TYPE_EASY508XX:
         ret = BOARD_Easy508XX_Register(pBoard);
      break;
#endif /* AR9 */
#if EASY336
      case TYPE_SVIP:
         ret = BOARD_Easy336_Register(pBoard);
      break;
#ifdef WITH_VXT
      case TYPE_XT16:
         ret = BOARD_XT_16_Register(pBoard);
      break;
#endif /* WITH_VXT */
#endif /* EASY336 */
#if XT16
      case TYPE_XT16:
         ret = BOARD_XT_16_Register(pBoard);
      break;
#endif /* XT16 */
#if TERIDIAN_FXO
      case TYPE_TERIDIAN:
         ret = BOARD_Teridian73M1866_Register(pBoard);
      break;
#endif /* TERIDIAN_FXO */
#if VR9
      case TYPE_VR9:
         ret = BOARD_Easy80910_Register(pBoard);
      break;
#endif /* VR9 */
      default:
         /* Unknown board type */
         ret = IFX_ERROR;
      break;
   }

   return ret;
} /* Common_RegisterBoard() */

/**
   Handles the TAPI error codes and stack.

   \param nFD -    device file descriptor of system, which generated
                   the error (must not be fd of channel)
   \param pBoard - board where the error occured
   \param nDev   - device number where the error occured             
   \remarks
*/
IFX_void_t Common_ErrorHandle(IFX_int32_t nFD, BOARD_t* pBoard, 
   IFX_uint16_t nDev)
{
   IFX_TAPI_Error_t error = {0};
   IFX_int32_t i = 0;
   IFX_return_t ret = IFX_SUCCESS;
   IFX_uint32_t code = 0;
   TAPIDEMO_DEVICE_t *pDevice = IFX_NULL;
   TAPIDEMO_DEVICE_CPU_t *pCpuDevice = IFX_NULL;

   if (pBoard == IFX_NULL)
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return;
   }
#ifdef TAPI_VERSION4
   error.nDev = nDev;
#endif   
   ret = ioctl(nFD, IFX_TAPI_LASTERR, (IFX_int32_t) &error);

   if (ret != IFX_ERROR && error.nCode > 0)
   {
      /* we have additional information */
#ifdef TAPI_VERSION4
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
         ("%s: Error on device %2d\n"
         "%sError code 0x%X.\n",
         pBoard->pszBoardName, error.nDev,
         pBoard->pIndentBoard, error.nCode));
#else
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
         ("%s: Error code 0x%X.\n",
         pBoard->pszBoardName, error.nCode));
#endif
      if ((pDevice = Common_GetDevice_CPU(pBoard->pDevice)) == IFX_NULL)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Err, CPU device not found. (File: %s, line: %d)\n",
               __FILE__, __LINE__));
         return;
      }
      pCpuDevice = &pDevice->uDevice.stCpu;

      for (i = 0; i < error.nCnt; ++i)
      {
         code = (error.stack[i].nHlCode << 16) |
                 error.stack[i].nLlCode;

         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("%s:%d HL Code 0x%X, LL Code 0x%X\n",
               error.stack[i].sFile,
               error.stack[i].nLine,
               error.stack[i].nHlCode,
               error.stack[i].nLlCode));
         pCpuDevice->DecodeErrno(code);
      } /* for */
      if (error.nCnt >= IFX_TAPI_MAX_ERROR_ENTRIES)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Buffer overflow, array index of 'error.stack' "
               "may be outside the bounds. (File: %s, line: %d)\n",
               __FILE__, __LINE__));
      }
   } /* if (ret != IFX_ERROR && error.nCode > 0) */
} /* Common_ErrorHandle */

#ifndef STREAM_1_1
/**
   Get capabilities, basically number of phone, data and pcm channels.

   \param pBoard - pointer to board
   \param nPhoneChCnt - pointer to phone channels count
   \param nDataChCnt  - pointer to data channels count
   \param nPCM_ChCnt  - pointer to pcm channels count

   \return IFX_SUCCESS on ok, otherwise IFX_ERROR
*/
IFX_return_t Common_GetCapabilities(BOARD_t* pBoard,
                                    IFX_int32_t *nPhoneChCnt,
                                    IFX_int32_t *nDataChCnt,
                                    IFX_int32_t *nPCM_ChCnt)
{
   IFX_return_t ret = IFX_SUCCESS;
#ifdef TAPI_VERSION4
   IFX_TAPI_CAP_t CapList;
   IFX_int32_t nIores;
#else /* TAPI_VERSION4 */
   IFX_int32_t nCapNr = 0;
   IFX_TAPI_CAP_t *pCapList = IFX_NULL;
   IFX_int32_t nT38ChCnt = 0;
#endif /* TAPI_VERSION4 */
   IFX_int32_t i = 0;

   if ((IFX_NULL == nPhoneChCnt) || (IFX_NULL == nDataChCnt)
       || (IFX_NULL == nPCM_ChCnt) || (pBoard == IFX_NULL))
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }

#ifndef TAPI_VERSION4
   /* Get the cap list size */
   ret = ioctl(pBoard->nDevCtrl_FD, IFX_TAPI_CAP_NR, (int) &nCapNr);
   if (IFX_ERROR == ret)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("IFX_TAPI_CAP_LIST returned error. (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      Common_ErrorHandle(pBoard->nDevCtrl_FD, pBoard, 0);
      return IFX_ERROR;
   }

   TRACE(TAPIDEMO, DBG_LEVEL_LOW,
        ("Number of capabilities %d.\n", (int) nCapNr));

#ifdef VXWORKS
   if (0 == nCapNr)
   {
      /** \todo Under VxWorks a count 0 is received, strange ??? */
      nCapNr = 20;
   }
#endif /* VXWORKS */

   if (0 == nCapNr)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("0 capabilities available read from fd %d of ctrl ch, "
              " was TAPI initialized? \n",
              (int) pBoard->nDevCtrl_FD));
      return IFX_ERROR;
   }

   /* \todo VOS_Malloc() check for not enough memory?, yes but return NULL */
   pCapList = malloc(nCapNr * sizeof(IFX_TAPI_CAP_t));
   if (IFX_NULL == pCapList)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("pCapList is null. (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }

   memset(pCapList, 0, sizeof(pCapList));

   /* Get the cap list */
   ret = ioctl(pBoard->nDevCtrl_FD, IFX_TAPI_CAP_LIST, (int) &(*pCapList));
   if (IFX_ERROR == ret)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("IFX_TAPI_CAP_LIST returned error. (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      free(pCapList);
      return IFX_ERROR;
   }

   for (i = 0; i < nCapNr; i++)
   {

#ifdef ABACUS_TEST
      COM_SendCapabilityData(&pCapList[i]);
#endif
      switch (pCapList[i].captype)
      {
         case IFX_TAPI_CAP_TYPE_PCM:
               *nPCM_ChCnt = pCapList[i].cap;
            break;
         case IFX_TAPI_CAP_TYPE_CODECS:
               *nDataChCnt = pCapList[i].cap;
            break;
         case IFX_TAPI_CAP_TYPE_PHONES:
               *nPhoneChCnt = pCapList[i].cap;
            break;
          case IFX_TAPI_CAP_TYPE_T38:
               nT38ChCnt = pCapList[i].cap;
            break;
         default:
            break;
      }
   } /* for all cap.. */

   TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
         ("Available channels: phone %d, data %d, PCM %d, T.38 %d.\n",
          (int) *nPhoneChCnt, (int) *nDataChCnt, (int) *nPCM_ChCnt, nT38ChCnt));

   if (nT38ChCnt > 0)
   { /* T.38 supported*/
      pBoard->nT38_Support = IFX_TRUE;
   }
   /* Free the allocated memory */
   free(pCapList);
   pCapList = IFX_NULL;
#else /* TAPI_VERSION4 */
   *nDataChCnt = 0;
   *nPhoneChCnt = 0;
   *nPCM_ChCnt = 0;
   for (i = 0; i < pBoard->nChipCnt; i++)
   {
      /*list codecs*/
      memset(&CapList, 0, sizeof(IFX_TAPI_CAP_t));
      CapList.dev = i;
      CapList.captype = IFX_TAPI_CAP_TYPE_CODECS;
      TD_ASSERT((nIores = ioctl(pBoard->nDevCtrl_FD, IFX_TAPI_CAP_CHECK,
         (IFX_int32_t) &CapList)) >= 0, pBoard->nDevCtrl_FD, pBoard,
         CapList.dev, 0);
      switch (nIores)
      {
         case 1:
#ifdef EASY336
            if (pBoard->nType == TYPE_SVIP && CapList.cap == 0)
               CapList.cap = MAX_CODER_CH_SVIP;
#endif
            *nDataChCnt += (IFX_uint8_t) CapList.cap;
            TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("%d data channels available on %s device %d.\n",
               CapList.cap, pBoard->pszBoardName, i));
            break;
         case 0:
            TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("Capability type IFX_TAPI_CAP_TYPE_CODECS not supported.\n"));
            break;
         case -1:
            /* Error reading codec number */
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Error reading data channel number. (File: %s, line: %d)\n",
               __FILE__, __LINE__));
               return IFX_ERROR;
      }
      /*list ALMs*/
      memset(&CapList, 0, sizeof(IFX_TAPI_CAP_t));
      CapList.dev = i;
      CapList.captype = IFX_TAPI_CAP_TYPE_PHONES;
      TD_ASSERT ((nIores = ioctl(pBoard->nDevCtrl_FD, IFX_TAPI_CAP_CHECK,
         (IFX_int32_t) &CapList)) >= 0, pBoard->nDevCtrl_FD, pBoard,
         CapList.dev, 0);
      switch (nIores)
      {
         case 1:
            *nPhoneChCnt += CapList.cap;
            TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("%d phone channels available on %s device %d.\n",
               CapList.cap, pBoard->pszBoardName, i));
            break;
         case 0:
            TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("Capability type IFX_TAPI_CAP_TYPE_PHONES not supported.\n"));
            break;
         case -1:
            /* Error reading ALM number */
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                 ("Error reading phone channel number. (File: %s, line: %d)\n",
                  __FILE__, __LINE__));
            return IFX_ERROR;
      }
      /*list PCMs*/
      memset(&CapList, 0, sizeof(IFX_TAPI_CAP_t));
      CapList.dev = i;
      CapList.captype = IFX_TAPI_CAP_TYPE_PCM;
      TD_ASSERT ((nIores = ioctl(pBoard->nDevCtrl_FD, IFX_TAPI_CAP_CHECK,
         (IFX_int32_t) &CapList)) >= 0, pBoard->nDevCtrl_FD, pBoard, 
         CapList.dev, 0);
      switch (nIores)
      {
         case 1:
#ifdef EASY336
            if (pBoard->nType == TYPE_SVIP &&
               CapList.cap == 0)
               CapList.cap = MAX_PCM_CH_SVIP;
#endif
            *nPCM_ChCnt += CapList.cap;
            TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("%d PCM channels available on %s device %d.\n",
               CapList.cap, pBoard->pszBoardName, i));
            break;
         case 0:
            TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("Capability type IFX_TAPI_CAP_TYPE_PCM not supported.\n"));
            break;
         case -1:
            /* Error reading PCM number */
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                 ("Error reading PCM channel number. (File: %s, line: %d)\n",
                  __FILE__, __LINE__));
            return IFX_ERROR;
      }
   }
#endif /* TAPI_VERSION4 */

   return ret;
} /* Common_GetCapabilities() */

#else

IFX_return_t Common_GetCapabilities(BOARD_t* pBoard,
                                    IFX_int32_t *nPhoneChCnt,
                                    IFX_int32_t *nDataChCnt,
                                    IFX_int32_t *nPCM_ChCnt)
{
   return IFX_SUCCESS;
} /* Common_GetCapabilities() */
#endif /* STREAM_1_1 */

/**
   Get trace indention pointer.

   \param nSize - size of indention 

   \return string with indention
*/

IFX_char_t* Common_GetTraceIndention(IFX_int32_t nSize)
{
   IFX_int32_t i;

   /* Fill string with spaces - should be done only once, but because this
      function is used only during initialization this ia acceptable */
   for(i=0; i < TD_MAX_INDENT; i++)
   {
      acTraceIndentTable[i] = ' ';
   }
   /* set end of string */
   acTraceIndentTable[TD_MAX_INDENT - 1]='\0';

   /* check if valid size */
   if (0 > nSize)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Invalid indentation size %d, setting Indetation size to 0\n",
             nSize));
      nSize = 0;
   }
   /* indentation must be less than TD_MAX_INDENT */
   if (nSize >= TD_MAX_INDENT)
   {
      return &acTraceIndentTable[0];
   }
   else
   {
      /* table elements are counted from 0 so decrease char address by 1 */
      return &acTraceIndentTable[TD_MAX_INDENT - nSize - 1];
   }
}

/*
    Sets codec according to dialed sequence.

    \param pBoard - pointer to the board
    \param pPhone - pointer to the phone
    \param nAction - which action was selected (dialed number)
    \param fConfigureAll - true if all phone on this baord will be configured,
                           otherwise (false) only this one will be configured,
    \param fConfigureConn - configure both phones in connection

    \return IFX_SUCCESS if ok, otherwise IFX_ERROR
*/
IFX_return_t Common_SetCodec(BOARD_t* pBoard, PHONE_t* pPhone,
                             IFX_int32_t nAction, IFX_boolean_t fConfigureAll,
                             IFX_boolean_t fConfigureConn)
{
   IFX_int32_t nCodec = 0;
   IFX_int32_t nPacketLen = 0;
   IFX_int32_t i = 0;
#ifdef EASY336
   VOIP_DATA_CH_t* pCodec = IFX_NULL;
#endif
#if (!defined EASY336) && (!defined XT16)
   IFX_int32_t nDataCh = -1;
#endif /* (!defined EASY336) && (!defined XT16) */
#ifndef XT16
   CONNECTION_t* pConn = IFX_NULL;
#endif /* XT16 */
   IFX_TAPI_COD_AAL2_BITPACK_t nBitPack = IFX_TAPI_COD_RTP_BITPACK;

   /* check input parameters */
   TD_PTR_CHECK(pBoard, IFX_ERROR);
   TD_PTR_CHECK(pPhone, IFX_ERROR);
   TD_PTR_CHECK(pBoard->pCtrl, IFX_ERROR);

   /* check if coder channels are avaible and structure allocated */
   if ((0 >= pBoard->nMaxCoderCh || IFX_NULL == pBoard->pDataChStat)
#ifdef EASY336
       /* if xt16 board is connected to svip board then resource manager
         alows external voip calls for phones on XT16 board */
       && (TYPE_XT16 != pBoard->nType)
#endif /* EASY336 */
      )
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("%s: Feature *%d - codec change not avaible - no coder channels\n",
             pBoard->pszBoardName, nAction));
      return IFX_ERROR;
   }
   /* check feature number */
   if ((nAction < 1000) || (nAction > 1999))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Invalid set codec feature number %d\n", nAction));
      return IFX_ERROR;
   }

   /* get number of codec */
   i = (nAction%1000)/10;
   i = i - (i%2);
   /* because more packet lengths can be added
      there are 20 DTMF codes for each codec */
   switch(i)
   {
      /* G.711 codecs */
      /* DTMF code *100x */
      case 0:
         nCodec = IFX_TAPI_ENC_TYPE_MLAW;
         break;
      /* DTMF code *102x */
      case 2:
         nCodec = IFX_TAPI_ENC_TYPE_ALAW;
         break;
      /* G.726 codecs */
      /* DTMF code *104x */
      case 4:
         nCodec = IFX_TAPI_ENC_TYPE_G726_16;
         break;
      /* DTMF code *106x */
      case 6:
         nCodec = IFX_TAPI_ENC_TYPE_G726_24;
         break;
      /* DTMF code *108x */
      case 8:
         nCodec = IFX_TAPI_ENC_TYPE_G726_32;
         break;
      /* DTMF code *110x */
      case 10:
         nCodec = IFX_TAPI_ENC_TYPE_G726_40;
         break;
      /* G.729 codec */
      case 12:
         nCodec = IFX_TAPI_ENC_TYPE_G729;
         break;
      /* AMR codecs */
      case 14:
         nCodec = IFX_TAPI_ENC_TYPE_AMR_4_75;
         break;
      case 16:
         nCodec = IFX_TAPI_ENC_TYPE_AMR_5_15;
         break;
      case 18:
         nCodec = IFX_TAPI_ENC_TYPE_AMR_5_9;
         break;
      case 20:
         nCodec = IFX_TAPI_ENC_TYPE_AMR_6_7;
         break;
      case 22:
         nCodec = IFX_TAPI_ENC_TYPE_AMR_7_4;
         break;
      case 24:
         nCodec = IFX_TAPI_ENC_TYPE_AMR_7_95;
         break;
      case 26:
         nCodec = IFX_TAPI_ENC_TYPE_AMR_10_2;
         break;
      case 28:
         nCodec = IFX_TAPI_ENC_TYPE_AMR_12_2;
         break;
      /* iLBC codecs */
      case 30:
         nCodec = IFX_TAPI_ENC_TYPE_ILBC_133;
         break;
      case 32:
         nCodec = IFX_TAPI_ENC_TYPE_ILBC_152;
         break;
      /* G.722 codec */
      case 34:
         nCodec = IFX_TAPI_ENC_TYPE_G722_64;
         break;
      /* G.722.1 codecs */
      case 36:
         nCodec = IFX_TAPI_ENC_TYPE_G7221_24;
         break;
      case 38:
         nCodec = IFX_TAPI_ENC_TYPE_G7221_32;
         break;
      /* linear codecs */
      case 40:
         nCodec = IFX_TAPI_ENC_TYPE_LIN16_8;
         break;
      case 42:
         nCodec = IFX_TAPI_ENC_TYPE_LIN16_16;
         break;
      case 44:
         nCodec = IFX_TAPI_ENC_TYPE_G723_53;
         break;
      case 46:
         nCodec = IFX_TAPI_ENC_TYPE_G723_63;
         break;
      case 48:
         nCodec = IFX_TAPI_ENC_TYPE_MLAW_VBD;
         break;
      case 50:
         nCodec = IFX_TAPI_ENC_TYPE_ALAW_VBD;
         break;
      case 52:
         nCodec = IFX_TAPI_ENC_TYPE_G728;
         break;
      /* DTMF code *1540x */
      case 54:
         nCodec = IFX_TAPI_ENC_TYPE_G729_E;
         break;
      /* G.726 codecs */
      /* DTMF code *156x */
      case 56:
         nCodec = IFX_TAPI_ENC_TYPE_G726_16;
         nBitPack = IFX_TAPI_COD_AAL2_BITPACK;
         break;
      /* DTMF code *158x */
      case 58:
         nCodec = IFX_TAPI_ENC_TYPE_G726_24;
         nBitPack = IFX_TAPI_COD_AAL2_BITPACK;
         break;
      /* DTMF code *160x */
      case 60:
         nCodec = IFX_TAPI_ENC_TYPE_G726_32;
         nBitPack = IFX_TAPI_COD_AAL2_BITPACK;
         break;
      /* DTMF code *162x */
      case 62:
         nCodec = IFX_TAPI_ENC_TYPE_G726_40;
         nBitPack = IFX_TAPI_COD_AAL2_BITPACK;
         break;
      default:
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("%s: Feature *%d - codec not avaible\n",
                pBoard->pszBoardName, nAction));
         return IFX_ERROR;
         break;
   } /* switch() */

   /* check if codec supported */
   if (IFX_SUCCESS != VOIP_CodecAvailabilityCheck(pBoard, nCodec))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("%s: Feature *%d - codec not supported\n",
             pBoard->pszBoardName, nAction));
      return IFX_ERROR;
   }

   /* get packetisation length */
   i = nAction %20;
   switch (i)
   {
      case 0:
         nPacketLen = IFX_TAPI_COD_LENGTH_10;
         break;
      case 1:
         nPacketLen = IFX_TAPI_COD_LENGTH_20;
         break;
      case 2:
         nPacketLen = IFX_TAPI_COD_LENGTH_30;
         break;
      case 3:
         nPacketLen = IFX_TAPI_COD_LENGTH_40;
         break;
      case 4:
         nPacketLen = IFX_TAPI_COD_LENGTH_50;
         break;
      case 5:
         nPacketLen = IFX_TAPI_COD_LENGTH_60;
         break;
      default:
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("%s: Feature *%d - packet length not avaible\n",
                pBoard->pszBoardName, nAction));
         return IFX_ERROR;
   } /* switch (i) */

   if ((nCodec != 0) && (nPacketLen != 0))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("Feature DTMF code: *%d - Codec change.\n", nAction));
      /* Configue all channels on this board */
      if (fConfigureAll)
      {
         /* check if board has data channels */
         if (0 >= pBoard->nMaxCoderCh)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("%s: Unable to change codec - no data channels avaible."
                   " (File: %s, line: %d)\n",
                   pBoard->pszBoardName, __FILE__, __LINE__));
            return IFX_ERROR;
         }
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("%s: Set codec for all data channels on this board:\n"
                "%s%s, %s\n",
                pBoard->pszBoardName,
                pBoard->pIndentBoard, 
                pCodecName[nCodec], oFrameLen[nPacketLen]));

         for (i = 0; i < pBoard->nMaxCoderCh; ++i)
         {
            pBoard->pDataChStat[i].nCodecType = nCodec;
            pBoard->pDataChStat[i].nFrameLen = nPacketLen;
            pBoard->pDataChStat[i].fStartCodec = IFX_TRUE;
            pBoard->pDataChStat[i].nBitPack = nBitPack;
         } /* for */
      }
      /* Configure only this one */
      else
      {
         /* this feature doesn't work for conference */
#if (!defined EASY336) && (!defined XT16)
         pBoard->pDataChStat[pPhone->nDataCh].nCodecType = nCodec;
         pBoard->pDataChStat[pPhone->nDataCh].nFrameLen = nPacketLen;
         pBoard->pDataChStat[pPhone->nDataCh].fStartCodec = IFX_TRUE;
         pBoard->pDataChStat[pPhone->nDataCh].nBitPack = nBitPack;

         pConn = &pPhone->rgoConn[0];

         if ((IFX_NULL != pConn) &&
             (NO_CONFERENCE == pPhone->nConfIdx))
         {
            if ((EXTERN_VOIP_CALL == pConn->nType) ||
                ((LOCAL_CALL == pConn->nType) &&
                 (pBoard->pCtrl->pProgramArg->oArgFlags.nUseCodersForLocal)))
            {
               nDataCh = pPhone->rgoConn[0].nUsedCh;
               VOIP_StopCodec(nDataCh, pConn, pPhone->pBoard);
               if (IFX_SUCCESS == VOIP_StartCodec(nDataCh, pConn,
                                                  pPhone->pBoard))
               {
                  if(IFX_TAPI_ENC_TYPE_MAX > nCodec)
                  {
                     TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
                           ("Phone No %d: Voice encoding: %s, %s\n",
                            pPhone->nPhoneNumber,
                            pCodecName[nCodec],
                            oFrameLen[nPacketLen]));
                  }
                  else
                  {
                     TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                           ("Buffer overflow, array index of 'pCodecName' "
                            "may be outside the bounds. (File: %s, line: %d)\n",
                            __FILE__, __LINE__));
                  }
               }
               pConn->nFeatureID = nAction;
               if (IFX_TRUE == fConfigureConn)
               {
                  TAPIDEMO_SetAction(pBoard->pCtrl, pPhone, pConn,
                                     IE_CONFIG_PEER);
               }
            }
         } /* if ((IFX_NULL != conn) */
#elif EASY336
         /* check if using Voip ID, connection in progress
            and no conference */
         if ((pPhone->nVoipIDs > 0) &&
             (NO_CONFERENCE == pPhone->nConfIdx) &&
             (0 < pPhone->nConnCnt))
         {
            /* get connection */
            pConn = &pPhone->rgoConn[0];
            /* check if extern call or local call using codecs */
            if ((EXTERN_VOIP_CALL == pConn->nType) ||
                ((LOCAL_CALL == pConn->nType) &&
                 (pBoard->pCtrl->pProgramArg->oArgFlags.nUseCodersForLocal)))
            {
               /* get pointer to codec structure */
               pCodec = Common_GetCodec(pBoard->pCtrl, pConn->voipID);
               if (pCodec != IFX_NULL)
               {
                  /* change codec and packetization */
                  pCodec->nCodecType = nCodec;
                  pCodec->nFrameLen = nPacketLen;
                  pCodec->fStartCodec = IFX_TRUE;
                  pCodec->nBitPack = nBitPack;
               }
               else
               {
                  TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                        ("Phone No %d: Failed to get codec. "
                         "(File: %s, line: %d)\n",
                         pPhone->nPhoneNumber, __FILE__, __LINE__));
                  return IFX_ERROR;
               }
               /* for TAPI V4 number of data channel is not needed */
               VOIP_StopCodec(0, pConn, ABSTRACT_GetBoard(pBoard->pCtrl,
                  IFX_TAPI_DEV_TYPE_VIN_SUPERVIP));
               if (IFX_SUCCESS == VOIP_StartCodec(0, pConn,
                   ABSTRACT_GetBoard(pBoard->pCtrl,
                                     IFX_TAPI_DEV_TYPE_VIN_SUPERVIP)))
               {
                  /* printout codec data */
                  TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
                        ("Phone No %d: Voice encoding: %s, %s\n",
                         pPhone->nPhoneNumber,
                         pCodecName[pCodec->nCodecType],
                         oFrameLen[pCodec->nFrameLen]));
               }
               else
               {
                  TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                        ("Phone No %d: Failed to change codec. "
                         "(File: %s, line: %d)\n",
                         pPhone->nPhoneNumber, __FILE__, __LINE__));
                  return IFX_ERROR;
               }
               /* check if second phone should be configured */
               if (IFX_TRUE == fConfigureConn)
               {
                  pConn->nFeatureID = nAction;
                  TAPIDEMO_SetAction(pBoard->pCtrl, pPhone, pConn,
                                     IE_CONFIG_PEER);
               }
            }
         } /* if ((pPhone->nVoipIDs > 0) && ... */
#endif /* EASY336 */
      } /* if (fConfigureAll) */
      return IFX_SUCCESS;
   }
   else
   {
      return IFX_ERROR;
   }
}

/*
    Configures additional features according to dialing sequence.

    \param pCtrl - pointer to the control status structure
    \param pBoard - pointer to the board
    \param pPhone - pointer to the phone
    \param nAction - which action was selected (dialed number)
    \param nDataCh_FD - file descriptor of the data channel
    \param fConfigureAll - true if all phone on this baord will be configured
                           (selected feature must suppotr this),
                           otherwise (false) only this one will be configured,
    \param fConfigureConn - configure boh phones in connection
                            (must be supported by selected feature)

    \return IFX_SUCCESS if ok, otherwise IFX_ERROR
*/
IFX_return_t Common_ConfigureTAPI(CTRL_STATUS_t* pCtrl,
                                  BOARD_t* pBoard,
                                  PHONE_t* pPhone,
                                  IFX_int32_t nAction,
                                  IFX_int32_t nDataCh_FD,
                                  IFX_boolean_t fConfigureAll,
                                  IFX_boolean_t fConfigureConn)
{
   IFX_int32_t ret = IFX_SUCCESS;
   IFX_int32_t i = 0;
   CONNECTION_t* pConn = IFX_NULL;
   IFX_boolean_t bFeatureExist = IFX_FALSE;

    /* check input parameters */
   TD_PTR_CHECK(pCtrl, IFX_ERROR);
   TD_PTR_CHECK(pBoard, IFX_ERROR);
   TD_PTR_CHECK(pPhone, IFX_ERROR);

   TRACE(TAPIDEMO, DBG_LEVEL_LOW,
         ("Common_ConfigureTAPI() called with feature number %4d\n",
          nAction));

   if ((nAction >= 1000) && (nAction <= 1999))
   {
      /* feature exists */
      bFeatureExist = IFX_TRUE;
      /* Set codec */
      ret = Common_SetCodec(pBoard, pPhone, nAction,
                            fConfigureAll, fConfigureConn);
   } /* if nAction between 1000..1999 */
   else if ((nAction >= 2000) && (nAction <= 2099))
   {
      switch (nAction)
      {
         case 2000:
            {
               /* feature exists */
               bFeatureExist = IFX_TRUE;
               TRACE(TAPIDEMO, DBG_LEVEL_LOW,
                     ("Feature DTMF code: *%d - Set speech range.\n", nAction));
#if defined(EASY336) || defined(XT16)
               /* check if single connection and Voip ID is assigned */
               if ((pPhone->nVoipIDs <= 0) ||
                   (NO_CONFERENCE != pPhone->nConfIdx) ||
                   (0 >= pPhone->nConnCnt))
               {
                  TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                        ("Phone No %d: Configure NarrowBand can be only done "
                         "during single call(no conference), codec must be used. "
                         "(File: %s, line: %d)\n",
                         pPhone->nPhoneNumber, __FILE__, __LINE__));
                  return IFX_ERROR;
               }
#endif /* defined(EASY336) || defined(XT16) */
               /* Turn this channel to narrowband, coder is set to G.711 */
               ret = FEATURE_SetSpeechRange(pPhone, nDataCh_FD,
                                            IFX_TAPI_LINE_TYPE_FXS_NB,
                                            IFX_TAPI_COD_TYPE_MLAW, IFX_FALSE);

               if (ret != IFX_SUCCESS)
               {
                  TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                        ("Err to configure NarrowBand. "
                         "(File: %s, line: %d)\n", __FILE__, __LINE__));
                  return IFX_ERROR;
               }
               TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                     ("Switch one party to NARROWBAND\n"));

/* tone No 51 is defined for danube and duslic xt */
#if defined(DANUBE) || defined(DXT)
#ifdef TAPI_VERSION4
               if (pBoard->fSingleFD)
               {
                  Common_TonePlay(pCtrl, pPhone, 51);
               } /* if (pBoard->fSingleFD) */
               else
#endif /* TAPI_VERSION4 */
               {
                  ioctl(nDataCh_FD, IFX_TAPI_TONE_LOCAL_PLAY, 51);
               } /* if (pBoard->fSingleFD) */
#endif /* defined(DANUBE) || defined(DXT) */

               pPhone->fWideBand_Cfg = IFX_FALSE;
            }
            break;
         case 2001:
            {
               /* feature exists */
               bFeatureExist = IFX_TRUE;
               TRACE(TAPIDEMO, DBG_LEVEL_LOW,
                     ("Feature DTMF code: *%d - Set speech range.\n", nAction));
#if defined(EASY336) || defined(XT16)
               /* check if single connection and Voip ID is assigned */
               if ((pPhone->nVoipIDs <= 0) ||
                   (NO_CONFERENCE != pPhone->nConfIdx) ||
                   (0 >= pPhone->nConnCnt))
               {
                  TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                        ("Phone No %d: Configure WideBand can be only done "
                         "during call that uses codecs. "
                         "(File: %s, line: %d)\n",
                         pPhone->nPhoneNumber, __FILE__, __LINE__));
                  return IFX_ERROR;
               }
#endif /* defined(EASY336) || defined(XT16) */
               /* Turn this channel to wideband, coder is set to G.722 (64k)*/
               ret = FEATURE_SetSpeechRange(pPhone, nDataCh_FD,
                                            IFX_TAPI_LINE_TYPE_FXS_WB,
                                            IFX_TAPI_COD_TYPE_G722_64, IFX_TRUE);
               if (ret != IFX_SUCCESS)
               {
                  TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                        ("Err to configure WideBand. "
                         "(File: %s, line: %d)\n", __FILE__, __LINE__));
                  return IFX_ERROR;
               }
               TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                     ("Switch one party to WIDEBAND\n"));
/* tone No 50 is defined for danube and duslic xt */
#if defined(DANUBE) || defined(DXT)
#ifdef TAPI_VERSION4
               if (pBoard->fSingleFD)
               {
                  Common_TonePlay(pCtrl, pPhone, 50);
               } /* if (pBoard->fSingleFD) */
               else
#endif /* TAPI_VERSION4 */
               {
                  ioctl(nDataCh_FD, IFX_TAPI_TONE_LOCAL_PLAY, 50);
               } /* if (pBoard->fSingleFD) */
#endif /* defined(DANUBE) || defined(DXT) */

               pPhone->fWideBand_Cfg = IFX_TRUE;
            }
            break;
         case 2010:
            {
               /* feature exists */
               bFeatureExist = IFX_TRUE;
               TRACE(TAPIDEMO, DBG_LEVEL_LOW,
                     ("Feature DTMF code: *%d - Disable AGC.\n", nAction));
               /* check if single call */
               if ((NO_CONFERENCE == pPhone->nConfIdx) &&
                   (0 < pPhone->nConnCnt))
               {
                  pConn = &pPhone->rgoConn[0];
                  /* check if call uses codecs */
                  if ((EXTERN_VOIP_CALL == pConn->nType) ||
                      ((LOCAL_CALL == pConn->nType) &&
                       (pCtrl->pProgramArg->oArgFlags.nUseCodersForLocal)))
                  {
                     /* everything alright - enable AGC */
                  }
                  else
                  {
                     TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                           ("Phone No %d: Disabling Automatic Gain Control can "
                            "be only done during single call that uses codecs. "
                            "(File: %s, line: %d)\n",
                            pPhone->nPhoneNumber, __FILE__, __LINE__));
                        return IFX_ERROR;
                  }
               }
               else
               {
                  TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                        ("Phone No %d: Disabling Automatic Gain Control can "
                         "be only done during single call that uses codecs. "
                         "(File: %s, line: %d)\n",
                         pPhone->nPhoneNumber, __FILE__, __LINE__));
                     return IFX_ERROR;
               }

               /* Disable Automatic Gain Control (AGC) */
               ret = FEATURE_AGC_Enable(nDataCh_FD, pPhone, IFX_FALSE);

               if (ret != IFX_SUCCESS)
               {
                  TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                        ("Err with disable AGC. "
                         "(File: %s, line: %d)\n", __FILE__, __LINE__));
                  return IFX_ERROR;
               }

#ifndef TAPI_VERSION4
               TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
                     ("Automatic Gain Control (AGC) was disabled on data ch %d\n",
                      (int) pPhone->nDataCh));
#else/* TAPI_VERSION4 */
               TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
                     ("Automatic Gain Control (AGC) was disabled on phone "
                      "dev %d, channel %d\n",
                      pPhone->nDev,
                      pPhone->nPhoneCh));
#endif /* TAPI_VERSION4 */

               pPhone->fAGC_Cfg = IFX_FALSE;
            }
            break;
         case 2011:
            {
               /* feature exists */
               bFeatureExist = IFX_TRUE;
               TRACE(TAPIDEMO, DBG_LEVEL_LOW,
                     ("Feature DTMF code: *%d - Enable AGC.\n", nAction));
               /* check if single call */
               if ((NO_CONFERENCE == pPhone->nConfIdx) &&
                   (0 < pPhone->nConnCnt))
               {
                  pConn = &pPhone->rgoConn[0];
                  /* check if call uses codecs */
                  if ((EXTERN_VOIP_CALL == pConn->nType) ||
                      ((LOCAL_CALL == pConn->nType) &&
                       (pCtrl->pProgramArg->oArgFlags.nUseCodersForLocal)))
                  {
                     /* everything alright - enable AGC */
                  }
                  else
                  {
                     TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                           ("Phone No %d: Enabling Automatic Gain Control can "
                            "be only done during single call that uses codecs. "
                            "(File: %s, line: %d)\n",
                            pPhone->nPhoneNumber, __FILE__, __LINE__));
                        return IFX_ERROR;
                  }
               }
               else
               {
                  TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                        ("Phone No %d: Enabling Automatic Gain Control can "
                         "be only done during single call that uses codecs. "
                         "(File: %s, line: %d)\n",
                         pPhone->nPhoneNumber, __FILE__, __LINE__));
                     return IFX_ERROR;
               }
               /* Enable Automatic Gain Control (AGC) */
               ret = FEATURE_AGC_Cfg(nDataCh_FD, pPhone,
                                     -30,
                                     10,
                                     -20,
                                     -25);

               if (ret != IFX_SUCCESS)
               {
                  TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                        ("Err with configure AGC. "
                         "(File: %s, line: %d)\n", __FILE__, __LINE__));
                  return IFX_ERROR;
               }

               ret = FEATURE_AGC_Enable(nDataCh_FD, pPhone, IFX_TRUE);
               if (ret != IFX_SUCCESS)
               {
                  TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                        ("Err with enable AGC. "
                         "(File: %s, line: %d)\n", __FILE__, __LINE__));
                  return IFX_ERROR;
               }
#ifndef EASY336
               TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
                     ("Automatic Gain Control (AGC) was enabled on data ch %d\n",
                      (int) pPhone->nDataCh));
#else /* EASY336 */
               TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
                     ("Automatic Gain Control (AGC) was enabled on phone "
                      "dev %d, channel %d\n",
                      pPhone->nDev,
                      pPhone->nPhoneCh));
#endif /* EASY336 */

               pPhone->fAGC_Cfg = IFX_TRUE;
            }
            break;
         /** No voice activity detection. */
         case 2020:
         /** Voice activity detection on, in this case also comfort noise and
            spectral information (nicer noise) is switched on.*/
         case 2021:
         /** Voice activity detection on with comfort noise generation without
            spectral information. */
         case 2022:
         /** Voice activity detection on with comfort noise generation without
            silence compression */
         case 2023:
         /** Voice activity detection on with silence compression without
            comfort noise generation. */
         case 2024:
            /* feature exists */
            bFeatureExist = IFX_TRUE;
            TRACE(TAPIDEMO, DBG_LEVEL_LOW,
                  ("Feature DTMF code: *%d - Change VAD setting.\n",
                   nAction));
         case 2025:
         case 2026:
         case 2027:
         case 2028:
         case 2029:
            {
               ret = Common_SetVad(pPhone, pBoard, nAction % 10);
            }
            break;
         default:
            ret = IFX_ERROR;
            break;
      }
   } /* else if ((nAction >= 2000) && (nAction <= 2099)) */
   else if ((nAction >= 2100) && (nAction <= 2199))
   {
#ifdef ABACUS_TEST
      /*   actions needed by ABACUS test of call progress tone detection:
        2101 - stop playing tone
        2101 - play dialtone
        2102 - play busy tone
        2103 - play ringback tone
        if number of action changes also it should be changed in abacus scripts.
       */
#endif
      if (nAction == 2100)
      {
         /* feature exists */
         bFeatureExist = IFX_TRUE;
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("Feature DTMF code: *%d - Stop playing tone.\n",
                nAction));
         /* stop playing tone */
         ret = Common_ToneStop(pCtrl, pPhone);
         if (IFX_SUCCESS == ret)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
                  ("Phone No %d: Stop playing tone.\n",
                   (int) pPhone->nPhoneNumber));
         }
      }
      else if (nAction == 2101)
      {
         /* feature exists */
         bFeatureExist = IFX_TRUE;
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("Feature DTMF code: *%d - Play dialtone.\n",
                nAction));
         Common_ToneStop(pCtrl, pPhone);
         /* play dialtone */
         TD_CPTG_PLAY_CHECK(ret, pCtrl, pPhone, DIALTONE_IDX, "Dialtone");
      }
      else if (nAction == 2102)
      {
         /* feature exists */
         bFeatureExist = IFX_TRUE;
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("Feature DTMF code: *%d - Play busy tone.\n",
                nAction));
         Common_ToneStop(pCtrl, pPhone);
         /* play busy tone */
         TD_CPTG_PLAY_CHECK(ret, pCtrl, pPhone, BUSY_TONE_IDX, "Busy tone");
      }
      else if (nAction == 2103)
      {
         /* feature exists */
         bFeatureExist = IFX_TRUE;
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("Feature DTMF code: *%d - Play ringback tone.\n",
                nAction));
         Common_ToneStop(pCtrl, pPhone);
         /* play ringback tone */
         TD_CPTG_PLAY_CHECK(ret, pCtrl, pPhone, RINGBACK_TONE_IDX,
                            "Ringback tone");
      } /* if (nAction == 2101) */
   } /* else if ((nAction >= 2100) && (nAction <= 2199)) */
#ifdef ABACUS_TEST
   else if ((nAction >= 3000) && (nAction <= 3999))
   {
      if (nAction == 3110)
      {
         /* FAX/MODEM reset */
      }
      else if (nAction == 3111)
      {
         /* FAX/MODEM reset */
      }
   }
#endif /* ABACUS_TEST */
#ifdef ABACUS_TEST
   else if ((nAction >= 4000) && (nAction <= 4999))
   {
      if (nAction == 4001)
      {
         DEV_PrintConnections(pBoard);
      }
      else if (nAction == 4002)
      {
         DEV_PrintMappedDataCh(pBoard);
      }
      /* temporary code for changing country settings */
      else if (nAction == 4020)
      {

      }
      else if (nAction == 4100)
      {
         /* use default CPTs */
         pCtrl->nUseCustomCpt = IFX_FALSE;
      }
      else if (nAction == 4101)
      {
         /* use custom CPTs */
         pCtrl->nUseCustomCpt = IFX_TRUE;
      }
      else if (nAction == 4102)
      {
         COM_CPT_Set(pCtrl, IFX_NULL);
      }
      else if ( (nAction >= 4110) && (nAction <= 4119) )
      {
         i = nAction % 10;
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Play CPT %d using ioctl().\n", i));
         ioctl(pPhone->nDataCh_FD, IFX_TAPI_TONE_LOCAL_PLAY,
               (IFX_int32_t) (CPT_INDEX + i));
      }
      else if ( (nAction >= 4120) && (nAction <= 4129) )
      {
         i = nAction % 10;
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Play CPT %d using Common_CPT_Play.\n", i));
         Common_TonePlay(pCtrl, pPhone, i);
      }
      else if ( (nAction >= 4201) && (nAction <= 4399) )
      {
         i = nAction % 200;
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Play CPT %d using ioctl().\n", i));
         ioctl(pPhone->nDataCh_FD, IFX_TAPI_TONE_LOCAL_PLAY, i);
      }
      else
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("No feature defined for %d\n", nAction));

      }
   }
#endif /* ABACUS_TEST */
   else if ((nAction >= 8000) && (nAction <= 8999))
   {
      if ((nAction >= 8000) && (nAction <= 8010))
      {
         /* get new debug level */
         i = nAction%10;
         /* check if valid value */
         if ((i > DBG_LEVEL_OFF) || (i < DBG_LEVEL_LOW))
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("Wrong parameter of change debug level DTMF command "
                   "(File: %s, line: %d)\n", __FILE__, __LINE__));
         }
         else
         {
            /* feature exists */
            bFeatureExist = IFX_TRUE;
            TRACE(TAPIDEMO, DBG_LEVEL_LOW,
                  ("Feature DTMF code: *%d - Change Debug Trace Level.\n",
                   nAction));
            /* change trace level */
            SetTraceLevel(TAPIDEMO, i);
#ifdef EASY336
            SVIP_RM_SetTraceLevel(i);
#endif
         }

      }
      else if ((nAction >= 8100) && (nAction <= 8299))
      {
         /* feature exists */
         bFeatureExist = IFX_TRUE;
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("Feature DTMF code: *%d - %s JB statistics.\n",
                 nAction, (8200 > nAction) ? "Get" : "Reset"));
         i = nAction % 100;
         if (pBoard->nMaxCoderCh < i)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("There in no data channel %d. (File: %s, line: %d)\n",
                   i, __FILE__, __LINE__));
            return IFX_ERROR;
         }
         if (8200 > nAction)
         {
            Common_JBStatistic(i, pPhone->nDev, IFX_FALSE, IFX_TRUE, pBoard);
         }
         else
         {
            Common_JBStatistic(i, pPhone->nDev, IFX_TRUE, IFX_FALSE, pBoard);
         }
      }
   }
   else if ((nAction >= 9000) && (nAction <= 9999))
   {
#if defined(EASY336) || defined(XT16)
      /* check if single connection and Voip ID is assigned */
      if ((pPhone->nVoipIDs <= 0) ||
          (NO_CONFERENCE != pPhone->nConfIdx) ||
          (0 >= pPhone->nConnCnt))
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Phone No %d: Configure NarrowBand and WideBand can be only "
                "done during call that uses codecs. "
                "(File: %s, line: %d)\n",
                pPhone->nPhoneNumber, __FILE__, __LINE__));
         return IFX_ERROR;
      }
#endif /* defined(EASY336) || defined(XT16) */
      /* Go through all conections of this phone and send them
      to all the parties to set this feature also. */
      for (i = 0; i < pPhone->nConnCnt; i++)
      {
         if (i < MAX_PEERS_IN_CONF)
         {
            pConn = &pPhone->rgoConn[i];
            if ((pConn != IFX_NULL) &&
                (CALL_DIRECTION_TX == pPhone->nCallDirection))
            {
               pConn->nFeatureID = nAction;
               ret = TAPIDEMO_SetAction(pCtrl, pPhone, pConn,
                                        IE_CONFIG_PEER);
            }
         }
         else
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("Invalid number of connections - %d. "
                   "(File: %s, line: %d)\n", i, __FILE__, __LINE__));
            return IFX_ERROR;
         }
      }
      switch (nAction)
      {
         case 9000:
            {
               /* feature exists */
               bFeatureExist = IFX_TRUE;
               TRACE(TAPIDEMO, DBG_LEVEL_LOW,
                     ("Feature DTMF code: *%d - "
                      "Set speech range for both peers.\n", nAction));
               /* Turn this channel to narrowband, coder is set to G.711 */
               ret = FEATURE_SetSpeechRange(pPhone, nDataCh_FD,
                                            IFX_TAPI_LINE_TYPE_FXS_NB,
                                            IFX_TAPI_COD_TYPE_MLAW,
                                            IFX_FALSE);

               if (ret != IFX_SUCCESS)
               {
                  TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                        ("Err to configure NarrowBand. "
                         "(File: %s, line: %d)\n", __FILE__, __LINE__));
                  return IFX_ERROR;
               }
               TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                     ("Switch all parties to NARROWBAND\n"));

/* tone No 51 is defined for danube and duslic xt */
#if defined(DANUBE) || defined(DXT)
#ifdef TAPI_VERSION4
               if (pBoard->fSingleFD)
               {
                  Common_TonePlay(pCtrl, pPhone, 51);
               } /* if (pBoard->fSingleFD) */
               else
#endif /* TAPI_VERSION4 */
               {
                  ioctl(nDataCh_FD, IFX_TAPI_TONE_LOCAL_PLAY, 51);
               } /* if (pBoard->fSingleFD) */
#endif /* defined(DANUBE) || defined(DXT) */

               pPhone->fWideBand_Cfg = IFX_FALSE;
            }
            break;
         case 9001:
            {
               /* feature exists */
               bFeatureExist = IFX_TRUE;
               TRACE(TAPIDEMO, DBG_LEVEL_LOW,
                     ("Feature DTMF code: *%d - "
                      "Set speech range for both peers.\n", nAction));
               /* Turn this channel to wideband,
                  coder is set to G.722 (64k)*/
               ret = FEATURE_SetSpeechRange(pPhone, nDataCh_FD,
                                            IFX_TAPI_LINE_TYPE_FXS_WB,
                                            IFX_TAPI_COD_TYPE_G722_64,
                                            IFX_TRUE);

               if (ret != IFX_SUCCESS)
               {
                  TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                        ("Err to configure WideBand. "
                         "(File: %s, line: %d)\n", __FILE__, __LINE__));
                  return IFX_ERROR;
               }
               TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                     ("Switch all parties to WIDEBAND\n"));

/* tone No 50 is defined for danube and duslic xt */
#if defined(DANUBE) || defined(DXT)
#ifdef TAPI_VERSION4
               if (pBoard->fSingleFD)
               {
                  Common_TonePlay(pCtrl, pPhone, 50);
               } /* if (pBoard->fSingleFD) */
               else
#endif /* TAPI_VERSION4 */
               {
                  ioctl(nDataCh_FD, IFX_TAPI_TONE_LOCAL_PLAY, 50);
               } /* if (pBoard->fSingleFD) */
#endif /* defined(DANUBE) || defined(DXT) */

               pPhone->fWideBand_Cfg = IFX_TRUE;
            }
            break;
         case 9010:
         case 9011:
            /* feature exists */
            bFeatureExist = IFX_TRUE;
            TRACE(TAPIDEMO, DBG_LEVEL_LOW,
                  ("Feature DTMF code: *%d - "
                   "Set PCM speech range for both peers.\n", nAction));
            /* must be Duslic-XT */
            if (pPhone->pBoard->nType == TYPE_DUSLIC_XT)
            {
               /* must be local PCM call */
               if (pPhone->rgoConn[0].nType == LOCAL_PCM_CALL)
               {
                  if (9010 == nAction)
                  {
                     /* Set NarrowBand */
                      ret = FEATURE_SetPCM_SpeechRange(pPhone, IFX_FALSE);
                  } /* else if (9010 == nAction) */
                  else if (9011 == nAction)
                  {
                     /* Set WideBand */
                     ret = FEATURE_SetPCM_SpeechRange(pPhone, IFX_TRUE);
                  } /* else if (9011 == nAction) */
                  else
                  {
                     TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                           ("Common_ConfigureTAPI(): Err there is no option %d."
                            " (File: %s, line: %d)\n",
                            nAction, __FILE__, __LINE__));
                     return IFX_ERROR;
                  } /* else */
               } /* if local PCM call */
            } /* if Duslic-XT */
            break;
         default:
            ret = IFX_ERROR;
            break;
      }
   }
   else
   {
      ret = IFX_ERROR;
   }
   /* check if selected feature exists */
   if (IFX_FALSE == bFeatureExist)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
            ("No feature for DTMF code: *%d.\n", nAction));
   }
#ifdef ABACUS_TEST
   ret = COM_ExecuteDTMFCommand(pBoard, nAction, ret);
#endif /* ABACUS_TEST */

   return ret;
} /* Common_ConfigureTAPI() */

/**
   Make ip address from phone number.

   \param pCtrl   - handle to control structure
   \param pPhone  - pointer to PHONE
   \param pConn    - pointer to phone connection
   \param nIP_Addr - ip address
   \param fIsPCM_Num - IFX_TRUE we have phone number of PCM phone,
                       otherwise normal phone number.

   \return IFX_ERROR if an error, otherwise IFX_SUCCESS (no error occurred)

   \remark This address will be used to call the external phone.
*/
IFX_return_t Common_FillAddrByNumber(CTRL_STATUS_t* pCtrl, PHONE_t* pPhone,
                                     CONNECTION_t* pConn,
                                     IFX_uint32_t nIP_Addr,
                                     IFX_boolean_t fIsPCM_Num)
{
   IFX_int32_t ip = 0;
   struct in_addr tmp_ip_addr = {0};


   if ((IFX_NULL == pCtrl) || (IFX_NULL == pPhone) || (IFX_NULL == pConn))
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }

   /* Makes addres between network address (xxx.xxx.xxx.0)
      and local host address (0.0.0.xxx). In VxWorks we must remove last
      number, otherwise will not make address. */
   tmp_ip_addr = inet_makeaddr((pCtrl->nMy_IP_Addr >> 8), 0);

   /* Calling boards from different network */
   if (pCtrl->pProgramArg->oArgFlags.nDiffNet)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("Phone No %d: Calling board on different network.\n",
             pPhone->nPhoneNumber));
      tmp_ip_addr.s_addr = pCtrl->pProgramArg->oDiffNetIP.s_addr;
   }

   /* set ip */
   if (nIP_Addr <= 0xff)
   {
      /* set low part of IP address */
      ip = nIP_Addr;
   }
   else
   {
      /* set low part of IP address - only last bit is used */
      ip = 0x000000FF & nIP_Addr;
      /* set high part of IP address - IP address was specified
         with dialing sequence */
      tmp_ip_addr = inet_makeaddr((nIP_Addr >> 8), 0);
   }

   TRACE(TAPIDEMO, DBG_LEVEL_LOW,
         ("Phone No %d: IP low part = %u, netmask %s\n", 
          pPhone->nPhoneNumber, nIP_Addr, inet_ntoa(tmp_ip_addr)));


   /* Check if ip is valid */
   if ((0 > ip) || (255 < ip))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Invalid IP-number %d.(File: %s, line: %d)\n",
             (int)ip, __FILE__, __LINE__));
      return IFX_ERROR;
   }

   if (IFX_TRUE == fIsPCM_Num)
   {
      bzero((char*) &pConn->oConnPeer.oPCM.oToAddr,
            (int) sizeof(struct sockaddr_in));
      pConn->oConnPeer.oPCM.oToAddr.sin_family = AF_INET;
      pConn->oConnPeer.oPCM.oToAddr.sin_port = htons(ADMIN_UDP_PORT);

#if defined(VXWORKS)
          /* Only for VxWorks inet_makeaddr
              1st Parameter is Network Part, Ex: 10.1.1.0 should be used 
              0x000A0101
              2nd Parameter is Host Part, depend on how many bits have been used
                 by Network Part
              In Our Case, we always regard the mask is 255.255.255.0, so
                 shift right 8 bits
          */
          pConn->oConnPeer.oPCM.oToAddr.sin_addr =
              inet_makeaddr((ntohl(tmp_ip_addr.s_addr) >> 8), ip);
#else /* VXWORKS */
          pConn->oConnPeer.oPCM.oToAddr.sin_addr =
              inet_makeaddr((ntohl(tmp_ip_addr.s_addr) >> 8), ip);
#endif /* VXWORKS */

      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("Phone No %d: Peer address %s:%d\n",
             pPhone->nPhoneNumber,
             inet_ntoa(pConn->oConnPeer.oPCM.oToAddr.sin_addr),
             ntohs((int)pConn->oConnPeer.oPCM.oToAddr.sin_port)));
   }
   else
   {
      bzero((char*) &pConn->oConnPeer.oRemote.oToAddr,
            (int) sizeof(struct sockaddr_in));
      pConn->oConnPeer.oRemote.oToAddr.sin_family = AF_INET;

#if defined(VXWORKS)
          /* Only for VxWorks inet_makeaddr
              1st Parameter is Network Part, Ex: 10.1.1.0 should be used 
              0x000A0101
              2nd Parameter is Host Part, depend on how mane bits have been used
                 by Network Part
              In Our Case, we always regard the mask is 255.255.255.0, so
                 shift right 8 bits
          */
          pConn->oConnPeer.oRemote.oToAddr.sin_addr =
              inet_makeaddr((ntohl(tmp_ip_addr.s_addr) >> 8), ip);
#else /* VXWORKS */
          pConn->oConnPeer.oRemote.oToAddr.sin_addr =
              inet_makeaddr((ntohl(tmp_ip_addr.s_addr) >> 8), ip);
#endif /* VXWORKS */

      if (pCtrl->pProgramArg->oArgFlags.nQos)
      {
         /* QoS support */
         QOS_InitializePairStruct(pCtrl, pConn, pPhone->pBoard->nBoard_IDX);
      }
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("Phone No %d: Peer address = %s\n", pPhone->nPhoneNumber,
             inet_ntoa(pConn->oConnPeer.oRemote.oToAddr.sin_addr)));
   }

   return IFX_SUCCESS;
} /* Common_FillAddrByNumber() */


/**
   Converts digits to number, changes digits to chars

   \param prgnDialNum - pointer to array of digits
   \param nFirstDigit - first digit to make number from
   \param nDialNrCnt  - number of digits

   \return Dialed number (if 0 error)
*/
IFX_int32_t Common_DigitsToNum(IFX_char_t* prgnDialNum, IFX_int32_t nFirstDigit,
                               IFX_int32_t nDialNrCnt)
{
   const IFX_int32_t digit2number[32] = {
       -1,  1,  2,  3,  4,  5,  6,  7,  8,  9, -1,  0, -1, -1, -1, -1,
       -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 };

   const IFX_char_t digit2char[32] = {
      ' ', '1', '2', '3', '4', '5', '6', '7',
      '8', '9', '*', '0', '#', ' ', ' ', ' ',
      ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ',
      ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '};

   IFX_int32_t i = 0;
   IFX_int32_t dialed_number = 0;

   /* if converting more than 9 chars to number then number
      could be too big for IFX_int32_t */
   if ((IFX_NULL == prgnDialNum) || (0 > nDialNrCnt) ||
       (9 < (nDialNrCnt - nFirstDigit)))
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return dialed_number;
   }

   for (i = 0; i < nDialNrCnt; i++)
   {
      if ((0 > (IFX_int8_t) prgnDialNum[i])
          || (32 < (IFX_int8_t) prgnDialNum[i]))
      {
         /* Error, our digit outside conversion table */
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Digit %d outside transform table. (File: %s, line: %d)\n",
               prgnDialNum[i], __FILE__, __LINE__));
         return dialed_number;
      }
      /* check if first digit of number occured */
      if (i >= nFirstDigit)
      {
         /* get number */
         dialed_number = dialed_number * 10 +
            digit2number[(int) prgnDialNum[i]];
      }
      /* change digit number to char */
      prgnDialNum[i] = digit2char[(int) prgnDialNum[i]];
   }

   prgnDialNum[nDialNrCnt] = 0;
   return dialed_number;
} /* Common_DigitsToNum() */

/**
   Check dialed number for IP address if dialing started with "0*" sequence.

   \param prgnDialNum - array of dialed digits
   \param nDialNrCnt  - number of digits
   \param nIpAddr     - ip address, used with sockets

   \return UNKNOWN_CALL_TYPE if wrong number or EXTERN_VOIP_CALL if dialing
           is completed.
*/
CALL_TYPE_t Common_CheckDialedNumForFull_IP(IFX_char_t* prgnDialNum,
                                            IFX_int32_t nDialNrCnt,
                                            IFX_uint32_t* nIpAddr)
{
   IFX_int32_t i, j = 0, nNumber;
   IFX_int32_t prgnStarLocation[NUM_OF_STARS_IN_IP];
   IFX_int32_t prgnIP_address[NUM_OF_STARS_IN_IP];

   /* array to convert dialed digits to numbers */
   const IFX_int32_t digit2number[32] = {
       -1,  1,  2,  3,  4,  5,  6,  7,  8,  9, -1,  0, -1, -1, -1, -1,
       -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 };

   /* check all dialed digits */
   for (i=0; i<nDialNrCnt; i++)
   {
      /* check if dialed digit is a star digit - '*' */
      if (DIGIT_STAR == prgnDialNum[i])
      {
         /* check if not too many star digits found */
         if (NUM_OF_STARS_IN_IP > j)
         {
            /* save star digit location */
            prgnStarLocation[j] = i;
         }
         /* increas star digit count */
         j++;
      }
   }
   /* clear table */
   memset(prgnIP_address, 0, NUM_OF_STARS_IN_IP * sizeof(IFX_int32_t));
   /* check if right amount of star digits was found and if phone number has
      right length */
   if ((NUM_OF_STARS_IN_IP == j) &&
       ((prgnStarLocation[NUM_OF_STARS_IN_IP - 1] + 1) ==
        nDialNrCnt))
   {
      /* set star digit counter */
      j = 0;
      for (i=2; i<nDialNrCnt; i++)
      {
         if (32 <= (IFX_int32_t) prgnDialNum[i])
         {
            TRACE (TAPIDEMO, DBG_LEVEL_HIGH,
                   ("Invalid digit number %d. (File: %s, line: %d)\n",
                    i, __FILE__, __LINE__));
            return UNKNOWN_CALL_TYPE;
         }
         /* get number of one dialed digit */
         nNumber = digit2number[(IFX_int32_t)prgnDialNum[i]];
         /* TAPIDEMO_DigitsToNum() should return -1 for star digit */
         if (-1 == nNumber)
         {
            /* check if star digit was detected in this loaction */
            if (NUM_OF_STARS_IN_IP > (j+1))
            {
               if ((j >= NUM_OF_STARS_IN_IP) ||
                   (i != prgnStarLocation[j+1]))
               {
                  /* change dialed digits table to printable string */
                  Common_DigitsToNum(prgnDialNum, nDialNrCnt, nDialNrCnt);
                  TRACE (TAPIDEMO, DBG_LEVEL_HIGH,
                         ("Dialed number \"%s\" is invalid."
                          " (File: %s, line: %d)\n",
                          prgnDialNum, __FILE__, __LINE__));
                  return UNKNOWN_CALL_TYPE;
               }
               /* increase star digit counter */
               j++;
            }
            else
            {
               TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                     ("Buffer overflow, array index of 'prgnStarLocation' "
                      "may be outside the bounds. (File: %s, line: %d)\n",
                      __FILE__, __LINE__));
            }
         } /* if (-1 == nNumber) */
         else
         {
            /* get IP address, last element of table is phone number */
            prgnIP_address[j] = prgnIP_address[j] * 10 + nNumber;
         } /* if (-1 == nNumber) */
      } /* for (i=2; i<nDialNrCnt; i++) */
   } /* if valid number */
   else
   {
      return UNKNOWN_CALL_TYPE;
   }
   /* reset IP address */
   *nIpAddr = 0;
   /* get IP address */
   for (j=0; j<4; j++)
   {
      /* check if valid part of IP address */
      if (0xFF < prgnIP_address[j])
      {
         TRACE (TAPIDEMO, DBG_LEVEL_HIGH,
                ("Dialed number \"%s\" is invalid - wrong IP address!"
                 " (File: %s, line: %d)\n",
                 prgnDialNum, __FILE__, __LINE__));
         return UNKNOWN_CALL_TYPE;
      }
      /* add next part of ip address */
      *nIpAddr = *nIpAddr * 0x100 + prgnIP_address[j];
   }
   /* change dialed digits table to printable string */
   Common_DigitsToNum(prgnDialNum, nDialNrCnt, nDialNrCnt);

   /* making extern voip call */
   return EXTERN_VOIP_CALL;
} /* Common_CheckDialedNumForFull_IP() */

/**
   Check dialed number.

   \param pPhone        - pointer to phone
   \param nDialedNum    - dialed number
   \param nIpAddr       - ip address, used with sockets
   \param pCtrl         - handle to control status structure
   \param pConn         - pointer to connection structure

   \return UNKNOWN_CALL_TYPE if wrong number or call type and dialed number
           and also channel number if phone is called.
   \remarks nDevNum is used only for TAPI_VERSION4
*/
CALL_TYPE_t Common_CheckDialedNum(PHONE_t* pPhone,
                                  IFX_int32_t* nDialedNum,
                                  IFX_uint32_t* nIpAddr,
                                  CTRL_STATUS_t* pCtrl)
{
   if ((IFX_NULL == pPhone) || (0 > pPhone->nDialNrCnt) ||
       (IFX_NULL == nDialedNum) || (IFX_NULL == nIpAddr))
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return UNKNOWN_CALL_TYPE;
   }

#if (!defined(EASY336) && !defined(XT16))
   /* 0x0b = '0' => DIGIT_ZERO */
   if ((1 <= pPhone->nDialNrCnt) && (2 >= pPhone->nDialNrCnt) &&
       (DIGIT_ZERO != pPhone->pDialedNum[0]) &&
       (DIGIT_STAR != pPhone->pDialedNum[0]))
   {
      /* Number <xx>; but first must not be 0. Phone channel counting
         from 0, but we count from 1. */
      /* If the first digit is a star - DTMF command is requested*/
      *nDialedNum = Common_DigitsToNum(pPhone->pDialedNum, 0,
                                       pPhone->nDialNrCnt);
      if (pCtrl->nSumAnalogCh < *nDialedNum)
      {
         /* Calling unknown phone channel */
#ifdef ABACUS_TEST
         /* If it's not a test of digits... */
         if (!((pCtrl->pProgramArg->nTestType == COM_MODULAR_TEST) &&
               (pCtrl->pProgramArg->nSubTestType &
                (COM_SUBTEST_PULSEDIGIT | COM_SUBTEST_DTMFDIGIT))))
         {
            /* ...raport error */
#endif /* ABACUS_TEST */
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("Invalid number entered (%d),"
                   " phone chanel does not exists.\n",
                   (int) *nDialedNum));
#ifdef ABACUS_TEST
         }
#endif /* ABACUS_TEST */
         return UNKNOWN_CALL_TYPE;
      }
      else
      {
         return LOCAL_CALL;
      }
   }
#else /* (!defined(EASY336) && !defined(XT16)) */
   if ((pPhone->nDialNrCnt == 3) &&
       (DIGIT_ZERO != pPhone->pDialedNum[0]) &&
       (DIGIT_STAR != pPhone->pDialedNum[0]))
   {
      /* Number <xxx>; but first must not be 0. Device and phone channel 
       * counting from 0, but we count from 1. */
      /* If the first digit is a star - DTMF command is requested*/
      *nDialedNum = Common_DigitsToNum(pPhone->pDialedNum, 0,
                                       pPhone->nDialNrCnt);

#ifdef ABACUS_TEST
      /* If it's a test of digits... */
      if ((pCtrl->pProgramArg->nTestType == COM_MODULAR_TEST) &&
          (pCtrl->pProgramArg->nSubTestType &
           (COM_SUBTEST_PULSEDIGIT | COM_SUBTEST_DTMFDIGIT)))
      {
         return UNKNOWN_CALL_TYPE;
      }
#endif /* ABACUS_TEST */

      return LOCAL_CALL;
   }
#endif /* (!defined(EASY336) && !defined(XT16)) */

   if ((4 == pPhone->nDialNrCnt) &&
       (DIGIT_ZERO == pPhone->pDialedNum[0]) &&
       ( (DIGIT_ZERO == pPhone->pDialedNum[1]) ||
         (DIGIT_ONE == pPhone->pDialedNum[1]) ||
         (DIGIT_TWO == pPhone->pDialedNum[1]) ))
   {
      *nDialedNum = Common_DigitsToNum(pPhone->pDialedNum, 1,
                                       pPhone->nDialNrCnt);
      *nIpAddr = *nDialedNum;

      return EXTERN_VOIP_CALL;
   }

   /* EXTERN_VOIP_CALL with whole ip address */
   if ((DIGIT_ZERO == pPhone->pDialedNum[0]) &&
       (DIGIT_STAR == pPhone->pDialedNum[1]))
   {
      /* Dial 0*<aaa>*<bbb>*<ccc>*<ddd>*xxx  to
         call ip address aaa.bbb.ccc.ddd, phone number xxx */
      if (EXTERN_VOIP_CALL == Common_CheckDialedNumForFull_IP(
         pPhone->pDialedNum, pPhone->nDialNrCnt, nIpAddr))
      {
         return EXTERN_VOIP_CALL;
      }
   }

#if (!defined(EASY336) && !defined(XT16))
   if ((6 <= pPhone->nDialNrCnt) && (7 >= pPhone->nDialNrCnt) &&
       (DIGIT_ZERO == pPhone->pDialedNum[0]) &&
       (DIGIT_NINE == pPhone->pDialedNum[1]))
   {
      if ((pCtrl->pProgramArg->oArgFlags.nPCM_Master) ||
          (pCtrl->pProgramArg->oArgFlags.nPCM_Slave))
      {
         /* Number 09<xxx><yy>; xxx - 0..255 (ip addr), yy - pcm channel.
            Phone channel counting from 0, but we type from 1. */
         *nDialedNum = Common_DigitsToNum(pPhone->pDialedNum, 2,
                                          pPhone->nDialNrCnt);
         *nIpAddr = *nDialedNum / 10;
         /* Subract 1, because if calling channel 0, 1 was typed */

         if (7 == pPhone->nDialNrCnt)
         {
            /* Two digits */
            *nDialedNum = (*nDialedNum) % 100;
         }
         else
         {
            /* One digit */
            *nDialedNum = (*nDialedNum) % 10;
         }
         return PCM_CALL;
      }
      else
      {
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("Phone No %d, %s: Unable to start PCM call.\n",
                pPhone->nPhoneNumber, pPhone->pBoard->pszBoardName));
      }
   }

   if ((3 <= pPhone->nDialNrCnt) &&
       (DIGIT_ZERO == pPhone->pDialedNum[0]) &&
       (DIGIT_FIVE == pPhone->pDialedNum[1]))
   {
      /* FXO call between \todo some mapping table between FXO and phones
         at start. */
      /* Number 05<xxx>; xxx - 0..999 - channel on other board. */
      *nDialedNum = Common_DigitsToNum(pPhone->pDialedNum, 2,
                                       pPhone->nDialNrCnt);

      return FXO_CALL;
   }
#else /* (!defined(EASY336) && !defined(XT16)) */
   if ((pPhone->nDialNrCnt == 8) &&
       (DIGIT_ZERO == pPhone->pDialedNum[0]) &&
       (DIGIT_NINE == pPhone->pDialedNum[1]))
   {
      if ((pCtrl->pProgramArg->oArgFlags.nPCM_Master) ||
          (pCtrl->pProgramArg->oArgFlags.nPCM_Slave))
      {
        /* Number 09<xxx><yzz>; xxx - 0..255 (ip addr), y - device number,
            zz - phone channel number. Device and
            phone channel counting from 0, but we type from 1.
            Example.
            09<112><104>
            IP = xxx.xxx.xxx.112
            Device number = 0
            Channel number = 3 */
         *nDialedNum = Common_DigitsToNum(pPhone->pDialedNum, 2,
                                          pPhone->nDialNrCnt);

         return PCM_CALL;
      }
   }
#endif /* (!defined(EASY336) && !defined(XT16)) */

   if ((5 == pPhone->nDialNrCnt) &&
       (DIGIT_STAR == pPhone->pDialedNum[0]))
   {
       /* If the first digit is a star and 4 digits were pressed
          - DTMF command is requested*/
       *nDialedNum = Common_DigitsToNum(pPhone->pDialedNum, 1,
                                        pPhone->nDialNrCnt);
       return DTMF_COMMAND;
   }
   return UNKNOWN_CALL_TYPE;
} /* Common_CheckDialedNum() */

/**
   Print call data, like start and end of connection and call type.

   \param current_state   - current state of phone
   \param new_event  - new internal event of phone
   \param pPhone    - pointer to phone structure
   \param pConn    - pointer to phone connection
   \param nUseCodersForLocal - IFX_TRUE if codecs are used for local connection

   \return none
*/
IFX_void_t Common_PrintCallProgress(STATE_MACHINE_STATES_t current_state,
                                    INTERNAL_EVENTS_t new_event,
                                    CALL_TYPE_t nOldCallType,
                                    PHONE_t* pPhone, CONNECTION_t* pConn,
                                    IFX_int32_t nUseCodersForLocal)
{
   IFX_char_t call_string[PCP_MAX_STR_LEN];
   IFX_char_t phone_string[PCP_MAX_STR_LEN];

   CALL_TYPE_t nCallType;
   IFX_return_t ret = IFX_SUCCESS;

   /* set action according to state transition */
   switch (current_state)
   {
      case S_IN_RINGING:
         if (IE_HOOKOFF == new_event)
         {
            strcpy(call_string,
                   "Phone No %d: Established %s with calling %s\n");
            break;
         }
         ret = IFX_ERROR;
         break;
      case S_ACTIVE:
         if (IE_END_CONNECTION == new_event ||
             IE_HOOKON == new_event)
         {
            strcpy(call_string,
                   "Phone No %d: Disconnected %s with %s\n");
            break;
         }
         ret = IFX_ERROR;
      case S_OUT_RINGBACK:
         if (IE_ACTIVE == new_event)
         {
            strcpy(call_string,
                   "Phone No %d: Established %s with called %s\n");
            break;
         }
         ret = IFX_ERROR;
         break;
      default :
         ret = IFX_ERROR;
         break;
   }

   /* if none of supported state transitions then leave function */
   if (IFX_ERROR == ret)
   {
      return;
   }
   /* check if call type is set */
   if ((UNKNOWN_CALL_TYPE == pConn->nType) && (pConn->nType != nOldCallType))
   {
      /* if pConn->nType was resetted on connection end,
         use remembered call type */
      nCallType = nOldCallType;
   }
   else
   {
      nCallType = pConn->nType;
   }

   /* set phones data depending on call type */
   switch (nCallType)
   {
      case EXTERN_VOIP_CALL:
         sprintf(phone_string, "PhoneNo %d\n%son %s",
                 pConn->oConnPeer.nPhoneNum,
                 pPhone->pBoard->pIndentPhone,
                 inet_ntoa(pConn->oConnPeer.oRemote.oToAddr.sin_addr)
                 );
         break;
      case LOCAL_CALL:
      case LOCAL_BOARD_CALL:
         if (IFX_FALSE == nUseCodersForLocal)
         {
            sprintf(phone_string, "PhoneNo %d, PhoneCh %d",
                    pConn->oConnPeer.nPhoneNum,
                    pConn->oConnPeer.oLocal.pPhone->nPhoneCh
                    );
         }
         else
         {
            sprintf(phone_string, "PhoneNo %d, DataCh %d",
                    pConn->oConnPeer.nPhoneNum,
                    pConn->oConnPeer.oLocal.pPhone->nDataCh
                    );
         }
         break;
      case PCM_CALL:
         sprintf(phone_string, "PhoneNo %d, on %s",
                 pConn->oConnPeer.nPhoneNum,
                 inet_ntoa(pConn->oConnPeer.oPCM.oToAddr.sin_addr)
                 );
         break;
      case LOCAL_PCM_CALL:
      case LOCAL_BOARD_PCM_CALL:
         sprintf(phone_string, "PhoneNo %d DataCh %d",
                 pConn->oConnPeer.nPhoneNum,
                 pConn->oConnPeer.oLocal.pPhone->nDataCh
                 );
         break;
      default:
         ret = IFX_ERROR;
         break;
   }

   /* if none of supported call types then leave function */
   if (IFX_ERROR == ret)
   {
      return;
   }
   /* print connection call progress data */
   TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
         (call_string, pPhone->nPhoneNumber, 
          Common_Enum2Name(nCallType, TD_rgCallTypeName),
          phone_string));
   return;

} /* Common_PrintCallProgres() */

/**
   Print encoder type and packets length.

   \param pCtrl - pointer to cotrol structure
   \param pPhone - pointer to phone structure
   \param pConn - pointer to connection structure

   \return none
*/
IFX_void_t Common_PrintEncoderType(CTRL_STATUS_t* pCtrl, PHONE_t* pPhone,
                                   CONNECTION_t* pConn)
{

   IFX_TAPI_ENC_TYPE_t nCodecType = IFX_TAPI_ENC_TYPE_MAX;
   IFX_int32_t nFrameLen = 0;
#ifdef TAPI_VERSION4
#ifdef EASY336
   VOIP_DATA_CH_t *pOurCodec;
#endif /* EASY336 */
#endif /* TAPI_VERSION4 */
   /* check input */
   if ((IFX_NULL == pCtrl) || (IFX_NULL == pPhone) || (IFX_NULL ==  pConn))
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return;
   }

#ifdef TAPI_VERSION4
   if (pPhone->pBoard->fSingleFD)
   {
#ifdef EASY336
      /* get codec data structure */
      pOurCodec = Common_GetCodec(pCtrl, pConn->voipID);
      if (pOurCodec != IFX_NULL)
      {
         /* set codec type and packets length */
         nCodecType = pOurCodec->nCodecType;
         nFrameLen = pOurCodec->nFrameLen;
      }
      else
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Failed to get codec. "
                "(File: %s, line: %d)\n",
                __FILE__, __LINE__));
      }
#endif /* EASY336 */
   }
   else
#endif /* TAPI_VERSION4 */
   {
      /* set codec type and packets length */
      nCodecType = pPhone->pBoard->pDataChStat[pConn->nUsedCh].nCodecType;
      nFrameLen = pPhone->pBoard->pDataChStat[pConn->nUsedCh].nFrameLen;
   } /* if (pPhone->pBoard->fSingleFD) */

   /* check if codec data obtained */
   if (IFX_TAPI_ENC_TYPE_MAX > nCodecType)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
            ("Phone No %d: Voice encoding: %s, %s\n",
             pPhone->nPhoneNumber, pCodecName[nCodecType],
             oFrameLen[nFrameLen]));
   }
   else
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Failed to get codec name. (File: %s, line: %d)\n",
             __FILE__, __LINE__));
   }

   return;
} /* Common_PrintEncoderType() */

/**
   Function to show and reset jitter buffer statistics.

   \param nDataCh - data channel number
   \param nDev    - device number
   \param fReset  - if not IFX_FALSE function resets JB statistics
   \param fShow   - if not IFX_FALSE function prints JB statistics
   \param pBoard  - pointer to board structure

   \return IFX_SUCCESS if successful, otherwise IFX_ERROR.
   \remarks nDev is used only for TAPI_VERSION4
*/
IFX_return_t Common_JBStatistic(IFX_int32_t nDataCh,
                                IFX_int32_t nDev,
                                IFX_boolean_t fReset,
                                IFX_boolean_t fShow,
                                BOARD_t* pBoard)
{
   IFX_int32_t ret;
   IFX_int32_t fd;
   IFX_TAPI_JB_STATISTICS_t jbStatistic;
#ifdef TAPI_VERSION4
   IFX_TAPI_JB_STATISTICS_RESET_t jbStatisticReset;
#else
   PHONE_t* pPhone;
   CONNECTION_t* pConn;
#endif /* TAPI_VERSION4 */
   IFX_char_t aDataChannelInfo[TD_MAX_NAME];
   IFX_int32_t param = 0;

   fd = Common_GetDeviceOfCh(pBoard, nDataCh);
   if (IFX_ERROR == fd)
   {
      /* Couldn't get file descriptor */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Couldn't get file descriptor of data channel %d."
            "(File: %s, line: %d)\n",
            nDataCh, __FILE__, __LINE__));
      return IFX_ERROR;
   }
#ifndef TAPI_VERSION4
   pPhone = ABSTRACT_GetPHONE_OfDataCh(nDataCh, &pConn, pBoard);
   if (IFX_NULL == pPhone)
   {
      /* Couldn't get phone of channel */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Data channel %d is not mapped to phone channel. "
            "(File: %s, line: %d)\n",
            nDataCh, __FILE__, __LINE__));
      return IFX_ERROR;
   }
#endif /* TAPI_VERSION4 */
   /* get info string */
#ifndef TAPI_VERSION4
   snprintf(aDataChannelInfo, TD_MAX_NAME, "Phone No %d, data channel %d",
            pPhone->nPhoneNumber, nDataCh);
#else
   snprintf(aDataChannelInfo, TD_MAX_NAME, "data channel %d, device %d",
            nDataCh, nDev);
#endif /* TAPI_VERSION4 */
   if (IFX_FALSE != fShow)
   {

      memset(&jbStatistic, 0, sizeof(IFX_TAPI_JB_STATISTICS_t));
#ifdef TAPI_VERSION4
      /* set device and channel */
      jbStatistic.ch = nDataCh;
      jbStatistic.dev = nDev;
#endif /* TAPI_VERSION4 */
      ret = ioctl(fd, IFX_TAPI_JB_STATISTICS_GET, (IFX_int32_t)&jbStatistic);

      if (IFX_SUCCESS == ret)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("Jitter buffer statistics for %s:\n",
                aDataChannelInfo));

         /* jitter buffer type*/
         switch (jbStatistic.nType)
         {
            case 1:
               TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
                     ("  - jitter buffer type fixed,\n"));
               break;
            case 2:
               TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
                     ("  - jitter buffer type adaptive,\n"));
               break;
            default:
               TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
                     ("  - jitter buffer type unknown,\n"));
               break;
         }
         /* current jitter buffer size */
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("  - current jitter buffer size 0x%hx(%uus),\n",
                jbStatistic.nBufSize,
                (jbStatistic.nBufSize * 125)));

         /* minimum and maximum estimated buffer size */
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("  - estimated jitter buffer size minimum 0x%hx(%uus),\n"
                "    maximum 0x%hx(%uus),\n",
                jbStatistic.nMinBufSize,
                (jbStatistic.nMinBufSize * 125),
                jbStatistic.nMaxBufSize,
                (jbStatistic.nMaxBufSize * 125)));

         /* playout delay */
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("  - playout delay %hu,\n",
                jbStatistic.nPODelay));

         /* minimum and maximum playout delay */
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("  - playout delay minimum %hu, maximum %hu,\n",
                jbStatistic.nMinDelay,
                jbStatistic.nMaxPODelay));

         /* received packet number */
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("  - received packet number %u,\n",
                jbStatistic.nPackets));

         /*  invalid packet number */
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("  - invalid packet number %hu,\n",
                jbStatistic.nInvalid));

         /*  late packets number */
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("  - late packets number %hu,\n",
                jbStatistic.nLate));

         /*  early packets number */
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("  - early packet number %hu,\n",
                jbStatistic.nEarly));

         /*  resynchronizations number */
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("  - resynchronizations number %hu,\n",
                jbStatistic.nResync));

         /* nIsUnderflow - total number of injected samples since the beginning
            of the connection or since the last statistic reset
            due to jitter buffer underflows.*/

         /* nIsNoUnderflow total number of injected samples since the beginning
            of the connection or since the last statistic reset in case
             of normal jitter buffer operation,
            which means when there is not a jitter buffer underflow. */

         /* nIsIncrement total number of injected samples since the beginning
            of the connection or
            since the last statistic reset in case of jitter buffer 
            increments.*/
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("  - nIsUnderflow %u, nIsNoUnderflow %u, nIsIncrement %u\n",
                jbStatistic.nIsUnderflow,
                jbStatistic.nIsNoUnderflow,
                jbStatistic.nIsIncrement));

         /* nSkDecrement total number of skipped lost samples since
            the beginning of the connection or since the last statistic
            reset in case of jitter buffer decrements. */

         /* nDsDecrement total number of dropped samples since the beginning
            of the connection or since the last statistic reset in case
            of jitter buffer decrements.*/

         /* nDsOverflow total number of dropped samples since the beginning
            of the connection  or since the last statistic reset in case
            of jitter buffer overflows.*/
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("  - nSkDecrement %u, nDsDecrement %u, nDsOverflow %u\n",
                jbStatistic.nSkDecrement,
                jbStatistic.nDsDecrement,
                jbStatistic.nDsOverflow));

         /* nSid total number of comfort noise samples since the beginning
            of the connection or since the last statistic reset. */
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("  - total number of comfort noise samples %u,\n",
                jbStatistic.nSid));

         /* nRecBytesH number of received bytes high part
            including event packets. */
         /* nRecBytesL number of received bytes low part
            including event packets. */
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("  - number of received bytes including event packets:\n"
                "    high part %u, low part %u.\n",
                jbStatistic.nRecBytesH,
                jbStatistic.nRecBytesL));

      }
      else
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Failed to get jitter buffer statistics for %s.\n",
                aDataChannelInfo));
         return IFX_ERROR;
      }
   }
   if (IFX_FALSE != fReset)
   {
#ifdef TAPI_VERSION4
      param = (IFX_int32_t)&jbStatisticReset;
      jbStatisticReset.ch = nDataCh;
      jbStatisticReset.dev = nDev;
#endif

      ret = ioctl(fd, IFX_TAPI_JB_STATISTICS_RESET, param);
      if (IFX_ERROR == ret)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Failed to reset jitter buffer statistics %s.\n",
                aDataChannelInfo));
         return IFX_ERROR;
      }
      else
      {
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("Jitter buffer statistics for %s are reseted.\n",
                aDataChannelInfo));
      } /* if (IFX_ERROR == ret) */

   } /* if (IFX_FALSE != fReset) */

   return IFX_SUCCESS;
} /* Common_JBStatistic() */

/**
   Function to show and reset RTCP statistics

   \param fd      - data channel file descriptor
   \param fReset  - if not IFX_FALSE function resets RTCP statistics
   \param fShow   - if not IFX_FALSE function prints RTCP statistics

   \return IFX_SUCCESS if successful, otherwise IFX_ERROR.
*/
IFX_return_t Common_RTCPStatistic(IFX_int32_t fd,
                                  IFX_boolean_t fReset,
                                  IFX_boolean_t fShow)
{
   IFX_int32_t ret;
   IFX_TAPI_PKT_RTCP_STATISTICS_t rtcpStatistic;

   if (IFX_FALSE != fShow)
   {

      memset(&rtcpStatistic, 0, sizeof(IFX_TAPI_PKT_RTCP_STATISTICS_t));

      ret = ioctl(fd, IFX_TAPI_PKT_RTCP_STATISTICS_GET,
                  (IFX_int32_t)&rtcpStatistic);

      if (IFX_SUCCESS == ret)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL, ("RTCP statistic for fd %d:\n",fd));


         /* Sender generating this report. */
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
                     ("  - sender generating this report %u,\n",
                      rtcpStatistic.ssrc));

         /* RTP time stamp. */
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
                     ("  - RTP time stamp %u,\n",
                      rtcpStatistic.rtp_ts));

         /* Sent packet count. */
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
                     ("  - sent packet count %u,\n",
                      rtcpStatistic.psent));

         /* Sent octets count. */
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
                     ("  - sent octets count %u,\n",
                      rtcpStatistic.osent));

         /* Data source. */
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
                     ("  - data source %u,\n",
                      rtcpStatistic.rssrc));

         /* Receivers fraction loss. */
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
                     ("  - receivers fraction loss %u,\n",
                      rtcpStatistic.fraction));

         /* Receivers packet loss. */
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
                     ("  - receivers packet loss %d,\n",
                      rtcpStatistic.lost));
         /*TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
                     ("  - receivers packet loss %d,\n",
                      ((rtcpStatistic.lost&&0x00FFFFFF)));*/

         /* Extended last seq nr. received.*/
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
                     ("  - extended last seq nr. received %u,\n",
                      rtcpStatistic.last_seq));

         /* Receives interarrival jitter. */
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
                     ("  - receives interarrival jitter %u.\n",
                      rtcpStatistic.jitter));

      }
      else
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Failed to get RTCP statistics.\n"));
         return IFX_ERROR;
      }
   }
   if (IFX_FALSE != fReset)
   {

      ret = ioctl(fd, IFX_TAPI_PKT_RTCP_STATISTICS_RESET, 0);
      if (IFX_ERROR == ret)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Failed to reset RTCP statistics.\n"));
         return IFX_ERROR;
      }
      else
      {
          TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("RTCP statistics for fd %d are reseted.\n",fd));
      }

   }
   return IFX_SUCCESS;
} /* Common_RTCPStatistic */

#ifdef TAPI_VERSION4
#if (defined(EASY336) || defined(XT16))
/**
   Set PCM configuration.

   \param pBoard - pointer to board
   \param pcm_cfg - pointer to PCM configuration.

   \return IFX_SUCCESS if successful, otherwise IFX_ERROR.
   \remark
      Memory for member "pDevs..." of pcm_cfg allocated by procedure and
      should be freed by caller afterwards.

   \return IFX_SUCCESS if successful, otherwise IFX_ERROR.
*/
IFX_return_t Common_PCM_ConfigSet(BOARD_t* pBoard,
                                  RM_SVIP_PCM_CFG_t* pcm_cfg)
{
   if (pBoard == IFX_NULL)
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
           __FILE__, __LINE__));
      return IFX_ERROR;
   }

   switch (pBoard->nType)
   {
#ifdef EASY336
      case TYPE_SVIP:
         return BOARD_Easy336_PCM_ConfigSet(pBoard, pcm_cfg);
#ifdef WITH_VXT
      case TYPE_XT16:
         return BOARD_XT_16_PCM_ConfigSet(pBoard, pcm_cfg);
#endif /* WITH_VXT */
#endif /* EASY336 */
      default:
         return IFX_ERROR;
   }
} /* Common_PCM_ConfigSet */
#endif /* (defined(EASY336) || defined(XT16)) */

#ifdef EASY336
/**
   Get resources (signal generator, signal detector, coder etc.) from resource
   manager.
   \param pPhone pointer to PHONE.
          pResources Pointer to resources structure to get.

   \return IFX_SUCCESS on ok, otherwise IFX_ERROR
*/
IFX_return_t Common_ResourcesGet(PHONE_t* pPhone, const RESOURCES_t* pResources)
{
   RM_SVIP_RU_t RU, RU_Tone;
   SVIP_RM_Status_t ret;
   IFX_int32_t voipID, connID, i;

   TRACE(TAPIDEMO, DBG_LEVEL_LOW,
     ("Phone No %d (%s): getting\n",
        pPhone->nPhoneNumber, pPhone->pBoard->pszBoardName));
   if (pResources->nType & RES_VOIP_ID)
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
        ("\tVoipID\n"));
   if (pResources->nType & RES_CONN_ID)
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
        ("\tConnID\n"));
   else
      if (pResources->nConnID_Nr != SVIP_RM_UNUSED)
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
        ("\tConnecting to connID %d\n", pResources->nConnID_Nr));
   if (pResources->nType & RES_CODEC)
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
        ("\tCodec\n"));
   if (pResources->nType & RES_GEN)
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
        ("\tSignal generator\n"));
   if (pResources->nType & RES_DET)
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
        ("\tSignal detector\n"));
   if (pResources->nType & RES_LEC)
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
        ("\tLEC\n"));

   if (pResources->nType & RES_CONN_ID && pPhone->connID != SVIP_RM_UNUSED)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
        ("Err: connID already assigned to phone. (File: %s, "
         "line: %d)\n",
         __FILE__, __LINE__));
      return IFX_ERROR;
   }

   /* get connID */
   if (pResources->nType & RES_CONN_ID)
   {
      /* get connID, assign lineID and voipID to it */
      ret = SVIP_RM_ConnIdAlloc(&connID);
      if (ret != SVIP_RM_Success)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Err: SVIP_RM_ConnIdAlloc returned %d (File: %s, "
            "line: %d)\n",
            ret, __FILE__, __LINE__));
         return IFX_ERROR;
      }

      ret = SVIP_RM_LineIdAssign(connID, pPhone->lineID);

      if (ret != SVIP_RM_Success)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Err: SVIP_RM_LineIdAssign returned %d (File: %s, "
            "line: %d)\n",
            ret, __FILE__, __LINE__));
         return IFX_ERROR;
      }
      pPhone->connID = connID;
   }
   else
   {
      if (pResources->nConnID_Nr != SVIP_RM_UNUSED)
      {
         ret = SVIP_RM_LineIdAssign(pResources->nConnID_Nr, pPhone->lineID);
         if (ret != SVIP_RM_Success)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                    ("Err: SVIP_RM_LineIdAssign returned %d (File: %s, "
               "line: %d)\n",
               ret, __FILE__, __LINE__));
            return IFX_ERROR;
         }
         pPhone->connID = pResources->nConnID_Nr;
      }
   }

   /* get voipID */
   if (pResources->nType & RES_VOIP_ID)
   {
      ret = SVIP_RM_VoipIdAlloc(&voipID);
      if (ret != SVIP_RM_Success)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Err: SVIP_RM_VoipIdAlloc returned %d (File: %s, "
            "line: %d)\n",
            ret, __FILE__, __LINE__));
         return IFX_ERROR;
      }
      Common_VoipID_Add(pPhone, voipID);
   }

   /* if connID allocation done in this step, assign all voipIDs to connID */
   if (pResources->nType & RES_CONN_ID)
   for (i = 0; i < pPhone->nVoipIDs; i++)
   {
      ret = SVIP_RM_VoipIdAssign(pPhone->connID,
                                 pPhone->pVoipIDs[i]);
      if (ret != SVIP_RM_Success)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: SVIP_RM_VoipIdAssign returned %d (File: %s, "
            "line: %d)\n",
            ret, __FILE__, __LINE__));
         return IFX_ERROR;
      }
   }

   /* if voipID allocation done in this call, assign connID to it */
   if (pResources->nType & RES_VOIP_ID && pPhone->connID != SVIP_RM_UNUSED)
   {
      ret = SVIP_RM_VoipIdAssign(pPhone->connID,
         pPhone->pVoipIDs[pPhone->nVoipIDs - 1]);
      if (ret != SVIP_RM_Success)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: SVIP_RM_VoipIdAssign returned %d (File: %s, "
            "line: %d)\n",
            ret, __FILE__, __LINE__));
         return IFX_ERROR;
      }
   }

   /* get codec */
   if (pResources->nType & RES_CODEC)
   {
      ret = SVIP_RM_RUCodAlloc(pPhone->pVoipIDs[pPhone->nVoipIDs - 1],
         eCurrEncType, pPhone->nDev, &RU);
      if (ret != SVIP_RM_Success)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: SVIP_RM_RUCodAlloc returned %d (File: %s, "
            "line: %d)\n",
            ret, __FILE__, __LINE__));
         Common_ErrorHandle (pPhone->pBoard->nDevCtrl_FD, pPhone->pBoard,
            pPhone->nDev);
         return IFX_ERROR;
      }
   }

   RU.module = IFX_TAPI_MODULE_TYPE_ALM;
   RU.nDev = pPhone->nDev;
   RU.nCh = pPhone->nPhoneCh;
   if (pPhone->pBoard->nType == TYPE_SVIP)
      RU.devType = IFX_TAPI_DEV_TYPE_VIN_SUPERVIP;
   else if (pPhone->pBoard->nType == TYPE_XT16)
      RU.devType = IFX_TAPI_DEV_TYPE_VIN_XT16;

   /* get signal detector */
   if (pResources->nType & RES_DET)
   {
      ret = SVIP_RM_RUSigDetAlloc(&RU, &RU_Tone);
      if (ret != SVIP_RM_Success)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: SVIP_RM_RUSigDetAlloc returned %d (File: %s, "
            "line: %d)\n",
            ret, __FILE__, __LINE__));
         return IFX_ERROR;
      }
   }

   /* get singal generator */
   if (pResources->nType & RES_GEN)
   {
      ret = SVIP_RM_RUSigGenAlloc(&RU, &RU_Tone);
      if (ret != SVIP_RM_Success)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: SVIP_RM_RUSigGenAlloc returned %d (File: %s, "
            "line: %d)\n",
            ret, __FILE__, __LINE__));
         return IFX_ERROR;
      }
   }

   /* get LEC */
   if (pResources->nType & RES_LEC)
   {
      ret = SVIP_RM_RULecAlloc(pPhone->lineID, &RU_Tone);
      if (ret != SVIP_RM_Success)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Err: SVIP_RM_RULecAlloc returned %d (File: %s, "
            "line: %d)\n",
            ret, __FILE__, __LINE__));
         return IFX_ERROR;
      }
   }
   return IFX_SUCCESS;
}


/**
   Return resources to resource manager.
   \param pPhone pointer to PHONE.
          pResources Pointer to resources structure to release.

   \return IFX_SUCCESS on ok, otherwise IFX_ERROR
*/
IFX_return_t Common_ResourcesRelease(PHONE_t* pPhone,
                                     const RESOURCES_t* pResources)
{
   RM_SVIP_RU_t RU;
   SVIP_RM_Status_t ret;
   IFX_return_t funcRet = IFX_SUCCESS;
   IFX_int32_t i;
   TRACE(TAPIDEMO, DBG_LEVEL_LOW,
     ("Phone No %d (%s): releasing\n",
        pPhone->nPhoneNumber, pPhone->pBoard->pszBoardName));
   if (pResources->nType & RES_VOIP_ID)
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
        ("\tVoipID\n"));
   if (pResources->nType & RES_CONN_ID)
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
        ("\tConnID\n"));
   if (pResources->nType & RES_GEN)
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
        ("\tSignal generator\n"));
   if (pResources->nType & RES_DET)
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
        ("\tSignal detector\n"));
   if (pResources->nType & RES_LEC)
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
        ("\tLEC\n"));
   if (pResources->nType & RES_CODEC)
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
        ("\tCodec\n"));

   if ((pResources->nType & RES_VOIP_ID) && (pPhone->nVoipIDs == 0))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: voipID not assigned to phone. (File: %s, "
            "line: %d)\n",
            __FILE__, __LINE__));
      funcRet = IFX_ERROR;
   }
   if ((pResources->nType & RES_CONN_ID) && (pPhone->connID == SVIP_RM_UNUSED))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: connID not assigned to phone. (File: %s, "
            "line: %d)\n",
            __FILE__, __LINE__));
      funcRet = IFX_ERROR;
   }

   if (pResources->nType & RES_GEN)
   {
   /* release signal generator */
   ret = SVIP_RM_LineIdRUSigGenGet(pPhone->lineID, &RU);
   if (ret == SVIP_RM_Success)
   {
      ret = SVIP_RM_RUSigGenRelease(&RU);
      if (ret != SVIP_RM_Success)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: SVIP_RM_RUSigGenRelease returned %d (File: %s, "
            "line: %d)\n",
            ret, __FILE__, __LINE__));
         funcRet = IFX_ERROR;
      }
   }
   else
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
        ("Err: SVIP_RM_LineIdRUSigGenGet returned %d (File: %s, "
         "line: %d)\n",
         ret, __FILE__, __LINE__));
      funcRet = IFX_ERROR;
   }
   }

   if (pResources->nType & RES_DET)
   {
   /* release signal detector */
   ret = SVIP_RM_LineIdRUSigDetGet(pPhone->lineID, &RU);
   if (ret == SVIP_RM_Success)
   {
      ret = SVIP_RM_RUSigDetRelease(&RU);
      if (ret != SVIP_RM_Success)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: SVIP_RM_RUSigDetRelease returned %d (File: %s, "
            "line: %d)\n",
            ret, __FILE__, __LINE__));
         funcRet = IFX_ERROR;
      }
   }
   else
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
        ("Err: SVIP_RM_LineIdRUSigDetGet returned %d (File: %s, "
         "line: %d)\n",
         ret, __FILE__, __LINE__));
      funcRet = IFX_ERROR;
   }
   }

   if (pResources->nType & RES_LEC)
   {
      /* release LEC */
      ret = SVIP_RM_RULecRelease(pPhone->lineID);
   if (ret != SVIP_RM_Success)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: SVIP_RM_RULecRelease returned %d (File: %s, "
         "line: %d)\n",
         ret, __FILE__, __LINE__));
      funcRet = IFX_ERROR;
   }
   }

   if (pResources->nType & RES_CODEC)
   {
      /* release codec */
      if (pResources->nVoipID_Nr != SVIP_RM_UNUSED)
      {
         ret = SVIP_RM_RUCodRelease(pResources->nVoipID_Nr);
         if (ret != SVIP_RM_Success)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Err: SVIP_RM_RUCodRelease returned %d (File: %s, "
               "line: %d)\n",
               ret, __FILE__, __LINE__));
            Common_ErrorHandle (pPhone->pBoard->nDevCtrl_FD, pPhone->pBoard,
               pPhone->nDev);
            funcRet = IFX_ERROR;
         }
      }
      else
      /* release all codecs allocated for this phone */
      for (i = pPhone->nVoipIDs - 1; i >= 0; i--)
      {
         ret = SVIP_RM_RUCodRelease(pPhone->pVoipIDs[i]);
         if (ret != SVIP_RM_Success)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Err: SVIP_RM_RUCodRelease returned %d (File: %s, "
               "line: %d)\n",
               ret, __FILE__, __LINE__));
            funcRet = IFX_ERROR;
         }
      }
   }

   /* deassign voipID form connID, release voipID */
   if (pResources->nType & RES_VOIP_ID)
   {
      if (pPhone->connID != SVIP_RM_UNUSED)
      {
         if (pResources->nVoipID_Nr != SVIP_RM_UNUSED)
         {
            ret = SVIP_RM_VoipIdDeassign(pPhone->connID,
               pResources->nVoipID_Nr);
            if (ret != SVIP_RM_Success)
            {
               TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                    ("Err: SVIP_RM_VoipIdDeassign returned %d (File: %s, "
                     "line: %d)\n",
                     ret, __FILE__, __LINE__));
               funcRet = IFX_ERROR;
            }
         }
         else
         {
            for (i = pPhone->nVoipIDs - 1; i >= 0; i--)
            {
               ret = SVIP_RM_VoipIdDeassign(pPhone->connID,
                  pPhone->pVoipIDs[i]);
               if (ret != SVIP_RM_Success)
               {
                  TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                       ("Err: SVIP_RM_VoipIdDeassign returned %d (File: %s, "
                        "line: %d)\n",
                        ret, __FILE__, __LINE__));
                  funcRet = IFX_ERROR;
               }
            }
         }
      }
      /* release voipID */
      if (pResources->nVoipID_Nr != SVIP_RM_UNUSED)
      {
         ret = SVIP_RM_VoipIdRelease(pResources->nVoipID_Nr);
         if (ret != SVIP_RM_Success)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Err: SVIP_RM_VoipIdRelease returned %d (File: %s, "
               "line: %d)\n",
               ret, __FILE__, __LINE__));
            funcRet = IFX_ERROR;
         }
         Common_VoipID_Remove(pPhone, pResources->nVoipID_Nr);
      }
      else
      {
         for (i = pPhone->nVoipIDs - 1; i >= 0; i--)
         {
            ret = SVIP_RM_VoipIdRelease(pPhone->pVoipIDs[i]);
            if (ret != SVIP_RM_Success)
            {
               TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                 ("Err: SVIP_RM_VoipIdRelease returned %d (File: %s, "
                  "line: %d)\n",
                  ret, __FILE__, __LINE__));
               funcRet = IFX_ERROR;
            }
            Common_VoipID_Remove(pPhone, pPhone->pVoipIDs[i]);
         }
      }
   }

   /* deassign lineID and voipIDs form connID, release connID */
   if (pResources->nType & RES_CONN_ID)
   {
      for (i = 0; i < pPhone->nVoipIDs; i++)
      {
         ret = SVIP_RM_VoipIdDeassign(pPhone->connID,
            pPhone->pVoipIDs[i]);
         if (ret != SVIP_RM_Success)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Err: SVIP_RM_VoipIdDeassign returned %d (File: %s, "
               "line: %d)\n",
               ret, __FILE__, __LINE__));
            funcRet = IFX_ERROR;
         }
      }

      ret = SVIP_RM_LineIdDeassign(pPhone->connID, pPhone->lineID);
      if (ret != SVIP_RM_Success)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: SVIP_RM_LineIdDeassign returned %d (File: %s, "
            "line: %d)\n",
            ret, __FILE__, __LINE__));
         funcRet = IFX_ERROR;
      }

      /* release connID */
      ret = SVIP_RM_ConnIdRelease(pPhone->connID);
      if (ret != SVIP_RM_Success)
      {
#if 0
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Err: SVIP_RM_ConnIdRelease returned %d (File: %s, "
            "line: %d)\n",
            ret, __FILE__, __LINE__));
#endif
         funcRet = IFX_ERROR;
      }
      pPhone->connID = SVIP_RM_UNUSED;
   }

   return funcRet;
} /* Common_ResourcesRelease() */

/**
   Get pointer to codec status structure from voipID.

   \param pCtrl - handle to connection control structure
   \param voipID - voipID of codec of interest.

   \return pointer to structure on success, otherwise IFX_NULL
*/
VOIP_DATA_CH_t* Common_GetCodec(CTRL_STATUS_t* pCtrl, IFX_int32_t voipID)
{
   IFX_int32_t i;
   RM_SVIP_RU_t RU;
   SVIP_RM_Status_t ret;
   BOARD_t *pBoard;

   if (voipID == SVIP_RM_UNUSED)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: invalid voipID. (File: %s, "
            "line: %d)\n",
            __FILE__, __LINE__));
      return IFX_NULL;
   }
   ret = SVIP_RM_VoipIdRUCodGet(voipID, &RU);
   if (ret == SVIP_RM_Success)
   {
      /* codec not allocated ? */
      if (RU.devType == IFX_TAPI_DEV_TYPE_NONE)
         /* Y */
         return IFX_NULL;

      pBoard = ABSTRACT_GetBoard(pCtrl, RU.devType);
      if (pBoard == IFX_NULL)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: Invalid device type (File: %s, "
            "line: %d)\n", __FILE__, __LINE__));
         return IFX_NULL;
      }

      for (i = 0; i < pBoard->nMaxCoderCh; i++)
         if (pBoard->pDataChStat[i].nDev == RU.nDev &&
            pBoard->pDataChStat[i].nCh == RU.nCh)
            return &pBoard->pDataChStat[i];
   }
   else
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
        ("Err: SVIP_RM_VoipIdRUCodGet returned %d (File: %s, "
         "line: %d)\n",
         ret, __FILE__, __LINE__));
      return IFX_NULL;
   }

   return IFX_NULL;
} /* Common_GetCodec() */
#endif /* EASY336 */
#endif /* TAPI_VERSION4 */
 /**
   Start playing tone on phone.

   \param pCtrl - handle to connection control structure
   \param pPhone Pointer to PHONE.
          nTone Tone to play.
   \return IFX_SUCCESS on ok, otherwise IFX_ERROR
*/
IFX_return_t Common_TonePlay(CTRL_STATUS_t* pCtrl, PHONE_t* pPhone,
                             IFX_uint32_t nTone)
{
   if ((IFX_NULL == pCtrl) || (IFX_NULL == pPhone) || (0 > nTone))
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }

#ifdef ABACUS_TEST
   if (IFX_TRUE == pCtrl->nUseCustomCpt)
   {
      /* get CPT index*/
      nTone = COM_GetCustomCPT(nTone);
   }
#endif /* ABACUS_TEST */

#ifdef TAPI_VERSION4
   if (!pPhone->pBoard->fSingleFD)
#endif /* TAPI_VERSION4 */
   {
      /* check if data channel avaible */
      if (0 < pPhone->pBoard->nMaxCoderCh)
      {
         /* play tone */
         return ioctl(pPhone->nDataCh_FD, IFX_TAPI_TONE_LOCAL_PLAY, nTone);
      }
      else
      {
         /* use phone channel if no data channel avaible on board */
         return ioctl(pPhone->nPhoneCh_FD, IFX_TAPI_TONE_LOCAL_PLAY, nTone);
      }
   }
#ifdef TAPI_VERSION4
   IFX_TAPI_TONE_PLAY_t tone;
   IFX_return_t ioctlRet;
#ifdef EASY336
   SVIP_RM_Status_t ret;
   RM_SVIP_RU_t RU;
   BOARD_t *pBoard;
   RESOURCES_t resources;

   ret = SVIP_RM_LineIdRUSigGenGet(pPhone->lineID, &RU);
   if (ret == SVIP_RM_Success)
   {
      /* siggen not allocated ? */
      if (RU.devType == IFX_TAPI_DEV_TYPE_NONE)
      {
         /* Y */
         resources.nType = RES_GEN;
         resources.nConnID_Nr = SVIP_RM_UNUSED;
         if (Common_ResourcesGet(pPhone, &resources) !=
            IFX_SUCCESS)
            return IFX_ERROR;
         ret = SVIP_RM_LineIdRUSigGenGet(pPhone->lineID, &RU);
     }

      if (ret == SVIP_RM_Success)
      {
         pBoard = ABSTRACT_GetBoard(pCtrl, RU.devType);
         if (pBoard == IFX_NULL)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Err: Invalid device type (File: %s, "
               "line: %d)\n", __FILE__, __LINE__));
            return IFX_ERROR;
         }

         memset (&tone, 0, sizeof (IFX_TAPI_TONE_PLAY_t));
         tone.module = RU.module;
         tone.dev = RU.nDev;
         tone.ch = RU.nCh;
         tone.index = nTone;
         tone.external = 1;
         tone.internal = 0;
         tone.reserved = 0;
         TD_ASSERT ((ioctlRet = ioctl(pBoard->nDevCtrl_FD,
            IFX_TAPI_TONE_PLAY, (IFX_int32_t) &tone)) == IFX_SUCCESS,
            pBoard->nDevCtrl_FD, pBoard, tone.dev, tone.ch);
         if (ioctlRet == IFX_ERROR)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Err: IFX_TAPI_TONE_PLAY failed (File: %s, "
               "line: %d)\n",
               __FILE__, __LINE__));
            return IFX_ERROR;
         }
      }
   }
   else
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
        ("Err: SVIP_RM_LineIdRUSigGenGet returned %d (File: %s, "
         "line: %d)\n",
         ret, __FILE__, __LINE__));
      return IFX_ERROR;
   }
#else /* EASY336 */
   memset (&tone, 0, sizeof (IFX_TAPI_TONE_PLAY_t));
   tone.dev = pPhone->nDev;
   tone.ch = pPhone->nPhoneCh;
   tone.module = IFX_TAPI_MODULE_TYPE_ALM;
   tone.index = nTone;
   tone.external = 1;
   tone.internal = 0;
   tone.reserved = 0;
   TD_ASSERT ((ioctlRet = ioctl(pPhone->pBoard->nDevCtrl_FD,
      IFX_TAPI_TONE_PLAY, (IFX_int32_t) &tone)) == IFX_SUCCESS,
      pPhone->pBoard->nDevCtrl_FD, pPhone->pBoard, tone.dev, tone.ch);
   if (ioctlRet == IFX_ERROR)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
        ("Err: IFX_TAPI_TONE_PLAY failed (File: %s, "
         "line: %d)\n",
         __FILE__, __LINE__));
      return IFX_ERROR;
   }
#endif /* EASY336 */
   return IFX_SUCCESS;
#endif /* TAPI_VERSION4 */
}

/**
   Stop playing tone on phone.

   \param pCtrl - handle to connection control structure
   \param pPhone Pointer to PHONE.
   \return IFX_SUCCESS on ok, otherwise IFX_ERROR
*/
IFX_return_t Common_ToneStop(CTRL_STATUS_t* pCtrl, PHONE_t* pPhone)
{
   if ((IFX_NULL == pCtrl) || (IFX_NULL == pPhone))
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }
#ifdef TAPI_VERSION4
   if (!pPhone->pBoard->fSingleFD)
#endif /* TAPI_VERSION4 */
   {
      /* check if data channel avaible */
      if (0 < pPhone->pBoard->nMaxCoderCh)
      {
         /* play tone */
         return ioctl(pPhone->nDataCh_FD, IFX_TAPI_TONE_STOP, 0);
      }
      else
      {
         /* use phone channel if no data channel avaible on board */
         return ioctl(pPhone->nPhoneCh_FD, IFX_TAPI_TONE_STOP, 0);
      }
   }
#ifdef TAPI_VERSION4
   IFX_TAPI_TONE_PLAY_t tone;
   IFX_return_t ioctlRet;
#ifdef EASY336
   SVIP_RM_Status_t ret;
   RM_SVIP_RU_t RU;
   BOARD_t *pBoard;

#ifdef ABACUS_TEST
   /* if DTMF digit test do not use this function
      because resource manager is deinitialized */
   if (pCtrl->pProgramArg->nTestType == COM_MODULAR_TEST &&
       pCtrl->pProgramArg->nSubTestType & COM_SUBTEST_DTMFDIGIT)
   {
      /* end function */
      return IFX_SUCCESS;
   }
#endif /* ABACUS_TEST */

   ret = SVIP_RM_LineIdRUSigGenGet(pPhone->lineID, &RU);
   if (ret == SVIP_RM_Success)
   {
      pBoard = ABSTRACT_GetBoard(pCtrl, RU.devType);
      if (pBoard == IFX_NULL)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: Invalid device type (File: %s, "
            "line: %d)\n", __FILE__, __LINE__));
         return IFX_ERROR;
      }

      memset (&tone, 0, sizeof (IFX_TAPI_TONE_PLAY_t));
      tone.module = RU.module;
      tone.dev = RU.nDev;
      tone.ch = RU.nCh;
      tone.external = 1;
      tone.internal = 0;
      tone.reserved = 0;
      TD_ASSERT ((ioctlRet = ioctl(pBoard->nDevCtrl_FD,
         IFX_TAPI_TONE_STOP, (IFX_int32_t) &tone)) == IFX_SUCCESS,
         pBoard->nDevCtrl_FD, pBoard, tone.dev, tone.ch);
      if (ioctlRet == IFX_ERROR)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Err: IFX_TAPI_TONE_STOP failed (File: %s, "
               "line: %d)\n",
               __FILE__, __LINE__));
         return IFX_ERROR;
      }
   }
   else
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
        ("Err: SVIP_RM_LineIdRUSigGenGet returned %d (File: %s, line: %d)\n",
         ret, __FILE__, __LINE__));
      return IFX_ERROR;
   }
#else /* EASY336 */
   memset (&tone, 0, sizeof (IFX_TAPI_TONE_PLAY_t));
   tone.module = IFX_TAPI_MODULE_TYPE_ALM;
   tone.dev = pPhone->nDev;
   tone.ch = pPhone->nPhoneCh;
   tone.external = 1;
   tone.internal = 0;
   tone.reserved = 0;
   TD_ASSERT ((ioctlRet = ioctl(pPhone->pBoard->nDevCtrl_FD,
      IFX_TAPI_TONE_STOP, (IFX_int32_t) &tone)) == IFX_SUCCESS,
      pPhone->pBoard->nDevCtrl_FD, pPhone->pBoard, tone.dev, tone.ch);
   if (ioctlRet == IFX_ERROR)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: IFX_TAPI_TONE_STOP failed (File: %s, "
            "line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }
#endif /* EASY336 */
   return IFX_SUCCESS;
#endif /* TAPI_VERSION4 */
} /* Common_ToneStop() */

#ifdef TAPI_VERSION4
/**
   Enable DTMF detection on phone.

   \param pCtrl - handle to connection control structure
   \param pPhone Pointer to PHONE.
   \return IFX_SUCCESS on ok, otherwise IFX_ERROR
*/
IFX_return_t Common_DTMF_Enable(CTRL_STATUS_t* pCtrl, PHONE_t* pPhone)
{
   IFX_TAPI_EVENT_t tapiEvent;
   IFX_return_t ioctlRet;
#ifdef EASY336
   SVIP_RM_Status_t ret = SVIP_RM_Success;
   RM_SVIP_RU_t RU;
   BOARD_t *pBoard;
   RESOURCES_t resources;

   ret = SVIP_RM_LineIdRUSigDetGet(pPhone->lineID, &RU);
   if (ret == SVIP_RM_Success)
   {
      /* sigdet not allocated ? */
      if (RU.devType == IFX_TAPI_DEV_TYPE_NONE)
      {
         /* Y */
         resources.nType = RES_DET;
         resources.nConnID_Nr = SVIP_RM_UNUSED;
         if (Common_ResourcesGet(pPhone, &resources) !=
            IFX_SUCCESS)
            return IFX_ERROR;
         ret = SVIP_RM_LineIdRUSigDetGet(pPhone->lineID, &RU);
     }

      if (ret == SVIP_RM_Success)
      {
         pBoard = ABSTRACT_GetBoard(pCtrl, RU.devType);
         if (pBoard == IFX_NULL)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Err: Invalid device type (File: %s, "
               "line: %d)\n", __FILE__, __LINE__));
            return IFX_ERROR;
         }

         memset((void*)&tapiEvent, 0, sizeof(IFX_TAPI_EVENT_t));
         tapiEvent.dev = RU.nDev;
         tapiEvent.ch = RU.nCh;
         tapiEvent.module = RU.module;
         /* DTMF tone detection done from the analog line side. */
         tapiEvent.id = IFX_TAPI_EVENT_DTMF_DIGIT;
#ifdef EASY336
         tapiEvent.data.dtmf.external = 1;
         tapiEvent.data.dtmf.internal = 0;
#else
         tapiEvent.data.dtmf.local = 1;
         tapiEvent.data.dtmf.network = 0;
#endif
         TD_ASSERT ((ioctlRet = ioctl(pBoard->nDevCtrl_FD,
            IFX_TAPI_EVENT_ENABLE, (IFX_int32_t) &tapiEvent)) == IFX_SUCCESS,
            pBoard->nDevCtrl_FD, pBoard, tapiEvent.dev, tapiEvent.ch);
         if (ioctlRet != IFX_SUCCESS)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Err: IFX_TAPI_EVENT_ENABLE failed (File: %s, "
               "line: %d)\n",
               __FILE__, __LINE__));
            return IFX_ERROR;
         }
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
              ("Phone No %d: Tone detection started on %s dev %d, "
               "ch %d \n",
               pPhone->nPhoneNumber,
               pBoard->pszBoardName, tapiEvent.dev, tapiEvent.ch));
      }
   }
   else
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
        ("Err: SVIP_RM_LineIdRUSigDetGet returned %d (File: %s, "
         "line: %d)\n",
         ret, __FILE__, __LINE__));
      return IFX_ERROR;
   }
#else /* EASY336 */
   memset((void*)&tapiEvent, 0, sizeof(IFX_TAPI_EVENT_t));
   tapiEvent.dev = pPhone->nDev;
   tapiEvent.ch = pPhone->nPhoneCh;
   tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
   /* DTMF tone detection done from the analog line side. */
   tapiEvent.id = IFX_TAPI_EVENT_DTMF_DIGIT;
#if (defined(EASY336) || defined(XT16))
         tapiEvent.data.dtmf.external = 1;
         tapiEvent.data.dtmf.internal = 0;
#else
         tapiEvent.data.dtmf.local = 1;
         tapiEvent.data.dtmf.network = 0;
#endif
   TD_ASSERT ((ioctlRet = ioctl(pPhone->pBoard->nDevCtrl_FD,
      IFX_TAPI_EVENT_ENABLE, (IFX_int32_t) &tapiEvent)) == IFX_SUCCESS,
      pPhone->pBoard->nDevCtrl_FD, pPhone->pBoard, tapiEvent.dev,
      tapiEvent.ch);
   if (ioctlRet != IFX_SUCCESS)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
        ("Err: IFX_TAPI_EVENT_ENABLE failed (File: %s, "
         "line: %d)\n",
         __FILE__, __LINE__));
      return IFX_ERROR;
   }
   TRACE(TAPIDEMO, DBG_LEVEL_LOW,
        ("Phone No %d: Tone detection started on %s dev %d, "
         "ch %d \n",
         pPhone->nPhoneNumber,
         pPhone->pBoard->pszBoardName, tapiEvent.dev, tapiEvent.ch));
#endif /* EASY336 */
   return IFX_SUCCESS;
}

/**
   Disable DTMF detection on phone.

   \param pCtrl - handle to connection control structure
   \param pPhone Pointer to PHONE.
   \return IFX_SUCCESS on ok, otherwise IFX_ERROR
*/
IFX_return_t Common_DTMF_Disable(CTRL_STATUS_t* pCtrl, PHONE_t* pPhone)
{
   IFX_TAPI_EVENT_t tapiEvent;
   IFX_return_t ioctlRet;
#ifdef EASY336
   SVIP_RM_Status_t ret = SVIP_RM_Success;
   RM_SVIP_RU_t RU;
   BOARD_t *pBoard;

   ret = SVIP_RM_LineIdRUSigDetGet(pPhone->lineID, &RU);
   if (ret == SVIP_RM_Success)
   {
      pBoard = ABSTRACT_GetBoard(pCtrl, RU.devType);
      if (pBoard == IFX_NULL)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: Invalid device type (File: %s, "
            "line: %d)\n", __FILE__, __LINE__));
         return IFX_ERROR;
      }

      memset((void*)&tapiEvent, 0, sizeof(IFX_TAPI_EVENT_t));
      tapiEvent.dev = RU.nDev;
      tapiEvent.ch = RU.nCh;
      tapiEvent.module = RU.module;
      /* DTMF tone detection done from the analog line side. */
      tapiEvent.id = IFX_TAPI_EVENT_DTMF_DIGIT;
#ifdef EASY336
      tapiEvent.data.dtmf.external = 1;
      tapiEvent.data.dtmf.internal = 0;
#else
      tapiEvent.data.dtmf.local = 1;
      tapiEvent.data.dtmf.network = 0;
#endif
      TD_ASSERT ((ioctlRet = ioctl(pBoard->nDevCtrl_FD,
         IFX_TAPI_EVENT_DISABLE, (IFX_int32_t) &tapiEvent)) == IFX_SUCCESS,
         pBoard->nDevCtrl_FD, pBoard, tapiEvent.dev, tapiEvent.ch);
      if (ioctlRet != IFX_SUCCESS)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: IFX_TAPI_EVENT_DISABLE failed (File: %s, "
            "line: %d)\n",
            __FILE__, __LINE__));
         return IFX_ERROR;
      }

      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
           ("Phone No %d: Tone detection stopped on "
            "%s dev %d, channel %d \n",
            pPhone->nPhoneNumber, pBoard->pszBoardName,
            tapiEvent.dev, tapiEvent.ch));
   }
   else
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
        ("Err: SVIP_RM_LineIdRUSigDetGet returned %d (File: %s, "
         "line: %d)\n",
         ret, __FILE__, __LINE__));
      return IFX_ERROR;
   }
#else /* EASY336 */
   memset((void*)&tapiEvent, 0, sizeof(IFX_TAPI_EVENT_t));
   tapiEvent.dev = pPhone->nDev;
   tapiEvent.ch = pPhone->nPhoneCh;
   tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
   /* DTMF tone detection done from the analog line side. */
   tapiEvent.id = IFX_TAPI_EVENT_DTMF_DIGIT;
#if (defined(EASY336) || defined(XT16))
   tapiEvent.data.dtmf.external = 1;
   tapiEvent.data.dtmf.internal = 0;
#else
   tapiEvent.data.dtmf.local = 1;
   tapiEvent.data.dtmf.network = 0;
#endif
   TD_ASSERT ((ioctlRet = ioctl(pPhone->pBoard->nDevCtrl_FD,
      IFX_TAPI_EVENT_DISABLE, (IFX_int32_t) &tapiEvent)) == IFX_SUCCESS,
      pPhone->pBoard->nDevCtrl_FD, pPhone->pBoard, tapiEvent.dev,
      tapiEvent.ch);
   if (ioctlRet != IFX_SUCCESS)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
        ("Err: IFX_TAPI_EVENT_DISABLE failed (File: %s, "
         "line: %d)\n",
         __FILE__, __LINE__));
      return IFX_ERROR;
   }
   TRACE(TAPIDEMO, DBG_LEVEL_LOW,
        ("Phone No %d: Tone detection stopped on "
         "%s dev %d, ch %d \n",
         pPhone->nPhoneNumber, pPhone->pBoard->pszBoardName,
         tapiEvent.dev, tapiEvent.ch));
#endif /* EASY336 */
   return IFX_SUCCESS;
}

#ifdef EASY336
/**
   Enable LEC on phone.

   \param pCtrl - handle to connection control structure
   \param pPhone Pointer to PHONE.
   \return IFX_SUCCESS on ok, otherwise IFX_ERROR
*/
IFX_return_t Common_LEC_Enable(CTRL_STATUS_t* pCtrl, PHONE_t* pPhone)
{
   SVIP_RM_Status_t ret;
   IFX_TAPI_WLEC_CFG_t wlecConf;
   RM_SVIP_RU_t RU;
   BOARD_t *pBoard;
   IFX_return_t ioctlRet;

   /* check if lec should be used */
   if (pCtrl->pProgramArg->oArgFlags.nNoLEC)
      return IFX_SUCCESS;

   /* start LEC on callee */
   ret = SVIP_RM_LineIdRULecGet(pPhone->lineID, &RU);
   if (ret == SVIP_RM_Success)
   {
      pBoard = ABSTRACT_GetBoard(pCtrl, RU.devType);
      if (pBoard == IFX_NULL)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: Invalid device type (File: %s, "
            "line: %d)\n", __FILE__, __LINE__));
         return IFX_ERROR;
      }
      memset(&wlecConf, 0, sizeof(IFX_TAPI_WLEC_CFG_t));

      wlecConf.nType = IFX_TAPI_WLEC_TYPE_NFE;
      wlecConf.bNlp = IFX_TAPI_WLEC_NLP_DEFAULT;
      wlecConf.nNBFEwindow = IFX_TAPI_WLEN_WSIZE_8;
      wlecConf.nNBNEwindow = IFX_TAPI_WLEN_WSIZE_8;
      wlecConf.dev = RU.nDev;
      wlecConf.ch = RU.nCh;
      if (RU.module == IFX_TAPI_MODULE_TYPE_ALM)
      {
         TD_ASSERT ((ioctlRet = ioctl(pBoard->nDevCtrl_FD,
            IFX_TAPI_WLEC_PHONE_CFG_SET, (IFX_int32_t) &wlecConf)) ==
            IFX_SUCCESS,
            pBoard->nDevCtrl_FD, pBoard, wlecConf.dev, wlecConf.ch);
      }
      else
      {
         TD_ASSERT ((ioctlRet = ioctl(pBoard->nDevCtrl_FD,
            IFX_TAPI_WLEC_PCM_CFG_SET, (IFX_int32_t) &wlecConf)) ==
            IFX_SUCCESS,
            pBoard->nDevCtrl_FD, pBoard, wlecConf.dev, wlecConf.ch);
      }
      if (ioctlRet != IFX_SUCCESS)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("IFX_TAPI_WLEC_PHONE_CFG_SET failed for dev %d channel %d, "
                "(File: %s, line: %d)\n",
                wlecConf.dev, wlecConf.ch, __FILE__, __LINE__));
         return IFX_ERROR;
      }
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
        ("Phone No %d: LEC started on %s dev %d, ch %d\n",
         pPhone->nPhoneNumber,
         pBoard->pszBoardName, wlecConf.dev, wlecConf.ch));
   }
   else
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
        ("Err: SVIP_RM_LineIdRULecGet returned %d (File: %s, "
         "line: %d)\n",
         ret, __FILE__, __LINE__));
      return IFX_ERROR;
   }
   return IFX_SUCCESS;
}

/**
   Disable LEC on phone.

   \param pCtrl - handle to connection control structure
   \param pPhone Pointer to PHONE.
   \return IFX_SUCCESS on ok, otherwise IFX_ERROR
*/
IFX_return_t Common_LEC_Disable(CTRL_STATUS_t* pCtrl, PHONE_t* pPhone)
{
   SVIP_RM_Status_t ret;
   IFX_TAPI_WLEC_CFG_t wlecConf;
   RM_SVIP_RU_t RU;
   BOARD_t *pBoard;
   IFX_return_t ioctlRet;

   /* check if lec should be used */
   if (pCtrl->pProgramArg->oArgFlags.nNoLEC)
      return IFX_SUCCESS;

   ret = SVIP_RM_LineIdRULecGet(pPhone->lineID, &RU);
   if (ret == SVIP_RM_Success)
   {
      pBoard = ABSTRACT_GetBoard(pCtrl, RU.devType);
      if (pBoard == IFX_NULL)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: Invalid device type (File: %s, "
            "line: %d)\n", __FILE__, __LINE__));
         return IFX_ERROR;
      }

      memset(&wlecConf, 0, sizeof(IFX_TAPI_WLEC_CFG_t));

      wlecConf.nType = IFX_TAPI_WLEC_TYPE_OFF;
      wlecConf.bNlp = IFX_TAPI_WLEC_NLP_OFF;
      wlecConf.dev = RU.nDev;
      wlecConf.ch = RU.nCh;
      if (RU.module == IFX_TAPI_MODULE_TYPE_ALM)
      {
         TD_ASSERT ((ioctlRet = ioctl(pBoard->nDevCtrl_FD,
            IFX_TAPI_WLEC_PHONE_CFG_SET, (IFX_int32_t) &wlecConf)) ==
            IFX_SUCCESS,
            pBoard->nDevCtrl_FD, pBoard, wlecConf.dev, wlecConf.ch);
      }
      else
      {
         TD_ASSERT ((ioctlRet = ioctl(pBoard->nDevCtrl_FD,
            IFX_TAPI_WLEC_PCM_CFG_SET, (IFX_int32_t) &wlecConf)) ==
            IFX_SUCCESS,
            pBoard->nDevCtrl_FD, pBoard, wlecConf.dev, wlecConf.ch);
      }
      if (ioctlRet != IFX_SUCCESS)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("IFX_TAPI_WLEC_CFG_SET failed for dev %d, ch %d, "
                "(File: %s, line: %d)\n",
                wlecConf.dev, wlecConf.ch, __FILE__, __LINE__));
         return IFX_ERROR;
      }
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
        ("Phone No %d: LEC stopped on %s dev %d, ch %d\n",
         pPhone->nPhoneNumber,
         pBoard->pszBoardName, wlecConf.dev, wlecConf.ch));
   }
   else
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
        ("Err: SVIP_RM_LineIdRULecGet returned %d (File: %s, "
         "line: %d)\n",
         ret, __FILE__, __LINE__));
      return IFX_ERROR;
   }
   return IFX_SUCCESS;
}
#endif /* EASY336 */
#else /* TAPI_VERSION4 */

/**
   Enable LEC on phone.

   \param pCtrl - handle to connection control structure
   \param pPhone Pointer to PHONE.
   \return IFX_SUCCESS on ok, otherwise IFX_ERROR
*/
IFX_return_t Common_LEC_Enable(CTRL_STATUS_t* pCtrl, PHONE_t* pPhone)
{
#if (!defined DXT && !defined XT16)
   IFX_return_t ret = IFX_SUCCESS;
   IFX_TAPI_WLEC_CFG_t wlecConf = {0};

   if ((pCtrl == IFX_NULL) || (pPhone == IFX_NULL))
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }

   /* check if lec should be used */
   if (pCtrl->pProgramArg->oArgFlags.nNoLEC)
      return IFX_SUCCESS;

   /* Activate WLEC for Phone */
   memset(&wlecConf, 0, sizeof(IFX_TAPI_WLEC_CFG_t));

   wlecConf.nType = IFX_TAPI_WLEC_TYPE_OFF;
   wlecConf.bNlp = IFX_TAPI_WLEC_NLP_ON;
   wlecConf.nNBNEwindow = 0 /* IFX_TAPI_WLEN_WSIZE_8 */;
   wlecConf.nNBFEwindow = 0 /* IFX_TAPI_WLEN_WSIZE_8 */;
   wlecConf.nWBNEwindow = 0 /* IFX_TAPI_WLEN_WSIZE_8 */;

   ret = ioctl(pPhone->nPhoneCh_FD, IFX_TAPI_WLEC_PHONE_CFG_SET, 
      (IFX_int32_t) &wlecConf);
   if (ret != IFX_SUCCESS)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
         ("IFX_TAPI_WLEC_PHONE_CFG_SET failed for board: %s, phone ch %d, "
         "(File: %s, line: %d)\n",
         pPhone->pBoard->pszBoardName, pPhone->nPhoneCh, __FILE__, __LINE__));
      return IFX_ERROR;
   }

   TRACE(TAPIDEMO, DBG_LEVEL_LOW,
         ("Phone No %d: LEC started on %s, PhoneCh %d\n",
          pPhone->nPhoneNumber, pPhone->pBoard->pszBoardName, pPhone->nPhoneCh));
#endif /* !DXT && !XT16 */

   return IFX_SUCCESS;
}

/**
   Disable LEC on phone.

   \param pCtrl - handle to connection control structure
   \param pPhone Pointer to PHONE.
   \return IFX_SUCCESS on ok, otherwise IFX_ERROR
*/
IFX_return_t Common_LEC_Disable(CTRL_STATUS_t* pCtrl, PHONE_t* pPhone)
{
#if (!defined DXT && !defined XT16)
   IFX_return_t ret = IFX_SUCCESS;
   IFX_TAPI_WLEC_CFG_t wlecConf = {0};

   if ((pCtrl == IFX_NULL) || (pPhone == IFX_NULL))
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }

   /* check if lec should be used */
   if (pCtrl->pProgramArg->oArgFlags.nNoLEC)
      return IFX_SUCCESS;

   /* Activate WLEC for Phone */
   memset(&wlecConf, 0, sizeof(IFX_TAPI_WLEC_CFG_t));

   wlecConf.nType = IFX_TAPI_WLEC_TYPE_OFF;
   wlecConf.bNlp = IFX_TAPI_WLEC_NLP_OFF;
   wlecConf.nNBNEwindow = 0 /* IFX_TAPI_WLEN_WSIZE_8 */;
   wlecConf.nNBFEwindow = 0 /* IFX_TAPI_WLEN_WSIZE_8 */;
   wlecConf.nWBNEwindow = 0 /* IFX_TAPI_WLEN_WSIZE_8 */;

   ret = ioctl(pPhone->nPhoneCh_FD, IFX_TAPI_WLEC_PHONE_CFG_SET, 
      (IFX_int32_t) &wlecConf);
   if (ret != IFX_SUCCESS)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
         ("IFX_TAPI_WLEC_PHONE_CFG_SET failed for board: %s, phone ch %d, "
         "(File: %s, line: %d)\n",
         pPhone->pBoard->pszBoardName, pPhone->nPhoneCh, __FILE__, __LINE__));
      return IFX_ERROR;
   }

   TRACE(TAPIDEMO, DBG_LEVEL_LOW,
         ("Phone No %d: LEC stopped on %s, phone ch %d\n",
          pPhone->nPhoneNumber, pPhone->pBoard->pszBoardName, pPhone->nPhoneCh));

#endif /* !DXT && !XT16 */
   return IFX_SUCCESS;
}
#endif /* TAPI_VERSION4 */

/**
   Enable LEC on phone.

   \param nPCM_Ch - PCM channel number
   \param pEnable - if IFX_ENABLE then enable LEC on PCM, if else disable LEC
   \param pBoard  - poiter to board

   \return IFX_SUCCESS on ok, otherwise IFX_ERROR
*/
IFX_return_t Common_LEC_On_PCM_Ch(IFX_int32_t nPCM_Ch, IFX_enDis_t nEnable,
                                  BOARD_t* pBoard)
{
   IFX_return_t ret = IFX_SUCCESS;
   IFX_TAPI_WLEC_CFG_t wlecConf = {0};
   IFX_int32_t nPCM_Fd;

   if ((IFX_NULL == pBoard) || (0 > nPCM_Ch))
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Invalid input argument(s). (File: %s, line: %d)\n",
             __FILE__, __LINE__));
      return IFX_ERROR;
   }

   /* check if lec should be used */
   if (pBoard->pCtrl->pProgramArg->oArgFlags.nNoLEC)
      return IFX_SUCCESS;

   /* get channel file descriptor */
   nPCM_Fd = PCM_GetFD_OfCh(nPCM_Ch, pBoard);
   if (NO_DEVICE_FD == nPCM_Fd)
   {
      /* PCM channel */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Failed to get FD of PCM ch %d, board %s. "
             "(File: %s, line: %d)\n",
             nPCM_Ch, pBoard->pszBoardName, __FILE__, __LINE__));
      return IFX_ERROR;
   }

   /* Prepare structure */
   memset(&wlecConf, 0, sizeof(IFX_TAPI_WLEC_CFG_t));

   if (IFX_ENABLE == nEnable)
   {
      /* enable settings */
      wlecConf.nType = IFX_TAPI_WLEC_TYPE_NE;
      wlecConf.bNlp = IFX_TAPI_WLEC_NLP_ON;
      wlecConf.nNBNEwindow = IFX_TAPI_WLEN_WSIZE_16;
      wlecConf.nNBFEwindow = 0;
      wlecConf.nWBNEwindow = 0;
   }
   else
   {
      /* disable settings */
      wlecConf.nType = IFX_TAPI_WLEC_TYPE_OFF;
      wlecConf.bNlp = IFX_TAPI_WLEC_NLP_OFF;
      wlecConf.nNBNEwindow = 0;
      wlecConf.nNBFEwindow = 0;
      wlecConf.nWBNEwindow = 0;
   }
#ifdef TAPI_VERSION4
   wlecConf.ch = nPCM_Ch;
   wlecConf.dev = 0;
#endif /*  TAPI_VERSION4 */
   /* activate LEC */
   ret = ioctl(nPCM_Fd, IFX_TAPI_WLEC_PCM_CFG_SET, (IFX_int32_t) &wlecConf);
   if (ret != IFX_SUCCESS)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("PCM ch No %d, Board %s: LEC failed to %s on PCM channel "
             "(File: %s, line: %d)\n",
             nPCM_Ch, pBoard->pszBoardName,
             IFX_ENABLE == nEnable ? "start" : "stop", __FILE__, __LINE__));
   }
   else
   {
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("PCM ch No %d, Board %s: LEC %s on PCM channel\n",
             nPCM_Ch, pBoard->pszBoardName,
             IFX_ENABLE == nEnable ? "started" : "stopped"));
   }
   return ret;
}

#ifdef TAPI_VERSION4
/**
   Enable pulse detection on phone.

   \param pCtrl - handle to connection control structure
   \param pPhone Pointer to PHONE.
   \return IFX_SUCCESS on ok, otherwise IFX_ERROR
*/
IFX_return_t Common_PulseEnable(CTRL_STATUS_t* pCtrl, PHONE_t* pPhone)
{
#ifdef EASY336
   SVIP_RM_Status_t ret = SVIP_RM_Success;
   RM_SVIP_RU_t RU;
#endif
   IFX_TAPI_EVENT_t tapiEvent;
   IFX_return_t ioctlRet;
   BOARD_t *pBoard;

#ifdef EASY336
   ret = SVIP_RM_LineIdRUSigDetGet(pPhone->lineID, &RU);
   if (ret == SVIP_RM_Success)
#endif
   {
#ifdef XT16
      pBoard = ABSTRACT_GetBoard(pCtrl, IFX_TAPI_DEV_TYPE_VIN_XT16);
#elif EASY336
      pBoard = ABSTRACT_GetBoard(pCtrl, RU.devType);
#else
      pBoard = pPhone->pBoard;
#endif
      if (pBoard == IFX_NULL)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: Invalid device type (File: %s, "
            "line: %d)\n", __FILE__, __LINE__));
         return IFX_ERROR;
      }

      memset((void*)&tapiEvent, 0, sizeof(IFX_TAPI_EVENT_t));
#ifdef EASY336
      tapiEvent.dev = RU.nDev;
      tapiEvent.ch = RU.nCh;
      tapiEvent.module = RU.module;
#else
      tapiEvent.dev = pPhone->nDev;
      tapiEvent.ch = pPhone->nPhoneCh;
      tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
#endif
      tapiEvent.id = IFX_TAPI_EVENT_PULSE_DIGIT;
      TD_ASSERT ((ioctlRet = ioctl(pBoard->nDevCtrl_FD,
         IFX_TAPI_EVENT_ENABLE, (IFX_int32_t) &tapiEvent)) == IFX_SUCCESS,
         pBoard->nDevCtrl_FD, pBoard, tapiEvent.dev, tapiEvent.ch);
      if (ioctlRet != IFX_SUCCESS)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: IFX_TAPI_EVENT_ENABLE failed (File: %s, "
            "line: %d)\n",
            __FILE__, __LINE__));
         return IFX_ERROR;
      }
   }
#ifdef EASY336
   else
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
        ("Err: SVIP_RM_LineIdRUSigDetGet returned %d (File: %s, "
         "line: %d)\n",
         ret, __FILE__, __LINE__));
      return IFX_ERROR;
   }
#endif
   return IFX_SUCCESS;
}

/**
   Disable pulse detection on phone.

   \param pCtrl - handle to connection control structure
   \param pPhone Pointer to PHONE.
   \return IFX_SUCCESS on ok, otherwise IFX_ERROR
*/
IFX_return_t Common_PulseDisable(CTRL_STATUS_t* pCtrl, PHONE_t* pPhone)
{
#ifdef EASY336
   SVIP_RM_Status_t ret = SVIP_RM_Success;
   RM_SVIP_RU_t RU;
#endif
   IFX_TAPI_EVENT_t tapiEvent;
   IFX_return_t ioctlRet;
   BOARD_t *pBoard;

#ifdef EASY336
   ret = SVIP_RM_LineIdRUSigDetGet(pPhone->lineID, &RU);
   if (ret == SVIP_RM_Success)
#endif
   {
#ifdef EASY336
      pBoard = ABSTRACT_GetBoard(pCtrl, RU.devType);
#elif XT16
      pBoard = ABSTRACT_GetBoard(pCtrl, IFX_TAPI_DEV_TYPE_VIN_XT16);
#else
      pBoard = pPhone->pBoard;
#endif
      if (pBoard == IFX_NULL)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: Invalid device type (File: %s, "
            "line: %d)\n", __FILE__, __LINE__));
         return IFX_ERROR;
      }

      memset((void*)&tapiEvent, 0, sizeof(IFX_TAPI_EVENT_t));
#ifdef EASY336
      tapiEvent.dev = RU.nDev;
      tapiEvent.ch = RU.nCh;
      tapiEvent.module = RU.module;
#else
      tapiEvent.dev = pPhone->nDev;
      tapiEvent.ch = pPhone->nDataCh;
      tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
#endif
      /* DTMF tone detection done from the analog line side. */
      tapiEvent.id = IFX_TAPI_EVENT_PULSE_DIGIT;
      TD_ASSERT ((ioctlRet = ioctl(pBoard->nDevCtrl_FD,
         IFX_TAPI_EVENT_DISABLE, (IFX_int32_t) &tapiEvent)) == IFX_SUCCESS,
         pBoard->nDevCtrl_FD, pBoard, tapiEvent.dev, tapiEvent.ch);
      if (ioctlRet != IFX_SUCCESS)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: IFX_TAPI_EVENT_DISABLE failed (File: %s, "
            "line: %d)\n",
            __FILE__, __LINE__));
         return IFX_ERROR;
      }
   }
#ifdef EASY336
   else
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
        ("Err: SVIP_RM_LineIdRUSigDetGet returned %d (File: %s, "
         "line: %d)\n",
         ret, __FILE__, __LINE__));
      return IFX_ERROR;
   }
#endif
   return IFX_SUCCESS;
}

/**
   Enable CID detection on phone.

   \param pPhone Pointer to PHONE.
   \param bCID_AS Boolean switch to request CID with acknowledge signal
                  detection
   \return IFX_SUCCESS on ok, otherwise IFX_ERROR
*/
IFX_return_t Common_CID_Enable(PHONE_t* pPhone, IFX_boolean_t bCID_AS)
{
#ifdef EASY336
   SVIP_RM_Status_t ret = SVIP_RM_Success;
   RM_SVIP_RU_t RU;
   if (IFX_TRUE == pPhone->fCID_Enabled)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Err: Phone No %d: CID already enabled. (File: %s, line: %d)\n",
             pPhone->nPhoneNumber, __FILE__, __LINE__));
      return IFX_ERROR;
   }

   /* request CID from resource manager */
   ret = SVIP_RM_CidAlloc (pPhone->lineID, bCID_AS, &RU);
   if (ret != SVIP_RM_Success)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Err: SVIP_RM_CidAlloc returned %d (File: %s, line: %d)\n",
             ret, __FILE__, __LINE__));
      return IFX_ERROR;
   }
   pPhone->fCID_Enabled = IFX_TRUE;
#endif
   return IFX_SUCCESS;
}

/**
   Disable CID detection on phone.

   \param pPhone Pointer to PHONE.
   \return IFX_SUCCESS on ok, otherwise IFX_ERROR
*/
IFX_return_t Common_CID_Disable(PHONE_t* pPhone)
{
#ifdef EASY336
   SVIP_RM_Status_t ret = SVIP_RM_Success;
   if (IFX_FALSE == pPhone->fCID_Enabled)
   {
      /* CID already disabled on this phone */
      return IFX_SUCCESS;
   }

   ret = SVIP_RM_CidRelease(pPhone->lineID);
   if (ret != SVIP_RM_Success)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Err: SVIP_RM_CidRelease returned %d (File: %s, line: %d)\n",
             ret, __FILE__, __LINE__));
      return IFX_ERROR;
   }
   pPhone->fCID_Enabled = IFX_FALSE;
#endif
   return IFX_SUCCESS;
}
#endif /* TAPI_VERSION4 */

#ifndef TAPI_VERSION4
/**
   Enable DTMF signal detection on phone.

   \param pCtrl - handle to connection control structure
   \param pPhone Pointer to PHONE.
   \return IFX_SUCCESS on ok, otherwise IFX_ERROR
*/
IFX_return_t Common_DTMF_SIG_Enable(CTRL_STATUS_t* pCtrl, PHONE_t* pPhone)
{
   IFX_return_t ret = IFX_SUCCESS;
   IFX_TAPI_SIG_DETECTION_t dtmfDetection;

   if ((pCtrl == IFX_NULL) || (pPhone == IFX_NULL))
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }

   if (pPhone->nDataCh_FD > 0)
   {
      memset(&dtmfDetection, 0, sizeof (IFX_TAPI_SIG_DETECTION_t));

      /* Stop detection of DTMF tones from local interface */
      dtmfDetection.sig = IFX_TAPI_SIG_DTMFTX | IFX_TAPI_SIG_DTMFRX;
      ret = ioctl(pPhone->nDataCh_FD, IFX_TAPI_SIG_DETECT_ENABLE,
                  (IFX_int32_t) &dtmfDetection);
      if (ret != IFX_SUCCESS)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: IFX_TAPI_SIG_DETECT_ENABLE failed, %s, DataCh %d, fd %d"
            "(File: %s, line: %d)\n",
             pPhone->pBoard->pszBoardName, pPhone->nDataCh, pPhone->nDataCh_FD,
            __FILE__, __LINE__));
         return IFX_ERROR;
      }
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
           ("Phone No %d: DTMF detection started on %s, DataCh %d, fd %d\n",
            pPhone->nPhoneNumber, pPhone->pBoard->pszBoardName,
            pPhone->nDataCh, pPhone->nDataCh_FD));
   }

   return IFX_SUCCESS;
}

/**
   Disable DTMF signal detection on phone.

   \param pCtrl - handle to connection control structure
   \param pPhone Pointer to PHONE.
   \return IFX_SUCCESS on ok, otherwise IFX_ERROR
*/
IFX_return_t Common_DTMF_SIG_Disable(CTRL_STATUS_t* pCtrl, PHONE_t* pPhone)
{
   IFX_return_t ret = IFX_SUCCESS;
   IFX_TAPI_SIG_DETECTION_t dtmfDetection;

   if ((pCtrl == IFX_NULL) || (pPhone == IFX_NULL))
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }

   if (pPhone->nDataCh_FD > 0)
   {
      memset(&dtmfDetection, 0, sizeof (IFX_TAPI_SIG_DETECTION_t));

      /* Stop detection of DTMF tones from local interface */
      dtmfDetection.sig = IFX_TAPI_SIG_DTMFTX | IFX_TAPI_SIG_DTMFRX;
      ret = ioctl(pPhone->nDataCh_FD, IFX_TAPI_SIG_DETECT_DISABLE, 
         (IFX_int32_t) &dtmfDetection);
      if (ret != IFX_SUCCESS)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: IFX_TAPI_SIG_DETECT_DISABLE failed %s, data ch %d, fd %d (File: %s, "
            "line: %d)\n",
            pPhone->pBoard->pszBoardName, pPhone->nDataCh, pPhone->nDataCh_FD, 
            __FILE__, __LINE__));
         return IFX_ERROR;
      }
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
           ("Phone No %d: DTMF detection stopped on %s, data ch %d, fd %d \n",
            pPhone->nPhoneNumber, pPhone->pBoard->pszBoardName, pPhone->nDataCh, pPhone->nDataCh_FD));
   }
   return IFX_SUCCESS;

}

#endif /* TAPI_VERSION4 */

#ifdef TAPI_VERSION4
#ifdef EASY336
/**
   Get array of phones with the specified connID.

   \param pCtrl - handle to connection control structure
   \param connID - connID
   \param pPhones - pointer to array of pointers to phones (array allocated
   and filled by function)

   \return Number of phones with specified connID
*/
IFX_int32_t Common_ConnID_PhonesGet(CTRL_STATUS_t* pCtrl,
                                    IFX_int32_t connID, PHONE_t** pPhones[])
{
   IFX_int32_t i, j, k = 0, nPhones = 0;

   for (i = 0; i < pCtrl->nBoardCnt; i++)
      for (j = 0; j < pCtrl->rgoBoards[i].nMaxAnalogCh; j++)
         if (pCtrl->rgoBoards[i].rgoPhones[j].connID == connID)
            nPhones++;

   if (nPhones > 0)
   {
      *pPhones = malloc(sizeof(PHONE_t*) * nPhones);
      for (i = 0; i < pCtrl->nBoardCnt; i++)
         for (j = 0; j < pCtrl->rgoBoards[i].nMaxAnalogCh; j++)
            if (pCtrl->rgoBoards[i].rgoPhones[j].connID == connID)
               (*pPhones)[k++] = &pCtrl->rgoBoards[i].rgoPhones[j];
   }
   else
      *pPhones = IFX_NULL;

   return nPhones;
}
#endif /* EASY336 */

#ifdef LINUX
/**
   Writes a byte to the specified memory address.

   \param nAddr - address.
   \param nValue - value.

   \return IFX_SUCCESS on ok, otherwise IFX_ERROR
*/
IFX_return_t Common_WriteByte(IFX_uint32_t nAddr, IFX_uint8_t nValue)
{
   IFX_int32_t mfd;
   IFX_void_t *pRealIO;
   off_t real_addr, offset;

   real_addr = nAddr & ~4095;
   if (real_addr == 0xfffff000)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
        ("Err: cannot map the top 4K page (File: %s, "
         "line: %d)\n",
         __FILE__, __LINE__));
         return IFX_ERROR;
      }
   offset = nAddr - real_addr;
   if (real_addr + 0x1000 < real_addr)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
        ("Err: Aligned addr+len exceeds top of address space (File: %s, "
         "line: %d)\n",
         __FILE__, __LINE__));
      return IFX_ERROR;
   }
   mfd = open("/dev/mem", O_RDWR);
   if (mfd < 0)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
        ("Err: /dev/mem open failed (File: %s, "
         "line: %d)\n",
         __FILE__, __LINE__));
      return IFX_ERROR;
   }
   pRealIO = mmap(IFX_NULL, 0x1000, PROT_READ|PROT_WRITE, MAP_SHARED, mfd,
      real_addr);
   if (pRealIO == (void *)(-1))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
        ("Err: mmap failed (File: %s, "
         "line: %d)\n",
         __FILE__, __LINE__));
      close (mfd);
      return IFX_ERROR;
   }
   *(IFX_uint8_t *)((IFX_uint8_t *)pRealIO + offset) = nValue;
   munmap(pRealIO, 0x1000);
   close (mfd);
   return IFX_SUCCESS;
}
#endif

/**
   Reads chip count and sets board field "nChipCnt" of passed BOARD_t structure
   pointer.

   \param pBoard - pointer to board structure.

   \return IFX_SUCCESS on ok, otherwise IFX_ERROR
*/
IFX_return_t Common_SetChipCount(BOARD_t* pBoard)
{
   IFX_TAPI_CAP_t CapList;
   IFX_int32_t nIO_Ret;
   IFX_TAPI_Error_t error = {0};

   memset(&CapList, 0, sizeof(IFX_TAPI_CAP_t));
   switch (pBoard->nType)
   {
      case TYPE_VINETIC:
         CapList.cap = IFX_TAPI_DEV_TYPE_VINETIC;
         break;
      case TYPE_DANUBE:
         CapList.cap = IFX_TAPI_DEV_TYPE_INCA_IPP;
         break;
      case TYPE_DUSLIC_XT:
         CapList.cap = IFX_TAPI_DEV_TYPE_DUSLIC_XT;
         break;
      case TYPE_DUSLIC:
         CapList.cap = IFX_TAPI_DEV_TYPE_DUSLIC;
         break;
      case TYPE_SVIP:
         CapList.cap = IFX_TAPI_DEV_TYPE_VIN_SUPERVIP;
         break;
      case TYPE_XT16:
         CapList.cap = IFX_TAPI_DEV_TYPE_VIN_XT16;
         break;
      default:
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err: Invalid board type (File: %s, "
         "line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }
   CapList.captype = IFX_TAPI_CAP_TYPE_DEVTYPE;
   do
   {
      CapList.dev = (IFX_uint16_t) pBoard->nChipCnt;
      nIO_Ret = ioctl(pBoard->nDevCtrl_FD, IFX_TAPI_CAP_CHECK,
         (IFX_int32_t) &CapList);
      if (nIO_Ret == 1)
         pBoard->nChipCnt++;
   }
   while (nIO_Ret == 1);

   /* clear error stack after previous failure of IFX_TAPI_CAP_CHECK */
   ioctl(pBoard->nDevCtrl_FD, IFX_TAPI_LASTERR, (IFX_int32_t) &error);
   
   return IFX_SUCCESS;
}
#endif /* TAPI_VERSION4 */

/**
   Sets line feed mode.

   \param pPhone - pointer to phone structure.
   \param mode - desired line feed mode.

   \return IFX_SUCCESS on ok, otherwise IFX_ERROR
*/
IFX_return_t Common_LineFeedSet(PHONE_t* pPhone, TD_LINE_MODE_t mode)
{
   IFX_return_t ret = IFX_ERROR;
#ifdef TAPI_VERSION4
   IFX_TAPI_LINE_FEED_t lineFeed;

   if (pPhone->pBoard->fSingleFD)
   {
      lineFeed.dev = pPhone->nDev;
      lineFeed.ch = pPhone->nPhoneCh;
      lineFeed.lineMode = mode;
      TD_ASSERT ((ret = ioctl(pPhone->nPhoneCh_FD, IFX_TAPI_LINE_FEED_SET,
         (IFX_int32_t) &lineFeed)) == IFX_SUCCESS, pPhone->pBoard->nDevCtrl_FD,
         pPhone->pBoard, lineFeed.dev, lineFeed.ch);
      if (IFX_SUCCESS != ret)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Line feed mode set to %s failed for %s dev %d, ch %d, "
                "(File: %s, line: %d)\n",
                LINE_MODES[mode], pPhone->pBoard->pszBoardName, lineFeed.dev,
                lineFeed.ch, __FILE__, __LINE__));
         return IFX_ERROR;
      }
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("Phone No %d: Line mode changed to %s \n"
             "%son %s dev %d, ch %d \n",
             pPhone->nPhoneNumber, LINE_MODES[mode],
             pPhone->pBoard->pIndentPhone,
             pPhone->pBoard->pszBoardName, lineFeed.dev, lineFeed.ch));
   }
   else
#endif /* TAPI_VERSION4 */
   {
      if (pPhone->nPhoneCh_FD > 0)
      {
         ret = ioctl(pPhone->nPhoneCh_FD, IFX_TAPI_LINE_FEED_SET, mode);
      }
   }
   return ret;
}

/**
   Moves string pointer to the filename part if passed string.

   \param sPath - string.

   \return none
*/
IFX_void_t Common_StripPath (IFX_char_t** sFileName)
{
   IFX_char_t* sBegin = *sFileName;

   if (*sFileName == IFX_NULL || strlen(*sFileName) == 0)
      return;

   /* move to the last character */
   *sFileName += strlen(*sFileName) - 1;

   /* only name of file without path */
   while (*sFileName > sBegin && *(*sFileName - 1) != 0x5C &&
      *(*sFileName - 1) != '/')
   {
      *sFileName -= 1;
   }
}

/**
   Return pointer to the CPU device structure from given board.

   \param pDevice - pointer to the device list
   \return pointer to the CPU device structure or IFX_NULL in case of error
*/
TAPIDEMO_DEVICE_t * Common_GetDevice_CPU(TAPIDEMO_DEVICE_t* pDevice)
{
   TAPIDEMO_DEVICE_t* pDev = pDevice;

   if (pDevice == IFX_NULL)
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
           __FILE__, __LINE__));
      return IFX_NULL;
   }

   do
   {
      if (pDev->nType == TAPIDEMO_DEV_CPU)
      {
         /* CPU device found. Return pointer to structure. */
         return pDev;
      }
      pDev = pDev->pNext;
   } while(pDev != IFX_NULL);

   return IFX_NULL;
}

/**
   Return pointer to the FXO device structure from given board.

   \param pDevice - pointer to the device list
   \param nFxoType - FXO type (Teridian or DuSLIC)
   \return pointer to the CPU device structure or IFX_NULL in case of error
*/
TAPIDEMO_DEVICE_t * Common_GetDevice_FXO(TAPIDEMO_DEVICE_t* pDevice,
                                          FXO_TYPE_t nFxoType)
{
   TAPIDEMO_DEVICE_t* pDev = pDevice;

   if (pDevice == IFX_NULL)
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
           __FILE__, __LINE__));
      return IFX_NULL;
   }

   do
   {
      if (pDev->nType == TAPIDEMO_DEV_FXO)
      {
         if (pDev->uDevice.stFxo.nFxoType == nFxoType)
         {
            /* FXO device found. Return pointer to structure. */
            return pDev;
         }
      }
      pDev = pDev->pNext;
   } while(pDev != IFX_NULL);

   return IFX_NULL;
}

/**
   Add device to the board.

   \param pBoard - pointer to the board
   \param pDevice - pointer to the device to add
   \return IFX_SUCCESS if ok, otherwise IFX_ERROR
*/
IFX_return_t Common_AddDevice(BOARD_t* pBoard, TAPIDEMO_DEVICE_t* pDevice)
{
   TAPIDEMO_DEVICE_t* pDev = IFX_NULL;

   if ((pBoard == IFX_NULL) || (pDevice == IFX_NULL))
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
           __FILE__, __LINE__));
      return IFX_ERROR;
   }

   if (pBoard->pDevice == IFX_NULL)
   {
      pBoard->pDevice = pDevice;
      return IFX_SUCCESS;
   }
   pDev = pBoard->pDevice;

   /* look for last element in the list */
   while (pDev->pNext != IFX_NULL)
   {
      pDev = pDev->pNext;
   }

   /* add device to the end of list */
   pDev->pNext = pDevice;
   return IFX_SUCCESS;
}

/**
   Remove device list from the board.

   \param pBoard - pointer to the board
   \return IFX_SUCCESS if ok, otherwise IFX_ERROR
*/
IFX_return_t Common_RemoveDeviceList(BOARD_t* pBoard)
{
   TAPIDEMO_DEVICE_t *pDevice = IFX_NULL;
   TAPIDEMO_DEVICE_t *pNextDevice = IFX_NULL;

   if (pBoard == IFX_NULL)
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
           __FILE__, __LINE__));
      return IFX_ERROR;
   }

   pDevice = pBoard->pDevice;
   while (pDevice != IFX_NULL)
   {
      pNextDevice = pDevice->pNext;
      free(pDevice);
      pDevice = pNextDevice;
   }
   return IFX_SUCCESS;
}

/**
   Set Voice Activity Detection (VAD).

   \param pPhone - pointer to phone
   \param pBoard - pointer to the board
   \param nVadCfg - vad configuration to set,
        if RESTORE_DEFAULT then restore default settings

   \return IFX_SUCCESS if ok, otherwise IFX_ERROR
*/
IFX_return_t Common_SetVad(PHONE_t* pPhone, BOARD_t* pBoard,
                            IFX_int32_t nVadCfg)
{
   IFX_return_t ret = IFX_SUCCESS;
   VOIP_DATA_CH_t* pDataChStat = IFX_NULL;
#ifdef TAPI_VERSION4
   IFX_TAPI_ENC_VAD_CFG_t oVadCfg;
#endif /* TAPI_VERSION4 */

   if ((IFX_NULL == pPhone) || (IFX_NULL == pBoard))
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
           __FILE__, __LINE__));
      return IFX_ERROR;
   }

#ifdef TAPI_VERSION4
#ifdef EASY336
   /* get codec data */
   if (pPhone->nVoipIDs > 0)
   {
      pDataChStat = Common_GetCodec(pBoard->pCtrl,
                       pPhone->pVoipIDs[pPhone->nVoipIDs - 1]);
      if (IFX_NULL == pDataChStat)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Phone No %d: Failed to get codec. "
                "(File: %s, line: %d)\n",
                pPhone->nPhoneNumber, __FILE__, __LINE__));
      }
   }
   else if (RESTORE_DEFAULT != nVadCfg)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Phone No %d: No Voip ID - unable to change VAD settings  "
             "(File: %s, line: %d)\n",
             pPhone->nPhoneNumber, __FILE__, __LINE__));
      return IFX_ERROR;
   }
   else
   {
      /* we are trying to restore default settings but no codec available,
         settings should be restored before releasing voip id */
      return IFX_SUCCESS;
   }
#endif /* EASY336 */
#else /* TAPI_VERSION4 */
   if (0 > pPhone->nDataCh || 0 >= pPhone->pBoard->nMaxCoderCh)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Failed to get channel number of phone No %d "
             "(File: %s, line: %d)\n",
             pPhone->nPhoneNumber, __FILE__, __LINE__));
      return IFX_ERROR;
   }
   /* get used data channel statistics */
   pDataChStat = &pBoard->pDataChStat[pPhone->nDataCh];
#endif /* TAPI_VERSION4 */

   if (RESTORE_DEFAULT == nVadCfg)
   {
      if (IFX_TRUE == pDataChStat->fDefaultVad)
      {
         return IFX_SUCCESS;
      }
      nVadCfg = pBoard->pCtrl->pProgramArg->nVadCfg;
   } /* if (IFX_FALSE == nVadCfg) */
   else
   {
      switch (nVadCfg)
      {
         case IFX_TAPI_ENC_VAD_NOVAD:
            TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
                  ("Phone No %d: Set VAD: %d - IFX_TAPI_ENC_VAD_NOVAD\n"
                   "%s(no voice activity detection).\n",
                   pPhone->nPhoneNumber, nVadCfg, pBoard->pIndentPhone));
            break;
         case IFX_TAPI_ENC_VAD_ON:
            TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
                  ("Phone No %d: Set VAD: %d - IFX_TAPI_ENC_VAD_ON\n"
                   "%s(voice activity detection on).\n",
                   pPhone->nPhoneNumber, nVadCfg, pBoard->pIndentPhone));
            break;
         case IFX_TAPI_ENC_VAD_G711:
            TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
                  ("Phone No %d: Set VAD: %d - IFX_TAPI_ENC_VAD_G711\n"
                   "%s(voice activity detection on without "
                   "spectral information).\n",
                   pPhone->nPhoneNumber, nVadCfg, pBoard->pIndentPhone));
            break;
         case IFX_TAPI_ENC_VAD_SC_ONLY:
            TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
                  ("Phone No %d: Set VAD: %d - IFX_TAPI_ENC_VAD_SC_ONLY\n"
                   "%s(voice activity detection on without "
                   "comfort noise generation).\n",
                   pPhone->nPhoneNumber, nVadCfg, pBoard->pIndentPhone));
            break;
         default:
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("Phone No %d: Invalid vad configuration - %d.\n",
                   pPhone->nPhoneNumber, nVadCfg));
            return IFX_ERROR;
            break;
      } /* switch (nVadCfg) */
   } /* if (IFX_FALSE == nVadCfg) */

#ifdef TAPI_VERSION4
   if (pBoard->fSingleFD)
   {
      BOARD_t *pSvipBoard = pBoard;
      /* Voice Activity Detection (VAD) configuration */
      oVadCfg.dev = pDataChStat->nDev;
      oVadCfg.ch = pDataChStat->nCh;
      oVadCfg.vadMode = nVadCfg;

      if (TYPE_XT16 == pBoard->nType)
      {
         pSvipBoard = ABSTRACT_GetBoard(pBoard->pCtrl,
                                        IFX_TAPI_DEV_TYPE_VIN_SUPERVIP);
         if (IFX_NULL == pSvipBoard)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("Failed to get svip board structure. (File: %s, line: %d)\n",
                   __FILE__, __LINE__));
            return IFX_ERROR;
         }
      }
      TD_ASSERT ((ret = ioctl(pSvipBoard->nDevCtrl_FD, IFX_TAPI_ENC_VAD_CFG_SET,
         (IFX_int32_t) &oVadCfg)) == IFX_SUCCESS, pSvipBoard->nDevCtrl_FD,
         pSvipBoard, oVadCfg.dev, oVadCfg.ch);
      if (ret != IFX_SUCCESS)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("IFX_TAPI_ENC_VAD_CFG_SET failed for device %d channel %d."
                " (File: %s, line: %d)\n",
                oVadCfg.dev, oVadCfg.ch, __FILE__, __LINE__));
         return IFX_ERROR;
      }
   }
   else
#endif /* TAPI_VERSION4 */
   {
      /* Voice Activity Detection (VAD) configuration */
      ret =  ioctl(pPhone->nDataCh_FD, IFX_TAPI_ENC_VAD_CFG_SET,
                   (IFX_int32_t) nVadCfg);
      if (ret != IFX_SUCCESS)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("IFX_TAPI_ENC_VAD_CFG_SET failed for fd %d to set %d"
                "(File: %s, line: %d)\n",
                pPhone->nDataCh_FD, nVadCfg, __FILE__, __LINE__));
         return IFX_ERROR;
      }
   }

   /* set vad configuration flag to restore vad settings */
   if (nVadCfg == pBoard->pCtrl->pProgramArg->nVadCfg)
   {
      pDataChStat->fDefaultVad = IFX_TRUE;
   }
   else
   {
      pDataChStat->fDefaultVad = IFX_FALSE;
   }

   return IFX_SUCCESS;
}

/**
    function to return the name assigned to a enum value

   \param   nEnum  -  enum value (id)
   \param   pEnumName -  pointer to array with enum name

   \return  pointer to enum name. If enum value is not found or in case of 
            error, pointer to string with error description is returned.
*/
IFX_char_t* Common_Enum2Name(IFX_int32_t nEnum, TD_ENUM_2_NAME_t *pEnumName)
{
   IFX_char_t *pName = IFX_NULL;
   IFX_int32_t nDebugLevel = DBG_LEVEL_HIGH;

   /* check pointer, on error return pointer to error string */
   TD_PTR_CHECK(pEnumName, TD_rgEnum2NameErrorCode[ENUM_NULL_POINTER]);

   /* search for nEnum value */
   do
   {
      /* check if value is the same */
      if (pEnumName->nID == nEnum)
      {
         /* set enum name string */
         pName = pEnumName->rgName;
      }
      pEnumName++;
   } while ((pName == IFX_NULL) && (pEnumName->nID != TD_MAX_ENUM_ID));

#ifdef ABACUS_TEST
   /* sometimes nEnum is not set to valid enum value to show that some value is
      not set, to avoid ABACUS TEST error set trace debug to normal. */
   nDebugLevel = DBG_LEVEL_NORMAL;
#endif /* ABACUS_TEST */

   if (pName == IFX_NULL)
   {
      /* error - enum name was not found */
      TRACE(TAPIDEMO, nDebugLevel,
            ("Name for enum value (0x%x) type %s not found in the table.\n",
             nEnum, pEnumName->rgName));
      /* if pName is IFX_NULL, it means that pEnumName points to last 
         element in the table. Assign string with error info. */
      pName = TD_rgEnum2NameErrorCode[ENUM_VALUE_NOT_FOUND];
   }
   return pName;
}

/**
   Function to return the number of enum assigned to a enum name.

   \param   pString  -  name of enum to be found
   \param   pEnumName -  pointer to array with enum name
   \param   bPrintError -  if IFX_TRUE error messahe is printed
                           if name not found

   \return  number of enum from pEnumName table or
            IFX_ERROR if enum number not found.
*/
IFX_int32_t Common_Name2Enum(IFX_char_t* pString, TD_ENUM_2_NAME_t *pEnumName,
                             IFX_boolean_t bPrintError)
{
   /* check pointer, on error return pointer to error string */
   TD_PTR_CHECK(pEnumName, IFX_ERROR);
   TD_PTR_CHECK(pString, IFX_ERROR);
   TD_PARAMETER_CHECK((0 >= strlen(pString)), strlen(pString), IFX_ERROR);

   /* search for nEnum value */
   do
   {
      /* check if value is the same */
      if (0 == strcmp(pString, pEnumName->rgName))
      {
         /* return enum number */
         return pEnumName->nID;
      }
      pEnumName++;
   } while (pEnumName->nID != TD_MAX_ENUM_ID);

   if (IFX_TRUE == bPrintError)
   {
      /* error - enum number was not found */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Number of enum %s for string %s was not found.\n",
             pEnumName->rgName, pString));
   }

   return IFX_ERROR;
}

/**
    Function check if trace is not too long

   \param   pBuffer2  -  trace

   \return
*/
IFX_void_t Common_CheckTraceLen(IFX_char_t* pBuffer2)
{
   IFX_char_t* pPresentCh;
   IFX_char_t* pPreviusCh;

   const IFX_char_t EOL = '\n';

   /* If there isn't IFX_char_t '\n' at the end of string, some data can be missed */
   IFX_char_t pBuffer[strlen(pBuffer2+1)];
   sprintf(pBuffer, "%s\n", pBuffer2);
  
   /* Locate first occurrence of character '\n' */
   pPresentCh = strchr(pBuffer, EOL);
     
   pPreviusCh = pBuffer;

   while (pPresentCh != NULL)
   {
      if (pPresentCh-pPreviusCh+1 > MAX_TRACE_LENGTH)
      {
         printf("Above trace is too long: %d\n", pPresentCh-pPreviusCh+1);
      }

      pPreviusCh = pPresentCh+1;
      /* Locate next occurrence of character '\n' */
      pPresentCh = strchr(pPreviusCh, EOL);
   } /* while */
}

#ifdef VXWORKS
size_t strnlen (__const char *__string, size_t __maxlen)
{
   return (strlen(__string) < __maxlen) ? strlen(__string) : __maxlen; 
}
#endif
